/*
 * Softing CAN PCI Driver
 * Copyright (C) 1994-2006, Softing AG, Werner Knorsch
 * 
 * See file CREDITS for list of people who contributed to this
 * project.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * email: support.automation@softing.com
 * Softing AG, Richard-Reitzner-Allee 6, 85540 Haar, Germany
 */

/**
 * Title:        Test Node CAN Driver
 * Description:  This driver contains ported Softing CAN driver from windows (CAN-ACX-104).
 *
 *               Driver code is implemented for a Linux Char driver
 *
 *               can_init() sets up each of the Softing CAN Interface Cards
 *               can_cleanup() release resources
 *               can_interrupt() is called when a CAN message is recived
 *               do_can_read() is called when reading data from driver
 *
 *               sample driver code was taken from:
 *                  Title       : Linux Device Drivers 2nd Edition
 *                  Authors     : Alessandro Rubini & Jonathan Corbet
 *                  published   : O'Reilly
 */

/*
 * $Id: can_driver.c,v 1.9 2003/11/04 12:48:16 kno Exp $
 */

#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif


/******************************************************************
**  include files
*******************************************************************/

#include "os.h"

#include "BaseTypes.h"

#include "can_device.h"
#include <linux/module.h>
#include <linux/proc_fs.h>

#include "os.h"
#include "sysdep.h"
#include "shared_usb_mem.h"
#include "card_def.h"

#ifdef CANCARD
#include "/usr/src/linux/drivers/pcmcia/cs_internal.h"
#endif

#include "can_def.h"

#include "can_loadcard.h"
#include "can_card.h"

#include "can_ioctl.h"

#ifdef CANCARD
#include "can_pcmcia.h"
#endif
#ifdef CANACXPCI
#include "can_pci.h"
#endif

#ifdef CANAC104
#include "can_104.h"
#endif

/******************************************************************
**  macros
*******************************************************************/

extern struct proc_dir_entry *procDir;


 int *ppram;
 unsigned long can_ioc_pram[10];
 unsigned long can_ioc_l;
 unsigned long *can_ioc_pl;
 unsigned int *can_ioc_pi;
 
 unsigned long *pCanIocPram = can_ioc_pram;
 unsigned long *pcan_resources = (unsigned long*)&can_resources;
 unsigned long *pcan_ioc_l = &can_ioc_l;


 CANPC_RESOURCES can_resources;

/******************************************************************
** protoypes  
*******************************************************************/
int can_open(struct inode *, struct file *);
int can_release(struct inode *, struct file *);
ssize_t do_can_read(struct inode *, struct file *, SCHAR *, size_t, loff_t *);
ssize_t can_read(struct file *, char *, size_t, loff_t *);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,33)
int can_ioctl( struct inode *, struct file *, unsigned int, unsigned long);
#else
long can_ioctl(struct file *, unsigned int, unsigned long);
#endif
VOID can_interrupt(int, VOID *, struct pt_regs *);
INT32 CanInstallInterrupt(PCDEVICES pDevice);
INT32 CancardInstallISR(PCDEVICES pDevice);
int can_intCanDriver(PCDEVICES pDevice);
PCDEVICES GetDeviceFromMinorNumber(USIGN32 u32MinorNumber);
PCDEVICES GetDeviceFromFilepointer(PVOID pFile);


/****************************************************************************
**  variables
****************************************************************************/
/* CAN globals*/

struct cdev *driver_object = NULL;

// TRACE LEVEL
int can_trace_level = DEBUGLEVEL; //_PRINT_ERROR;

struct semaphore cDeviceListSemaphore;
PCDEVICES pcCanDevices = NULL;  
typedef struct can_dev
{
    SCHAR name[15];
}
can_dev;

/* Unmap a region obtained with can_remap */
VOID can_unmap(VOID *virt_add)
{
  iounmap((VOID *) ((CAN_ADDRESS) virt_add & PAGE_MASK));
}




/*****************************************************************************/
/* can_poll                                                                  */
/*****************************************************************************/
/* FUNCTIONAL DESCRIPTION                                                    */
/*   blocks the usermode application till an CAN event is available          */
/*****************************************************************************/
/* PARAMETER DECRIPTION                                                      */
/*    struct file                                                            */
/*       file descriptor                                                     */
/*                                                                           */
/*    poll_table                                                             */
/*     poll table                                                            */
/*                                                                           */
/*****************************************************************************/


unsigned int can_poll(struct file *, poll_table *);
unsigned int can_poll(struct file *filp, poll_table * table)
{
    USIGN32 mask = 0;
    can_dev *dev;
    PCDEVICES pDevice;
   
    dev = (can_dev *) filp->private_data;

   if (down_interruptible(&cDeviceListSemaphore))
    {
        _ERRPRINT
            ("can_poll: down_interruptible(&cDeviceListSemaphore) was interrupted\n");
        return ENODEV;
    }
    pDevice = GetDeviceFromFilepointer(filp);
    if (!pDevice)
    {
        _ERRPRINT("Error in can_poll\n");
        up(&cDeviceListSemaphore);
        return -ENODEV;
    }
    
    atomic_inc(&pDevice->atomInUseCount); // one more app is waiting
    up(&cDeviceListSemaphore);

    if (!pDevice->bIsStarted)
    {
      // device is not started
      atomic_dec(&pDevice->atomInUseCount);
      return -ENOMEDIUM;
    }

    _TRACE5("waiting for interrupt ... \n");
    if(pDevice->l2_state == ST_CANL2_RUN)
    { 
      // only if the system is running and online!
      _TRACE5("blocking for interrupt ... \n");
      poll_wait(filp, &pDevice->can_info.read_queue, table);
    }
    
    if (pDevice->can_info.irq_flag == 1)
    { 
      // unblock user application from ISR (data available)
      mask = POLLIN | POLLRDNORM;
    }
    
    if (pDevice->can_info.irq_flag == 2)
    { 
      // unblocked from close, reinitialize, reset, .... app should terminate on POLLHUP
      mask = POLLHUP;
    }
    
    _TRACE5("... interrupt! flag=%u, state=%u, mask=0x%x\n",pDevice->can_info.irq_flag, pDevice->l2_state, mask);
    
    pDevice->can_info.irq_flag = 0;
    
    
    atomic_dec(&pDevice->atomInUseCount);
    return mask;
}
struct file_operations can_fops = 
{
    read:can_read,
    poll:can_poll,
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,32)
    unlocked_ioctl:can_ioctl,
#else
    ioctl:can_ioctl,
#endif
    open:can_open,
    release:can_release,
};

/**
 * *****************************************************************
 * @brief  check a candidate for a minor number, if it is already in use.   
 * @author Werner Knorsch, Softing Industrial Automation GmbH
 * @date   2013-09-18
 *
 * @param  minor:   minor number which should be checked. 
 * @return FALSE, if it is already used, TRUE otherwise.
 *
 ******************************************************************
 */

BOOL checkMinor(int minor)
{
  PCDEVICES pHelpDevice = pcCanDevices;
  
  while(pHelpDevice)
  {
    if(pHelpDevice->u32MinorNumber == minor)
    {
      return FALSE;
    }
    pHelpDevice = pHelpDevice->pNextDevice;
  }
  return TRUE;
}

/**
 * *****************************************************************
 * @brief  returns a minor number which is not in use.   
 * @author Werner Knorsch, Softing Industrial Automation GmbH
 * @date   2013-09-18
 *
 * @return minor number which is not in use.
 *
 ******************************************************************
 */

int GetFreeMinor(void)
{
  int i;
  for(i=0; i<256; i++)
  {
    if(checkMinor(i))
    {
      return i;
    }
  }
  
  return MINORUNDEFINED;
  
}

/**
 * *****************************************************************
 * @brief  looks for a device in the device list with a matching memoy address.   
 * @brief  the function returns NULL if no matching device can be found.   
 * @author Werner Knorsch, Softing Industrial Automation GmbH
 * @date   2013-09-18
 *
 * @param  u32Dpram:   memory address where the device data is stored. 
 * @return NULL, if no device can be found, the address of the device otherwise.
 *
 ******************************************************************
 */


PCDEVICES FindDevice(CAN_ADDRESS Dpram)
{
    _TRACE1("FindDevice: \n");

    if (pcCanDevices)
    {
        PCDEVICES pHelpDevice = pcCanDevices;

        while (pHelpDevice)
        {
            if (pHelpDevice->can_dpram_physaddress == Dpram)
            {
                return pHelpDevice;
            }
            pHelpDevice = pHelpDevice->pNextDevice;
        }


    }
    _ERRPRINT("device not found! \n");
    return NULL;

}


/**
 * *****************************************************************
 * @brief  looks for a device in the device list with a matching memoy address.   
 * @brief  the function returns NULL if no matching device can be found.   
 * @author Werner Knorsch, Softing Industrial Automation GmbH
 * @date   2013-09-18
 *
 * @param  u32Dpram:   memory address where the device data is stored. 
 * @return NULL, if no device can be found, the address of the device otherwise.
 *
 ******************************************************************
 */


PCDEVICES SecureFindDevice(CAN_ADDRESS u32Dpram)
{
    PCDEVICES pcHelpDevice;
    _TRACE1("SecureFindDevice: \n");
    if (down_interruptible(&cDeviceListSemaphore))
    {
        _ERRPRINT
            ("SecureFindDevice: down_interruptible(&cDeviceListSemaphore) was interrupted\n");
        return NULL;
    }
    pcHelpDevice = FindDevice(u32Dpram);
    up(&cDeviceListSemaphore);
    return pcHelpDevice;
}

/*****************************************************************************/
/* InsertDevice                                                              */
/*****************************************************************************/
/* FUNCTIONAL DESCRIPTION                                                    */
/*   inserts a device in the device list and returns a pointer to the device */
/*****************************************************************************/
/* PARAMETER DECRIPTION                                                      */
/*    PCDEVICES pNewDevice: pointer to the device which should be inserted   */
/*                                                                           */
/*****************************************************************************/

PCDEVICES InsertDevice(PCDEVICES pDevice)
{

    _TRACE1("InsertDevice: \n");

    if (down_interruptible(&cDeviceListSemaphore))
    {
        _ERRPRINT
            ("InsertDevice: down_interruptible(&cDeviceListSemaphore) was interrupted\n");
        return NULL;
    }
    if (pcCanDevices)
    {
        PCDEVICES pHelpDevice = pcCanDevices;
        // looking for the end of the list
        while (pHelpDevice->pNextDevice)
        {
            pHelpDevice = pHelpDevice->pNextDevice;
        }
        pHelpDevice->pNextDevice = pDevice;
        _TRACE2("InsertDevice: device inserted at: %lx\n", (CAN_ADDRESS) pDevice);
    }
    else
    {
        pcCanDevices = pDevice;
        _TRACE1("InsertDevice: first device inserted at: %lx\n",
                (CAN_ADDRESS) pDevice);
        up(&cDeviceListSemaphore);
        return pDevice;
    }
    up(&cDeviceListSemaphore);
    return pDevice;

}

/*****************************************************************************/
/* WaitForProcesses                                                          */
/*****************************************************************************/
/* FUNCTIONAL DESCRIPTION                                                    */
/*   waits for all processes which are using the device in the driver code   */
/*****************************************************************************/
/* PARAMETER DECRIPTION                                                      */
/*    PCDEVICES pDevice: pointer to the device                               */
/*                                                                           */
/*****************************************************************************/

VOID WaitForProcesses(PCDEVICES pDevice)
{
    while (atomic_read(&pDevice->atomInUseCount))
    {
        _TRACE2("WaitForProcesses atomInUseCount=%i\n",
                atomic_read(&pDevice->atomInUseCount));
        schedule();
    }
}

/*****************************************************************************/
/* RemoveDevice                                                              */
/*****************************************************************************/
/* FUNCTIONAL DESCRIPTION                                                    */
/*   removes a device from the device list                                   */
/*****************************************************************************/
/* PARAMETER DECRIPTION                                                      */
/*    USIGN32 Dpram: physical DPRAM address of the device                    */
/*                                                                           */
/*****************************************************************************/

INT32 RemoveDevice(CAN_ADDRESS Dpram)
{
    PCDEVICES pFatherDevice, pDevice;
    _TRACE1("RemoveDevice:\n");


    if (down_interruptible(&cDeviceListSemaphore))
    {
        _ERRPRINT
            ("RemoveDevice: down_interruptible(&cDeviceListSemaphore) was interrupted\n");
        return -1;
    }
    pFatherDevice = pcCanDevices;
    pDevice = FindDevice(Dpram);

    if (pDevice != NULL)
    {                           // device is a member of the list
        if (pDevice == pcCanDevices)
        {                       // first device must be removed!
            pcCanDevices = pDevice->pNextDevice;
            // form this point the device can no more be accesed
        }
        else
        {
            // find device before the device which must be removed
            while (pFatherDevice->pNextDevice != pDevice)
            {
                pFatherDevice = pFatherDevice->pNextDevice;
            }
            pFatherDevice->pNextDevice = pDevice->pNextDevice;
            // form this point the device can no more be accesed
        }
        _TRACE1("RemoveDevice: device removed from: %lx\n",
                (CAN_ADDRESS) pDevice);

        up(&cDeviceListSemaphore);
        WaitForProcesses(pDevice);
        kfree(pDevice);
        return 0;

    }
    else
    {
        // not in list ???
        _ERRPRINT("RemoveDevice: device not found!\n");
    }
    up(&cDeviceListSemaphore);
    return 0;

}

/*****************************************************************************/
/* GetFreeDevice                                                             */
/*****************************************************************************/
/* FUNCTIONAL DESCRIPTION                                                    */
/*   returns a pointer to a free device                                      */
/*   ATTENTION: USE COUNTER MUST BE INCREMENTED WHILE USING THIS FUNCTION!!! */
/*****************************************************************************/
/* PARAMETER DECRIPTION                                                      */
/*    PCDEVICES pcDevices: device list                                       */
/*    USIGN32 u32Minor   : minor number                                      */
/*                                                                           */
/*****************************************************************************/

PCDEVICES GetFreeDevice(USIGN32 u32Minor)
{
    PCDEVICES pHelpDevice = NULL;

    _TRACE1("GetFreeDevice: %u\n",u32Minor);
    if (down_interruptible(&cDeviceListSemaphore))
    {
        _ERRPRINT
            ("GetFreeDevice: down_interruptible(&cDeviceListSemaphore) was interrupted\n");
        return NULL;
    }
    pHelpDevice = pcCanDevices;
    while (pHelpDevice)
    {
        if (pHelpDevice->u32MinorNumber == u32Minor)
        {
            _ERRPRINT
                ("GetFreeDevice: two devices cannot be opened with one minor number %u\n",u32Minor);
            up(&cDeviceListSemaphore);
            return NULL;
        }

        pHelpDevice = pHelpDevice->pNextDevice;
    }

    pHelpDevice = pcCanDevices;

    while (pHelpDevice)
    {
        _TRACE2("pHelpDevice=%lx, bIsOpen=%x, bIsStarted=%x\n",
                (CAN_ADDRESS) pHelpDevice, pHelpDevice->bIsOpen,
                pHelpDevice->bIsStarted);
        if ((!pHelpDevice->bIsOpen)
            && (pHelpDevice->u32User == MINORUNDEFINED))
        {
            up(&cDeviceListSemaphore);
            return pHelpDevice;
        }

        pHelpDevice = pHelpDevice->pNextDevice;
    }
    up(&cDeviceListSemaphore);
    return NULL;
}

/*****************************************************************************/
/* ResetUser                                                                 */
/*****************************************************************************/
/* FUNCTIONAL DESCRIPTION                                                    */
/*   sets the u32User flag of the device which fits the minor number to      */
/*   MINORUNDEFINED                                                          */
/*****************************************************************************/
/* PARAMETER DECRIPTION                                                      */
/*    PCDEVICES pcDevices: device list                                       */
/*    USIGN32 u32Minor   : minor number                                      */
/*                                                                           */
/*****************************************************************************/

VOID ResetUser(USIGN32 u32MinorNumber)
{
    PCDEVICES pHelpDevice = NULL;

    if (down_interruptible(&cDeviceListSemaphore))
    {
        _ERRPRINT
            ("ResetUser: down_interruptible(&cDeviceListSemaphore) was interrupted\n");
        return;
    }
    pHelpDevice = pcCanDevices;
    while (pHelpDevice)
    {
      _TRACE2("pHelpDevice=%lx, bIsOpen=%x\n", (CAN_ADDRESS) pHelpDevice,
                pHelpDevice->bIsOpen);
        if (pHelpDevice->u32User == u32MinorNumber)
        {
            pHelpDevice->u32User = MINORUNDEFINED;
        }

        pHelpDevice = pHelpDevice->pNextDevice;
    }
    up(&cDeviceListSemaphore);
    return;
}

/*****************************************************************************/
/* GetDeviceFromSlot                                                         */
/*****************************************************************************/
/* FUNCTIONAL DESCRIPTION                                                    */
/*   returns a pointer to the device which matches the slot number           */
/*   ATTENTION: USE COUNTER MUST BE INCREMENTED WHILE USING THIS FUNCTION!!! */
/*****************************************************************************/
/* PARAMETER DECRIPTION                                                      */
/*    PCDEVICES pcDevices: device list                                       */
/*    USIGN32 u32Minor   : slot number                                       */
/*                                                                           */
/*****************************************************************************/

PCDEVICES GetDeviceFromSlot(PCDEVICES pcDevices, USIGN16 u32Slot)
{                               // semaphore in initialize_interface !!!!
    PCDEVICES pHelpDevice = pcDevices;

    while (pHelpDevice)
    {
      _TRACE2("pHelpDevice=%lx, bIsOpen=%x\n", (CAN_ADDRESS) pHelpDevice,
                pHelpDevice->bIsOpen);
        if (pHelpDevice->u16Socket == u32Slot)
        {
            return pHelpDevice;
        }

        pHelpDevice = pHelpDevice->pNextDevice;
    }
    return NULL;
}

/*****************************************************************************/
/* GetDeviceFromMinorNumber                                                  */
/*****************************************************************************/
/* FUNCTIONAL DESCRIPTION                                                    */
/*   returns a pointer to the device which is connected to the minor number  */
/*   ATTENTION: USE COUNTER MUST BE INCREMENTED WHILE USING THIS FUNCTION!!! */
/*****************************************************************************/
/* PARAMETER DECRIPTION                                                      */
/*    PCDEVICES pcDevices: device list                                       */
/*    USIGN32 u32Minor   : minor number                                      */
/*                                                                           */
/*****************************************************************************/

PCDEVICES GetDeviceFromMinorNumber(USIGN32 u32MinorNumber)
{
    PCDEVICES pHelpDevice;

    pHelpDevice = pcCanDevices;

    while (pHelpDevice)
    {
      _TRACE4("pHelpDevice=%lx, bIsOpen=%x\n", (CAN_ADDRESS) pHelpDevice,
                pHelpDevice->bIsOpen);
        if (pHelpDevice->u32MinorNumber == u32MinorNumber)
        {
            return pHelpDevice;
        }

        pHelpDevice = pHelpDevice->pNextDevice;
    }
    return NULL;
}

/*****************************************************************************/
/* GetDeviceFromFilepointer                                                  */
/*****************************************************************************/
/* FUNCTIONAL DESCRIPTION                                                    */
/*   returns a pointer to the device which is connected to the file pointer  */
/*   ATTENTION: USE COUNTER MUST BE INCREMENTED WHILE USING THIS FUNCTION!!! */
/*****************************************************************************/
/* PARAMETER DECRIPTION                                                      */
/*    PCDEVICES pcDevices: device list                                       */
/*    USIGN32 pFile   : file pointer                                         */
/*                                                                           */
/*****************************************************************************/

PCDEVICES GetDeviceFromFilepointer(PVOID pFile)
{
  PCDEVICES pHelpDevice;

  pHelpDevice = pcCanDevices;

  while (pHelpDevice)
  {
    _TRACE4("pHelpDevice=%lx, bIsOpen=%x\n", (CAN_ADDRESS) pHelpDevice,
            pHelpDevice->bIsOpen);
    if (pHelpDevice->pFile == pFile)
    {
      return pHelpDevice;
    }

    pHelpDevice = pHelpDevice->pNextDevice;
  }
  return NULL;
}




/*****************************************************************************/
/* can_open                                                                  */
/*****************************************************************************/
/* FUNCTIONAL DESCRIPTION                                                    */
/*   will be called while opening a device                                   */
/*****************************************************************************/
/* PARAMETER DECRIPTION                                                      */
/*    struct inode *inode : inode                                            */
/*    struct file *filp   : file descriptor                                  */
/*                                                                           */
/*****************************************************************************/

int can_open(struct inode *inode, struct file *filp)
{
    can_dev *dev;
    PCDEVICES pMyDevice;
    USIGN32 u32Minor = 0;

//     u32Minor = (MINOR(inode->i_rdev) >> 4) | (MINOR(inode->i_rdev) & 0x0F);
    u32Minor = MINOR(inode->i_rdev);
    _TRACE4("CANpro USB can_open: minor=%u\n",u32Minor);
    
   pMyDevice = GetDeviceFromMinorNumber(u32Minor);

    if (!pMyDevice)
    {
        _ERRPRINT("device not found!\n");
      //  MOD_DEC_USE_COUNT;
        return -ENODEV;
    }
    else
    {
        atomic_inc(&pMyDevice->atomInUseCount);
        if (!pMyDevice->bIsStarted)
        {
          //  MOD_DEC_USE_COUNT;
            atomic_dec(&pMyDevice->atomInUseCount);
            _ERRPRINT("A problem happend while starting the device!\n");
            return -ENOMEDIUM;
        }

        if (pMyDevice->bIsOpen)
        {
            _ERRPRINT("device is already open!\n");
         //   MOD_DEC_USE_COUNT;
            atomic_dec(&pMyDevice->atomInUseCount);
            return -EBUSY;
        }
    }

    if (!
        (filp->private_data =
         (can_dev *) kmalloc(sizeof(can_dev), GFP_KERNEL)))
    {
        filp->private_data = NULL;
      //  MOD_DEC_USE_COUNT;
        _ERRPRINT("out of memory!\n");
        atomic_dec(&pMyDevice->atomInUseCount);
        return -ENOMEM;
    }

    dev = (can_dev *) filp->private_data;

    pMyDevice->bIsOpen = TRUE;
    pMyDevice->pFile = filp;
   // pMyDevice->u32MinorNumber = u32Minor;
    //pMyDevice->u32User = u32Minor;

    sprintf(dev->name, "CAN device %02d", pMyDevice->u32MinorNumber);
    atomic_dec(&pMyDevice->atomInUseCount);
    _TRACE4("open was successful!\n");
    return 0;
}

/*****************************************************************************/
/* FreeCancardResources                                                      */
/*****************************************************************************/
/* FUNCTIONAL DESCRIPTION                                                    */
/*   frees the resources, unhooks the interrupt service routine              */
/*****************************************************************************/
/* PARAMETER DECRIPTION                                                      */
/*    PVOID pDpram: pointer to the DPRAM                                     */
/*                                                                           */
/*****************************************************************************/
INT32 FreeResources(PCDEVICES pDevice)
{

    BOOLEAN   bIsChannel1;
    
    CAN_ADDRESS Dpram = pDevice->can_dpram_card_address, Config = pDevice->can_config_space;

    pDevice->bIsStarted = FALSE;

    if (!pDevice->bIsOpen)
    {
        bIsChannel1 = pDevice->can_dpram_card_address == pDevice->can_dpram_address;
        
        RemoveDevice(pDevice->can_dpram_physaddress);
        if(bIsChannel1)
        {  // only ressources of channel1 are mapped.
           if (Config)
           {
              _TRACE2("free PCI configuration space at %lx\n",
                      (CAN_ADDRESS) Config);
              can_unmap((PVOID) Config);

           }

           if (Dpram)
           {
            _TRACE2("free DPRAM at %lx\n",
                    (CAN_ADDRESS) Dpram);
            can_unmap((PVOID) Dpram);

           }
        }
        if(pDevice->procFile)
        {
          USIGN8 name[100];
          sprintf(name, "%09u_%u",pDevice->Serialnumber,(bIsChannel1) ? 1 : 2);
          _TRACE2("removing /proc/%s\n",name);
          remove_proc_entry(name, procDir);
          
        }

    }

    return 0;

}


/*****************************************************************************/
/* can_release                                                               */
/*****************************************************************************/
/* FUNCTIONAL DESCRIPTION                                                    */
/*   will be called while closing a device                                   */
/*****************************************************************************/
/* PARAMETER DECRIPTION                                                      */
/*    struct inode *inode : inode                                            */
/*    struct file *filp   : file descriptor                                  */
/*                                                                           */
/*****************************************************************************/

int can_release(struct inode *inode, struct file *filp)
{
    PCDEVICES pMyDevice,pOtherDevice;
    USIGN32 u32Minor = 0;
    

    _TRACE4("can_release:\n");
    u32Minor = MINOR(inode->i_rdev);
//     u32Minor = (MINOR(inode->i_rdev) >> 4) | (MINOR(inode->i_rdev) & 0x0F);
    if (down_interruptible(&cDeviceListSemaphore))
    {
        _ERRPRINT
            ("can_poll: down_interruptible(&cDeviceListSemaphore) was interrupted\n");
        return ENODEV;
    }
    pMyDevice = GetDeviceFromMinorNumber(u32Minor);
    
    up(&cDeviceListSemaphore);
    

    if (!pMyDevice)
    {                           // device not found
        return -ENODEV;
    }

    pOtherDevice = pMyDevice->pOtherDevice;
    
    if (!pMyDevice->bIsStarted)
    {                           // device is not started (removed) but the resources are present. 
        kfree(filp->private_data);
      //  MOD_DEC_USE_COUNT;
        return -ENOMEDIUM;
    }
    
    INIDR_close_board(pMyDevice);
    
    if(pOtherDevice)
    {
      if(!pOtherDevice->bIsOpen)
      {
        // disable interrupt
        CanEnableIrq(FALSE, pMyDevice);
        
        // reset card
        CanResetCard(TRUE, pMyDevice);
      }
    }
    else
    {
      // reset card
      CanResetCard(TRUE, pMyDevice);
    }

    kfree(filp->private_data);

    pMyDevice->pFile = 0;
    pMyDevice->bIsOpen = FALSE;
    // decrement use counter
 //   MOD_DEC_USE_COUNT;
    return 0;
}

/*
 *  read from driver
 */
/*****************************************************************************/
/* do_can_read                                                               */
/*****************************************************************************/
/* FUNCTIONAL DESCRIPTION                                                    */
/*   read handler                                                            */
/*****************************************************************************/
/* PARAMETER DECRIPTION                                                      */
/*    struct inode *inode : inode                                            */
/*    struct file *filp   : file descriptor                                  */
/*                                                                           */
/*****************************************************************************/

ssize_t do_can_read(struct inode * inode, struct file * filp, SCHAR *buf,
            size_t count, loff_t * f_pos)
{

    int retval = 0;
    int minor_num;
    USIGN32 u32Minor;
    can_dev *dev;
    PCDEVICES pDevice;


    dev = (can_dev *) filp->private_data;

    minor_num = MINOR(inode->i_rdev);
    u32Minor = MINOR(inode->i_rdev);
//     u32Minor = (MINOR(inode->i_rdev) >> 4) | (MINOR(inode->i_rdev) & 0x0F);
    if (down_interruptible(&cDeviceListSemaphore))
    {
        _ERRPRINT
            ("can_poll: down_interruptible(&cDeviceListSemaphore) was interrupted\n");
        return ENODEV;
    }
    pDevice = GetDeviceFromFilepointer(filp);

    
    _TRACE4("do_can_read: pDevice=%lx, Minor=%u\n", (CAN_ADDRESS) pDevice,
            u32Minor);
    if (!pDevice)
    {
        _ERRPRINT("do_can_read: no CAN device found!\n");
        up(&cDeviceListSemaphore);
        return -ENODEV;
    }
    atomic_inc(&pDevice->atomInUseCount);
    up(&cDeviceListSemaphore);

    if (!pDevice->bIsStarted)
    {
      atomic_dec(&pDevice->atomInUseCount);
        return -ENOMEDIUM;
    }

    if(pDevice->l2_state != ST_CANL2_RUN)
    {
      
      atomic_dec(&pDevice->atomInUseCount);
      return -EINTR;
      
    }

    if (count <= 0)
    {
        atomic_dec(&pDevice->atomInUseCount);
        return -EFAULT;
    }

    retval = copy_to_user(buf, "1", 1);

    if (!count)
    {
        retval = -EFAULT;
        atomic_dec(&pDevice->atomInUseCount);
        return retval;
    }

    retval = 1;
    atomic_dec(&pDevice->atomInUseCount);
    return retval;
}

ssize_t can_read(struct file * filp, char *buf, size_t count, loff_t * f_pos)
{
    return do_can_read(filp->f_dentry->d_inode, filp, buf, count, f_pos);
}

int initialize_board(CANPC_RESOURCES * pcrResources, PCDEVICES pDevice)
{
    PCDEVICES pLocalDevice = pDevice;

    if (down_interruptible(&cDeviceListSemaphore))
    {
        _ERRPRINT
            ("initialize_board: down_interruptible(&cDeviceListSemaphore) was interrupted\n");
        return INIPC_IB_ERR_VC_IOCANCELLED;
    }

        _TRACE5("initialize_board: socket auto\n");
        pcrResources->uSocket = pLocalDevice->u16Socket;
        pcrResources->uInterrupt = pLocalDevice->can_irq;
        pcrResources->ulDPRMemBase = pLocalDevice->can_dpram_physaddress;
        pcrResources->ulDPRMemSize = 0x1000;
        pcrResources->uChip =
            ((pLocalDevice->u16DeviceId & 0xFF) <=
             2) ? NEC005_CHIP : SJA1000_CHIP;

    up(&cDeviceListSemaphore);
    return 0;
}

/*****************************************************************************/
/* can_ioctl                                                                 */
/*****************************************************************************/
/* FUNCTIONAL DESCRIPTION                                                    */
/*   ioctl dispatcher                                                        */
/*****************************************************************************/
/* PARAMETER DECRIPTION                                                      */
/*    struct inode *inode : inode                                            */
/*    struct file *filp   : file descriptor                                  */
/*    unsigned int cmd    : ioctl code                                       */
/*    unsigned int arg   : pointer to the parameter                         */
/*                                                                           */
/*****************************************************************************/

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,32)
long can_ioctl(
  struct file *filp,
  unsigned int cmd, 
  unsigned long arg)
#else
int can_ioctl(
  struct inode *inode,
  struct file *filp,
  unsigned int cmd, 
  unsigned long arg)
#endif
{

    int err = 0;
    int ret = 0;
    USIGN32 pram[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
    USIGN32 lpram = 0;
    PARAM_STRUCT ac_struct;
    USIGN32 u32Minor;
    PCDEVICES pDevice;
    int dummy;

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,32) 
    struct inode *inode = filp->f_path.dentry->d_inode;

    _TRACE4("IOCTRL:\n");
   
//     u32Minor = (MINOR(inode->i_rdev) >> 4) | (MINOR(inode->i_rdev) & 0x0F);
    u32Minor = MINOR(inode->i_rdev);
#else
    _TRACE4("IOCTRL:\n");
    u32Minor = (MINOR(inode->i_rdev) >> 4) | (MINOR(inode->i_rdev) & 0x0F);
#endif
    
    if (down_interruptible(&cDeviceListSemaphore))
    {
        _ERRPRINT
            ("can_poll: down_interruptible(&cDeviceListSemaphore) was interrupted\n");
        return ENODEV;
    }

    pDevice = GetDeviceFromFilepointer(filp);
    
    
    //_TRACE2("can_ioctl: pDevice=%lx, Minor=%u\n",(USIGN32)pDevice,u32Minor);
    if (!pDevice)
    {
      up(&cDeviceListSemaphore);
        return -ENODEV;
    }
    atomic_inc(&pDevice->atomInUseCount);
    up(&cDeviceListSemaphore);
    
    
    
    _TRACE4("can_ioctl: cmd:0x%x, name:%s file:0x%lx\n", cmd, pDevice->au8IrqName, (unsigned long)filp);
    //_TRACE2("dpram=%lx,config=%lx,device=%lx\n",pDevice->can_dpram_address,pDevice->can_config_space,pDevice);

    if (!pDevice->bIsStarted)
    {
      atomic_dec(&pDevice->atomInUseCount);
        return -ENOMEDIUM;
    }


    /*
     * extract the type and number bitfields, and don't decode
     * wrong cmds: return ENOTTY (inappropriate ioctl) before access_ok()
     */
    if (_IOC_TYPE(cmd) != CAN_IOC_MAGIC)
    {
      atomic_dec(&pDevice->atomInUseCount);
        return -ENOTTY;
    }
    if (_IOC_NR(cmd) > CAN_IOC_MAXNR)
    {
      atomic_dec(&pDevice->atomInUseCount);
        return -ENOTTY;
    }


    /*
     * the direction is a bitmask, and VERIFY_WRITE catches R/W
     * transfers. `Type' is user-oriented, while
     * access_ok is kernel-oriented, so the concept of "read" and
     * "write" is reversed
     */
    if (_IOC_DIR(cmd) & _IOC_READ)
        err = !access_ok(VERIFY_WRITE, (VOID *) arg, _IOC_SIZE(cmd));
    else if (_IOC_DIR(cmd) & _IOC_WRITE)
        err = !access_ok(VERIFY_READ, (VOID *) arg, _IOC_SIZE(cmd));
    if (err)
    {
      atomic_dec(&pDevice->atomInUseCount);
        return -EFAULT;
    }

    

    if (down_interruptible(&pDevice->cDpramSemaphore))
    {
        _ERRPRINT
        ("can_ioctl: down_interruptible() was interrupted\n");
        atomic_dec(&pDevice->atomInUseCount);
        return (0x10000 - INIPC_IB_ERR_VC_IOCANCELLED);
    }
    switch (cmd)
    {

    case CAN_IOC_CANL2_reset_board:
        _TRACE4("CAN_IOC_CANL2_reset_board:\n");
        
        ret = CANDR_reset_board(pDevice);
        break;

      case CAN_IOC_CANL2_reset_chip:
        _TRACE4("CAN_IOC_CANL2_reset_chip:\n");

        ret = CANDR_reset_chip(pDevice);
        pDevice->can_info.irq_flag = 2;
        pDevice->l2_state = ST_CANL2_INIT;
        wake_up_interruptible(&pDevice->can_info.read_queue);          
        

        break;

    case (CAN_IOC_CANL2_initialize_chip):
       _TRACE4("CAN_IOC_CANL2_initialize_chip:\n");
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        ret = CANDR_initialize_chip((int) pram[0],
                                    (int) pram[1],
                                    (int) pram[2],
                                    (int) pram[3], (int) pram[4], pDevice);
        break;

  

    case CAN_IOC_CANL2_set_mode:
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        ret = CANDR_set_mode((int) pram[0], (int) pram[1], pDevice);

        break;


    case CAN_IOC_CANL2_set_acceptance:
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        ret = CANDR_set_acceptance((USIGN32) pram[0],
                                   (USIGN32) pram[1],
                                   (USIGN32) pram[2],
                                   (USIGN32) pram[3], pDevice);
        break;



    case CAN_IOC_CANL2_set_output_control:
        ret = CANDR_set_output_control((int) arg, pDevice);

        break;


    case CAN_IOC_CANL2_initialize_interface:
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        ret = CANDR_initialize_interface((int) pram[0],
                                         (int) pram[1],
                                         (int) pram[2],
                                         (int) pram[3],
                                         (int) pram[4],
                                         (int) pram[5],
                                         (int) pram[6],
                                         (int) pram[7],
                                         (int) pram[8], (int) pram[9],
                                         pDevice);

        break;

    case CAN_IOC_CANL2_define_object:
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        ret = CANDR_define_object((USIGN32) pram[0],
                                  (int *) &pram[1],
                                  (int) pram[2],
                                  (int) pram[3],
                                  (int) pram[4], (int) pram[5], pDevice);
        dummy = copy_to_user((long *) arg, (long *) pram, sizeof(pram));

        break;


    case CAN_IOC_CANL2_define_cyclic:
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        ret = CANDR_define_cyclic((int) pram[0],
                                  (USIGN32) pram[1],
                                  (USIGN32) pram[2], pDevice);
        break;



    case CAN_IOC_CANL2_enable_fifo:
        ret = CANDR_enable_fifo(pDevice);
        break;



    case CAN_IOC_CANL2_enable_dyn_obj_buf:
        ret = CANDR_enable_dyn_obj_buf(pDevice);
        break;

    case CAN_IOC_CANL2_enable_fifo_transmit_ack:
        ret = CANDR_enable_fifo_transmit_ack(pDevice);
        break;


    case CAN_IOC_CANL2_get_version:
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        ret = CANDR_get_version((int *) &pram[0],
                                (int *) &pram[1],
                                (int *) &pram[2],
                                (int *) &pram[3], (int *) &pram[4], pDevice);
        dummy = copy_to_user((long *) arg, (long *) pram, sizeof(pram));

        break;
  

      case CAN_IOC_CANL2_get_serial_number:
        ret = CANDR_get_serial_number(&lpram, pDevice);
        __put_user(lpram, (USIGN32 *) arg);

        break;
        
      case CAN_IOC_CANL2_get_phys_channel:
        ret = 0;
        lpram = pDevice->u8PhysCh;
        __put_user(lpram, (USIGN32 *) arg);
        break;
        
      case CAN_IOC_CANL2_get_device_type:
        ret = 0; 
        lpram = pDevice->u16DeviceId;
        __put_user(lpram, (USIGN32 *) arg);
        break;

      case CAN_IOC_CANL2_start_chip:
        pDevice->can_info.irq_flag = 0;
        
        ret = CANDR_start_chip(pDevice);
        
        if(!ret) pDevice->l2_state = ST_CANL2_RUN;
        
        break;

  
    case CAN_IOC_CANL2_send_remote_object:
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        ret = CANDR_send_remote_object((int) pram[0], (int) pram[1], pDevice);

        break;


    case CAN_IOC_CANL2_supply_object_data:
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        ret = CANDR_supply_object_data((int) pram[0],
                                       (int) pram[1],
                                       (USIGN8 *) &pram[2], pDevice);

        dummy = copy_to_user((long *) arg, (long *) pram, sizeof(pram));

        break;


    case CAN_IOC_CANL2_supply_rcv_object_data:
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        ret = CANDR_supply_rcv_object_data((int) pram[0],
                                           (int) pram[1],
                                           (USIGN8 *) &pram[2],
                                           pDevice);

        dummy = copy_to_user((long *) arg, (long *) pram, sizeof(pram));

        break;



    case CAN_IOC_CANL2_send_object:
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        ret = CANDR_send_object((int) pram[0], (int) pram[1], pDevice);

        break;


    case CAN_IOC_CANL2_write_object:
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        ret = CANDR_write_object((int) pram[0],
                                 (int) pram[1], (USIGN8 *) &pram[2],
                                 pDevice);

        dummy = copy_to_user((long *) arg, (long *) pram, sizeof(pram));

        break;


    case CAN_IOC_CANL2_read_rcv_data:
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        ret = CANDR_read_rcv_data((int) pram[0],
                                  (USIGN8 *) &pram[1],
                                  (USIGN32 *) &pram[3], pDevice);

        dummy = copy_to_user((long *) arg, (long *) pram, sizeof(pram));

        break;


    case CAN_IOC_CANL2_read_xmt_data:
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        ret = CANDR_read_xmt_data((int) pram[0],
                                  (int *) &pram[1],
                                  (USIGN8 *) &pram[2], pDevice);

        dummy = copy_to_user((long *) arg, (long *) pram, sizeof(pram));

        break;


    case CAN_IOC_CANL2_read_ac:
        dummy = copy_from_user((PARAM_STRUCT *) & ac_struct, (PARAM_STRUCT *) arg,
                       sizeof(ac_struct));
        ret = CANDR_read_ac((PARAM_STRUCT *) & ac_struct, pDevice);
        _TRACE4("can_ioctl: CAN_IOC_CANL2_read_ac ret=%x\n", ret);

        dummy = copy_to_user((PARAM_STRUCT *) arg, (PARAM_STRUCT *) & ac_struct,
                     sizeof(PARAM_STRUCT));

        break;

    case CAN_IOC_CANL2_send_data:
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        ret = CANDR_send_data((USIGN32) pram[0],
                              (int) pram[1],
                              (int) pram[2], (USIGN8 *) &pram[3],
                              pDevice);
        _TRACE4("can_ioctl: CAN_IOC_CANL2_send_data ret=%u\n", ret);

        dummy = copy_to_user((long *) arg, (long *) pram, sizeof(pram));

        break;



    case CAN_IOC_CANL2_send_remote:
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        ret =
            CANDR_send_remote((int) pram[0], (int) pram[1], (int) pram[2],
                              pDevice);

        break;




    case CAN_IOC_CANL2_reinitialize:
      
      pDevice->l2_state = ST_CANL2_INIT;
      pDevice->can_info.irq_flag = 2;
      wake_up_interruptible(&pDevice->can_info.read_queue);          

        ret = CANDR_reinitialize(pDevice);

        break;

    case CAN_IOC_CANL2_get_time:


        ret = CANDR_get_time((USIGN32 *) &lpram, pDevice);
        if (ret)
        {
            _ERRPRINT("CANDR_get_time failed: status=%x\n", ret);
        }
        __put_user((USIGN32) lpram, (USIGN32 *) arg);

        break;

   case CAN_IOC_CANL2_get_bus_state:
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        ret = CANDR_get_bus_state(pDevice);

        break;


    case CAN_IOC_CANL2_reset_rcv_fifo:
        ret = CANDR_reset_rcv_fifo(pDevice);

        break;

    case CAN_IOC_CANL2_reset_xmt_fifo:
        ret = CANDR_reset_xmt_fifo(pDevice);

        break;

    case CAN_IOC_CANL2_reset_lost_msg_counter:
        ret = CANDR_reset_lost_msg_counter(pDevice);

        break;

    case CAN_IOC_CANL2_read_rcv_fifo_level:
        ret = CANDR_read_rcv_fifo_level(pDevice);

        break;

    case CAN_IOC_CANL2_read_xmt_fifo_level:
        ret = CANDR_read_xmt_fifo_level(pDevice);

        break;

    case CAN_IOC_INIL2_initialize_board:
        {
            CANPC_RESOURCES crResources;
            _TRACE4("CAN_IOC_initialize_board:\n");
            dummy = copy_from_user((long *) &crResources, (long *) arg,
                           sizeof(CANPC_RESOURCES)); 
            up(&pDevice->cDpramSemaphore);      // no DPRAM access
            ret = initialize_board(&crResources, pDevice);
            dummy = copy_to_user((CANPC_RESOURCES *) arg,
                         (CANPC_RESOURCES *) & crResources,
                         sizeof(CANPC_RESOURCES));
       

            _TRACE4("CAN_IOC_initialize_board: socket=%u\n",
                    crResources.uSocket);
                    
            atomic_dec(&pDevice->atomInUseCount);
            return (0x10000 - ret);
        }
        break;


    case CAN_IOC_INIL2_close_board:
        ret = INIDR_close_board(pDevice);
        
        pDevice->l2_state = ST_CANL2_INIT;
        pDevice->can_info.irq_flag = 2;
        wake_up_interruptible(&pDevice->can_info.read_queue);          
        
        break;

    case CAN_IOC_CANL2_enable_error_frame_detection:
        ret = CANDR_enable_error_frame_detection(pDevice);

        break;

    case CAN_IOC_CANL2_get_device_id:
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        ret = CANDR_get_device_id((USIGN32 *) &pram[0], pDevice);
        dummy = copy_to_user((long *) arg, (long *) pram, sizeof(pram));
        break;

    case CAN_IOC_CANL2_init_signals:
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        ret = CANDR_init_signals((USIGN32) pram[0],
                                 (USIGN32) pram[1], pDevice);

        break;
  

    case CAN_IOC_CANL2_read_signals:
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        ret = CANDR_read_signals((USIGN32 *) &pram[0], pDevice);
        dummy = copy_to_user((long *) arg, (long *) pram, sizeof(pram));
        break;

    case CAN_IOC_CANL2_write_signals:
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        ret = CANDR_write_signals((USIGN32) pram[0],
                                  (USIGN32) pram[1], pDevice);



/*
    case CAN_IOCHARDRESET:
        // reset mod count to 1, to allow unloading in case
        // of problems
         
        while (MOD_IN_USE)
            MOD_DEC_USE_COUNT;
        MOD_INC_USE_COUNT;

        break;
*/
    case CAN_IOC_CANL2_trace_level:
        dummy = copy_from_user((long *) pram, (long *) arg, sizeof(pram));
        can_trace_level = (USIGN32) pram[0];
        _TRACE("new trace level = 0x%x\n", can_trace_level);
        _TRACE("_PRINT_DISABLE    0\n");
        _TRACE("_PRINT_ERROR      1\n");
        _TRACE("_PRINT_TRACE      2\n");
        _TRACE("_PRINT_MSG        4\n");
        _TRACE("_PRINT_INT        8\n");
        _TRACE("_PRINT_IO      0x10\n");
        _TRACE("_PRINT_API     0x20\n");


        _ERRPRINT("_PRINT_ERROR   active\n");
        _TRACE1("_PRINT_TRACE   active\n");
        _TRACE2("_PRINT_MSG     active\n");
        _TRACE3("_PRINT_INT     active\n");
        _TRACE4("_PRINT_IO      active\n");
        _TRACE5("_PRINT_API     active\n");


        break;
  

    default:                   /* redundant, as cmd was checked against MAXNR */
        _ERRPRINT("can_ioctl: wrong ioctrl code = %x\n", cmd);
        up(&pDevice->cDpramSemaphore);
        atomic_dec(&pDevice->atomInUseCount);
        return -ENOTTY;
        break;
    }

    up(&pDevice->cDpramSemaphore);

    /* hack to tunnel negative return codes from api to user space */
    if (ret < 0)
    {
        atomic_dec(&pDevice->atomInUseCount);
        
        return (0x10000 - ret);
    }
    atomic_dec(&pDevice->atomInUseCount);
    return ret;

}

/*****************************************************************************/
/* can_intCanDriver                                                          */
/*****************************************************************************/
/* FUNCTIONAL DESCRIPTION                                                    */
/*   loads the firmware to the card                                          */
/*****************************************************************************/
/* PARAMETER DECRIPTION                                                      */
/*    PCDEVICES           : pDevice                                          */
/*                                                                           */
/*****************************************************************************/

int can_intCanDriver(PCDEVICES pDevice)
{
  int result;
  USIGN32 SerialNumber;
  int fwV,swV,hwV,lic,chip[2];

//   if (INIDR_initialize_board(pDevice) == OK_)
//   {
      
    if ((result = CANDR_reset_board((PVOID) pDevice)) != 0)
    {
      _ERRPRINT(KERN_ERR "CANDR_reset_board failed\n");
      return result;
    }
    if ((result = CANDR_reset_chip((PVOID) pDevice)) != 0)
    {
      _ERRPRINT(KERN_ERR "CANDR_reset_chip failed\n");
      return result;
    }
        
    if ((result = CANDR_get_serial_number_fw(&SerialNumber, (PVOID) pDevice)) != 0)
    {
      _ERRPRINT(KERN_ERR "CANDR_get_serial_number failed\n");
      return result;
    }
        
    pDevice->Serialnumber = SerialNumber;
       
    if ((result = CANDR_get_version_fw(&swV, &fwV, &hwV, &lic, chip, (PVOID) pDevice)) != 0)
    {
      _ERRPRINT(KERN_ERR "CANDR_get_version failed\n");
      return result;
    }
  
    if(chip[1] == 0)
    {
      pDevice->u8NumOfCh = 1;
    }
    else
    {
      pDevice->u8NumOfCh = 2;
    }
    
    pDevice->hw_version = hwV;
    pDevice->fw_version = fwV;
    pDevice->license    = lic;
    pDevice->chip_type  = *chip;
        
 
    INIDR_close_board(pDevice);

    return 0;
//   }
// 
//   _ERRPRINT(KERN_ERR "INIDR_initialize_board failed\n");
//   return -1;
}




/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
