/*
 * Can-Driver for Softing Can-Card
 * Copyright (C) 1994-2003, Softing AG, Gunnar Larisch, Werner Knorsch
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * email: support.automation@softing.com
 * Softing AG, Richard-Reitzner-Allee 6, 85540 Haar, Germany
 */

/**************************************************************/
/*                                                            */
/* Filename  : SBIN.H                                         */
/*                                                            */
/* Funktion  : Stellt die Routinen zum Schreiben und Lesen    */
/*             von SBN-Records in/aus ein/em File zur         */
/*             Verfuegung.                                    */
/*                                                            */
/* Autor     : Thomas Lang                                    */
/* Datum     : 27.06.94                                       */
/**************************************************************/

/*
 * $Id: sbin.h,v 1.2 2003/10/31 15:16:38 kno Exp $
 */

/* Structured Binary Recordtypen */
#define _TypeSBINString_    0xffff
#define _TypeSBINData_      0x0000
#define _TypeSBINEndOfFile_ 0x0001
#define _TypeSBINStart_     0x0003

static const char *SBINKennung = "Structured Binary Format, Softing GmbH";

static union SBinHeaderUnion
{
    struct
    {
        unsigned short Type;
        USIGN32 Addr;
        unsigned short Len;
    }
    Struct;

    unsigned char Raw[2 * sizeof(unsigned short) + sizeof(USIGN32)];
};

#ifdef READRECORDS
/**************************************************************/
/*                                                            */
/* Name      : ReadSBINRecord                                 */
/*                                                            */
/* Funktion  : Lesen eines SBINData-Records aus dem Binaerfile*/
/*             incl. Ermitteln der 16-Bit-Additions-Checksumme*/
/*                                                            */
/* Parameter : BinFile: geoeffnetes OutputFile                */
/*             CalcChecksum: wenn = 1 wird die Checksumme     */
/*                                    geprueft.               */
/*             MaxDataLen: maximal zulaessige Recordlaenge    */
/*                         (entspricht Groesse von Data-Buffer*/
/*                          - 2 Byte (Checksumme), d.h.       */
/*                         Data-Buffer muss mindestens 2 Byte */
/*                         groesser sein als die Datenlaenge  */
/*                         des laengsten zulaessigen Records!)*/
/*                                                            */
/*             Type: Recordtype (Rueckgabe)                   */
/*             Data: Datenbytes (Rueckgabe)                   */
/*             Addr: Adresse (absolut, 32Bit!) (Rueckgabe)    */
/*             Len:  Anzahl der Datenbytes (Rueckgabe)        */
/*             DataChecksum: 16-Bit-Additionschecksumme ueber */
/*                           die Datenbytes.                  */
/*                                                            */
/* Returncode: 0 : wenn Lesevorgang erfolgreich               */
/*             -1: bei Lesefehler                             */
/*             -2: Record zu lang                             */
/*             -3: bei Checksummenfehler                      */
/*                                                            */
/* Autor     : Thomas Lang                                    */
/*                                                            */
/* Datum     : 21.06.94                                       */
/**************************************************************/
static int
ReadSBINRecord(FILE * BinFile,
               int CalcChecksum, unsigned int MaxDataLen,
               unsigned int *Type, unsigned char *Data,
               USIGN32 *Addr, unsigned int *Len,
               unsigned int *DataCheckSum)
{
    unsigned short i;
    unsigned short CheckSum, HeaderCheckSum;
    union SBinHeaderUnion Header;
    unsigned short ByteCount;

    /* Header des Records auslesen */
    if (1 != fread(&Header, sizeof(Header), 1, BinFile))
        return (-1);

    /* Record zu gross? */
    if (Header.Struct.Len > MaxDataLen)
        return (-2);

    /* Daten incl. Checksumme auslesen */
    ByteCount =
        fread(Data, 1, Header.Struct.Len + sizeof(unsigned short), BinFile);
    if (ByteCount != Header.Struct.Len + sizeof(unsigned short))
        return (-1);

    if (CalcChecksum)
    {                           /* CheckSumme fuer Header ermitteln */
        for (i = 0, HeaderCheckSum = 0; i < sizeof(Header.Raw);
             HeaderCheckSum += Header.Raw[i], i++);

        /* CheckSumme ueber Daten berechnen */
        for (i = 0, CheckSum = 0; i < Header.Struct.Len;
             CheckSum += Data[i], i++);

        /* Stimmt Checksumme mit in File abgelegter Summe ueberein? */
        if (((HeaderCheckSum + CheckSum) & 0xFFFF) !=
            *(unsigned short *) (Data + Header.Struct.Len))
            return (-3);

        *DataCheckSum = CheckSum;
    }

    /* Headerinformation ausgeben */
    *Type = Header.Struct.Type;
    *Addr = Header.Struct.Addr;
    *Len = Header.Struct.Len;

    return (0);
}
#endif /* READRECORDS */

#ifdef WRITERECORDS
/**************************************************************/
/*                                                            */
/* Name      : WriteSBINRecord                                */
/*                                                            */
/* Funktion  : Schreiben eines SBINData-Records in das BinFile*/
/*             incl. Ermitteln der 16-Bit-Additions-Checksumme*/
/*                                                            */
/* Parameter : BinFile: geoeffnetes OutputFile                */
/*             Type: Recordtype                               */
/*             Data: Datenbytes                               */
/*             Addr: Adresse (absolut, 32Bit!)                */
/*             Len:  Anzahl der Datenbytes                    */
/*                                                            */
/* Returncode: 0 wenn Schreibvorgang erfolgreich              */
/*             -1 bei Fehler.                                 */
/*                                                            */
/* Autor     : Thomas Lang                                    */
/*                                                            */
/* Datum     : 21.06.94                                       */
/**************************************************************/
static int
WriteSBINRecord(FILE * BinFile,
                unsigned int Type, const unsigned char *Data,
                USIGN32 Addr, unsigned int Len)
{
    unsigned int i;
    unsigned int CheckSum;
    union SBinHeaderUnion Header;

    /* Header eintragen in Struktur */
    Header.Struct.Type = Type;
    Header.Struct.Addr = Addr;
    Header.Struct.Len = Len;

    /* CheckSumme fuer Header ermitteln */
    for (i = 0, CheckSum = 0; i < sizeof(Header.Raw);
         CheckSum += Header.Raw[i], i++);

    /* CheckSumme ueber Daten weiterrechnen */
    for (i = 0; i < Len; CheckSum += Data[i], i++);

    /* Header in Datei schreiben */
    if (1 != fwrite(&Header, sizeof(Header), 1, BinFile))
    {
        return (-1);
    }

#if 0
    printf("Data-Record: Addr:%06lx Len:%03x Data: %02x %02x %02x ... %02x %02x %02x\n",
         Addr, Len, Data[0], Data[1], Data[2], Data[Len - 3], Data[Len - 2],
         Data[Len - 1]);
#endif

    /* Daten in Datei schreiben */
    if (Len > 0)
    {
        if (1 != fwrite(Data, Len, 1, BinFile))
        {
            return (-1);
        }
    }

    /* CheckSumme in Datei schreiben */
    if (1 != fwrite(&CheckSum, sizeof(CheckSum), 1, BinFile))
    {
        return (-1);
    }

    return (0);
}

#endif /* WRITERECORDS */
