/*****************************************************************************/
/*                                                                           */
/*   Softing Industrial Automation GmbH   Richard-Reitzner-Allee 6   85540 Haar         */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*  Copyright (c) Softing Industrial Automation GmbH, All rights reserved.   */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*                        I N T E R R U P T . C                              */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* PROJECT:            CANopen Client API C-Samples                          */
/*                                                                           */
/* MODULE:             Interrupt.c                                           */
/*                                                                           */
/* AUTHOR:             Softing Industrial Automation GmbH                               */
/*                                                                           */
/* DATE                01.04.2012                                           */
/*                                                                           */
/*****************************************************************************/
/* MODULE_DESCRIPTION                                                        */ 
/*                                                                           */
/*  GENERAL        This is a demo application to show how to use             */
/*                 the CANopen Client API in interrupt mode.                 */
/*                                                                           */
/*  CONTENT        The CANopen Client API will be configured to run in       */
/*                 interrupt mode. The demo program initiates actions which  */
/*                 result in the generation of interrupt events.             */
/*                 These events will be evaluated in an extra interrupt      */
/*                 thread.                                                   */
/*                                                                           */
/*  DEMO STRUCTURE The demo is maintaned by a state machine. The global      */
/*                 variable <ucDemoState> contains the current state.        */
/*                 Following a short description of the state machine        */
/*                 especially which action is performed in each state:       */
/*                                                                           */
/*                   state 1.0:  Initialising the Softing CAN interface to   */
/*                               run CANopen Client API                      */
/*                                 - success -> state 1.1                    */
/*                                 - failure -> state 5.2                    */
/*                   state 1.1:  Initialising of interrupt mode              */
/*                                 - success -> state 2.0                    */
/*                                 - failure -> state 4.1                    */
/*                                                                           */
/*                   states 2.x: Initialising the CANopen Master             */
/*                   state 2.1:  Initialising CAN parameters                 */
/*                                 - success -> state 2.2                    */
/*                                 - failure -> state 6.0                    */
/*                   state 2.2:  Initialising the Node manager               */
/*                                 - success -> state 3.0                    */
/*                                 - failure -> state 6.0                    */
/*                                                                           */
/*                   state 3.0:  Initiate Startup of CANopen Master          */
/*                                 - success -> state 3.1                    */
/*                                 - failure -> state 5.0                    */
/*                   state 3.1:  Check for the completion of the Startup.    */
/*                                 - success -> state 3.2                    */
/*                                 - failure -> state 5.0                    */
/*                   state 3.2:  Check if CANopen Master could send its      */
/*                               bootup message.                             */
/*                                 - success -> state 4.0                    */
/*                                 - failure -> state 3.3                    */
/*                   state 3.3:  Dummy state                                 */
/*                                 - next    -> state 4.0                    */
/*                                                                           */
/*                   state 4.0:  Initiate Shutdown of CANopen Master         */
/*                                 - success -> state 4.1                    */
/*                                 - failure -> state 5.0                    */
/*                   state 4.1:  Check for the completion of the Shutdown.   */
/*                                 - success -> state 5.0                    */
/*                                 - failure -> state 4.2                    */
/*                   state 4.2:  Dummy state                                 */
/*                                 - next    -> state 5.0                    */
/*                                                                           */
/*                   state 5.0:  Release resources for interrupt mode        */
/*                                 - success -> state 5.1                    */
/*                                 - failure -> state 5.1                    */
/*                                                                           */
/*                   state 5.1:  Free CAN interface ressources               */
/*                                 - success -> state 5.2                    */
/*                                 - failure -> state 5.2                    */
/*                                                                           */
/*                   state 4.2:  End of program                              */
/*                                                                           */
/*                 Each action can be configured by setting constants.       */
/*                 Please refer to the >Constants to be changed by the user< */
/*                 section below.                                            */
/*                                                                           */
/* SPECIAL                                                                   */
/* CONSIDERATIONS  If Your application calls API functions in the main       */
/*                 thread and the interrupt thread please consider to use    */
/*                 functions of the Win32 API Critical-Section mechanism.    */
/*                 For a detailed description of the difficulty please refer */
/*                 to the CANopen Client API manual.                         */
/*                 The mechanism is used in the demo i.e. at state 2.0 and   */
/*                 2.1 which belongs to the Startup of the master.           */
/*                                                                           */
/*                 In this demo application, the WIN32 API function          */
/*                 CreateThread is used for compatibility reason to create   */
/*                 the interrupt thread. If the thread uses Microsoft        */
/*                 C run-time library functions it is recommended to use     */
/*                 the macro _BEGINTHREADEX. The macro expands to the        */
/*                 function _beginthreadex which offers advantages over the  */
/*                 function CreateThread(). For further information please   */
/*                 see the Microsoft MSDN library online help                */
/*                                                                           */
/*****************************************************************************/   
#include <stdio.h>
#include <sys/time.h>
#include <string.h>
#include <pthread.h>

// CANopen Master API include files 
#include "BaseTypes.h"
#include "kbhit.h"
#include "cmadef.h"
#include "cmaret.h"

/*****************************************************************************/
/*                                                                           */
/* Constants to be changed by the user                                       */
/*                                                                           */
/*****************************************************************************/
// constants used in state 2.1:
                              // baud rate, possible values are:
                              //  - 0: 1   Mbit/s
                              //  - 1: 800 Kbit/s
                              //  - 2: 500 Kbit/s
                              //  - 3: 250 Kbit/s
                              //  - 4: 125 Kbit/s
                              //  - 5: 50  Kbit/s
                              //  - 6: 20  Kbit/s
                              //  - 7: 10  Kbit/s
#define BAUD_RATE             1

// constants used in state 2.2:
                              // node-ID of CANopen Master
#define MASTER_NODE_ID        127

// constants used in state 3.1:
                             // timeout in ms for the master to startup
#define STARTUP_TIME          500

// constants used in state 3.2:
                             // time we wait after startup if the 
                             // CAN controller changes it's bus state,
                             // indication that bootup message could
                             // be sent
#define BOOT_MSG_TIME         10

// constants used in state 4.1
                             // timeout in ms for the master to shutdown
#define SHUTDOWN_TIME         100

/*****************************************************************************/
/*                                                                           */
/* Constants which should not be changed                                     */
/*                                                                           */
/*****************************************************************************/
                             // CAN1 runs in CANopen Master mode
#define CAN1_FUNCTIONALITY     1
                             // CAN2 is not used
#define CAN2_FUNCTIONALITY     0
                             // license to check by the master
                             // (CANopen Master mode)
#define LIC                    0

                             // index values of the event handles array
                             // in the ThreadDataStruct structure
#define INT_EVENT 0
#define END_EVENT 1

#define FALSE 0
#define TRUE !FALSE



/*****************************************************************************/
/*                                                                           */
/* Global structures and variables                                           */
/*                                                                           */
/*****************************************************************************/
// constants used in state 2.1:
                              // baud rate, possible values are:
                              //  - 0: 1   Mbit/s
                              //  - 1: 800 Kbit/s
                              //  - 2: 500 Kbit/s
                              //  - 3: 250 Kbit/s
                              //  - 4: 125 Kbit/s
                              //  - 5: 50  Kbit/s
                              //  - 6: 20  Kbit/s
                              //  - 7: 10  Kbit/s
#define BAUD_RATE             0

// constants used in state 2.2:
                              // node-ID of CANopen Master
#define MASTER_NODE_ID        127

// constants used in state 3.1:
                             // timeout in ms for the master to startup
#define STARTUP_TIME          500

// constants used in state 3.2:
                             // time we wait after startup if the 
                             // CAN controller changes it's bus state,
                             // indication that bootup message could
                             // be sent
#define BOOT_MSG_TIME         10

// constants used in state 4.1
                             // timeout in ms for the master to shutdown
#define SHUTDOWN_TIME         100

pthread_t interrupt_pthread;


                             // data structure used by Win32 API Critical-Section
                             // mechanism

pthread_mutex_t g_CriticalSection = PTHREAD_MUTEX_INITIALIZER;

                             // state of the demo program
static unsigned char    ucDemoState;
                             // channel handles
static unsigned char    ucCANHandle;
static unsigned char    ucCANHandle2;
// start time for timeout timer
static unsigned long    ulTimerStartTime;



/*****************************************************************************/
/*                                                                           */
/* Function prototypes                                                       */
/*                                                                           */
/*****************************************************************************/
short ReadAndCheckEvent(void);
void  PrintUnexpectedEvent(CMA_EventStruct stEvent);
short EvaluateEventInState_31(CMA_EventStruct stEvent);
short EvaluateEventInState_32(CMA_EventStruct stEvent);
short EvaluateEventInState_41(CMA_EventStruct stEvent);
void  PrintCMAFunctReturnError(int ret);
void  StartTimeoutTimer (void);
short CheckTimeoutTimer (unsigned long ulTimeoutInMs);
int   InitIRQEvents (unsigned char *can_handle);
void  ExitIRQEvents (void);
int InterruptThread(void *context);

unsigned long GetTickCount(void)
{
  unsigned long count;
  struct timeval tv;
  struct timezone tz;
  int ret;
  
  ret =  gettimeofday(&tv, &tz);
  
  if(ret) return 0;
  
  count = tv.tv_usec/1000 + tv.tv_sec*1000;
  
  return count;
}

int EnterCriticalSection(pthread_mutex_t *pMutex)
{
  int ret;
  ret = pthread_mutex_lock(pMutex);
  if(ret)
  {
    printf("pthread_mutex_lock returned 0x%x\n",ret);
  }
  return ret;
}

int LeaveCriticalSection(pthread_mutex_t *pMutex)
{
  int ret;
  ret = pthread_mutex_unlock(pMutex);
  if(ret)
  {
    printf("pthread_mutex_unlock returned 0x%x\n",ret);
  }
  return ret;
}

/*****************************************************************************/
/*                                                                           */
/* Functions                                                                 */
/*                                                                           */
/*****************************************************************************/
/* ------------------------------------------------------------------------- */
/*                                                                           */
/* Main-Function                                                             */
/*                                                                           */
/* ------------------------------------------------------------------------- */
int main(int argc, char **argv)
{
  // function return codes
  short           ret;

  // structures needed for function call in:
  // state 2.1
  CMA_CAN_CONFIG    stCANConfig;

  printf("***************************************************************\n");
  printf("*           Softing Industrial Automation GmbH  2012          *\n");
  printf("*                                                             *\n");
  printf("*                        Interrupt                            *\n");
  printf("*            (Demo Program for CANopen Client API)            *\n");
  printf("*                                                             *\n");
  printf("***************************************************************\n");

  // initialise state machine of demo program
  ucDemoState = 10;


  printf("\n");

  // State 1.0) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 1.0 Initialising the Softing CAN interface to run CANopen Client API    */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 10)
  {

    printf("1.0) Initialising the Softing CAN channel \"CANopen\" to run\n");
    printf("     CANopen Client API.\n"); 
    printf("     This will be done by using the function CMA_InitializeCard().\n");
    printf("     Call of CMA_InitializeCard() function.\n");

    ret = CMA_InitializeCard (0xFFFFFFFF, 
      CAN1_FUNCTIONALITY,
      &ucCANHandle,
      LIC
      ,(unsigned char*)"CANopen");
    
    if (ret!=0)
    { 
      // error which leads to an abort of the demo program,
      // next state is 6.1) 
      ucDemoState = 52;
      
      printf("ERROR: Please rename a CAN channel to \"CANopen\" first!\n");
      printf("You can use the SCIM to rename a channel!\n");
      
      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else 
    { 
      printf("     1.0) successfully                                           - done.\n");
      
      // next state is 2.1)
      ucDemoState = 11;
      
      printf("\n");
    }
  }
  

  // State 1.1) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 1.1 Initialising the interrupt mode                                     */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 11)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf("1.1) Initialising the interrupt mode.\n");
    printf("     This will be done by:\n");
    printf("     - creating events used for the interrupt thread          and\n");
    printf("     - creating the interrupt thread and assigning the events\n");
    printf("       to it                                                  and\n");
	printf("     - announcing the interrupt thread and related events to\n");
    printf("       the driver (Function CMA_SetIntEvent()).\n\n");

    // Initialise resources for interrupt mode
	if(InitIRQEvents(&ucCANHandle) == TRUE)
    {
      // next state is 2.1)
      ucDemoState = 21;

      printf("     1.1) successfully                                           - done.\n");
    }
	else
    {
      // error which leads to an abort of the demo program,
      // next state is 4.1)
      ucDemoState = 41;

      printf("     1.1)                                                        - failed !\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
	}
  }

  // States 2.x) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 2. Initialising CANopen Master                                          */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 21)
  {
    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf("2) Initialising CANopen Master.\n");
  }

  // State 2.1) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 2.1 Initialising CAN parameters                                         */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 21)
  {

    printf(" 2.1) Initialising of CAN parameters.\n");
    printf("      This will be done by using the function\n");
    printf("      CMA_ConfigCANChannel().\n");
    printf("      Following function parameter is specified:\n");
    // print the baud rate
    switch(BAUD_RATE)
    {
      case 0:
        printf("      - baud rate = 1 MBit/s\n");
        break;
      case 1:
        printf("      - baud rate = 800 KBit/s\n");
        break;
      case 2:
        printf("      - baud rate = 500 KBit/s\n");
        break;
      case 3:
        printf("      - baud rate = 250 KBit/s\n");
        break;
      case 4:
        printf("      - baud rate = 125 KBit/s\n");
        break;
      case 5:
        printf("      - baud rate = 50 KBit/s\n");
        break;
      case 6:
        printf("      - baud rate = 20 KBit/s\n");
        break;
      case 7:
        printf("      - baud rate = 10 KBit/s\n");
        break;
      default:
        printf("      - no valid baud rate !\n");
        break;
    }

    // get CAN parameter from constants section
    stCANConfig.ucBaudRate = BAUD_RATE;

    stCANConfig.ucAccCode = 0;
    stCANConfig.ucAccMask = 0;

    printf("      Call of CMA_ConfigCANChannel() function.\n");

    ret = CMA_ConfigCANChannel (ucCANHandle,stCANConfig);

    if (ret != 0)
    {
      // error which leads to an abort of the demo program,
      // next state is 6.0)
      ucDemoState = 60;

      PrintCMAFunctReturnError(ret);
      printf("      2.1)                                                       - failed.\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else
    {
      printf("      2.1) successfully                                          - done.\n");

      // next state is 2.2)
      ucDemoState = 22;
    }
  }

  // State 2.2) of the demo program:
  /* --------------------------------------------------------------------- */
  /* 2.2 Initialising the Node manager                                     */
  /* --------------------------------------------------------------------- */
  if(ucDemoState == 22)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf(" 2.2) Initialising the Node manager.\n");
    printf("      This will be done by using the function\n");
    printf("      CMA_InitNodeMan().\n");
    printf("      Following function parameter is specified:\n");
    printf("      - Node-ID of the master = %d\n",MASTER_NODE_ID);
    printf("      Call of CMA_InitNodeMan() function.\n");

    ret = CMA_InitNodeMan(ucCANHandle,MASTER_NODE_ID);

    if (ret != 0)
    {
      // error which leads to an abort of the demo program,
      // next state is 6.0)
      ucDemoState = 60;

      PrintCMAFunctReturnError(ret);
      printf("      2.2)                                                       - failed.\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else
    {
      // next state is 2.3)
      ucDemoState = 30;

      printf("      2.2) successfully                                          - done.\n");
    }
    printf("\n");
  }


  // State 3.0) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 3.0 Initiate the Startup of CANopen Master                              */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 30)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf("3.0) Initiate the Startup of CANopen Master.\n");
    printf("     The startup will be initiated by calling the function\n");
    printf("     CMA_Startup().\n");
    printf("     Call of CMA_Startup() function.\n");


    ret = CMA_Startup (ucCANHandle);

    if(ret != 0)
    {
      // error which leads to an abort of the demo program,
      // next state is 4.0)
      ucDemoState = 50;

      PrintCMAFunctReturnError(ret);
      printf("     3.0)                                                        - failed !\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");


    }
    else
    {
      // next state is 3.1)
      ucDemoState = 31;
      printf("     3.0) Startup of CANopen Master successfully                 - initiated.\n");
    }
  }

  // State 3.1) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 3.1 Check for the completion of the Startup                             */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 31)
  {
    printf("\n");
    printf("3.1) Check for the completion of the Startup\n");
    printf("     This will be done by waiting %d ms for the occurance\n",STARTUP_TIME);
    printf("     a StartupCompleted event.\n");
    printf("     - Now we wait %d ms while checking the Event-Fifo for\n",
             STARTUP_TIME);
    printf("     a StartupCompleted event.\n");

    // start timeout timer
    StartTimeoutTimer();


    do
    {
      // check a timeout
      if(CheckTimeoutTimer(STARTUP_TIME))
      {
        // error which leads to an abort of the demo program,
        // next state is 5.0)
        ucDemoState = 50;

        printf("          - Startup of CANopen Master timed out !\n");
        printf("     3.1) Startup of CANopen Master                              - failed !\n");

        printf("-----------------------------------------------------------------\n");
        printf("--> Error occured which leads to an abort of the demo program.\n");
        printf("-----------------------------------------------------------------\n");
      }
    }
    while((ucDemoState != 32) && (ucDemoState != 50));


    if(ucDemoState == 32)
    {


      printf("     3.1) Startup of CANopen Master successfully                 - done.\n");
        printf("          - CANopen Master is now in state >>Pre-Operational<<.\n");
    }
  }

  // State 3.2) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 3.2 Check if CANopen Master could send its bootup message               */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 32)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf("3.2) Check if CANopen Master could send it's bootup message.\n");
    printf("     This will be done by waiting %d ms for the occurance\n",BOOT_MSG_TIME);
    printf("     of a CAN Error event wich indicates a state change of the\n");
    printf("     CAN controller.\n");
    printf("     - Now we wait %d ms while checking the Event-Fifo for\n",
           BOOT_MSG_TIME);
    printf("       a CANError event, Error Code: 0x101 or 0x102.\n");

    // start timeout timer
    StartTimeoutTimer();


    do
    {
      // check a timeout of BOOT_MSG_TIME ms
      if(CheckTimeoutTimer(BOOT_MSG_TIME))
      {
        // next state is 4.0)
        ucDemoState = 40;
        printf("     3.2) successfully                                           - done.\n");
      }
    }
    while((ucDemoState != 33) && (ucDemoState != 40));
  }


  // State 3.3) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 3.3 Dummy state                                                         */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 33)
  {
    // error which leads to an abort of the demo program,
    // next state is 4.0)
    ucDemoState = 40;

    printf("     3.2)                                                        - failed !\n");
    printf("-----------------------------------------------------------------\n");
    printf("--> Error occured which leads to an abort of the demo program.\n");
    printf("-----------------------------------------------------------------\n");
  }

  // State 4.0) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 4.0 Initiate Shutdown of CANopen Master                                 */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 40)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf("4.0) Initiate Shutdown of CANopen Master\n");

    printf("     The shutdown will be initiated by calling the function\n");
    printf("     CMA_Shutdown().\n");
    printf("     Call of function CMA_Shutdown().\n");


    ret = CMA_Shutdown(ucCANHandle);

    if(ret != 0)
    {
      // error which leads to an abort of the demo program,
      // next state is 5.0)
      ucDemoState = 50;
      PrintCMAFunctReturnError(ret);

      printf("     4.0)                                                       - failed !\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else
    {
      // next state is 4.1)
      ucDemoState = 41;
      printf("     4.0) successfully                                           - initiated.\n");
    }
  }

  // State 4.1) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 4.1 Check for the completion of the Shutdown                            */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 41)
  {
    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf("4.1) Check for the completion of the Shutdown\n");
    printf("     This will be done by waiting %d ms for the occurance\n",SHUTDOWN_TIME);
    printf("     a ShutdownCompleted event.\n");
    printf("     - Now we wait %d ms while checking the Event-Fifo for\n",
           SHUTDOWN_TIME);
    printf("       ShutdownCompleted event.\n");

    // start timeout timer
    StartTimeoutTimer();


    do
    {
      // check a timeout
      if(CheckTimeoutTimer(SHUTDOWN_TIME))
      {
        // error which leads to an abort of the demo program,
        // next state is 5.0)
        ucDemoState = 50;

        printf("     - Shutdown of CANopen Master timed out !\n");
        printf("     4.1) Shutdown of CANopen Master                             - failed !\n");

        printf("-----------------------------------------------------------------\n");
        printf("--> Error occured which leads to an abort of the demo program.\n");
        printf("-----------------------------------------------------------------\n");
      }
    }
    while((ucDemoState != 42) && (ucDemoState != 50));
  }

  // State 4.2) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 4.2 Dummy state                                                         */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 42)
  {
    // next state is 5.0)
    ucDemoState = 50;

    printf("     4.1) Shutdown of CANopen Master successfully                - done.\n");
  }

  // State 5.0) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 5.0 Release resources for interrupt mode                                */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 50)
  {

    printf("\n");
    printf("5.0) Release resources for interrupt mode.\n");
    printf("     This will be done by:\n");
    printf("     - closing all event handles related to the interrupt\n");
    printf("       thread                                             and\n");
    printf("     - terminating the interrupt thread.\n");


    // Release resources for interrupt mode
    ExitIRQEvents();

    // next state is 5.1)
    ucDemoState = 51;

    printf("     5.0) successfully                                           - done.\n");

  }

  // State 5.1) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 5.1 Free CAN interface ressources                                       */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 51)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf("5.1) Freeing of CAN interface resources.\n");

    printf("     This will be done by calling the function\n");
    printf("     CMA_CloseCard().\n");
    printf("     Call of function CMA_CloseCard().\n");

    ret = CMA_CloseCard(ucCANHandle);

    if(ret != 0)
    {
      PrintCMAFunctReturnError(ret);
      printf("     5.1)                                                        - failed !\n");
      // next state is 5.2)
      ucDemoState = 52;

    }
    else
    {
      printf("     5.1) successfully                                           - done.\n");
      // next state is 5.2)
      ucDemoState = 52;
    }
  }

  // State 5.2) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 5.2 End of program                                                      */
  /* ----------------------------------------------------------------------- */
  printf("\nEnd of program. -> Press any key.\n");
  getch();

  return(0);
}

/* ------------------------------------------------------------------------- */
/* Function: InitIRQEvents()                                                 */
/*           - Function creates events used for the interrupt thread. Then   */
/*             the interrupt thread will be created and the events will be   */
/*             assigned to it.                                               */
/*             At least the interrupt thread and related events will be      */
/*             announced to the driver.                                      */
/*                                                                           */
/* Output:   ret                                                             */
/*           - TRUE  success                                                 */ 
/*           - FALSE failure                                                 */
/* ------------------------------------------------------------------------- */
int InitIRQEvents (unsigned char *can_handle)
{
  int   ret;

  printf("starting interrupt thread!\n");
  if ((ret =
       pthread_create(&interrupt_pthread, 0, (void *) InterruptThread, can_handle)))
  {
    return (-1);
  }


  return TRUE;
}

/* ------------------------------------------------------------------------- */
/* Function: ExitIRQEvents()                                                 */
/*           - Function fires the interrupt thread end event. The interrupt  */
/*             will then be terminated. At least it closes all event handles */
/*             related to the interrupt thread.                              */
/*                                                                           */
/* ------------------------------------------------------------------------- */
void ExitIRQEvents(void)
{
  pthread_cancel(interrupt_pthread);
  printf("interrupt thread is stopped now!\n");
}

/* ------------------------------------------------------------------------- */
/* Function: InterruptThread()                                               */
/*           - Interrupt thread.                                             */
/*                                                                           */
/* Intput:   context                                                         */
/*           - Context of interrupt thread.                                  */ 
/* ------------------------------------------------------------------------- */
int InterruptThread(void *context)
{   
  unsigned char can_handle = *(unsigned char*)context;
    BOOLEAN       end         = FALSE;
    int                   ret;

    while(!end)
    {
       ret = CMA_WaitForEvent(can_handle);

        // enter critical section
        EnterCriticalSection(&g_CriticalSection);

        switch (ret)
        {
            case 0:
              ReadAndCheckEvent();
              break;
            case -1:
              end = TRUE;
              printf("     - Interrupt thread is to be terminated.\n");
              break;
            default:
              end = TRUE;
              printf("     - Bad interrupt thread wait result 0x%x,\n",ret);
              printf("       interrupt thread will be terminated.\n");
            break;
        }

        // leave critical section
        LeaveCriticalSection(&g_CriticalSection);
    }
    return 0;
}

/* ------------------------------------------------------------------------- */
/* Function: ReadAndCheckEvent()                                             */
/*           - Reads new events from the event-fifo. Dependend on the        */
/*             current program state it branches to evaluation routines      */
/*             related to each the state.                                    */
/*                                                                           */
/* Output:   ret                                                             */
/*           - 0 success                                                     */ 
/* ------------------------------------------------------------------------- */
short ReadAndCheckEvent(void)
{

  CMA_EventStruct  stEvent;
  unsigned short   ret = 0;

  ret = CMA_ReadEvent (ucCANHandle,&stEvent);
  if(ret)
  {
    PrintCMAFunctReturnError(ret);
  }
  else
  {
    if(stEvent.ucEventType != 0)
    {
      switch(ucDemoState)
      {
        case 31: // Function to evaluate events in state 3.1)
                 // of the demo program.
          ret = EvaluateEventInState_31(stEvent);
          break;
        case 32: // Function to evaluate events in state 3.2)
                 // of the demo program.
          ret = EvaluateEventInState_32(stEvent);
          break;
        case 41: // Function to evaluate events in state 4.1)
                 // of the demo program.
          ret = EvaluateEventInState_41(stEvent);
          break;
        default:
          PrintUnexpectedEvent(stEvent);
        break;
      }
    }
  }
  return(ret);
}

/* ------------------------------------------------------------------------- */
/* Function: EvaluateEventInState_31()                                       */
/*           - Function evaluates events in state 3.1) of the demo program.  */
/*             Called from function ReadAndCheckEvent().                     */
/*                                                                           */
/* Input:    stEvent                                                         */
/*           - last read event                                               */
/* Output:   ret                                                             */
/*           - 0 success                                                     */
/* ------------------------------------------------------------------------- */
short EvaluateEventInState_31(CMA_EventStruct stEvent)
{
  unsigned short   ret = 0;

  switch(stEvent.ucEventType)
  {
    case StartupCompleted:
      // next state is 3.2)
      ucDemoState = 32;
      printf("     - StartupCompleted event received.\n");
      break;
    default:
      PrintUnexpectedEvent(stEvent);
      break;
  }
  return(ret);
}

/* ------------------------------------------------------------------------- */
/* Function: EvaluateEventInState_32()                                       */
/*           - Function evaluates events in state 3.2) of the demo program.  */
/*             Called from function ReadAndCheckEvent().                     */
/*                                                                           */
/* Input:    stEvent                                                         */
/*           - last read event                                               */
/* Output:   ret                                                             */
/*           - 0 success                                                     */
/* ------------------------------------------------------------------------- */
short EvaluateEventInState_32 (CMA_EventStruct stEvent)
{
  unsigned short   ret = 0;
  unsigned short   usTempErrCode;

  switch(stEvent.ucEventType)
  {
    case ErrorEvent:
      switch(stEvent.ucaEventData[0])
      {
        case CANError:
          // copy low part of Error Code
          usTempErrCode = (unsigned short)stEvent.ucaEventData[2];
          // copy high part of Error Code
          usTempErrCode |= (unsigned short)stEvent.ucaEventData[3] << 8;
          if(usTempErrCode == 0x0102)
          {
            // error which leads to an abort of the demo program,
            // next state is 3.3)
            ucDemoState = 33;
            printf("     - CANError event 0x%x received.\n",usTempErrCode);
            printf("       - CAN controller went ERROR PASSIVE !\n");
          }
          else if(usTempErrCode == 0x0101)
          {
            // error which leads to an abort of the demo program,
            // next state is 3.3)
            ucDemoState = 33;
            printf("     - CANError event 0x%x received.\n",usTempErrCode);
            printf("       - CAN controller went BUS OFF !\n");
          }
          break;
        default:
          {
            PrintUnexpectedEvent(stEvent);
          }
          break;
      }
      break;
    default:
      {
        PrintUnexpectedEvent(stEvent);
      }
      break;
  }
  return(ret);
}

/* ------------------------------------------------------------------------- */
/* Function: EvaluateEventInState_41()                                       */
/*           - Function evaluates events in state 4.0) of the demo program.  */
/*             Called from function ReadAndCheckEvent().                     */
/*                                                                           */
/* Input:    stEvent                                                         */
/*           - last read event                                               */
/* Output:   ret                                                             */
/*           - 0 success                                                     */
/* ------------------------------------------------------------------------- */
short EvaluateEventInState_41(CMA_EventStruct stEvent)
{
  unsigned short   ret = 0;

  switch(stEvent.ucEventType)
  {
    case ShutdownCompleted:
      // next state is 4.2)
      ucDemoState = 42;
      printf("     - ShutdownCompleted event received.\n");
      break;
    default:
      PrintUnexpectedEvent(stEvent);
      break;
  }
  return(ret);
}


/* ------------------------------------------------------------------------- */
/* Function: PrintUnexpectedEvent()                                          */
/*           - Function evaluates each possible event and prints it out.     */
/*             Called if an event is read from event-fifo which was not      */
/*             expected.                                                     */
/*                                                                           */
/* Input:    stEvent                                                         */
/*           - last read event                                               */
/* ------------------------------------------------------------------------- */
void PrintUnexpectedEvent(CMA_EventStruct stEvent)
{

  int              i;
  unsigned short   usTempCode;

  printf("-----------------------------------------------------------------\n");
  printf("--> Unexpected event received ! : \n");

  switch(stEvent.ucEventType)
  {
    case DownloadCompleted:
      printf("--> Event: DownloadCompleted:\n");
      // copy low part of Index
      usTempCode = (unsigned short)stEvent.ucaEventData[2];
      // copy high part of Index
      usTempCode |= (unsigned short)stEvent.ucaEventData[3] << 8;
      printf("-->        NodeID: %d, Index: 0x%x, ",stEvent.ucaEventData[0],
             usTempCode);
      printf("SubIndex: %d\n",stEvent.ucaEventData[1]);
      break;
    case UploadCompleted:
      printf("--> Event: UploadCompleted:\n");
      // copy low part of Index
      usTempCode = (unsigned short)stEvent.ucaEventData[2];
      // copy high part of Index
      usTempCode |= (unsigned short)stEvent.ucaEventData[3] << 8;
      printf("-->        NodeID: %d, Index: 0x%x, ",stEvent.ucaEventData[0],
             usTempCode);
      printf("SubIndex: %d, \n",stEvent.ucaEventData[1]);
      if(stEvent.ucaEventData[4] != 0xFF)
      {
        // expedited SDO transfer data is included
        printf("-->        DLC: %d, data: ",stEvent.ucaEventData[4]);
        for(i=0; i < stEvent.ucaEventData[4]; i++)
          printf("%d ",stEvent.ucaEventData[5+i]);
        printf("\n");
      }
      break;
    case EMCYReceived:
      printf("--> Event: EMCYReceived:\n");
      printf("-->        NodeID: %d, Data: ",stEvent.ucaEventData[0]);
      for(i=0; i<8; i++)
        printf("%d ",stEvent.ucaEventData[i+1]);
      printf("\n");
      break;
    case ErrorEvent:
      printf("--> Event: ErrorEvent:\n");
      switch(stEvent.ucaEventData[0])
      {
        case APIError:
          // copy low part of Error Code
          usTempCode = (unsigned short)stEvent.ucaEventData[2];
          // copy high part of Error Code
          usTempCode |= (unsigned short)stEvent.ucaEventData[3] << 8;
          printf("-->        EventType: APIError, ErrCode: 0x%x\n",usTempCode);
          break;
        case CANError:
          // copy low part of Error Code
          usTempCode = (unsigned short)stEvent.ucaEventData[2];
          // copy high part of Error Code
          usTempCode |= (unsigned short)stEvent.ucaEventData[3] << 8;
          printf("-->        EventType: CANError, ErrCode: 0x%x\n",usTempCode);
          break;
        case SDOError:
          printf("-->        EventType: SDOError, NodeID: %d,\n",stEvent.ucaEventData[1]);
          // copy low part of Additional Code
          usTempCode = (unsigned short)stEvent.ucaEventData[4];
          // copy high part of Additional Code
          usTempCode |= (unsigned short)stEvent.ucaEventData[5] << 8;
          printf("-->        ErrClass-ErrCode-AddCode: %d-%d-%d\n",stEvent.ucaEventData[2],
                 stEvent.ucaEventData[3],usTempCode);
          break;
        case GuardError:
          printf("-->        EventType: GuardError, NodeID: %d,\n",stEvent.ucaEventData[1]);
          printf("-->        ErrCode: %d\n",stEvent.ucaEventData[2]);
          break;
        default:
          printf("-->        Unknown error event type, value: %d received.\n",
                 stEvent.ucaEventData[0]);
          break;
      }
      break;
   case PDOReceived:
      printf("--> Event: PDOReceived:\n");
      // copy low part of Index
      usTempCode = (unsigned short)stEvent.ucaEventData[0];
      // copy high part of Index
      usTempCode |= (unsigned short)stEvent.ucaEventData[1] << 8;
      printf("-->        PDOHandle: %d, NumOfBytes: %d\n",usTempCode,
             stEvent.ucaEventData[2]);
      printf("-->        Data: ");
      for(i=0; i<stEvent.ucaEventData[2]; i++)
        printf("%d ",stEvent.ucaEventData[i+3]);
      printf("\n");
      break;
    case ShutdownCompleted:
      printf("--> Event: ShutdownCompleted\n");
      break;
    case StartupCompleted:
      printf("--> Event: StartupCompleted\n");
      break;
    case NoEvent:
      break;
    default:
      printf("--> Unknown event type, value %d received.\n",stEvent.ucEventType);
      break;
  }
  printf("-----------------------------------------------------------------\n");
}

/* ------------------------------------------------------------------------- */
/* Function: StartTimeoutTimer()                                             */
/*           - Starts the timeout timer. A timeout Timer will be checked     */
/*             with funtion CheckTimeoutTimer()                              */
/*                                                                           */
/* ------------------------------------------------------------------------- */
void StartTimeoutTimer (void)
{
  ulTimerStartTime = GetTickCount();
}

/* ------------------------------------------------------------------------- */
/* Function: CheckTimeoutTimer()                                             */
/*           - Checks if the timeout timer timed out. The Timer will be      */
/*             started with function StartTimeoutTimer().                    */
/*                                                                           */
/* Input:    ulTimeoutInMs                                                   */
/*           - Timeout value in ms.                                          */
/* ------------------------------------------------------------------------- */
short CheckTimeoutTimer (unsigned long ulTimeoutInMs)
{

  unsigned long ulActTime=0, ulDiffTime=0;

  ulActTime  = GetTickCount();
  ulDiffTime = (ulActTime >= ulTimerStartTime) ?
               ulActTime - ulTimerStartTime :
               (0xFFFFFFFFL - ulTimerStartTime) + ulActTime + 1;

  /* check timeout */
  if(ulDiffTime > ulTimeoutInMs)
  {
    // timeout occured
    return(-1);
  }
  else
    // no timeout occured yet
    return(0);
}

/* ------------------------------------------------------------------------- */
/* Function: PrintCMAFunctReturnError()                                      */
/*           - Prints function return codes in a textual form.               */
/*                                                                           */
/* Input:    ret                                                             */
/*           - Function return code                                          */
/* ------------------------------------------------------------------------- */
void PrintCMAFunctReturnError(int ret)
{
  switch(ret)
  {
    case CANPC_RB_INI_FILE:
      printf("      - CMA function error:\n");
      printf("        Can't open INI-file (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_FMT_INI:
      printf("      - CMA function error:\n");
      printf("        Format error in INI-file (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_OP_BIN:
      printf("      - CMA function error:\n");
      printf("        Error in opening binary-file (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_RD_BIN:
      printf("      - CMA function error:\n");
      printf("        Error in error reading binary-file (code: %d).\n",ret);
      break;
    case CANPC_RB_BIN_TOO_LONG:
      printf("      - CMA function error:\n");
      printf("        Binary-file too long (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_BIN_FMT:
      printf("      - CMA function error:\n");
      printf("        Binary-data format error (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_BIN_CS:
      printf("      - CMA function error:\n");
      printf("        Binary-data checksum error (code: %d).\n",ret);
      break;
    case CANPC_RB_NO_CARD:
      printf("      - CMA function error:\n");
      printf("        No interface present (code: %d).\n",ret);
      break;
    case CANPC_RB_NO_PHYS_MEM:
      printf("      - CMA function error:\n");
      printf("        No physical memory available (code: %d).\n",ret);
      break;
    case CANPC_RB_INVLD_IRQ:
      printf("      - CMA function error:\n");
      printf("        Invalid IRQ-number (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_DPRAM_ACCESS:
      printf("      - CMA function error:\n");
      printf("        Error in accessing dpram (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_CRD_RESP:
      printf("      - CMA function error:\n");
      printf("        Bad response from interface (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_SRAM:
      printf("      - CMA function error:\n");
      printf("        SRAM seems to be damaged (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_PRG:
      printf("      - CMA function error:\n");
      printf("        Invalid program start address (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_REC:
      printf("      - CMA function error:\n");
      printf("        Invalid record type (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_NORESP:
      printf("      - CMA function error:\n");
      printf("        No response after program start (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_BADRESP:
      printf("      - CMA function error:\n");
      printf("        Bad response after program start (code: %d).\n",ret);
      break;
    case CANPC_RB_PCMCIA_NSUPP:
      printf("      - CMA function error:\n");
      printf("        PCMCIA chip not supported (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_RD_PCMCIA:
      printf("      - CMA function error:\n");
      printf("        Error in reading PCMCIA parameters (code: %d).\n",ret);
      break;
    case CANPC_RB_INIT_CHIP:
      printf("      - CMA function error:\n");
      printf("        Error in initialising PCMCIA chip (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_NO_CANCARD:
      printf("      - CMA function error:\n");
      printf("        No CANcard plugged in (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_CRD_NORESP:
      printf("      - CMA function error:\n");
      printf("        Bad response from interface (code: %d).\n",ret);
      break;
    case CMA_ERR_FUNCTION_TIMEOUT:
      printf("      - CMA function error:\n");
      printf("        Function timeout occurred (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_FUNCTIONALITY:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucCANxFunctionality undefined (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_CHAN_HANDLE:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucChannelHandle undefined (code: %d).\n",ret);
      break;
    case CMA_ERR_SCT_BAUD_RATE:
      printf("      - CMA function error:\n");
      printf("        Struct parameter: ucBaudRate invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_TIMEOUT_MAX:
      printf("      - CMA function error:\n");
      printf("        Parameter: timeout too big (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_TIMEOUT_MIN:
      printf("      - CMA function error:\n");
      printf("        Parameter: timeout too small (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_NODE_ID:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucNodeID is invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_SCT_COB_ID:
      printf("      - CMA function error:\n");
      printf("        Struct parameter: COB-ID invalid or used (code: %d).\n",ret);
      break;
    case CMA_ERR_SCT_COMM_CYC_MAX:
      printf("      - CMA function error:\n");
      printf("        Struct parameter: ulSyncCycle is too big (code: %d).\n",ret);
      break;
    case CMA_ERR_SCT_NODE_ID:
      printf("      - CMA function error:\n");
      printf("        Struct parameter: ucNodeID invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_MASTER_STATE:
      printf("      - CMA function error:\n");
      printf("        Improper state of the master (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_STATE:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucState invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_INT_TX_FIFO_FULL:
      printf("      - CMA function error:\n");
      printf("        Internal transmit fifo is full (code: %d).\n",ret);
      break;
    case CMA_ERR_MASTER_GUARDING:
      printf("      - CMA function error:\n");
      printf("        Master cannot guard itself (code: %d).\n",ret);
      break;
    case CMA_ERR_COMMAND_FIFO_FULL:
      printf("      - CMA function error:\n");
      printf("        DPRAM command fifo is full (code: %d).\n",ret);
      break;
    case CMA_ERR_SDO_TX_FIFO_FULL:
      printf("      - CMA function error:\n");
      printf("        SDO transmit fifo is full (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_DATA_SIZE:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucDataSize is invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_INDEX:
      printf("      - CMA function error:\n");
      printf("        Parameter: usIndex is incorrect (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_SUBINDEX:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucSubIndex is incorrect (code: %d).\n",ret);
      break;
    case CMA_ERR_TOTAL_DATA_SIZE:
      printf("      - CMA function error:\n");
      printf("        Parameter: ulTotalDataSize exceeded (code: %d).\n",ret);
      break;
    case CMA_ERR_NO_TRANSF_ACTIVE:
      printf("      - CMA function error:\n");
      printf("        No normal SDO transfer active (code: %d).\n",ret);
      break;
    case CMA_ERR_SDO_RX_FIFO_EMPTY:
      printf("      - CMA function error:\n");
      printf("        SDO receive fifo is empty (code: %d).\n",ret);
      break;
    case CMA_ERR_SCT_PDO_ACC_TYPE:
      printf("      - CMA function error:\n");
      printf("        Struct parameter: ucAccType is invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_PDO_HANDLE:
      printf("      - CMA function error:\n");
      printf("        Parameter: usPDOHandle is undefined (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_BYTEPOS:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucDataOffset is invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_PDO_TOO_LONG:
      printf("      - CMA function error:\n");
      printf("        Parameter: DataOffset+Datalength>8 (code: %d).\n",ret);
      break;
    case CMA_ERR_PDO_ACCESS:
      printf("      - CMA function error:\n");
      printf("        Access to PDO in DPRAM failed (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_BITPOS:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucBitPosition is invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_UNKNOWN_EVENT:
      printf("      - CMA function error:\n");
      printf("        Unknown event in event fifo (code: %d).\n",ret);
      break;
    case CMA_ERR_WRONG_FW_VERSION:
      printf("      - CMA function error:\n");
      printf("        Wrong firmware version (code: %d).\n",ret);
      break;
    case CMA_ERR_SCT_TRANS_TYPE:
      printf("      - CMA function error:\n");
      printf("        Struct parameter: ucTransType invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_DPRAM_TX_PDO_MAX:
      printf("      - CMA function error:\n");
      printf("        Number of DPRAM-TX-PDOs exceeded (code: %d).\n",ret);
      break;
    case CMA_ERR_DPRAM_RX_PDO_MAX:
      printf("      - CMA function error:\n");
      printf("        Number of DPRAM-RX-PDOs exceeded (code: %d).\n",ret);
      break;
    case CMA_ERR_LAST_REQ_PENDING:
      printf("      - CMA function error:\n");
      printf("        Last request still pending (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_TOOMANYDEVICES:
      printf("      - CMA function error:\n");
      printf("        Maximum number of device exceeded (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_NOTENOUGHMEMORY:
      printf("      - CMA function error:\n");
      printf("        Not enough memory space to register a device (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_DRIVERNOTFOUND:
      printf("      - CMA function error:\n");
      printf("        Driver not found (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_DRIVERVERSIONERROR:
      printf("      - CMA function error:\n");
      printf("        Wrong driver version (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_VERSIONERROR:
      printf("      - CMA function error:\n");
      printf("        Wrong driver-dll version (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_NOTSUPPORTED:
      printf("      - CMA function error:\n");
      printf("        Internal error: Requested functionality not supported\n");
      printf("        (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_ILLEGALCALL:
      printf("      - CMA function error:\n");
      printf("        Internal error: Illegal driver call (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_IOPENDING:
      printf("      - CMA function error:\n");
      printf("        Internal error: Driver call pending (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_TIMEOUT:
      printf("      - CMA function error:\n");
      printf("        Internal error: Timeout during driver access (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_GENERALERROR:
      printf("      - CMA function error:\n");
      printf("        Internal error: General error in accessing the driver (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_INTERNALERROR:
      printf("      - CMA function error:\n");
      printf("        Internal driver error (code: %d).\n",ret);
      break;
    case INIPC_IB_PNP_NO_DEVICE_FOUND:
      printf("      - CMA function error:\n");
      printf("        No CAN interface found (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_UNKNOWNDEVICE:
      printf("      - CMA function error:\n");
      printf("        Access to an unknown device requested (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_DEVICEALREADYEXISTS:
      printf("      - CMA function error:\n");
      printf("        Device is already registered by the driver (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_DEVICEACCESSERROR:
      printf("      - CMA function error:\n");
      printf("        Device is already open (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_RESOURCEALREADYREGISTERED:
      printf("      - CMA function error:\n");
      printf("        A requested resource is already in use \n");
      printf("        by an other device (DPRAM, IRQ, I/O port) (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_RESOURCECONFLICT:
      printf("      - CMA function error:\n");
      printf("        Resource conflict with an other driver (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_RESOURCEACCESSERROR:
      printf("      - CMA function error:\n");
      printf("        Resource access error (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_PHYSMEMORYOVERRUN:
      printf("      - CMA function error:\n");
      printf("        Internal error: Invalid phys. mem-access (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_TOOMANYPORTS:
      printf("      - CMA function error:\n");
      printf("        Internal error: Too many I/O ports requested (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_UNKNOWNRESOURCE:
      printf("      - CMA function error:\n");
      printf("        Internal error: Unknown resource access requested (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_LOADDDLL_DEV:
      printf("      - CMA function error:\n");
      printf("        Driver DLL not found (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_CANT_DEVREGISTER:
      printf("      - CMA function error:\n");
      printf("        Interface cannot be registered (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_IRQ:
      printf("      - CMA function error:\n");
      printf("        Interrupt number invalid or device cannot\n");
      printf("        be assigned an interrupt (code: %d).\n",ret);
      break;
    case DDC_ERR_WRONG_DEVICEHANDLE:
      printf("      - CMA function error:\n");
      printf("        Internal error: Wrong device handle specified\n");
      printf("        (code: %d).\n",ret);
      break;
    default:
      printf("      - Unknown CMA function error (code: %d).\n",ret);
      break;
  }
}

