/*****************************************************************************/
/*                                                                           */
/*   Softing GmbH          Richard-Reitzner-Allee 6       85540 Haar         */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*   Copyright (C) by Softing GmbH 1999              All rights reserved.    */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*                        P D O T R A N S F E R . C                          */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* PROJECT:            CANopen Master API C-Samples                          */
/*                                                                           */
/* MODULE:             PDOTransfer.c                                         */
/*                                                                           */
/* AUTHOR:             Softing GmbH                                          */
/*                                                                           */
/* DATE                04.12.1999                                            */
/*                                                                           */
/*****************************************************************************/
/* MODULE_DESCRIPTION                                                        */ 
/*                                                                           */
/*  GENERAL:       This is a demo application to show how to use             */
/*                 CANopen Master API to maintain PDO data transfers.        */
/*                                                                           */
/*  CANOPEN                                                                  */
/*    BACKGROUND   In a CANopen network the real-time data transfer is       */
/*                 performed by means of "Process Data Objects(PDO)".        */
/*                 It is distinguished between Transmit-PDOs (TX-PDOs) and   */
/*                 Receive-PDOs (RX-PDOs). Devices supporting TX-PDOs are    */
/*                 PDO producer and devices which are able to receive PDOs   */
/*                 are called PDO consumer.                                  */
/*                 PDOs can have are different transmission modes            */
/*                 (synchronous, asynchronous) and different triggering      */
/*                 modes (event driven, remotely requested). Each            */
/*                 PDO has a transmission type (TT) parameter whose possible */
/*                 values define combinations of transmission modes and      */
/*                 triggering modes.                                         */
/*                 For more detailed information about PDOs please refer to  */
/*                 CiA DS 301 V3.0 (CANopen Application Layer and            */
/*                 Communication Profile)                                    */
/*                                                                           */
/*  CONTENT:       In the demo the master will be configured to maintain     */
/*                 as an example 2 PDOs. One of it is a synchronous TX-PDO   */
/*                 with a TT = 4 by default. For the transmission of data    */
/*                 via this PDO a Sync-Object is needed (The transmission of */
/*                 synchronous PDOs is related to the Sync-Object.).         */
/*                 The other PDO is a asynchronous RX-PDO (TT = 255 by       */
/*                 default). For this PDO a Remote Request is send out       */
/*                 activly initiate a PDO value update.                      */
/*                 Please adapt the COB-ID values of the 2 PDOs according    */
/*                 to the configuration of Your CANopen net. See parameters  */
/*                 TXPDO_COBID & RXPDO_COBID  below in section >Constants to */
/*                 be changed by the user<.                                  */
/*                                                                           */
/*  DEMO STRUCTURE The demo is maintaned by a state machine. The global      */
/*                 variable <ucDemoState> contains the current state.        */
/*                 Following a short description of the state machine        */
/*                 especially which action is performed in each state:       */
/*                                                                           */
/*                   state 1.0:  Initialising the Softing CAN interface to   */
/*                               run CANopen Master API                      */
/*                                 - success -> state 2.1                    */
/*                                 - failure -> state 6.1                    */
/*                                                                           */
/*                   states 2.x: Initialising the CANopen Master             */
/*                   state 2.1:  Initialising CAN parameters                 */
/*                                 - success -> state 2.2                    */
/*                                 - failure -> state 6.0                    */
/*                   state 2.2:  Initialising the Node manager               */
/*                                 - success -> state 3.0                    */
/*                                 - failure -> state 6.0                    */
/*                                                                           */
/*                   state 3.0:  Initiate Startup of CANopen Master          */
/*                                 - success -> state 3.1                    */
/*                                 - failure -> state 6.0                    */
/*                   state 3.1:  Check for the completion of the Startup.    */
/*                                 - success -> state 3.2                    */
/*                                 - failure -> state 6.0                    */
/*                   state 3.2:  Check if CANopen Master could send its      */
/*                               bootup message.                             */
/*                                 - success -> state 4.1                    */
/*                                 - failure -> state 5.0                    */
/*                                                                           */
/*                   states 4.x: CANopen Master is active.                   */
/*                   state 4.1:  Configuring the Sync manager                */
/*                                 - success -> state 4.2                    */
/*                                 - failure -> state 5.0                    */
/*                   state 4.2:  Installing a TX-PDO                         */
/*                                 - success -> state 4.3                    */
/*                                 - failure -> state 5.0                    */
/*                   state 4.3:  Installing a RX-PDO                         */
/*                                 - success -> state 4.4                    */
/*                                 - failure -> state 5.0                    */
/*                   state 4.4:  Set master and all connected slaves in      */
/*                               state >>Operational<<.                      */
/*                                 - success -> state 4.5                    */
/*                                 - failure -> state 5.0                    */
/*                   state 4.5:  Send a remote PDO request.                  */
/*                                 - success -> state 4.6                    */
/*                                 - failure -> state 5.0                    */
/*                   state 4.6:  Check if we received a RX-PDO on our        */
/*                               Remote Request.                             */
/*                                 - success -> state 4.7                    */
/*                                 - failure -> state 5.0                    */
/*                   state 4.7:  Update the TX-PDO.                          */
/*                                 - success -> state 4.8                    */
/*                                 - failure -> state 5.0                    */
/*                   state 4.8:  Set master and all connected slaves in      */
/*                               state >>Pre-Operational<<.                  */
/*                                 - success -> state 5.0                    */
/*                                 - failure -> state 5.0                    */
/*                                                                           */
/*                   state 5.0:  Initiate Shutdown of CANopen Master         */
/*                                 - success -> state 5.1                    */
/*                                 - failure -> state 6.0                    */
/*                   state 5.1:  Check for the completion of the Shutdown.   */
/*                                 - success -> state 5.2                    */
/*                                 - failure -> state 6.0                    */
/*                   state 5.2:  Dummy state                                 */
/*                                                                           */
/*                   state 6.0:  Free CAN interface ressources               */
/*                                 - success -> state 6.1                    */
/*                                 - failure -> state 6.1                    */
/*                                                                           */
/*                   state 6.1:  End of program                              */
/*                                                                           */
/*                 Each action can be configured by setting constants.       */
/*                 Please refer to the >Constants to be changed by the user< */
/*                 section below.                                            */
/*                                                                           */
/*****************************************************************************/   
#include <stdio.h>
#include <sys/time.h>
#include <string.h>

// CANopen Master API include files 
#include "kbhit.h"
#include "CmaDef.h"
#include "CmaRet.h"

/*****************************************************************************/
/*                                                                           */
/* Constants to be changed by the user                                       */
/*                                                                           */
/*****************************************************************************/
// constants used in state 2.1:
                              // baud rate, possible values are:
                              //  - 0: 1   Mbit/s
                              //  - 1: 800 Kbit/s
                              //  - 2: 500 Kbit/s
                              //  - 3: 250 Kbit/s
                              //  - 4: 125 Kbit/s
                              //  - 5: 50  Kbit/s
                              //  - 6: 20  Kbit/s
                              //  - 7: 10  Kbit/s
#define BAUD_RATE             0

// constants used in state 2.2:
                              // node-ID of CANopen Master
#define MASTER_NODE_ID        127

// constants used in state 3.0:
                              // timeout in ms for the master to startup
#define STARTUP_TIME          1000000

// constants used in state 3.1:
                              // time we wait after startup if the 
                              // CAN controller changes it's bus state,
                              // indication that bootup message could
                              // be sent
#define BOOT_MSG_TIME         10000

// constants used in state 4.1:
                              // definition if CANopen Master sends the 
                              // Sync-Object (= 1) or not (= 0)
#define SEND_MODE             1
                              // definition if CANopen Master itself consumes
                              // the Sync-object (= 1) or not (= 0)
#define CONSUME_MODE          1
                              // COB-ID of Sync-Object message 
#define SYNC_ID               128
                              // Communication Cycle Period in us
#define SYNC_CYCLE            300000 // 1s

// constants used in state 4.2:
//(installing a TX-PDO)
                              // COB-ID of TX-PDO
#define TXPDO_COBID           511
                              // Transmission type of TX-PDO
#define TXPDO_TT              4
                              // Default values of TX-PDO data bytes
#define TXPDO_DEFVAL_BYTE0    0
#define TXPDO_DEFVAL_BYTE1    0
#define TXPDO_DEFVAL_BYTE2    0
#define TXPDO_DEFVAL_BYTE3    0
#define TXPDO_DEFVAL_BYTE4    0
#define TXPDO_DEFVAL_BYTE5    0
#define TXPDO_DEFVAL_BYTE6    0
#define TXPDO_DEFVAL_BYTE7    0

// constants used in state 4.3:
//(installing a RX-PDO)
                              // COB-ID of RX-PDO
#define RXPDO_COBID           639
                              // Transmission type of RX-PDO
#define RXPDO_TT              255
                              // Default values of RX-PDO data bytes
#define RXPDO_DEFVAL_BYTE0    0
#define RXPDO_DEFVAL_BYTE1    0
#define RXPDO_DEFVAL_BYTE2    0
#define RXPDO_DEFVAL_BYTE3    0
#define RXPDO_DEFVAL_BYTE4    0
#define RXPDO_DEFVAL_BYTE5    0
#define RXPDO_DEFVAL_BYTE6    0
#define RXPDO_DEFVAL_BYTE7    0
                              // Defines if the application should be informed about
                              // the receive of the RX-PDO via a PDOReceived event,
                              // = 1 - the application will be informed
#define RXPDO_EVENT_NOTIF     1

// constant used in state 4.6:
                              // Time we wait after a PDO Remote Request if
                              // we receive the requested RX-PDO.
#define RXPDO_WAIT_EVENT      1000

// constants used in state 4.7:
                              // Byte values for update of TX-PDO data
#define TXPDO_UPD_BYTE0       1
#define TXPDO_UPD_BYTE1       2
#define TXPDO_UPD_BYTE2       3
#define TXPDO_UPD_BYTE3       4
#define TXPDO_UPD_BYTE4       5
#define TXPDO_UPD_BYTE5       6
#define TXPDO_UPD_BYTE6       7
#define TXPDO_UPD_BYTE7       8

// constants used in state 5.0
                              // timeout in ms for the master to shutdown
#define SHUTDOWN_TIME         20

/*****************************************************************************/
/*                                                                           */
/* Constants which should not be changed                                     */
/*                                                                           */
/*****************************************************************************/
                           // CAN1 runs in CANopen Master mode
#define CAN1_FUNCTIONALITY     1
                           // CAN2 is not used
#define CAN2_FUNCTIONALITY     0
                           // license to check by the master
                           // (CANopen Master mode)
#define LIC                    0
                           // index to which read acces is done
#define READ_INDEX             0x1000
                           // sub-index to which read acces is done
#define READ_SUB_INDEX         0

                           // values for parameter ucState(requested node state)
                           // of function CMA_ChangeState()
#define PREPARED               4
#define OPERATIONAL            5
#define RESET_NODE             6
#define RESET_COMMUNICATION    7
#define PRE_OPERATIONAL        127

                            // value of Node-ID parameter if all nodes
                            // are addressed
#define ALL_NODES_ADDRESSED    128


/*****************************************************************************/
/*                                                                           */
/* Global structures and variables                                           */
/*                                                                           */
/*****************************************************************************/
                           // state of the demo program
static unsigned char  ucDemoState;
                           // channel handles
static unsigned char  ucCAN1Handle;
                           // start time for timeout timer
static unsigned long  ulTimerStartTime;
                           // handle of TX-PDO
static unsigned short usTXPDOHandle;
                           // handle of RX-PDO
static unsigned short usRXPDOHandle;

/*****************************************************************************/
/*                                                                           */
/* Function prototypes                                                       */
/*                                                                           */
/*****************************************************************************/
short ReadAndCheckEvent(void);
void  PrintUnexpectedEvent(CMA_EventStruct stEvent);
short EvaluateEventInState_31(CMA_EventStruct stEvent);
short EvaluateEventInState_32(CMA_EventStruct stEvent);
short EvaluateEventInState_46 (CMA_EventStruct stEvent);
short EvaluateEventInState_51(CMA_EventStruct stEvent);
void  PrintCMAFunctReturnError(int ret);
void  StartTimeoutTimer (void);
short CheckTimeoutTimer (unsigned long ulTimeoutInMs);


unsigned long GetTickCount(void)
{
  unsigned long count;
  struct timeval tv;
  struct timezone tz;
  int ret;
  
  ret =  gettimeofday(&tv, &tz);
  
  if(ret) return 0;
  
  count = tv.tv_usec/1000 + tv.tv_sec*1000;
  
  return count;
}

/*****************************************************************************/
/*                                                                           */
/* Functions                                                                 */
/*                                                                           */
/*****************************************************************************/
/* ------------------------------------------------------------------------- */
/*                                                                           */
/* Main-Function                                                             */
/*                                                                           */
/* ------------------------------------------------------------------------- */
int main(int argc, char **argv)
{
  // function return codes
  short           ret; 

  // structures needed for function calls in:
  // state 2.1
  CMA_CAN_CONFIG    stCANConfig;
  // state 4.1
  CMA_SyncManConfig stSyncManConfig;
  // state 4.2
  CMA_PDOConfig     stPDOConfig;

  // variables needed in:
  // state 4.7
  unsigned char     TXPDO_UpdateData[8];

  memset(&stPDOConfig,0,sizeof(CMA_PDOConfig));
  memset(&stCANConfig,0,sizeof(CMA_CAN_CONFIG));
  memset(&stSyncManConfig,0,sizeof(CMA_SyncManConfig));
  


  printf("***************************************************************\n");
  printf("*                     SOFTING GMBH 1999                       *\n");
  printf("*                                                             *\n");
  printf("*                     PDO Transfer                            *\n");
  printf("*            (Demo Program for CANopen Master API)            *\n");
  printf("*                                                             *\n");
  printf("***************************************************************\n");

  // initialise state machine of demo program
  ucDemoState = 10;

  printf("\n");

  // State 1.0) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 1.0 Initialising the Softing CAN interface to run CANopen Master API    */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 10)
  {

    printf("1.0) Initialising the Softing CAN channel \"CANopen\" to run\n");
    printf("     CANopen Client API.\n"); 
    printf("     This will be done by using the function CMA_InitializeCard().\n");
    printf("     one channel must be renamed to CANopen to use this sample!\n");
    printf("     Call of CMA_InitializeCard() function to open the channel with the name \"CANopen\"\n");

    ret = CMA_InitializeCard (0xFFFFFFFF, // do not care
      CAN1_FUNCTIONALITY, // allways 1
      &ucCANHandle, // out: the channel handle
      LIC, // allways 0; 
      (unsigned char*)"CANopen"); // in: the channels name used to select the channel
 
  if (ret<0)
	{ 
      // error which leads to an abort of the demo program,
      // next state is 6.1)
      ucDemoState = 61;

      PrintCMAFunctReturnError(ret);
      printf("     1.0)                                                        - failed !\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
	}
    else
	{ 
      printf("     1.0) successfully                                           - done.\n");

      // next state is 2.1)
      ucDemoState = 21;

      if(ret==0)
      { 
        printf("     - Plugged interface is licensed to run CANopen Master API.\n");
      }
      else
      {  
        printf ("    ! - Still %d applications without license permitted !\n",ret);
      }

      printf("\n");
    }
  }


  // States 2.x) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 2. Initialising CANopen Master                                          */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 21)
  {
    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf("2) Initialising CANopen Master.\n");
  }

  // State 2.1) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 2.1 Initialising CAN parameters                                         */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 21)
  {

    printf(" 2.1) Initialising of CAN parameters.\n");
    printf("      This will be done by using the function\n");
    printf("      CMA_ConfigCANChannel().\n");
    printf("      Following function parameter is specified:\n");
    // print the baud rate
    switch(BAUD_RATE)
    {
      case 0:
        printf("      - baud rate = 1 MBit/s\n");
        break;
      case 1:
        printf("      - baud rate = 800 KBit/s\n");
        break;
      case 2:
        printf("      - baud rate = 500 KBit/s\n");
        break;
      case 3:
        printf("      - baud rate = 250 KBit/s\n");
        break;
      case 4:
        printf("      - baud rate = 125 KBit/s\n");
        break;
      case 5:
        printf("      - baud rate = 50 KBit/s\n");
        break;
      case 6:
        printf("      - baud rate = 20 KBit/s\n");
        break;
      case 7:
        printf("      - baud rate = 10 KBit/s\n");
        break;
      default:
        printf("      - no valid baud rate !\n");
        break;
    }

    // get CAN parameter from constants section
    stCANConfig.ucBaudRate = BAUD_RATE;

    printf("      Call of CMA_ConfigCANChannel() function.\n");

    ret = CMA_ConfigCANChannel (ucCAN1Handle,stCANConfig);

    if (ret != 0)
    {
      // error which leads to an abort of the demo program,
      // next state is 6.0)
      ucDemoState = 60;

      PrintCMAFunctReturnError(ret);
      printf("      2.1)                                                       - failed.\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else
    {
      printf("      2.1) successfully                                          - done.\n");

      // next state is 2.2)
      ucDemoState = 22;
    }
  }

  // State 2.2) of the demo program:
  /* --------------------------------------------------------------------- */
  /* 2.2 Initialising the Node manager                                     */
  /* --------------------------------------------------------------------- */
  if(ucDemoState == 22)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf(" 2.2) Initialising the Node manager.\n");
    printf("      This will be done by using the function\n");
    printf("      CMA_InitNodeMan().\n");
    printf("      Following function parameter is specified:\n");
    printf("      - Node-ID of the master = %d\n",MASTER_NODE_ID);
    printf("      Call of CMA_InitNodeMan() function.\n");

    ret = CMA_InitNodeMan(ucCAN1Handle,MASTER_NODE_ID);

    if (ret != 0)
    {
      // error which leads to an abort of the demo program,
      // next state is 6.0)
      ucDemoState = 60;

      PrintCMAFunctReturnError(ret);
      printf("      2.2)                                                       - failed.\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else
    {
      // next state is 3.0)
      ucDemoState = 30;

      printf("      2.2) successfully                                          - done.\n");
    }
    printf("\n");
  }

  // State 3.0) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 3.0 Initiate Startup of CANopen Master                                  */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 30)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf("3.0) Startup of CANopen Master.\n");
    printf("     The startup will be initiated by calling the function\n");
    printf("     CMA_Startup().\n");
    printf("     Call of CMA_Startup() function.\n");

    ret = CMA_Startup (ucCAN1Handle);

    if(ret != 0)
    {
      // error which leads to an abort of the demo program,
      // next state is 6.0)
      ucDemoState = 60;

      PrintCMAFunctReturnError(ret);
      printf("     3.0)                                                        - failed !\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else
    {
      printf("     3.0) Startup of CANopen Master successfully                 - initiated.\n");
      // next state is 3.1)
      ucDemoState = 31;
    }
  }

  // State 3.1) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 3.1 Check for the completion of the Startup                             */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 31)
  {
    printf("\n");
    printf("3.1) Check for the completion of the Startup\n");
    printf("     This will be done by waiting %d ms for the occurance\n",STARTUP_TIME);
    printf("     a StartupCompleted event.\n");
    printf("     - Now we wait %d ms while checking the Event-Fifo for\n",
           STARTUP_TIME);
    printf("     a StartupCompleted event.\n");

    // start timeout timer
    StartTimeoutTimer();

    do
    {
      // check a timeout
      if(CheckTimeoutTimer(STARTUP_TIME))
      {
        // error which leads to an abort of the demo program,
        // next state is 6.0)
        ucDemoState = 60;

        printf("     - Startup of CANopen Master timed out !\n");
        printf("     3.1) Startup of CANopen Master                              - failed !\n");

        printf("-----------------------------------------------------------------\n");
        printf("--> Error occured which leads to an abort of the demo program.\n");
        printf("-----------------------------------------------------------------\n");
      }
      // read and check event-fifo for events
      ReadAndCheckEvent();
    }
    while((ucDemoState != 32) && (ucDemoState != 60));

    if(ucDemoState == 32)
    {
      printf("     3.1) Startup of CANopen Master successfully                 - done.\n");
      printf("          - CANopen Master is now in state >>Pre-Operational<<.\n");
    }
  }


  // State 3.2) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 3.2 Check if CANopen Master could send its bootup message               */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 32)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf("3.2) Check if CANopen Master could send it's bootup message.\n");
    printf("     This will be done by waiting %d ms for the occurance\n",BOOT_MSG_TIME);
    printf("     of a CAN Error event wich indicates a state change of the\n");
    printf("     CAN controller.\n");
    printf("     - Now we wait %d ms while checking the Event-Fifo for\n",
           BOOT_MSG_TIME);
    printf("       a CANError event, Error Code: 0x101 or 0x102.\n");

    // start timeout timer
    StartTimeoutTimer();

    do
    {
      // check a timeout of 10ms
      if(CheckTimeoutTimer(BOOT_MSG_TIME))
      {
        // next state is 4.1)
        ucDemoState = 41;
        printf("     - No CANError event received.\n");
        printf("     3.2) successfully                                           - done.\n");
      }
      // read and check event-fifo for events
      ReadAndCheckEvent();
    }
    while((ucDemoState != 41) && (ucDemoState != 50));

    if(ucDemoState == 50)
    {
      printf("       3.2)                                                      - failed !\n");
      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
  }


  // States 4x) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 4.x CANopen Master is active.                                           */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 41)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf("4) CANopen Master is active.\n");
  }


  // State 4.1) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 4.1 Configuring the Sync manager.   */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 41)
  {

	printf(" 4.1) Configure the Sync manager.\n");
    printf("      The configuration is done via function CMA_ConfigSyncMan()\n");
    printf("      Following function parameter are specified:\n");

    if(SEND_MODE == 1)
    {
      printf("        - Send Mode                     = %d,\n",SEND_MODE);
	  printf("          (CANopen Master sends the Sync-Object.)\n");
    }
    else
    {
      printf("        - Send Mode                     = %d,\n",SEND_MODE);
	  printf("          (CANopen Master does not send the Sync-Object.)\n");
    }
    if(CONSUME_MODE == 1)
    {
      printf("        - Consume Mode                  = %d,\n",CONSUME_MODE);
	  printf("          (CANopen Master itself consumes the Sync-Object.)\n");
    }
    else
    {
      printf("        - Consume Mode                  = %d,\n",CONSUME_MODE);
	  printf("          (CANopen Master itself does not consume the Sync-Object.)\n");
    }

    printf("        - COB-ID of Sync-Object message = %d,\n",SYNC_ID);
    printf("        - Communication Cycle Period    = %d us,\n",SYNC_CYCLE);



    // get CMA_ConfigSyncMan() parameters from constants section
    stSyncManConfig.ucSendMode     = SEND_MODE;
    stSyncManConfig.ucConsumeMode  = CONSUME_MODE;
    stSyncManConfig.ulSyncID       = SYNC_ID;
    stSyncManConfig.ulSyncCycle    = SYNC_CYCLE;

	ret = CMA_ConfigSyncMan (ucCAN1Handle, stSyncManConfig); 

    if(ret != 0)
    {
      // error which leads to an abort of the demo program,
      // next state is 6.0)
      ucDemoState = 60;

      PrintCMAFunctReturnError(ret);
      printf("      4.1)                                                       - failed !\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else
    {
      // next state is 4.2)
      ucDemoState = 42;
      printf("      4.1) successfully                                          - done.\n");
    }
  }

  // State 4.2) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 4.2 Installing a TX-PDO                                                 */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 42)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf(" 4.2) Installing a TX-PDO.\n");
    printf("      This will be done by using the function CMA_InstallPDO().\n");
    printf("      Following configuration is specified:\n");
    printf("        - COB-ID of PDO              = %d,\n",TXPDO_COBID);
    printf("        - Transmission Type of PDO   = %d,\n",TXPDO_TT);
    printf("        - Default values of PDO data = %3d %3d %3d %3d %3d %3d %3d %3d\n",
		              TXPDO_DEFVAL_BYTE0,TXPDO_DEFVAL_BYTE1,TXPDO_DEFVAL_BYTE2,TXPDO_DEFVAL_BYTE3,
					  TXPDO_DEFVAL_BYTE4,TXPDO_DEFVAL_BYTE5,TXPDO_DEFVAL_BYTE6,TXPDO_DEFVAL_BYTE7);
    printf("      Call of CMA_InstallPDO() function.\n");

    // set function parameter for CMA_InstallPDO() 
	                             // TX-PDO
    stPDOConfig.ucPDOType    = 1;
	                             // TX-PDO COB-ID
    stPDOConfig.ulCOBID      = TXPDO_COBID;
	                             // Transmission type of PDO,
	                             // = 4, PDO data are transmitted after every 4th Sync-Object
    stPDOConfig.ucTransType  = TXPDO_TT;
	                             // PDO access type
	                             // until now only DPRAM PDOs are supported (= 0) 
	stPDOConfig.ucAccType    = 0;
	                             // default values of PDO data
	stPDOConfig.ucaDefVal[0]++; // = TXPDO_DEFVAL_BYTE0;
	stPDOConfig.ucaDefVal[1]++; //  = TXPDO_DEFVAL_BYTE1;
	stPDOConfig.ucaDefVal[2]++; //  = TXPDO_DEFVAL_BYTE2;
	stPDOConfig.ucaDefVal[3]++; //  = TXPDO_DEFVAL_BYTE3;
	stPDOConfig.ucaDefVal[4]++; //  = TXPDO_DEFVAL_BYTE4;
	stPDOConfig.ucaDefVal[5]++; //  = TXPDO_DEFVAL_BYTE5;
	stPDOConfig.ucaDefVal[6]++; //  = TXPDO_DEFVAL_BYTE6;
    stPDOConfig.ucaDefVal[7]++; //  = TXPDO_DEFVAL_BYTE7;


    ret = CMA_InstallPDO(ucCAN1Handle,stPDOConfig,&usTXPDOHandle);

    if (ret != 0)
    {
      // error which leads to an abort of the demo program,
      // next state is 5.0)
      ucDemoState = 50;

      PrintCMAFunctReturnError(ret);
      printf("      4.2)                                                       - failed !\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else
    {
      // next state is 4.3)
      ucDemoState = 43;
      printf("      4.2) successfully                                          - done.\n");
    }
  }

  // State 4.3) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 4.3 Installing a RX-PDO                                                 */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 43)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf(" 4.3) Installing a RX-PDO.\n");
    printf("      This will be done by using the function CMA_InstallPDO().\n");
    printf("      Following configuration is specified:\n");
    printf("        - COB-ID of PDO              = %d,\n",RXPDO_COBID);
    printf("        - Transmission Type of PDO   = %d,\n",RXPDO_TT);
    printf("        - Event Notification         = %d,\n",RXPDO_EVENT_NOTIF);
    if(RXPDO_EVENT_NOTIF == 0)
    {
	  printf("          (Application will not be informed about the receive\n");
	  printf("           of PDO data)\n");
    }
    else
    {
	  printf("          (Application will be informed about the receive\n");
	  printf("           of PDO data)\n");
    }
    printf("        - Default values of PDO data = %3d %3d %3d %3d %3d %3d %3d %3d\n",
		              RXPDO_DEFVAL_BYTE0,RXPDO_DEFVAL_BYTE1,RXPDO_DEFVAL_BYTE2,RXPDO_DEFVAL_BYTE3,
					  RXPDO_DEFVAL_BYTE4,RXPDO_DEFVAL_BYTE5,RXPDO_DEFVAL_BYTE6,RXPDO_DEFVAL_BYTE7);
    printf("      Call of CMA_InstallPDO() function.\n");

    // set function parameter for CMA_InstallPDO() 
	                             // RX-PDO
    stPDOConfig.ucPDOType    = 0;
	                             // RX-PDO COB-ID
    stPDOConfig.ulCOBID      = RXPDO_COBID;
	                             // Transmission type of PDO,
	                             // = 4, PDO data are transmitted after every 4th Sync-Object
    stPDOConfig.ucTransType  = RXPDO_TT;
	                             // PDO access type
	                             // until now only DPRAM PDOs are supported (= 0) 
	stPDOConfig.ucAccType    = 0;
                                 // defines if the application should be informed about
                                 // the receive of the RX-PDO via a PDOReceived event,
	stPDOConfig.ucEventNotif = RXPDO_EVENT_NOTIF;
	                             // default values of PDO data
	stPDOConfig.ucaDefVal[0] = RXPDO_DEFVAL_BYTE0;
	stPDOConfig.ucaDefVal[1] = RXPDO_DEFVAL_BYTE1;
	stPDOConfig.ucaDefVal[2] = RXPDO_DEFVAL_BYTE2;
	stPDOConfig.ucaDefVal[3] = RXPDO_DEFVAL_BYTE3;
	stPDOConfig.ucaDefVal[4] = RXPDO_DEFVAL_BYTE4;
	stPDOConfig.ucaDefVal[5] = RXPDO_DEFVAL_BYTE5;
	stPDOConfig.ucaDefVal[6] = RXPDO_DEFVAL_BYTE6;
    stPDOConfig.ucaDefVal[7] = RXPDO_DEFVAL_BYTE7;

    ret = CMA_InstallPDO(ucCAN1Handle,stPDOConfig,&usRXPDOHandle);

    if (ret != 0)
    {
      // error which leads to an abort of the demo program,
      // next state is 5.0)
      ucDemoState = 50;

      PrintCMAFunctReturnError(ret);
      printf("      4.3)                                                       - failed !\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else
    {
      // next state is 4.4)
      ucDemoState = 44;
      printf("      4.3) successfully                                          - done.\n");
    }
  }


  // State 4.4) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 4.4 Set master and all connected slaves in state >>Operational<<.       */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 44)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf(" 4.4) Set master and all connected slaves in state >>Operational<<.\n");
    printf("      This will be done by using the function CMA_ChangeState().\n");
    printf("      Call of CMA_ChangeState() function.\n");

    ret = CMA_ChangeState(ucCAN1Handle,ALL_NODES_ADDRESSED,OPERATIONAL);

    if (ret != 0)
    {
      // error which leads to an abort of the demo program,
      // next state is 5.0)
      ucDemoState = 50;

      PrintCMAFunctReturnError(ret);
      printf("      4.4)                                                       - failed !\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else
    {
      // next state is 4.5)
      ucDemoState = 45;
      printf("      4.4) successfully                                          - done.\n");
      printf("      - Master is now in state >>Operational<<.\n");
      printf("      - Sync-Object will be sent periodically every %d us.\n",
             SYNC_CYCLE);
	  printf("      - PDO transfer is active.\n");
	  if((TXPDO_TT > 1) && (TXPDO_TT < 241))
		printf("      - TX-PDO will be sent every %dth Sync-Object.\n",TXPDO_TT);
    }
  }


  // State 4.5) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 4.5 Send a remote PDO request.                                          */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 45)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf(" 4.5) Send a remote PDO request related to the RX-PDO.\n");
    printf("      This will be done by using the function CMA_SendRemotePDO().\n");
    printf("      Call of CMA_SendRemotePDO() function.\n");

    ret = CMA_SendRemotePDO(ucCAN1Handle,usRXPDOHandle);

    if (ret != 0)
    {
      // error which leads to an abort of the demo program,
      // next state is 5.0)
      ucDemoState = 50;

      PrintCMAFunctReturnError(ret);
      printf("      4.5)                                                       - failed !\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else
    {
      // next state is 4.6)
      ucDemoState = 46;
      printf("      4.5) successfully                                          - done.\n");
    }
  }


  // State 4.6) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 4.6)  Check if we received a RX-PDO on our Remote Request.              */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 46)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf(" 4.6) Check if we received a RX-PDO on our Remote Request.\n");
    printf("      This will be done by waiting %d ms for the occurance\n",
           RXPDO_WAIT_EVENT);
    printf("      of a PDOReceived event for our RX-PDO\n");
    printf("      - Now we wait %d ms while checking the Event-Fifo for\n",
           RXPDO_WAIT_EVENT);
    printf("        a PDOReceived event.\n");

    // start timeout timer
    StartTimeoutTimer();

    do
    {

      if(CheckTimeoutTimer(RXPDO_WAIT_EVENT))
      {
        // next state is 5.0)
        ucDemoState = 50;

        printf("      - No PDOReceived event for our RX-PDO received.\n");
        printf("      4.6)                                                       - failed !\n");
        printf("-----------------------------------------------------------------\n");
        printf("--> Error occured which leads to an abort of the demo program.\n");
        printf("-----------------------------------------------------------------\n");
      }

      // read and check event-fifo for events
      ReadAndCheckEvent(); 
    }
    while((ucDemoState != 50) && (ucDemoState != 47));

    if(ucDemoState == 47)
    {
      printf("      4.6) successfully                                          - done.\n");
    }
  }

  // State 4.7) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 4.7) Update the TX-PDO data.                                            */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 47)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf(" 4.7) Update the TX-PDO data.\n");
    printf("      This will be done by using the function CMA_WritePDO().\n");
    printf("      PDO data for update are: %3d %3d %3d %3d %3d %3d %3d %3d.\n",
                  TXPDO_UPD_BYTE0,TXPDO_UPD_BYTE1,TXPDO_UPD_BYTE2,TXPDO_UPD_BYTE3,
                  TXPDO_UPD_BYTE4,TXPDO_UPD_BYTE5,TXPDO_UPD_BYTE6,TXPDO_UPD_BYTE7);
    printf("      Call of CMA_WritePDO() function.\n");

    TXPDO_UpdateData[0] = TXPDO_UPD_BYTE0;
    TXPDO_UpdateData[1] = TXPDO_UPD_BYTE1;
    TXPDO_UpdateData[2] = TXPDO_UPD_BYTE2;
    TXPDO_UpdateData[3] = TXPDO_UPD_BYTE3;
    TXPDO_UpdateData[4] = TXPDO_UPD_BYTE4;
    TXPDO_UpdateData[5] = TXPDO_UPD_BYTE5;
    TXPDO_UpdateData[6] = TXPDO_UPD_BYTE6;
    TXPDO_UpdateData[7] = TXPDO_UPD_BYTE7;

    ret = CMA_WritePDO(ucCAN1Handle,usTXPDOHandle,0,8,TXPDO_UpdateData);

    if (ret != 0)
    {
      // error which leads to an abort of the demo program,
      // next state is 5.0)
      ucDemoState = 50;

      PrintCMAFunctReturnError(ret);
      printf("      4.7)                                                      - failed !\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else
    {
      // next state is 4.8)
      ucDemoState = 48;
      printf("      4.7) successfully                                          - done.\n");
      printf("      - TX-PDO should now be transmitted with the new data values.\n");

    }
  }

  // State 4.8) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 4.8) Set master and all connected slaves in state >>Pre-Operational<<.  */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 48)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf(" 4.8) Set the master and all connected slaves in state >>Pre-Operational<<.\n");
    printf("      This will be done by using the function CMA_ChangeState().\n");

    ret = CMA_ChangeState(ucCAN1Handle,ALL_NODES_ADDRESSED,PRE_OPERATIONAL);

    if (ret != 0)
    {
      // error which leads to an abort of the demo program,
      // next state is 5.0)
      ucDemoState = 50;

      PrintCMAFunctReturnError(ret);
      printf("      4.8)                                                       - failed !\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else
    {
      // Read all remaining Guard error events from the event-fifo especially
      // if guarding failed a step before. Otherwise we get unexpected 
      // events when calling function ReadAndCheckEvent() in the next state.
      ReadAndCheckEvent();

      // next state is 5.0)
      ucDemoState = 50;
      printf("      4.8) successfully                                          - done.\n");
      printf("      - Master is now in state >>Pre-Operational<<.\n");
      printf("      - Transmission of the Sync-Object is stopped.\n");
      printf("      - No PDO transfer is active anymore.\n");
    }
  }


  // State 5.0) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 5.0 Shutdown of CANopen Master                                          */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 50)
  {
    // we are in state 5.0)
    ucDemoState = 50;

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf("5.0) Shutdown of CANopen Master\n");

    printf("     The shutdown will be initiated by calling the function\n");
    printf("     CMA_Shutdown().\n");
    printf("     Call of function CMA_Shutdown().\n");

    ret = CMA_Shutdown(ucCAN1Handle);

    if(ret != 0)
    {
      // error which leads to an abort of the demo program,
      // next state is 6.0)
      ucDemoState = 60;
      PrintCMAFunctReturnError(ret);

      printf("     5.0)                                                        - failed !\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else
    {
      // next state is 5.1)
      ucDemoState = 51;

      printf("     5.0) successfully                                           - initiated.\n");
    }
  }


  // State 5.1) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 5.1 Check for the completion of the Shutdown                            */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 51)
  {
    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf("5.1) Check for the completion of the Startup\n");
    printf("     This will be done by waiting %d ms for the occurance\n",SHUTDOWN_TIME);
    printf("     a ShutdownCompleted event.\n");
    printf("     - Now we wait %d ms while checking the Event-Fifo for\n",
           SHUTDOWN_TIME);
    printf("       ShutdownCompleted event.\n");

    // start timeout timer
    StartTimeoutTimer();

    do
    {
      // check a timeout
      if(CheckTimeoutTimer(SHUTDOWN_TIME))
      {
        // error which leads to an abort of the demo program,
        // next state is 6.0)
        ucDemoState = 60;

        printf("     - Shutdown of CANopen Master timed out !\n");
        printf("     5.1) Shutdown of CANopen Master                             - failed !\n");

        printf("-----------------------------------------------------------------\n");
        printf("--> Error occured which leads to an abort of the demo program.\n");
        printf("-----------------------------------------------------------------\n");
      }

      // read and check event-fifo for events
      ReadAndCheckEvent(); 
    }
    while((ucDemoState != 52) && (ucDemoState != 60));
  }

  // State 5.2) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 5.2 Dummy state                                                         */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 52)
  {
     printf("     5.1) Shutdown of CANopen Master successfully                - done.\n");

    // next state is 6.0)
    ucDemoState = 60;
  }


  // State 6) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 6.0 Free CAN interface ressources                                       */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 60)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf("6.0) Freeing of CAN interface resources.\n");

    printf("     This will be done by calling the function\n");
    printf("     CMA_CloseCard().\n");
    printf("     Call of function CMA_CloseCard().\n");

    ret = CMA_CloseCard(ucCAN1Handle);

    if(ret != 0)
    {
      PrintCMAFunctReturnError(ret);
      printf("     6.0)                                                        - failed !\n");
      // next state is 6.1)
      ucDemoState = 61;

    }
    else
    {
      printf("     6.0) successfull                                            - done.\n");
      // next state is 6.1)
      ucDemoState = 61;
    }
  }

  // State 6.1) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 6.1 End of program                                                      */
  /* ----------------------------------------------------------------------- */
  printf("\nEnd of program. -> Press any key.\n");
  getch();

  return(0);
}

/* ------------------------------------------------------------------------- */
/* Function: ReadAndCheckEvent()                                             */
/*           - Reads new events from the event-fifo. Dependend on the        */
/*             current program state it branches to evaluation routines      */
/*             related to each the state.                                    */
/*                                                                           */
/* Output:   ret                                                             */
/*           - 0 success                                                     */ 
/* ------------------------------------------------------------------------- */
short ReadAndCheckEvent(void)
{

  CMA_EventStruct  stEvent;
  unsigned short   ret = 0;

  ret = CMA_ReadEvent (ucCAN1Handle,&stEvent);
  if(ret)
  {
    PrintCMAFunctReturnError(ret);
  }
  else
  {
    if(stEvent.ucEventType != 0)
    {
      switch(ucDemoState)
      {
        case 31: // Function to evaluate events in state 3.1)
                 // of the demo program.
          ret = EvaluateEventInState_31(stEvent);
          break;
        case 32: // Function to evaluate events in state 3.2)
                 // of the demo program.
          ret = EvaluateEventInState_32(stEvent);
          break;
        case 46: // Function to evaluate events in state 4.6)
                 // of the demo program.
          ret = EvaluateEventInState_46(stEvent);
          break;
        case 51: // Function to evaluate events in state 5.1)
                 // of the demo program.
          ret = EvaluateEventInState_51(stEvent);
          break;
        default:
          PrintUnexpectedEvent(stEvent);
        break;
      }
    }
  }
  return(ret);
}

/* ------------------------------------------------------------------------- */
/* Function: EvaluateEventInState_31()                                       */
/*           - Function evaluates events in state 3.1) of the demo program.  */
/*             Called from function ReadAndCheckEvent().                     */
/*                                                                           */
/* Input:    stEvent                                                         */
/*           - last read event                                               */
/* Output:   ret                                                             */
/*           - 0 success                                                     */
/* ------------------------------------------------------------------------- */
short EvaluateEventInState_31(CMA_EventStruct stEvent)
{
  unsigned short   ret = 0;

  switch(stEvent.ucEventType)
  {
    case StartupCompleted:
      // next state is 3.2)
      ucDemoState = 32;
      printf("     - StartupCompleted event received.\n");
      break;
    default:
      PrintUnexpectedEvent(stEvent);
      break;
  }
  return(ret);
}

/* ------------------------------------------------------------------------- */
/* Function: EvaluateEventInState_32()                                       */
/*           - Function evaluates events in state 3.2) of the demo program.  */
/*             Called from function ReadAndCheckEvent().                     */
/*                                                                           */
/* Input:    stEvent                                                         */
/*           - last read event                                               */
/* Output:   ret                                                             */
/*           - 0 success                                                     */
/* ------------------------------------------------------------------------- */
short EvaluateEventInState_32 (CMA_EventStruct stEvent)
{
  unsigned short   ret = 0;
  unsigned short   usTempErrCode;

  switch(stEvent.ucEventType)
  {
    case ErrorEvent:
      switch(stEvent.ucaEventData[0])
      {
        case CANError:
          // copy low part of Error Code
          usTempErrCode = (unsigned short)stEvent.ucaEventData[2];
          // copy high part of Error Code
          usTempErrCode |= (unsigned short)stEvent.ucaEventData[3] << 8;
          if(usTempErrCode == 0x0102)
          {
            // error which leads to an abort of the demo program,
            // next state is 5.0)
            ucDemoState = 50;
            printf("     - CANError event 0x%x received.\n",usTempErrCode);
            printf("       - CAN controller went ERROR PASSIVE !\n");
          }
          else if(usTempErrCode == 0x0101)
          {
            // error which leads to an abort of the demo program,
            // next state is 5.0)
            ucDemoState = 50;
            printf("     - CANError event 0x%x received.\n",usTempErrCode);
            printf("       - CAN controller went BUS OFF !\n");
          }
          break;
        default:
          {
            PrintUnexpectedEvent(stEvent);
          }
          break;
      }
      break;
    default:
      {
        PrintUnexpectedEvent(stEvent);
      }
      break;
  }
  return(ret);
}


/* ------------------------------------------------------------------------- */
/* Function: EvaluateEventInState_46()                                       */
/*           - Function evaluates events in state 4.6) of the demo program.  */
/*             Called from function ReadAndCheckEvent().                     */
/*                                                                           */
/* Input:    stEvent                                                         */
/*           - last read event                                               */
/* Output:   ret                                                             */
/*           - 0 success                                                     */
/* ------------------------------------------------------------------------- */
short EvaluateEventInState_46 (CMA_EventStruct stEvent)
{
  unsigned short   ret=0;
  unsigned short   usTempHandle;
  int              i;

  switch(stEvent.ucEventType)
  {
    case PDOReceived:
      // copy low part of PDO handle
      usTempHandle = (unsigned short)stEvent.ucaEventData[0];
      // copy high part of PDO handle
      usTempHandle |= (unsigned short)stEvent.ucaEventData[1] << 8;

      // Are the received data related to our RX-PDO?
	  if(usTempHandle == usRXPDOHandle)
      {
        // next state is 4.7)
        ucDemoState = 47;
        printf("      - Event PDOReceived occured.\n");
        printf("        - NumOfBytes: %d Data: ",stEvent.ucaEventData[2]);
        for(i=0; i<stEvent.ucaEventData[2]; i++)
          printf("%d ",stEvent.ucaEventData[i+3]);
        printf("\n");
      }
      else
        PrintUnexpectedEvent(stEvent);
      break;
    default:
      PrintUnexpectedEvent(stEvent);
      break;
  }
  return(ret);
}


/* ------------------------------------------------------------------------- */
/* Function: EvaluateEventInState_51()                                       */
/*           - Function evaluates events in state 5.1) of the demo program.  */
/*             Called from function ReadAndCheckEvent().                     */
/*                                                                           */
/* Input:    stEvent                                                         */
/*           - last read event                                               */
/* Output:   ret                                                             */
/*           - 0 success                                                     */
/* ------------------------------------------------------------------------- */
short EvaluateEventInState_51(CMA_EventStruct stEvent)
{
  unsigned short   ret = 0;

  switch(stEvent.ucEventType)
  {
    case ShutdownCompleted:
      // next state is 5.1)
      ucDemoState = 52;
      printf("     - ShutdownCompleted event received.\n");
      break;
    default:
      PrintUnexpectedEvent(stEvent);
      break;
  }
  return(ret);
}


/* ------------------------------------------------------------------------- */
/* Function: PrintUnexpectedEvent()                                          */
/*           - Function evaluates each possible event and prints it out.     */
/*             Called if an event is read from event-fifo which was not      */
/*             expected.                                                     */
/*                                                                           */
/* Input:    stEvent                                                         */
/*           - last read event                                               */
/* ------------------------------------------------------------------------- */
void PrintUnexpectedEvent(CMA_EventStruct stEvent)
{

  int              i;
  unsigned short   usTempCode;

  printf("-----------------------------------------------------------------\n");
  printf("--> Unexpected event received ! : \n");

  switch(stEvent.ucEventType)
  {
    case DownloadCompleted:
      printf("--> Event: DownloadCompleted:\n");
      // copy low part of Index
      usTempCode = (unsigned short)stEvent.ucaEventData[2];
      // copy high part of Index
      usTempCode |= (unsigned short)stEvent.ucaEventData[3] << 8;
      printf("-->        NodeID: %d, Index: 0x%x, ",stEvent.ucaEventData[0],
             usTempCode);
      printf("SubIndex: %d\n",stEvent.ucaEventData[1]);
      break;
    case UploadCompleted:
      printf("--> Event: UploadCompleted:\n");
      // copy low part of Index
      usTempCode = (unsigned short)stEvent.ucaEventData[2];
      // copy high part of Index
      usTempCode |= (unsigned short)stEvent.ucaEventData[3] << 8;
      printf("-->        NodeID: %d, Index: 0x%x, ",stEvent.ucaEventData[0],
             usTempCode);
      printf("SubIndex: %d, \n",stEvent.ucaEventData[1]);
      if(stEvent.ucaEventData[4] != 0xFF)
      {
        // expedited SDO transfer data is included
        printf("-->        DLC: %d, data: ",stEvent.ucaEventData[4]);
        for(i=0; i < stEvent.ucaEventData[4]; i++)
          printf("%d ",stEvent.ucaEventData[5+i]);
        printf("\n");
      }
      break;
    case EMCYReceived:
      printf("--> Event: EMCYReceived:\n");
      printf("-->        NodeID: %d, Data: ",stEvent.ucaEventData[0]);
      for(i=0; i<8; i++)
        printf("%d ",stEvent.ucaEventData[i+1]);
      printf("\n");
      break;
    case ErrorEvent:
      printf("--> Event: ErrorEvent:\n");
      switch(stEvent.ucaEventData[0])
      {
        case APIError:
          // copy low part of Error Code
          usTempCode = (unsigned short)stEvent.ucaEventData[2];
          // copy high part of Error Code
          usTempCode |= (unsigned short)stEvent.ucaEventData[3] << 8;
          printf("-->        EventType: APIError, ErrCode: 0x%x\n",usTempCode);
          break;
        case CANError:
          // copy low part of Error Code
          usTempCode = (unsigned short)stEvent.ucaEventData[2];
          // copy high part of Error Code
          usTempCode |= (unsigned short)stEvent.ucaEventData[3] << 8;
          printf("-->        EventType: CANError, ErrCode: 0x%x\n",usTempCode);
          break;
        case SDOError:
          printf("-->        EventType: SDOError, NodeID: %d,\n",stEvent.ucaEventData[1]);
          // copy low part of Additional Code
          usTempCode = (unsigned short)stEvent.ucaEventData[4];
          // copy high part of Additional Code
          usTempCode |= (unsigned short)stEvent.ucaEventData[5] << 8;
          printf("-->        ErrClass-ErrCode-AddCode: %d-%d-%d\n",stEvent.ucaEventData[2],
                 stEvent.ucaEventData[3],usTempCode);
          break;
        case GuardError:
          printf("-->        EventType: GuardError, NodeID: %d,\n",stEvent.ucaEventData[1]);
          printf("-->        ErrCode: %d\n",stEvent.ucaEventData[2]);
          break;
        default:
          printf("-->        Unknown error event type, value: %d received.\n",
                 stEvent.ucaEventData[0]);
          break;
      }
      break;
    case NextDownloadSegment:
      printf("--> Event: NextDownloadSegment:\n");
      // copy low part of Index
      usTempCode = (unsigned short)stEvent.ucaEventData[2];
      // copy high part of Index
      usTempCode |= (unsigned short)stEvent.ucaEventData[3] << 8;
      printf("-->        NodeID: %d, Index: 0x%x, ",stEvent.ucaEventData[0],
             usTempCode);
      printf("SubIndex: %d\n",stEvent.ucaEventData[1]);
      break;
    case NextUploadSegment:
      printf("--> Event: NextUploadSegment:\n");
      // copy low part of Index
      usTempCode = (unsigned short)stEvent.ucaEventData[2];
      // copy high part of Index
      usTempCode |= (unsigned short)stEvent.ucaEventData[3] << 8;
      printf("-->        NodeID: %x, Index: 0x%x, ",stEvent.ucaEventData[0],
             usTempCode);
      printf("SubIndex: %d\n",stEvent.ucaEventData[1]);
      break;
    case PDOReceived:
      printf("--> Event: PDOReceived:\n");
      // copy low part of PDO handle
      usTempCode = (unsigned short)stEvent.ucaEventData[0];
      // copy high part of PDO handle
      usTempCode |= (unsigned short)stEvent.ucaEventData[1] << 8;
      printf("-->        PDOHandle: %d, NumOfBytes: %d\n",usTempCode,
             stEvent.ucaEventData[2]);
      printf("-->        Data: ");
      for(i=0; i<stEvent.ucaEventData[2]; i++)
        printf("%d ",stEvent.ucaEventData[i+3]);
      printf("\n");
      break;
    case ShutdownCompleted:
      printf("--> Event: ShutdownCompleted\n");
      break;
    case StartupCompleted:
      printf("--> Event: StartupCompleted\n");
      break;
    case NoEvent:
      break;
    default:
      printf("--> Unknown event type, value %d received.\n",stEvent.ucEventType);
      break;
  }
  printf("-----------------------------------------------------------------\n");
}

/* ------------------------------------------------------------------------- */
/* Function: StartTimeoutTimer()                                             */
/*           - Starts the timeout timer. A timeout Timer will be checked     */
/*             with funtion CheckTimeoutTimer()                              */
/*                                                                           */
/* ------------------------------------------------------------------------- */
void StartTimeoutTimer (void)
{
  ulTimerStartTime = GetTickCount();
}

/* ------------------------------------------------------------------------- */
/* Function: CheckTimeoutTimer()                                             */
/*           - Checks if the timeout timer timed out. The Timer will be      */
/*             started with function StartTimeoutTimer().                    */
/*                                                                           */
/* Input:    ulTimeoutInMs                                                   */
/*           - Timeout value in ms.                                          */
/* ------------------------------------------------------------------------- */
short CheckTimeoutTimer (unsigned long ulTimeoutInMs)
{

  unsigned long ulActTime=0, ulDiffTime=0;

  ulActTime  = GetTickCount();
  ulDiffTime = (ulActTime >= ulTimerStartTime) ?
               ulActTime - ulTimerStartTime :
               (0xFFFFFFFFL - ulTimerStartTime) + ulActTime + 1;

  /* check timeout */
  if(ulDiffTime > ulTimeoutInMs)
  {
    // timeout occured
    return(-1);
  }
  else
    // no timeout occured yet
    return(0);
}

/* ------------------------------------------------------------------------- */
/* Function: PrintCMAFunctReturnError()                                      */
/*           - Prints function return codes in a textual form.               */
/*                                                                           */
/* Input:    ret                                                             */
/*           - Function return code                                          */
/* ------------------------------------------------------------------------- */
void PrintCMAFunctReturnError(int ret)
{
  switch(ret)
  {
    case CANPC_RB_INI_FILE:
      printf("      - CMA function error:\n");
      printf("        Can't open INI-file (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_FMT_INI:
      printf("      - CMA function error:\n");
      printf("        Format error in INI-file (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_OP_BIN:
      printf("      - CMA function error:\n");
      printf("        Error in opening binary-file (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_RD_BIN:
      printf("      - CMA function error:\n");
      printf("        Error in error reading binary-file (code: %d).\n",ret);
      break;
    case CANPC_RB_BIN_TOO_LONG:
      printf("      - CMA function error:\n");
      printf("        Binary-file too long (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_BIN_FMT:
      printf("      - CMA function error:\n");
      printf("        Binary-data format error (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_BIN_CS:
      printf("      - CMA function error:\n");
      printf("        Binary-data checksum error (code: %d).\n",ret);
      break;
    case CANPC_RB_NO_CARD:
      printf("      - CMA function error:\n");
      printf("        No interface present (code: %d).\n",ret);
      break;
    case CANPC_RB_NO_PHYS_MEM:
      printf("      - CMA function error:\n");
      printf("        No physical memory available (code: %d).\n",ret);
      break;
    case CANPC_RB_INVLD_IRQ:
      printf("      - CMA function error:\n");
      printf("        Invalid IRQ-number (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_DPRAM_ACCESS:
      printf("      - CMA function error:\n");
      printf("        Error in accessing dpram (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_CRD_RESP:
      printf("      - CMA function error:\n");
      printf("        Bad response from interface (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_SRAM:
      printf("      - CMA function error:\n");
      printf("        SRAM seems to be damaged (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_PRG:
      printf("      - CMA function error:\n");
      printf("        Invalid program start address (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_REC:
      printf("      - CMA function error:\n");
      printf("        Invalid record type (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_NORESP:
      printf("      - CMA function error:\n");
      printf("        No response after program start (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_BADRESP:
      printf("      - CMA function error:\n");
      printf("        Bad response after program start (code: %d).\n",ret);
      break;
    case CANPC_RB_PCMCIA_NSUPP:
      printf("      - CMA function error:\n");
      printf("        PCMCIA chip not supported (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_RD_PCMCIA:
      printf("      - CMA function error:\n");
      printf("        Error in reading PCMCIA parameters (code: %d).\n",ret);
      break;
    case CANPC_RB_INIT_CHIP:
      printf("      - CMA function error:\n");
      printf("        Error in initialising PCMCIA chip (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_NO_CANCARD:
      printf("      - CMA function error:\n");
      printf("        No CANcard plugged in (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_CRD_NORESP:
      printf("      - CMA function error:\n");
      printf("        Bad response from interface (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_DEV_LD:
      printf("      - CMA function error:\n");
      printf("        Interface not properly loaded (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_LICENCE:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucLicence invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_DEMOS_EXPIRED:
      printf("      - CMA function error:\n");
      printf("        Demo applications expired (code: %d).\n",ret);
      break;
    case CMA_ERR_LICENCE_ACCESS:
      printf("      - CMA function error:\n");
      printf("        Licence access impossible (code: %d).\n",ret);
      break;
    case CMA_ERR_FUNCTION_TIMEOUT:
      printf("      - CMA function error:\n");
      printf("        Function timeout occurred (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_FUNCTIONALITY:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucCANxFunctionality undefined (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_CHAN_HANDLE:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucChannelHandle undefined (code: %d).\n",ret);
      break;
    case CMA_ERR_SCT_BAUD_RATE:
      printf("      - CMA function error:\n");
      printf("        Struct parameter: ucBaudRate invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_TIMEOUT_MAX:
      printf("      - CMA function error:\n");
      printf("        Parameter: timeout too big (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_TIMEOUT_MIN:
      printf("      - CMA function error:\n");
      printf("        Parameter: timeout too small (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_NODE_ID:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucNodeID is invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_SCT_COB_ID:
      printf("      - CMA function error:\n");
      printf("        Struct parameter: COB-ID invalid or used (code: %d).\n",ret);
      break;
    case CMA_ERR_SCT_COMM_CYC_MAX:
      printf("      - CMA function error:\n");
      printf("        Struct parameter: ulSyncCycle is too big (code: %d).\n",ret);
      break;
    case CMA_ERR_SCT_NODE_ID:
      printf("      - CMA function error:\n");
      printf("        Struct parameter: ucNodeID invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_MASTER_STATE:
      printf("      - CMA function error:\n");
      printf("        Improper state of the master (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_STATE:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucState invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_INT_TX_FIFO_FULL:
      printf("      - CMA function error:\n");
      printf("        Internal transmit fifo is full (code: %d).\n",ret);
      break;
    case CMA_ERR_MASTER_GUARDING:
      printf("      - CMA function error:\n");
      printf("        Master cannot guard itself (code: %d).\n",ret);
      break;
    case CMA_ERR_COMMAND_FIFO_FULL:
      printf("      - CMA function error:\n");
      printf("        DPRAM command fifo is full (code: %d).\n",ret);
      break;
    case CMA_ERR_SDO_TX_FIFO_FULL:
      printf("      - CMA function error:\n");
      printf("        SDO transmit fifo is full (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_DATA_SIZE:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucDataSize is invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_INDEX:
      printf("      - CMA function error:\n");
      printf("        Parameter: usIndex is incorrect (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_SUBINDEX:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucSubIndex is incorrect (code: %d).\n",ret);
      break;
    case CMA_ERR_TOTAL_DATA_SIZE:
      printf("      - CMA function error:\n");
      printf("        Parameter: ulTotalDataSize exceeded (code: %d).\n",ret);
      break;
    case CMA_ERR_NO_TRANSF_ACTIVE:
      printf("      - CMA function error:\n");
      printf("        No normal SDO transfer active (code: %d).\n",ret);
      break;
    case CMA_ERR_SDO_RX_FIFO_EMPTY:
      printf("      - CMA function error:\n");
      printf("        SDO receive fifo is empty (code: %d).\n",ret);
      break;
    case CMA_ERR_SCT_PDO_ACC_TYPE:
      printf("      - CMA function error:\n");
      printf("        Struct parameter: ucAccType is invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_PDO_HANDLE:
      printf("      - CMA function error:\n");
      printf("        Parameter: usPDOHandle is undefined (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_BYTEPOS:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucDataOffset is invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_PDO_TOO_LONG:
      printf("      - CMA function error:\n");
      printf("        Parameter: DataOffset+Datalength>8 (code: %d).\n",ret);
      break;
    case CMA_ERR_PDO_ACCESS:
      printf("      - CMA function error:\n");
      printf("        Access to PDO in DPRAM failed (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_BITPOS:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucBitPosition is invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_UNKNOWN_EVENT:
      printf("      - CMA function error:\n");
      printf("        Unknown event in event fifo (code: %d).\n",ret);
      break;
    case CMA_ERR_WRONG_FW_VERSION:
      printf("      - CMA function error:\n");
      printf("        Wrong firmware version (code: %d).\n",ret);
      break;
    case CMA_ERR_SCT_TRANS_TYPE:
      printf("      - CMA function error:\n");
      printf("        Struct parameter: ucTransType invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_DPRAM_TX_PDO_MAX:
      printf("      - CMA function error:\n");
      printf("        Number of DPRAM-TX-PDOs exceeded (code: %d).\n",ret);
      break;
    case CMA_ERR_DPRAM_RX_PDO_MAX:
      printf("      - CMA function error:\n");
      printf("        Number of DPRAM-RX-PDOs exceeded (code: %d).\n",ret);
      break;
    case CMA_ERR_LAST_REQ_PENDING:
      printf("      - CMA function error:\n");
      printf("        Last request still pending (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_TOOMANYDEVICES:
      printf("      - CMA function error:\n");
      printf("        Maximum number of device exceeded (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_NOTENOUGHMEMORY:
      printf("      - CMA function error:\n");
      printf("        Not enough memory space to register a device (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_DRIVERNOTFOUND:
      printf("      - CMA function error:\n");
      printf("        Driver not found (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_DRIVERVERSIONERROR:
      printf("      - CMA function error:\n");
      printf("        Wrong driver version (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_VERSIONERROR:
      printf("      - CMA function error:\n");
      printf("        Wrong driver-dll version (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_NOTSUPPORTED:
      printf("      - CMA function error:\n");
      printf("        Internal error: Requested functionality not supported\n");
      printf("        (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_ILLEGALCALL:
      printf("      - CMA function error:\n");
      printf("        Internal error: Illegal driver call (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_IOCANCELLED:
      printf("      - CMA function error:\n");
      printf("        Internal error: Driver call cancelled (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_IOPENDING:
      printf("      - CMA function error:\n");
      printf("        Internal error: Driver call pending (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_TIMEOUT:
      printf("      - CMA function error:\n");
      printf("        Internal error: Timeout during driver access (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_GENERALERROR:
      printf("      - CMA function error:\n");
      printf("        Internal error: General error in accessing the driver (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_INTERNALERROR:
      printf("      - CMA function error:\n");
      printf("        Internal driver error (code: %d).\n",ret);
      break;
    case INIPC_IB_PNP_NO_DEVICE_FOUND:
      printf("      - CMA function error:\n");
      printf("        No CAN interface found (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_UNKNOWNDEVICE:
      printf("      - CMA function error:\n");
      printf("        Access to an unknown device requested (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_DEVICEALREADYEXISTS:
      printf("      - CMA function error:\n");
      printf("        Device is already registered by the driver (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_DEVICEACCESSERROR:
      printf("      - CMA function error:\n");
      printf("        Device is already open (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_RESOURCEALREADYREGISTERED:
      printf("      - CMA function error:\n");
      printf("        A requested resource is already in use \n");
      printf("        by an other device (DPRAM, IRQ, I/O port) (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_RESOURCECONFLICT:
      printf("      - CMA function error:\n");
      printf("        Resource conflict with an other driver (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_RESOURCEACCESSERROR:
      printf("      - CMA function error:\n");
      printf("        Resource access error (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_PHYSMEMORYOVERRUN:
      printf("      - CMA function error:\n");
      printf("        Internal error: Invalid phys. mem-access (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_TOOMANYPORTS:
      printf("      - CMA function error:\n");
      printf("        Internal error: Too many I/O ports requested (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_UNKNOWNRESOURCE:
      printf("      - CMA function error:\n");
      printf("        Internal error: Unknown resource access requested (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_LOADDDLL_DEV:
      printf("      - CMA function error:\n");
      printf("        Driver DLL not found (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_CANT_DEVREGISTER:
      printf("      - CMA function error:\n");
      printf("        Interface cannot be registered (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_IRQ:
      printf("      - CMA function error:\n");
      printf("        Interrupt number invalid or device cannot\n");
      printf("        be assigned an interrupt (code: %d).\n",ret);
      break;
    case DDC_ERR_WRONG_DEVICEHANDLE:
      printf("      - CMA function error:\n");
      printf("        Internal error: Wrong device handle specified\n");
      printf("        (code: %d).\n",ret);
      break;
    default:
      printf("      - Unknown CMA function error (code: %d).\n",ret);
      break;
  }
}

