/*****************************************************************************/
/*                                                                           */
/*   Softing Industrial Automation GmbH   Richard-Reitzner-Allee 6   85540 Haar         */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*  Copyright (c) Softing Industrial Automation GmbH, All rights reserved.   */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*                        S D O U P L O A D . C                              */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* PROJECT:            CANopen Client API C-Samples                          */
/*                                                                           */
/* MODULE:             SDOUpload.c                                           */
/*                                                                           */
/* AUTHOR:             Softing Industrial Automation GmbH                               */
/*                                                                           */
/* DATE                01.04.2012                                           */
/*                                                                           */
/*****************************************************************************/
/* MODULE_DESCRIPTION                                                        */ 
/*                                                                           */
/*  GENERAL:       This is a demo application to show how to use             */
/*                 CANopen Client API to perform a SDO Upload.               */
/*                                                                           */
/*  CANOPEN                                                                  */
/*    BACKGROUND   With Service Data Objects (SDOs) the access to entries of */
/*                 a Object Dictionary (OD) is provided.                     */
/*                 It is controlled via a multiplexor (Index and sub-index   */
/*                 of the OD entry) which data set is to be transferred.     */
/*                 The content of the data is defined within OD.             */
/*                 For a SDO transfer it is distinguished between an         */
/*                 initialisation phase and the phase of sending SDO         */
/*                 segments. Basically a SDO is transferred as a sequence of */
/*                 segments (Normal SDO transfer) but it is also possible to */
/*                 transfer a data set of up to 4 bytes during the           */
/*                 initialisation phase (Expedited SDO transfer).            */
/*                 During a SDO upload the CANopen Master functions as the   */
/*                 client and reads data from the OD of the server (slave).  */
/*                 For more detailed information about SDOs please refer to  */
/*                 CiA DS 301 V3.0 (CANopen Application Layer and            */
/*                 Communication Profile)                                    */
/*                                                                           */
/*  CONTENT:       In the demo the master performs one SDO upload from a     */
/*                 slave. Because the CANopen Master uses only Default SDOs  */
/*                 according to the Pre-Defined Connection Set an SDO        */
/*                 transfer is referenced by the node-ID of the addressed    */
/*                 slave.                                                    */
/*                 The configuration of the SDO upload can be done by        */
/*                 setting constants in the >Constants to be changed by the  */
/*                 user< section below. Following default configuration is   */
/*                 given at the delivery of this demo program:               */
/*                 - Node-ID of addressed slave(SLAVE_NODE_ID)    = 1        */
/*                 - OD index (UPLOAD_INDEX)                      = 0x1000   */
/*                                                             (device type) */
/*                 - OD sub-index (UPLOAD_SUB_INDEX)              = 0x0      */
/*                 - Timeout value for SDO transfers(SDO_TIMEOUT) = 100 ms   */
/*                 - Application timeout factor for SDO                      */
/*                   transfers(APPL_TIMEOUT_FACTOR)               = 1        */
/*                 The upload data is hold in the global data array          */
/*                 ucaUploadData[UPLOAD_MDS].                                */
/*                                                                           */
/*  DEMO STRUCTURE The demo is maintaned by a state machine. The global      */
/*                 variable <ucDemoState> contains the current state.        */
/*                 Following a short description of the state machine        */
/*                 especially which action is performed in each state:       */
/*                                                                           */
/*                   state 1.0:  Initialising the Softing CAN interface to   */
/*                               run CANopen Client API                      */
/*                                 - success -> state 2.1                    */
/*                                 - failure -> state 6.1                    */
/*                                                                           */
/*                   states 2.x: Initialising the CANopen Master             */
/*                   state 2.1:  Initialising CAN parameters                 */
/*                                 - success -> state 2.2                    */
/*                                 - failure -> state 6.0                    */
/*                   state 2.2:  Initialising the Node manager               */
/*                                 - success -> state 2.3                    */
/*                                 - failure -> state 6.0                    */
/*                   state 2.3:  Initialising the SDO manager                */
/*                                 - success -> state 3.0                    */
/*                                 - failure -> state 6.0                    */
/*                                                                           */
/*                   state 3.0:  Initiate Startup of CANopen Master          */
/*                                 - success -> state 3.1                    */
/*                                 - failure -> state 6.0                    */
/*                   state 3.1:  Check for the completion of the Startup.    */
/*                                 - success -> state 3.2                    */
/*                                 - failure -> state 6.0                    */
/*                   state 3.2:  Check if CANopen Master could send its      */
/*                               bootup message.                             */
/*                                 - success -> state 4.1                    */
/*                                 - failure -> state 5.0                    */
/*                                                                           */
/*                   states 4.x: CANopen Master is active.                   */
/*                   state 4.1:  Initiate a SDO Upload                       */
/*                                 - success -> state 4.2                    */
/*                                 - failure -> state 5.0                    */
/*                   state 4.2:  Check for occurance of the events           */
/*                               UploadCompleted and NextUploadSegment and   */
/*                               read next Upload segments from SDO-RX-Fifo. */
/*                                 - success -> state 4.3                    */
/*                                 - failure -> state 5.0                    */
/*                   state 4.3:  Dummy state                                 */
/*                                 - next    -> state 5.0                    */
/*                                                                           */
/*                   state 5.0:  Initiate Shutdown of CANopen Master         */
/*                                 - success -> state 5.1                    */
/*                                 - failure -> state 6.0                    */
/*                   state 5.1:  Check for the completion of the Shutdown.   */
/*                                 - success -> state 5.2                    */
/*                                 - failure -> state 6.0                    */
/*                   state 5.2:  Dummy state                                 */
/*                                 - next    -> state 6.0                    */
/*                                                                           */
/*                   state 6.0:  Free CAN interface ressources               */
/*                                 - success -> state 6.1                    */
/*                                 - failure -> state 6.1                    */
/*                                                                           */
/*                   state 6.1:  End of program                              */
/*                                                                           */
/*                 Each action can be configured by setting constants.       */
/*                 Please refer to the >Constants to be changed by the user< */
/*                 section below.                                            */
/*                                                                           */
/*****************************************************************************/   
#include <stdio.h>
#include <sys/time.h>
#include <string.h>

// CANopen Master API include files 
#include "kbhit.h"
#include "cmadef.h"
#include "cmaret.h"

/*****************************************************************************/
/*                                                                           */
/* Constants to be changed by the user                                       */
/*                                                                           */
/*****************************************************************************/
// constants used in state 2.1:
                              // baud rate, possible values are:
                              //  - 0: 1   Mbit/s
                              //  - 1: 800 Kbit/s
                              //  - 2: 500 Kbit/s
                              //  - 3: 250 Kbit/s
                              //  - 4: 125 Kbit/s
                              //  - 5: 50  Kbit/s
                              //  - 6: 20  Kbit/s
                              //  - 7: 10  Kbit/s
#define BAUD_RATE             0

// constants used in state 2.2:
                              // node-ID of CANopen Master
#define MASTER_NODE_ID        127

// constants used in state 3.0:
                              // timeout in ms for the master to startup
#define STARTUP_TIME          10000

// constants used in state 3.1:
                              // time we wait after startup if the 
                              // CAN controller changes it's bus state,
                              // indication that bootup message could
                              // be sent
#define BOOT_MSG_TIME         10

// constants used in state 4.1, 4.2 and 4.3:
                              // node-ID of the slave to be addressed by
                              // the SDO Download
#define SLAVE_NODE_ID         1


// example for segmented upload
                              // Index of object which is addressed by the
                              // SDO Upload
                              // manufacturer device name
#define UPLOAD_INDEX          0x1008 

                              // Sub-Index of object which is addressed by the
                              // SDO Upload
#define UPLOAD_SUB_INDEX      0 
                              
/*
// expedited upload example
                              // Index of object which is addressed by the
                              // SDO Upload
                              // idendity object
#define UPLOAD_INDEX          0x1018

                              // Sub-Index of object which is addressed by the
                              // SDO Upload
                              // vendor id
#define UPLOAD_SUB_INDEX      1 
                              */

                              // maximum number of bytes which the CANopen Master
                              // can store in the SDO Upload data array 
                              // (ucaDownloadData[])
                              // can store in the SDO Upload data array 
                              // (ucaDownloadData[])
#define UPLOAD_MDS            2000

                              // timeout value for SDO transfers (in ms)
#define SDO_TIMEOUT           1000
                              // Application timeout factor for SDO 
                              // transfers (related to SDO_TIMEOUT)
#define APPL_TIMEOUT_FACTOR   1
                              // timeout in ms for the master to shutdown
#define SHUTDOWN_TIME         20

/*****************************************************************************/
/*                                                                           */
/* Constants which should not be changed                                     */
/*                                                                           */
/*****************************************************************************/
                           // CAN1 runs in CANopen Master mode
#define CAN1_FUNCTIONALITY     1
                           // license to check by the master
                           // (CANopen Master mode)
#define LIC                    0


/*****************************************************************************/
/*                                                                           */
/* Global structures and variables                                           */
/*                                                                           */
/*****************************************************************************/
                           // state of the demo program
static unsigned char  ucDemoState;
                           // channel handles
static unsigned char  ucCANHandle;
                           // start time for timeout timer
static unsigned long  ulTimerStartTime;

unsigned int node_id = 1;

typedef struct 
  {
                           // number of transmitted bytes
    USIGN32 ulTransmittedBytes;
                           // data array for SDO Upload data
    BYTE ucaData[UPLOAD_MDS];
  } DEMO_UploadStruct;

static DEMO_UploadStruct     stUpload;
                           // current value of timeout value
static unsigned long  ulDemoTimeout;


/*****************************************************************************/
/*                                                                           */
/* Function prototypes                                                       */
/*                                                                           */
/*****************************************************************************/
short ReadAndCheckEvent(void);
void  PrintUnexpectedEvent(CMA_EventStruct stEvent);
short EvaluateEventInState_31(CMA_EventStruct stEvent);
short EvaluateEventInState_32(CMA_EventStruct stEvent);
short EvaluateEventInState_42 (CMA_EventStruct stEvent);
short EvaluateEventInState_51(CMA_EventStruct stEvent);
void  PrintCMAFunctReturnError(int ret);
void  StartTimeoutTimer (void);
short CheckTimeoutTimer (unsigned long ulTimeoutInMs);
void  PrintUploadedData(void);

unsigned long GetTickCount(void)
{
  unsigned long count;
  struct timeval tv;
  struct timezone tz;
  int ret;
  
  ret =  gettimeofday(&tv, &tz);
  
  if(ret) return 0;
  
  count = tv.tv_usec/1000 + tv.tv_sec*1000;
  
  return count;
}


/*****************************************************************************/
/*                                                                           */
/* Functions                                                                 */
/*                                                                           */
/*****************************************************************************/
/* ------------------------------------------------------------------------- */
/*                                                                           */
/* Main-Function                                                             */
/*                                                                           */
/* ------------------------------------------------------------------------- */
int main(int argc, char **argv)
{
  // function return codes
  short             ret;

  // structures needed for function calls in:
  // state 2.1
  CMA_CAN_CONFIG    stCANConfig;

  memset(&stCANConfig, 0, sizeof(CMA_CAN_CONFIG));

  printf("***************************************************************\n");
  printf("*           Softing Industrial Automation GmbH  2012          *\n");
  printf("*                                                             *\n");
  printf("*                     SDO Upload                              *\n");
  printf("*            (Demo Program for CANopen Client API)            *\n");
  printf("*                                                             *\n");
  printf("***************************************************************\n");

  // initialise state machine of demo program
  ucDemoState = 10;

  printf("\n");

  // State 1.0) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 1.0 Initialising the Softing CAN interface to run CANopen Client API    */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 10)
  {

    printf("1.0) Initialising the Softing CAN channel \"CANopen\" to run\n");
    printf("     CANopen Client API.\n"); 
    printf("     This will be done by using the function CMA_InitializeCard().\n");
    printf("     one channel must be renamed to CANopen to use this sample!\n");
    printf("     Call of CMA_InitializeCard() function to open the channel with the name \"CANopen\"\n");

    ret = CMA_InitializeCard (0xFFFFFFFF, // do not care
      CAN1_FUNCTIONALITY, // allways 1
      &ucCANHandle, // out: the channel handle
      LIC, // allways 0; 
      (unsigned char*)"CANopen"); // in: the channels name used to select the channel -> see SCIM
     
    
    if (ret!=0)
    { 
      // error which leads to an abort of the demo program,
      // next state is 6.1) 
      ucDemoState = 61;
      
      printf("ERROR: Please rename a CAN channel to \"CANopen\" first!\n");
      printf("You can use the SCIM to rename a channel!\n");
      
      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else 
    { 
      printf("     1.0) successfully                                           - done.\n");
      
      // next state is 2.1)
      ucDemoState = 21;
      
      printf("\n");
    }
  }
  


  // States 2.x) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 2. Initialising CANopen Master                                          */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 21)
  {
    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf("2) Initialising CANopen Master.\n");
  }

  // State 2.1) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 2.1 Initialising CAN parameters                                         */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 21)
  {
    printf(" 2.1) Initialising of CAN parameters.\n");
    printf("      This will be done by using the function\n");
    printf("      CMA_ConfigCANChannel().\n");
    printf("      Following function parameter is specified:\n");
    // print the baud rate
    switch(BAUD_RATE)
    {
      case 0:
        printf("      - baud rate = 1 MBit/s\n");
        break;
      case 1:
        printf("      - baud rate = 800 KBit/s\n");
        break;
      case 2:
        printf("      - baud rate = 500 KBit/s\n");
        break;
      case 3:
        printf("      - baud rate = 250 KBit/s\n");
        break;
      case 4:
        printf("      - baud rate = 125 KBit/s\n");
        break;
      case 5:
        printf("      - baud rate = 50 KBit/s\n");
        break;
      case 6:
        printf("      - baud rate = 20 KBit/s\n");
        break;
      case 7:
        printf("      - baud rate = 10 KBit/s\n");
        break;
      default:
        printf("      - no valid baud rate !\n");
        break;
    }

    // get CAN parameter from constants section
    stCANConfig.ucBaudRate = BAUD_RATE;
    stCANConfig.ucAccCode = 0;
    stCANConfig.ucAccMask = 0;

    printf("      Call of CMA_ConfigCANChannel() function.\n");

    ret = CMA_ConfigCANChannel (ucCANHandle,stCANConfig);

    if (ret != 0)
    {
      // error which leads to an abort of the demo program,
      // next state is 6.0)
      ucDemoState = 60;

      PrintCMAFunctReturnError(ret);
      printf("      2.1)                                                       - failed.\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else
    {
      printf("      2.1) successfully                                          - done.\n");

      // next state is 2.2)
      ucDemoState = 22;
    }
  }

  // State 2.2) of the demo program:
  /* --------------------------------------------------------------------- */
  /* 2.2 Initialising the Node manager                                     */
  /* --------------------------------------------------------------------- */
  if(ucDemoState == 22)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf(" 2.2) Initialising the Node manager.\n");
    printf("      This will be done by using the function\n");
    printf("      CMA_InitNodeMan().\n");
    printf("      Following function parameter is specified:\n");
    printf("      - Node-ID of the master = %d\n",MASTER_NODE_ID);
    printf("      Call of CMA_InitNodeMan() function.\n");

    ret = CMA_InitNodeMan(ucCANHandle,MASTER_NODE_ID);

    if (ret != 0)
    {
      // error which leads to an abort of the demo program,
      // next state is 6.0)
      ucDemoState = 60;

      PrintCMAFunctReturnError(ret);
      printf("      2.2)                                                       - failed.\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else
    {
      // next state is 3.0)
      ucDemoState = 23;

      printf("      2.2) successfully                                          - done.\n");
    }
    printf("\n");
  }

  // State 2.3) of the demo program:
  /* --------------------------------------------------------------------- */
  /* 2.3 Initialising the SDO manager                                      */
  /* --------------------------------------------------------------------- */
  if(ucDemoState == 23)
  {

    printf(" 2.3) Initialising the SDO manager.\n");
    printf("      This will be done by using the function\n");
    printf("      CMA_InitSDOMan().\n");
    printf("      Following function parameter are specified:\n");
    printf("      - SDO timeout                = %d ms\n",SDO_TIMEOUT);
    printf("      - Application timeout factor = %d\n",APPL_TIMEOUT_FACTOR);

    printf("      Call of CMA_InitSDOMan() function.\n");

    ret = CMA_InitSDOMan(ucCANHandle,SDO_TIMEOUT,APPL_TIMEOUT_FACTOR);

    if (ret != 0)
    {
      // error which leads to an abort of the demo program,
      // next state is 6.0)
      ucDemoState = 60;

      PrintCMAFunctReturnError(ret);

      printf("     2.3)                                                        - failed.\n");
      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else
    {
      // next state is 3.0)
      ucDemoState = 30;

      printf("      2.3) successfully                                          - done.\n");
    }
  }

  // State 3.0) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 3.0 Initiate Startup of CANopen Master                                  */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 30)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf("3.0) Startup of CANopen Master.\n");
    printf("     The startup will be initiated by calling the function\n");
    printf("     CMA_Startup().\n");
    printf("     Call of CMA_Startup() function.\n");

    ret = CMA_Startup (ucCANHandle);

    if(ret != 0)
    {
      // error which leads to an abort of the demo program,
      // next state is 6.0)
      ucDemoState = 60;

      PrintCMAFunctReturnError(ret);
      printf("     3.0)                                                        - failed !\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else
    {
      printf("     3.0) Startup of CANopen Master successfully                 - initiated.\n");
      // next state is 3.1)
      ucDemoState = 31;
    }
  }

  // State 3.1) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 3.1 Check for the completion of the Startup                             */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 31)
  {
    printf("\n");
    printf("3.1) Check for the completion of the Startup\n");
    printf("     This will be done by waiting %d ms for the occurance\n",STARTUP_TIME);
    printf("     a StartupCompleted event.\n");
    printf("     - Now we wait %d ms while checking the Event-Fifo for\n",
           STARTUP_TIME);
    printf("     a StartupCompleted event.\n");

    // start timeout timer
    StartTimeoutTimer();

    do
    {
      // check a timeout
      if(CheckTimeoutTimer(STARTUP_TIME))
      {
        // error which leads to an abort of the demo program,
        // next state is 6.0)
        ucDemoState = 60;

        printf("     - Startup of CANopen Master timed out !\n");
        printf("     3.1) Startup of CANopen Master                              - failed !\n");

        printf("-----------------------------------------------------------------\n");
        printf("--> Error occured which leads to an abort of the demo program.\n");
        printf("-----------------------------------------------------------------\n");
      }
      // read and check event-fifo for events
      ReadAndCheckEvent();
    }
    while((ucDemoState != 32) && (ucDemoState != 60));

    if(ucDemoState == 32)
    {
      printf("     3.1) Startup of CANopen Master successfully                 - done.\n");
      printf("          - CANopen Master is now in state >>Pre-Operational<<.\n");
    }
  }


  // State 3.2) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 3.2 Check if CANopen Master could send its bootup message               */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 32)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf("3.2) Check if CANopen Master could send it's bootup message.\n");
    printf("     This will be done by waiting %d ms for the occurance\n",BOOT_MSG_TIME);
    printf("     of a CAN Error event wich indicates a state change of the\n");
    printf("     CAN controller.\n");
    printf("     - Now we wait %d ms while checking the Event-Fifo for\n",
           BOOT_MSG_TIME);
    printf("       a CANError event, Error Code: 0x101 or 0x102.\n");

    // start timeout timer
    StartTimeoutTimer();

    do
    {
      // check a timeout of 10ms
      if(CheckTimeoutTimer(BOOT_MSG_TIME))
      {
        // next state is 4.1)
        ucDemoState = 41;
        printf("     - No CANError event received.\n");
        printf("     3.2) successfully                                           - done.\n");
      }
      // read and check event-fifo for events
      ReadAndCheckEvent();
    }
    while((ucDemoState != 41) && (ucDemoState != 50));

    if(ucDemoState == 50)
    {
      printf("       3.2)                                                      - failed !\n");
      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
  }


  // States 4x) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 4.x CANopen Master is active.                                           */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 41)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf("4) CANopen Master is active.\n");
  }


  // State 4.1) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 4.1 Initiate a SDO Upload.                                              */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 41)
  {
    printf("      please enter a node ID of an CANopen device in your network [1..127]: ");
    scanf("%u",&node_id);
    node_id &= 0x7F;


    printf(" 4.1) Initiate a SDO Upload from a object of the slave with\n");
    printf("      Node-ID = %d. To initiate the upload function \n", node_id);
    printf("      CMA_InitUpload() will be used.\n");
    printf("      Following parameters are specified for the SDO Upload:\n");
    printf("      - Index of object             = 0x%x\n",UPLOAD_INDEX);
    printf("      - Sub-Index of object         = 0x%x\n",UPLOAD_SUB_INDEX);

    // initialise variable which counts the transmitted bytes
    stUpload.ulTransmittedBytes = 0;

    printf("      Now function CMA_InitUpload() is called.\n");

    ret = CMA_InitUpload(ucCANHandle,(unsigned char)node_id,UPLOAD_INDEX,UPLOAD_SUB_INDEX);
    if(ret != 0)
    {
      // error which leads to an abort of the demo program,
      // next state is 5.0)
      ucDemoState = 50;

      PrintCMAFunctReturnError(ret);
      printf("      4.1)                                                       - failed !\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else
    {
      // next state is 4.2)
      ucDemoState = 42;
      printf("      4.1) successfully                                          - initiated.\n");
    }
  }


  // State 4.2) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 4.2 Check for occurance of the events UploadCompleted and               */
  /*     NextUploadSegment and read next Upload segments from SDO-RX-Fifo.   */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 42)
  {
    printf(" 4.2) Check for occurance of the events UploadCompleted\n");
    printf("      - Now we start with waiting %d ms (SDO timeout) while\n",
           SDO_TIMEOUT);
    printf("        for a UploadCompleted event.\n");

    // start timeout timer
    StartTimeoutTimer();
    // for the first run we wait SDO_TIMEOUT ms for a event-fifo event
    ulDemoTimeout = SDO_TIMEOUT;

    do
    {
      // check a timeout
      if(CheckTimeoutTimer(ulDemoTimeout))
      {
        printf("      - Check 4.2) timed out !\n");
        // error which leads to an abort of the demo program,
        // next state is 5.0)
        ucDemoState = 50;
      }

      // read and check event-fifo for events
      ReadAndCheckEvent();
    }
    while((ucDemoState != 43) && (ucDemoState != 50));

    if(ucDemoState == 50)
    {
      printf("      4.3)                                                       - failed !\n");
      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
  }

  // State 4.3) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 4.3 Dummy state                                                         */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 43)
  {
    printf("      4.2) successfully                                          - done.\n");
    // print uploaded data
    PrintUploadedData();

    // next state is 5.0)
    ucDemoState = 50;
  }


  // State 5.0) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 5.0 Shutdown of CANopen Master                                          */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 50)
  {
    // we are in state 5.0)
    ucDemoState = 50;

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf("5.0) Shutdown of CANopen Master\n");

    printf("     The shutdown will be initiated by calling the function\n");
    printf("     CMA_Shutdown().\n");
    printf("     Call of function CMA_Shutdown().\n");

    ret = CMA_Shutdown(ucCANHandle);

    if(ret != 0)
    {
      // error which leads to an abort of the demo program,
      // next state is 6.0)
      ucDemoState = 60;
      PrintCMAFunctReturnError(ret);

      printf("     5.0)                                                        - failed !\n");

      printf("-----------------------------------------------------------------\n");
      printf("--> Error occured which leads to an abort of the demo program.\n");
      printf("-----------------------------------------------------------------\n");
    }
    else
    {
      // next state is 5.1)
      ucDemoState = 51;

      printf("     5.0) successfully                                           - initiated.\n");
    }
  }


  // State 5.1) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 5.1 Check for the completion of the Shutdown                            */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 51)
  {
    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf("5.1) Check for the completion of the Startup\n");
    printf("     This will be done by waiting %d ms for the occurance\n",SHUTDOWN_TIME);
    printf("     a ShutdownCompleted event.\n");
    printf("     - Now we wait %d ms while checking the Event-Fifo for\n",
           SHUTDOWN_TIME);
    printf("       ShutdownCompleted event.\n");

    // start timeout timer
    StartTimeoutTimer();

    do
    {
      // check a timeout
      if(CheckTimeoutTimer(SHUTDOWN_TIME))
      {
        // error which leads to an abort of the demo program,
        // next state is 6.0)
        ucDemoState = 60;

        printf("     - Shutdown of CANopen Master timed out !\n");
        printf("     5.1) Shutdown of CANopen Master                             - failed !\n");

        printf("-----------------------------------------------------------------\n");
        printf("--> Error occured which leads to an abort of the demo program.\n");
        printf("-----------------------------------------------------------------\n");
      }

      // read and check event-fifo for events
      ReadAndCheckEvent(); 
    }
    while((ucDemoState != 52) && (ucDemoState != 60));
  }

  // State 5.2) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 5.2 Dummy state                                                         */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 52)
  {
     printf("     5.1) Shutdown of CANopen Master successfully                - done.\n");

    // next state is 6.0)
    ucDemoState = 60;
  }


  // State 6) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 6.0 Free CAN interface ressources                                       */
  /* ----------------------------------------------------------------------- */
  if(ucDemoState == 60)
  {

    printf("\n-> Press any key to continue demo program...\n\n");
    getch();

    printf("6.0) Freeing of CAN interface resources.\n");

    printf("     This will be done by calling the function\n");
    printf("     CMA_CloseCard().\n");
    printf("     Call of function CMA_CloseCard().\n");

    ret = CMA_CloseCard(ucCANHandle);

    if(ret != 0)
    {
      PrintCMAFunctReturnError(ret);
      printf("     6.0)                                                        - failed !\n");
      // next state is 6.1)
      ucDemoState = 61;

    }
    else
    {
      printf("     6.0) successfull                                            - done.\n");
      // next state is 6.1)
      ucDemoState = 61;
    }
  }

  // State 6.1) of the demo program:
  /* ----------------------------------------------------------------------- */
  /* 6.1 End of program                                                      */
  /* ----------------------------------------------------------------------- */
  printf("\nEnd of program. -> Press any key.\n");
  getch();

  return(0);
}

/* ------------------------------------------------------------------------- */
/* Function: ReadAndCheckEvent()                                             */
/*           - Reads new events from the event-fifo. Dependend on the        */
/*             current program state it branches to evaluation routines      */
/*             related to each the state.                                    */
/*                                                                           */
/* Output:   ret                                                             */
/*           - 0 success                                                     */ 
/* ------------------------------------------------------------------------- */
short ReadAndCheckEvent(void)
{

  CMA_EventStruct  stEvent;
  unsigned short   ret = 0;

  ret = CMA_ReadEvent (ucCANHandle,&stEvent);
  if(ret)
  {
    PrintCMAFunctReturnError(ret);
  }
  else
  {
    if(stEvent.ucEventType != 0)
    {
      switch(ucDemoState)
      {
        case 31: // Function to evaluate events in state 3.1)
                 // of the demo program.
          ret = EvaluateEventInState_31(stEvent);
          break;
        case 32: // Function to evaluate events in state 3.2)
                 // of the demo program.
          ret = EvaluateEventInState_32(stEvent);
          break;
        case 42: // Function to evaluate events in state 4.2)
                 // of the demo program.
          ret = EvaluateEventInState_42(stEvent);
          break;
        case 51: // Function to evaluate events in state 5.1)
                 // of the demo program.
          ret = EvaluateEventInState_51(stEvent);
          break;
        default:
          PrintUnexpectedEvent(stEvent);
        break;
      }
    }
  }
  return(ret);
}

/* ------------------------------------------------------------------------- */
/* Function: EvaluateEventInState_31()                                       */
/*           - Function evaluates events in state 3.1) of the demo program.  */
/*             Called from function ReadAndCheckEvent().                     */
/*                                                                           */
/* Input:    stEvent                                                         */
/*           - last read event                                               */
/* Output:   ret                                                             */
/*           - 0 success                                                     */
/* ------------------------------------------------------------------------- */
short EvaluateEventInState_31(CMA_EventStruct stEvent)
{
  unsigned short   ret = 0;

  switch(stEvent.ucEventType)
  {
    case StartupCompleted:
      // next state is 3.2)
      ucDemoState = 32;
      printf("     - StartupCompleted event received.\n");
      break;
    default:
      PrintUnexpectedEvent(stEvent);
      break;
  }
  return(ret);
}

/* ------------------------------------------------------------------------- */
/* Function: EvaluateEventInState_32()                                       */
/*           - Function evaluates events in state 3.2) of the demo program.  */
/*             Called from function ReadAndCheckEvent().                     */
/*                                                                           */
/* Input:    stEvent                                                         */
/*           - last read event                                               */
/* Output:   ret                                                             */
/*           - 0 success                                                     */
/* ------------------------------------------------------------------------- */
short EvaluateEventInState_32 (CMA_EventStruct stEvent)
{
  unsigned short   ret = 0;
  unsigned short   usTempErrCode;

  switch(stEvent.ucEventType)
  {
    case ErrorEvent:
      switch(stEvent.ucaEventData[0])
      {
        case CANError:
          // copy low part of Error Code
          usTempErrCode = (unsigned short)stEvent.ucaEventData[2];
          // copy high part of Error Code
          usTempErrCode |= (unsigned short)stEvent.ucaEventData[3] << 8;
          if(usTempErrCode == 0x0102)
          {
            // error which leads to an abort of the demo program,
            // next state is 5.0)
            ucDemoState = 50;
            printf("     - CANError event 0x%x received.\n",usTempErrCode);
            printf("       - CAN controller went ERROR PASSIVE !\n");
          }
          else if(usTempErrCode == 0x0101)
          {
            // error which leads to an abort of the demo program,
            // next state is 5.0)
            ucDemoState = 50;
            printf("     - CANError event 0x%x received.\n",usTempErrCode);
            printf("       - CAN controller went BUS OFF !\n");
          }
          break;
        default:
          {
            PrintUnexpectedEvent(stEvent);
          }
          break;
      }
      break;
    default:
      {
        PrintUnexpectedEvent(stEvent);
      }
      break;
  }
  return(ret);
}

/* ------------------------------------------------------------------------- */
/* Function: EvaluateEventInState_42()                                       */
/*           - Function evaluates events in state 4.2) of the demo program.  */
/*             Called from function ReadAndCheckEvent().                     */
/*                                                                           */
/* Input:    stEvent                                                         */
/*           - last read event                                               */
/* Output:   ret                                                             */
/*           - 0 success                                                     */
/* ------------------------------------------------------------------------- */
short EvaluateEventInState_42 (CMA_EventStruct stEvent)
{
  unsigned short   ret = 0;
  unsigned short   usTempErrCode;

  switch(stEvent.ucEventType)
  {
    case UploadCompleted:
      // is the UploadCompleted event related to the configured slave's Node-ID
      if(stEvent.ucaEventData[0] == node_id)
      {
         if(stEvent.ucaEventData[4] == 0xFF)
         {
           // segmented upload or block upload
           unsigned int i,j;

           if(UPLOAD_MDS > stEvent.ulDataLenSegmented)
           {
             
             
             ret = CMA_ReadData(
               ucCANHandle, 
               (unsigned char)node_id,
               UPLOAD_SUB_INDEX,
               UPLOAD_INDEX,stEvent.
               ulDataLenSegmented,
               stUpload.ucaData
               );
             if(ret) return -1;
           }

           printf("Downloaded values:\n");

           for(i=0; i<=stEvent.ulDataLenSegmented/8; i++)
           {
             for(j=i*16; (j<(i*16)+16); j++)
             {
               if(j<stEvent.ulDataLenSegmented)
               {
                 printf("%02x ",stUpload.ucaData[j]);
               }
               else
               {
                 printf("   ");
               }
             }
             
             printf(" ");

             for(j=i*16; (j<stEvent.ulDataLenSegmented) && (j<(i*16)+16); j++)
             {
               if((stUpload.ucaData[j] > 31) && (stUpload.ucaData[j] < 126))
               {
                 printf("%c",stUpload.ucaData[j]);
               }
               else
               {
                 printf(".");
               }
               
             }


             printf("\n");

           }
         }
         else
         {
           // expedited upload


           switch(stEvent.ucaEventData[4])
           {
           case 1:
              printf("downloaded value: 0x%x\n", *&stEvent.ucaEventData[5]);
             break;

           case 2:
              printf("downloaded value: 0x%x\n", *(USIGN16*)&stEvent.ucaEventData[5]);
             break;

           case 4:
              printf("downloaded value: 0x%x\n", *(USIGN32*)&stEvent.ucaEventData[5]);
             break;

           }

         }

          // Upload is completed

          // next state is 4.3)
          ucDemoState = 43;
      }
      else
        PrintUnexpectedEvent(stEvent);
      break;
    case ErrorEvent:
      switch(stEvent.ucaEventData[0])
      {
        case SDOError:
          // is the error related to the configured slave's Node-ID
          if(stEvent.ucaEventData[1] == node_id)
          {
            printf("      - SDO Error occured:  NodeID: %d,\n",stEvent.ucaEventData[1]);
            // copy low part of Additional Code
            usTempErrCode = (unsigned short)stEvent.ucaEventData[4];
            // copy high part of Additional Code
            usTempErrCode |= (unsigned short)stEvent.ucaEventData[5] << 8;
            printf("                            ErrClass-ErrCode-AddCode: %d-%d-%d\n",stEvent.ucaEventData[2],
                   stEvent.ucaEventData[3],usTempErrCode);
          }
          else
            PrintUnexpectedEvent(stEvent);
          break;
        default:
          PrintUnexpectedEvent(stEvent);
          break;
      }
      break;
    default:
      PrintUnexpectedEvent(stEvent);
      break;
  }
  return(ret);
}


/* ------------------------------------------------------------------------- */
/* Function: EvaluateEventInState_51()                                       */
/*           - Function evaluates events in state 5.1) of the demo program.  */
/*             Called from function ReadAndCheckEvent().                     */
/*                                                                           */
/* Input:    stEvent                                                         */
/*           - last read event                                               */
/* Output:   ret                                                             */
/*           - 0 success                                                     */
/* ------------------------------------------------------------------------- */
short EvaluateEventInState_51(CMA_EventStruct stEvent)
{
  unsigned short   ret = 0;

  switch(stEvent.ucEventType)
  {
    case ShutdownCompleted:
      // next state is 5.1)
      ucDemoState = 52;
      printf("     - ShutdownCompleted event received.\n");
      break;
    default:
      PrintUnexpectedEvent(stEvent);
      break;
  }
  return(ret);
}


/* ------------------------------------------------------------------------- */
/* Function: PrintUnexpectedEvent()                                          */
/*           - Function evaluates each possible event and prints it out.     */
/*             Called if an event is read from event-fifo which was not      */
/*             expected.                                                     */
/*                                                                           */
/* Input:    stEvent                                                         */
/*           - last read event                                               */
/* ------------------------------------------------------------------------- */
void PrintUnexpectedEvent(CMA_EventStruct stEvent)
{

  int              i;
  unsigned short   usTempCode;

  printf("-----------------------------------------------------------------\n");
  printf("--> Unexpected event received ! : \n");

  switch(stEvent.ucEventType)
  {
    case DownloadCompleted:
      printf("--> Event: DownloadCompleted:\n");
      // copy low part of Index
      usTempCode = (unsigned short)stEvent.ucaEventData[2];
      // copy high part of Index
      usTempCode |= (unsigned short)stEvent.ucaEventData[3] << 8;
      printf("-->        NodeID: %d, Index: 0x%x, ",stEvent.ucaEventData[0],
             usTempCode);
      printf("SubIndex: %d\n",stEvent.ucaEventData[1]);
      break;
    case UploadCompleted:
      printf("--> Event: UploadCompleted:\n");
      // copy low part of Index
      usTempCode = (unsigned short)stEvent.ucaEventData[2];
      // copy high part of Index
      usTempCode |= (unsigned short)stEvent.ucaEventData[3] << 8;
      printf("-->        NodeID: %d, Index: 0x%x, ",stEvent.ucaEventData[0],
             usTempCode);
      printf("SubIndex: %d, \n",stEvent.ucaEventData[1]);
      if(stEvent.ucaEventData[4] != 0xFF)
      {
        // expedited SDO transfer data is included
        printf("-->        DLC: %d, data: ",stEvent.ucaEventData[4]);
        for(i=0; i < stEvent.ucaEventData[4]; i++)
          printf("%d ",stEvent.ucaEventData[5+i]);
        printf("\n");
      }
      break;
    case EMCYReceived:
      printf("--> Event: EMCYReceived:\n");
      printf("-->        NodeID: %d, Data: ",stEvent.ucaEventData[0]);
      for(i=0; i<8; i++)
        printf("%d ",stEvent.ucaEventData[i+1]);
      printf("\n");
      break;
    case ErrorEvent:
      printf("--> Event: ErrorEvent:\n");
      switch(stEvent.ucaEventData[0])
      {
        case APIError:
          // copy low part of Error Code
          usTempCode = (unsigned short)stEvent.ucaEventData[2];
          // copy high part of Error Code
          usTempCode |= (unsigned short)stEvent.ucaEventData[3] << 8;
          printf("-->        EventType: APIError, ErrCode: 0x%x\n",usTempCode);
          break;
        case CANError:
          // copy low part of Error Code
          usTempCode = (unsigned short)stEvent.ucaEventData[2];
          // copy high part of Error Code
          usTempCode |= (unsigned short)stEvent.ucaEventData[3] << 8;
          printf("-->        EventType: CANError, ErrCode: 0x%x\n",usTempCode);
          break;
        case SDOError:
          printf("-->        EventType: SDOError, NodeID: %d,\n",stEvent.ucaEventData[1]);
          // copy low part of Additional Code
          usTempCode = (unsigned short)stEvent.ucaEventData[4];
          // copy high part of Additional Code
          usTempCode |= (unsigned short)stEvent.ucaEventData[5] << 8;
          printf("-->        ErrClass-ErrCode-AddCode: %d-%d-%d\n",stEvent.ucaEventData[2],
                 stEvent.ucaEventData[3],usTempCode);
          break;
        case GuardError:
          printf("-->        EventType: GuardError, NodeID: %d,\n",stEvent.ucaEventData[1]);
          printf("-->        ErrCode: %d\n",stEvent.ucaEventData[2]);
          break;
        default:
          printf("-->        Unknown error event type, value: %d received.\n",
                 stEvent.ucaEventData[0]);
          break;
      }
      break;
    case PDOReceived:
      printf("--> Event: PDOReceived:\n");
      // copy low part of PDO handle
      usTempCode = (unsigned short)stEvent.ucaEventData[0];
      // copy high part of PDO handle
      usTempCode |= (unsigned short)stEvent.ucaEventData[1] << 8;
      printf("-->        PDOHandle: %d, NumOfBytes: %d\n",usTempCode,
             stEvent.ucaEventData[2]);
      printf("-->        Data: ");
      for(i=0; i<stEvent.ucaEventData[2]; i++)
        printf("%d ",stEvent.ucaEventData[i+3]);
      printf("\n");
      break;
    case ShutdownCompleted:
      printf("--> Event: ShutdownCompleted\n");
      break;
    case StartupCompleted:
      printf("--> Event: StartupCompleted\n");
      break;
    case NoEvent:
      break;
    default:
      printf("--> Unknown event type, value %d received.\n",stEvent.ucEventType);
      break;
  }
  printf("-----------------------------------------------------------------\n");
}

/* ------------------------------------------------------------------------- */
/* Function: StartTimeoutTimer()                                             */
/*           - Starts the timeout timer. A timeout Timer will be checked     */
/*             with funtion CheckTimeoutTimer()                              */
/*                                                                           */
/* ------------------------------------------------------------------------- */
void StartTimeoutTimer (void)
{
  ulTimerStartTime = GetTickCount();
}

/* ------------------------------------------------------------------------- */
/* Function: CheckTimeoutTimer()                                             */
/*           - Checks if the timeout timer timed out. The Timer will be      */
/*             started with function StartTimeoutTimer().                    */
/*                                                                           */
/* Input:    ulTimeoutInMs                                                   */
/*           - Timeout value in ms.                                          */
/* ------------------------------------------------------------------------- */
short CheckTimeoutTimer (unsigned long ulTimeoutInMs)
{

  unsigned long ulActTime=0, ulDiffTime=0;

  ulActTime  = GetTickCount();
  ulDiffTime = (ulActTime >= ulTimerStartTime) ?
               ulActTime - ulTimerStartTime :
               (0xFFFFFFFFL - ulTimerStartTime) + ulActTime + 1;

  /* check timeout */
  if(ulDiffTime > ulTimeoutInMs)
  {
    // timeout occured
    return(-1);
  }
  else
    // no timeout occured yet
    return(0);
}



/* ------------------------------------------------------------------------- */
/* Function: PrintUploadedData()                                             */
/*           - Prints uploaded data bytes.                                   */
/*                                                                           */
/* ------------------------------------------------------------------------- */
void PrintUploadedData(void)
{
  unsigned long  i;
  unsigned char  ucByteCount;
  unsigned char  ucInputChar;

  // print uploaded data
  if(stUpload.ulTransmittedBytes > 0)
  {

    // print the uploaded data if wished by the user
    printf("      Do You want to see received upload data (y/n) ?");
    ucInputChar = getch(); printf("\n");
    if((ucInputChar=='Y') || (ucInputChar=='y'))
    {
      printf("      - Uploaded data (Number of bytes = %d):",stUpload.ulTransmittedBytes);
      ucByteCount = 0;
      for(i=0; i<stUpload.ulTransmittedBytes; i++)
      {
        if(ucByteCount == 0)
        {
          printf("\n       ");
          ucByteCount = 7;
        }
        ucByteCount--;
        printf("%3d ",stUpload.ucaData[i]);
      } 
      printf("\n");
    }
  }
}


/* ------------------------------------------------------------------------- */
/* Function: PrintCMAFunctReturnError()                                      */
/*           - Prints function return codes in a textual form.               */
/*                                                                           */
/* Input:    ret                                                             */
/*           - Function return code                                          */
/* ------------------------------------------------------------------------- */
void PrintCMAFunctReturnError(int ret)
{
  switch(ret)
  {
    case CANPC_RB_INI_FILE:
      printf("      - CMA function error:\n");
      printf("        Can't open INI-file (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_FMT_INI:
      printf("      - CMA function error:\n");
      printf("        Format error in INI-file (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_OP_BIN:
      printf("      - CMA function error:\n");
      printf("        Error in opening binary-file (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_RD_BIN:
      printf("      - CMA function error:\n");
      printf("        Error in error reading binary-file (code: %d).\n",ret);
      break;
    case CANPC_RB_BIN_TOO_LONG:
      printf("      - CMA function error:\n");
      printf("        Binary-file too long (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_BIN_FMT:
      printf("      - CMA function error:\n");
      printf("        Binary-data format error (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_BIN_CS:
      printf("      - CMA function error:\n");
      printf("        Binary-data checksum error (code: %d).\n",ret);
      break;
    case CANPC_RB_NO_CARD:
      printf("      - CMA function error:\n");
      printf("        No interface present (code: %d).\n",ret);
      break;
    case CANPC_RB_NO_PHYS_MEM:
      printf("      - CMA function error:\n");
      printf("        No physical memory available (code: %d).\n",ret);
      break;
    case CANPC_RB_INVLD_IRQ:
      printf("      - CMA function error:\n");
      printf("        Invalid IRQ-number (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_DPRAM_ACCESS:
      printf("      - CMA function error:\n");
      printf("        Error in accessing dpram (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_CRD_RESP:
      printf("      - CMA function error:\n");
      printf("        Bad response from interface (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_SRAM:
      printf("      - CMA function error:\n");
      printf("        SRAM seems to be damaged (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_PRG:
      printf("      - CMA function error:\n");
      printf("        Invalid program start address (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_REC:
      printf("      - CMA function error:\n");
      printf("        Invalid record type (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_NORESP:
      printf("      - CMA function error:\n");
      printf("        No response after program start (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_BADRESP:
      printf("      - CMA function error:\n");
      printf("        Bad response after program start (code: %d).\n",ret);
      break;
    case CANPC_RB_PCMCIA_NSUPP:
      printf("      - CMA function error:\n");
      printf("        PCMCIA chip not supported (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_RD_PCMCIA:
      printf("      - CMA function error:\n");
      printf("        Error in reading PCMCIA parameters (code: %d).\n",ret);
      break;
    case CANPC_RB_INIT_CHIP:
      printf("      - CMA function error:\n");
      printf("        Error in initialising PCMCIA chip (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_NO_CANCARD:
      printf("      - CMA function error:\n");
      printf("        No CANcard plugged in (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_CRD_NORESP:
      printf("      - CMA function error:\n");
      printf("        Bad response from interface (code: %d).\n",ret);
      break;
    case CANPC_RB_ERR_DEV_LD:
      printf("      - CMA function error:\n");
      printf("        Interface not properly loaded (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_LICENCE:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucLicence invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_LICENSE:
      printf("      - CMA function error:\n");
      printf("        Demo applications expired (code: %d).\n",ret);
      break;
    case CMA_ERR_LICENCE_ACCESS:
      printf("      - CMA function error:\n");
      printf("        Licence access impossible (code: %d).\n",ret);
      break;
    case CMA_ERR_FUNCTION_TIMEOUT:
      printf("      - CMA function error:\n");
      printf("        Function timeout occurred (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_FUNCTIONALITY:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucCANxFunctionality undefined (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_CHAN_HANDLE:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucChannelHandle undefined (code: %d).\n",ret);
      break;
    case CMA_ERR_SCT_BAUD_RATE:
      printf("      - CMA function error:\n");
      printf("        Struct parameter: ucBaudRate invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_TIMEOUT_MAX:
      printf("      - CMA function error:\n");
      printf("        Parameter: timeout too big (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_TIMEOUT_MIN:
      printf("      - CMA function error:\n");
      printf("        Parameter: timeout too small (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_NODE_ID:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucNodeID is invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_SCT_COB_ID:
      printf("      - CMA function error:\n");
      printf("        Struct parameter: COB-ID invalid or used (code: %d).\n",ret);
      break;
    case CMA_ERR_SCT_COMM_CYC_MAX:
      printf("      - CMA function error:\n");
      printf("        Struct parameter: ulSyncCycle is too big (code: %d).\n",ret);
      break;
    case CMA_ERR_SCT_NODE_ID:
      printf("      - CMA function error:\n");
      printf("        Struct parameter: ucNodeID invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_MASTER_STATE:
      printf("      - CMA function error:\n");
      printf("        Improper state of the master (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_STATE:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucState invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_INT_TX_FIFO_FULL:
      printf("      - CMA function error:\n");
      printf("        Internal transmit fifo is full (code: %d).\n",ret);
      break;
    case CMA_ERR_MASTER_GUARDING:
      printf("      - CMA function error:\n");
      printf("        Master cannot guard itself (code: %d).\n",ret);
      break;
    case CMA_ERR_COMMAND_FIFO_FULL:
      printf("      - CMA function error:\n");
      printf("        DPRAM command fifo is full (code: %d).\n",ret);
      break;
    case CMA_ERR_SDO_TX_FIFO_FULL:
      printf("      - CMA function error:\n");
      printf("        SDO transmit fifo is full (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_DATA_SIZE:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucDataSize is invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_INDEX:
      printf("      - CMA function error:\n");
      printf("        Parameter: usIndex is incorrect (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_SUBINDEX:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucSubIndex is incorrect (code: %d).\n",ret);
      break;
    case CMA_ERR_TOTAL_DATA_SIZE:
      printf("      - CMA function error:\n");
      printf("        Parameter: ulTotalDataSize exceeded (code: %d).\n",ret);
      break;
    case CMA_ERR_NO_TRANSF_ACTIVE:
      printf("      - CMA function error:\n");
      printf("        No normal SDO transfer active (code: %d).\n",ret);
      break;
    case CMA_ERR_SDO_RX_FIFO_EMPTY:
      printf("      - CMA function error:\n");
      printf("        SDO receive fifo is empty (code: %d).\n",ret);
      break;
    case CMA_ERR_SCT_PDO_ACC_TYPE:
      printf("      - CMA function error:\n");
      printf("        Struct parameter: ucAccType is invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_PDO_HANDLE:
      printf("      - CMA function error:\n");
      printf("        Parameter: usPDOHandle is undefined (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_BYTEPOS:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucDataOffset is invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_PDO_TOO_LONG:
      printf("      - CMA function error:\n");
      printf("        Parameter: DataOffset+Datalength>8 (code: %d).\n",ret);
      break;
    case CMA_ERR_PDO_ACCESS:
      printf("      - CMA function error:\n");
      printf("        Access to PDO in DPRAM failed (code: %d).\n",ret);
      break;
    case CMA_ERR_PAR_BITPOS:
      printf("      - CMA function error:\n");
      printf("        Parameter: ucBitPosition is invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_UNKNOWN_EVENT:
      printf("      - CMA function error:\n");
      printf("        Unknown event in event fifo (code: %d).\n",ret);
      break;
    case CMA_ERR_WRONG_FW_VERSION:
      printf("      - CMA function error:\n");
      printf("        Wrong firmware version (code: %d).\n",ret);
      break;
    case CMA_ERR_SCT_TRANS_TYPE:
      printf("      - CMA function error:\n");
      printf("        Struct parameter: ucTransType invalid (code: %d).\n",ret);
      break;
    case CMA_ERR_DPRAM_TX_PDO_MAX:
      printf("      - CMA function error:\n");
      printf("        Number of DPRAM-TX-PDOs exceeded (code: %d).\n",ret);
      break;
    case CMA_ERR_DPRAM_RX_PDO_MAX:
      printf("      - CMA function error:\n");
      printf("        Number of DPRAM-RX-PDOs exceeded (code: %d).\n",ret);
      break;
    case CMA_ERR_LAST_REQ_PENDING:
      printf("      - CMA function error:\n");
      printf("        Last request still pending (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_TOOMANYDEVICES:
      printf("      - CMA function error:\n");
      printf("        Maximum number of device exceeded (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_NOTENOUGHMEMORY:
      printf("      - CMA function error:\n");
      printf("        Not enough memory space to register a device (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_DRIVERNOTFOUND:
      printf("      - CMA function error:\n");
      printf("        Driver not found (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_DRIVERVERSIONERROR:
      printf("      - CMA function error:\n");
      printf("        Wrong driver version (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_VERSIONERROR:
      printf("      - CMA function error:\n");
      printf("        Wrong driver-dll version (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_NOTSUPPORTED:
      printf("      - CMA function error:\n");
      printf("        Internal error: Requested functionality not supported\n");
      printf("        (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_ILLEGALCALL:
      printf("      - CMA function error:\n");
      printf("        Internal error: Illegal driver call (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_IOCANCELLED:
      printf("      - CMA function error:\n");
      printf("        Internal error: Driver call cancelled (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_IOPENDING:
      printf("      - CMA function error:\n");
      printf("        Internal error: Driver call pending (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_TIMEOUT:
      printf("      - CMA function error:\n");
      printf("        Internal error: Timeout during driver access (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_GENERALERROR:
      printf("      - CMA function error:\n");
      printf("        Internal error: General error in accessing the driver (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_INTERNALERROR:
      printf("      - CMA function error:\n");
      printf("        Internal driver error (code: %d).\n",ret);
      break;
    case INIPC_IB_PNP_NO_DEVICE_FOUND:
      printf("      - CMA function error:\n");
      printf("        No CAN interface found (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_UNKNOWNDEVICE:
      printf("      - CMA function error:\n");
      printf("        Access to an unknown device requested (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_DEVICEALREADYEXISTS:
      printf("      - CMA function error:\n");
      printf("        Device is already registered by the driver (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_DEVICEACCESSERROR:
      printf("      - CMA function error:\n");
      printf("        Device is already open (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_RESOURCEALREADYREGISTERED:
      printf("      - CMA function error:\n");
      printf("        A requested resource is already in use \n");
      printf("        by an other device (DPRAM, IRQ, I/O port) (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_RESOURCECONFLICT:
      printf("      - CMA function error:\n");
      printf("        Resource conflict with an other driver (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_RESOURCEACCESSERROR:
      printf("      - CMA function error:\n");
      printf("        Resource access error (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_PHYSMEMORYOVERRUN:
      printf("      - CMA function error:\n");
      printf("        Internal error: Invalid phys. mem-access (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_TOOMANYPORTS:
      printf("      - CMA function error:\n");
      printf("        Internal error: Too many I/O ports requested (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_VC_UNKNOWNRESOURCE:
      printf("      - CMA function error:\n");
      printf("        Internal error: Unknown resource access requested (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_LOADDDLL_DEV:
      printf("      - CMA function error:\n");
      printf("        Driver DLL not found (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_CANT_DEVREGISTER:
      printf("      - CMA function error:\n");
      printf("        Interface cannot be registered (code: %d).\n",ret);
      break;
    case INIPC_IB_ERR_IRQ:
      printf("      - CMA function error:\n");
      printf("        Interrupt number invalid or device cannot\n");
      printf("        be assigned an interrupt (code: %d).\n",ret);
      break;
    case DDC_ERR_WRONG_DEVICEHANDLE:
      printf("      - CMA function error:\n");
      printf("        Internal error: Wrong device handle specified\n");
      printf("        (code: %d).\n",ret);
      break;
    default:
      printf("      - Unknown CMA function error (code: %d).\n",ret);
      break;
  }
}

