
/***************************************************************************
 *   canlib - Softing Library for API of Can-Cards                         *
 *   Copyright (C) 2007-2015 by Softing Industrial Automation GmbH         *
 *   support.automation@softing.com                                        *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this library; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


/*
 * $Id: canlib.c,v 1.11 2007/08/29 11:01:58 kno Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <linux/kdev_t.h>
#include <dirent.h>
#include <math.h>

#include <sys/ioctl.h>
#include <unistd.h>

#include <errno.h>
#include <string.h>

#include "BaseTypes.h"
#include "can_def.h"
#include "canL2.h"
#include "can_ioctl.h"

#define CANLIB_FAIL -100
#define INVALID_HANDLE_VALUE -1
#define DEVFILEPREFIX "/dev/Softing_AG_"
#define CAN_CONFIG_FILE "/etc/softing/candevices.conf"
#define LINE_SIZE 300




// get the channel configruration from config file
static INT32 getScimConfig(CAN_HANDLE Can, L2CONFIG *pScimCfg);

INT32 CANL2_get_device_type(CAN_HANDLE can_channel, USIGN32 *device_type);


/* +--- CANL2_init_signals         -------------------------------------------*/

INT32 CANL2_init_signals(CAN_HANDLE can_channel, USIGN32 ulChannelDirectionMask, USIGN32 ulChannelOutputDefaults)
{
  INT32 rt = CANLIB_FAIL;
    
    
  USIGN32 pram[10] = { (INT32) ulChannelDirectionMask,
    (INT32) ulChannelOutputDefaults,
    0, 0, 0, 0, 0, 0, 0, 0
  };

  rt = ioctl(can_channel,
             CAN_IOC_CANL2_init_signals, pram);
  
  if(rt > 0x10000)
  {
    rt = -(rt - 0x10000);
  }

  

  return rt;
}

/* +--- CANL2_read_signals         -------------------------------------------*/

INT32 CANL2_read_signals(CAN_HANDLE can_channel, USIGN32 *pulChannelRead)
{
  INT32 rt = CANLIB_FAIL;
    
    
  USIGN32 pram[10] = { *pulChannelRead,
    0,
    0, 0, 0, 0, 0, 0, 0, 0
  };

  rt = ioctl(can_channel,
             CAN_IOC_CANL2_read_signals, pram);
  
  *pulChannelRead = pram[0];
  
  if(rt > 0x10000)
  {
    rt = -(rt - 0x10000);
  }


  return rt;
}

/* +--- CANL2_write_signals         -------------------------------------------*/

INT32 CANL2_write_signals(CAN_HANDLE can_channel, USIGN32 ulChannelWrite, USIGN32 ulCareMask)
{
  INT32 rt = CANLIB_FAIL;
    
    
  USIGN32 pram[10] = { (INT32) ulChannelWrite,
    (INT32) ulCareMask,
    0, 0, 0, 0, 0, 0, 0, 0
  };

  rt = ioctl(can_channel,
             CAN_IOC_CANL2_write_signals, pram);
  
  if(rt > 0x10000)
  {
    rt = -(rt - 0x10000);
  }


  return rt;
}


/* +--- CANL2_enable_error_frame_detection         -------------------------------------------*/

INT32 CANL2_enable_error_frame_detection(CAN_HANDLE can_channel)
{
    INT32 rt = CANLIB_FAIL;

    rt = ioctl(can_channel,
               CAN_IOC_CANL2_enable_error_frame_detection, NULL);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }


    return rt;
}


/* +--- CANL2_reset_chip         -------------------------------------------*/
INT32 CANL2_reset_chip(CAN_HANDLE can_channel)
{
    INT32 rt = CANLIB_FAIL;


    rt = ioctl(can_channel,CAN_IOC_CANL2_reset_chip, NULL);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }


    return rt;
}

/*HEAD_BEG*******************************************************************/
/*                                                                          */
/* FUNCTION                       getPresc                                  */
/* FUNCTIONAL_DESCRIPTION         calculates the first possible presc.      */
/*                                >= StartPresc than meets the fBaudrate    */
/*                                and clock                                 */
/*                                                                          */
/* INPUTS                         StartPresc: start prescaler               */
/*                                fBaudrate:  baud rate                     */
/*                                clock:      clock rate of the CAN ctrl.   */
/*                                                                          */
/* OUTPUTS                        prescaler or StartPresc                   */
/* CHANGES                                                                  */
/*                                                                          */
/*HEAD_END*******************************************************************/


static USIGN8 getPresc(USIGN8 StartPresc, double fBaudrate, double clock)
{
  USIGN8 presc = StartPresc;
  USIGN32 IntTesg1AndTeseg2;
  double fTeseg, fTesegRound;

  if(presc == 0)
    presc++;

  while(presc <= 64)
{
    fTeseg = clock/(presc*fBaudrate)-1;
    IntTesg1AndTeseg2 = (USIGN32) fTeseg;
    fTesegRound = IntTesg1AndTeseg2;

  if(fTesegRound == fTeseg)
    return presc;

  presc++;
}

  return 0xFF;
}



/**
 ******************************************************************
 * @brief  calculates all possible CAN controller settings for a given sum of tseg1 and tseg2
 * @author Werner Knorsch, Softing AG
 * @date   2008-01-29
 *
 * @param  Tseg1AndTseg2Int: tseg1+tseg2
 * @return -
 *
 ******************************************************************
 */

static INT32 Sja1000ToTwinCAN(INT32 *ppresc, INT32 *psjw, INT32 *ptseg1, INT32 *ptseg2, INT32 *psam, USIGN32 deviceType)
{
  
  USIGN8 tseg1=*ptseg1, tseg2 = *ptseg2, sjw=*psjw, presc=*ppresc, Div8X=1, tseg12;
  double fTseg, fbaud, clock_rate = 40E6;
  INT32 baud1, baud2;
  USIGN32 prescaler,Tseg1AndTseg2Int;
  double samPt, lastSamPt=100.0, BitTime, samples,Tseg1AndTseg2, fsam;

  if((deviceType != CANPROPCIE) && (deviceType != CAN104PLUS))
    return CANL2_OK; // nothing to do (values have the right format)
  
  if((*ppresc==GET_FROM_SCIM) || (*ptseg1==GET_FROM_SCIM) || (*ptseg2==GET_FROM_SCIM))
    return CANL2_OK; // nothing to do - use values from SCIMCfg
  
  
  
  fbaud = 16E3 / (2 * presc * (1 + tseg1 + tseg2));
  fsam = ((1.0+tseg1)*100.0)/(1.0+tseg1+tseg2);
  
  if((fbaud < 6.0) || (fbaud > 1000.0))
  {
    return CANL2_WRONG_PARAM;
  }

  BitTime = 1.0/(fbaud * 1000);
  samples = BitTime * clock_rate;
 
   // start value for prescaler
  presc = (USIGN8)(INT32)(samples / 25);
  
 // as a floating point value the prescaler may be less than 1, but not in the real CAN controller.
  if(presc == 0)
    presc = 1;

  // presc is a BYTE: may be a prescaler of 255 is not enough to satisfy the following formula
  // to prevent an overrun from 255->0 (endless loop) we go only to max. 65
  while((((clock_rate/1000)/(presc*Div8X*fbaud)-1) > 24.0) && presc <= 65) presc++;
 
  if(presc > 64)
    presc = 64;
 
  presc = getPresc(presc, fbaud, clock_rate);
  
  if(presc == 0xFF)
  {
    return CANL2_WRONG_PARAM;
  }

  Tseg1AndTseg2 = (clock_rate/1000)/(presc*fbaud)-1;

  Tseg1AndTseg2Int = (INT32)Tseg1AndTseg2;
  
  prescaler = presc;
  
  for(Div8X=1; Div8X<=8; Div8X += 7)
  {
    tseg12 = (USIGN8)Tseg1AndTseg2Int;
    tseg2 = (USIGN8)(Tseg1AndTseg2Int/2);
    presc = prescaler/Div8X;
  
    if(!presc) break;

    do
    {
      do
      {
      

        if(tseg2 > 8) 
          tseg2 = 8;
      
        if(tseg2 < 2)
          tseg2 = 2;
      
        tseg1 = (USIGN8)(tseg12 - tseg2);
      
      
        if((tseg1 < 1) || (tseg1 > 16))
        { 
          // to do ERROR handling!
          break;
        }
      
 
        baud1 = (INT32) (fbaud * 100000.00);
        baud2 = (INT32) (clock_rate/((presc*Div8X)*(1+tseg1+tseg2))*100);

        if(baud1 != baud2)
        {
          break;
        }
      

        for(sjw = 1; sjw<=4; sjw++)
        {
          if((sjw <= tseg2) && (sjw <= (tseg1+tseg2)))
          {
           
            samPt = (1.0+tseg1)*100.0/(1.0+tseg1+tseg2);

            if(fabs(fsam-samPt) < fabs(fsam-lastSamPt))
            {
       // printf("presc=%u, tseg1=%u, tseg2=%u, realSamPt= %f, samPt=%f baud=%f\n", presc, tseg1, tseg2, SamplePoint, samPt, (fclock/1000)/(presc*(1+tseg1+tseg2)));
              lastSamPt = samPt;

              *ppresc  = presc;
              *ptseg1 = tseg1;
              *ptseg2 = tseg2;
              *psam = Div8X/8;
            }
          
          
          }
        }
      
        tseg2--;
      
      } while(tseg2 > 1);

      presc = getPresc(++presc, fbaud, clock_rate/Div8X);
      fTseg = clock_rate/(presc*fbaud*Div8X*1000)-1;
      tseg12 = (USIGN8)(INT32)fTseg;
      tseg2 = (USIGN8)(tseg12/2);
   
    } while ((fTseg > 8.0) && (presc <= 64));

  }
 
  return CANL2_OK;
}

/*HEAD_BEG***********************************************************************/
/*                                                                              */
/* FUNCTION                       getPresc                                      */
/* FUNCTIONAL_DESCRIPTION         calculates the bit timing values              */
/*                                                                              */
/* INPUTS                         Tseg1AndTseg2Int: tseg1+tseg2                 */
/*                                StartPresc:       start prescaler             */
/*                                SamplePoint:      sample point                */
/*                                fBaudrate:        baud rate                   */
/*                                fclock:           clock rate of the CAN ctrl. */
/*                                                                              */
/* OUTPUTS                        tseg1, tseg2 and prescaler                    */
/* CHANGES                                                                      */
/*                                                                              */
/*HEAD_END***********************************************************************/

static void calculateBitTimings(
  USIGN32 Tseg1AndTseg2Int,
  USIGN8  StartPresc,
  double SamplePoint,
  double fBaudrate,
  double fclock,
  INT32 *ptseg1,
  INT32 *ptseg2,
  INT32 *ppresc
  
)
{
  
  USIGN8 tseg1, tseg2 = 0, sjw, presc, tseg12 = (USIGN8)Tseg1AndTseg2Int;
  double fTseg, samPt, lastSamPt=100.0;
  
  tseg2 = (USIGN8)(Tseg1AndTseg2Int/2);
  presc = StartPresc;
  
  do
  {
    do
    {
      
      
      if(tseg2 > 8) 
        tseg2 = 8;
      
      if(tseg2 < 2)
        tseg2 = 2;
      
      tseg1 = (USIGN8)(tseg12 - tseg2);
      
      
      if((tseg1 < 1) || (tseg1 > 16))
      { // to do ERROR handling!
      break;
      }
      
      if(fBaudrate != (fclock/(presc*(1+tseg1+tseg2))))
      {
        break;
      }
      
      
      for(sjw = 1; sjw<=4; sjw++)
      {
        if((sjw <= tseg2) && ((sjw) <= (tseg1+tseg2)))
        {
          samPt = (1.0+tseg1)*100.0/(1.0+tseg1+tseg2);
          
          if(fabs(SamplePoint-samPt) < fabs(SamplePoint-lastSamPt))
          {
            // printf("presc=%u, tseg1=%u, tseg2=%u, realSamPt= %f, samPt=%f baud=%f\n", presc, tseg1, tseg2, SamplePoint, samPt, (fclock/1000)/(presc*(1+tseg1+tseg2)));
            lastSamPt = samPt;
            
            *ppresc  = presc;
            *ptseg1 = tseg1;
            *ptseg2 = tseg2;
          }
          
        }
      }
      
      tseg2--;
      
    } while(tseg2 > 1);
    
    presc = getPresc(++presc, fBaudrate, fclock);
    fTseg = fclock/(presc*fBaudrate)-1;
    tseg12 = (USIGN8)(int)fTseg;
    tseg2 = (USIGN8)(tseg12/2);
  } while ((fTseg > 8.0) && (presc <= 64));
  
  
}



/*HEAD_BEG*******************************************************************/
/*                                                                          */
/* FUNCTION                       canpro_initialize_chip                    */
/* FUNCTIONAL_DESCRIPTION         defines bittiming                         */
/* INPUTS                         Can:                  CAN handle          */
/*                                presc:                [1..64]             */
/*                                sjw:                  [1..4]              */
/*                                tseg1:                [1..16]             */
/*                                tseg2:                [1..8]              */
/*                                sam:               0: 1 Sample            */
/*                                                   1: 3 Sample            */
/*                                                                           */
/* OUTPUTS                        OK_       acknownledged by firmware       */
/*                                ERROR1    inconsistent parameters         */
/*                                CANL2_FW_TIMEOUT    timeout of firmware   */
/* CHANGES                                                                  */
/*                                                                          */
/*HEAD_END*******************************************************************/
static INT32 Sja1000ToMultiCAN(INT32 *ppresc, INT32 *psjw, INT32 *ptseg1, INT32 *ptseg2, INT32 *psam, USIGN32 deviceType)
{
  INT32 tseg1=*ptseg1, tseg2 = *ptseg2, sjw=*psjw, presc=*ppresc, div8x=1;
  USIGN8 u8Presc=(USIGN8)presc;
  INT32 ret = 0;
  double fBaud, fsamplePoint, fNewBaud, clock = 80E6;
  USIGN32 Tseg1AndTseg2Int;
  double Tseg1AndTseg2;
  
  
  if(deviceType != CANPROUSB)
    return CANL2_OK;
  
  if((*ppresc==GET_FROM_SCIM) || (*ptseg1==GET_FROM_SCIM) || (*ptseg2==GET_FROM_SCIM))
    return CANL2_OK; // nothing to do - use values from SCIMCfg
    
  
  if (u8Presc < 1 || u8Presc > 32 ||
    sjw   < 1 || sjw   > 4  ||
    tseg1 < 1 || tseg1 > 16 ||
    tseg2 < 1 || tseg2 >  8 )
  {
    return(CANL2_WRONG_PARAM);
  }
  
  
  fBaud = 16E6/(2*u8Presc*(1+tseg1+tseg2));
  fsamplePoint = (1.0+tseg1)*100.0/(1.0+tseg1+tseg2);
  
  div8x = 0;
  u8Presc=1;
  
  // find the smallest prescaler, which fits the baud rate
  while(((clock/(u8Presc*fBaud)-1) > 24.0) && u8Presc <= 65) u8Presc++;
  
  /*
   *    if(u8Presc > 64)
   *      u8Presc = 64;
   */
  
  if(u8Presc > 60)
  {
    div8x = 1;
    clock = clock/8.0;
    u8Presc=1;
    while(((clock/(u8Presc*fBaud)-1) > 24.0) && u8Presc <= 65) u8Presc++;
    
  }
  
  
  if(u8Presc > 64)
    u8Presc = 64;
  
  if(getPresc(u8Presc, fBaud, clock) == 0xFF)
  { 
    // no fitting prescaler found!
    if((fBaud < 50000.0) && (div8x == 0))
    { 
      // try it with div8x for small baud rates!
      div8x = 1;
      clock = clock/8.0;
      u8Presc = 1;
      while(((clock/(u8Presc*fBaud)-1) > 24.0) && u8Presc <= 65) u8Presc++;
      
      if(u8Presc > 64)
        u8Presc = 64;
      
      
      if(getPresc(u8Presc, fBaud, clock) != 0xFF)
      { 
        // ok, a valid prescaler was found: get it!
        u8Presc = getPresc(u8Presc, fBaud, clock);
      }
    }
  }
  else
  { 
    // ok, a valid prescaler was found: get it!
    u8Presc = getPresc(u8Presc, fBaud, clock);
  }
  
  Tseg1AndTseg2 = clock/(u8Presc*fBaud)-1;
  
  Tseg1AndTseg2Int = (int)Tseg1AndTseg2;
  
  calculateBitTimings(Tseg1AndTseg2Int, u8Presc, fsamplePoint, fBaud, clock, &tseg1, &tseg2, &presc); 
  
  fNewBaud = clock/(presc*(1+tseg1+tseg2));
  // check baudrate
  if(fabs(fNewBaud-fBaud) > 0.1)
  {  
    // baudrate does not fit
    if(fBaud < 30.0)
    { 
      // try div8x
      div8x = 1;
      clock = 40E3/8;
      u8Presc = 1;
      while(((clock/(u8Presc*fBaud)-1) > 24.0) && u8Presc <= 65) u8Presc++;
      
      if(u8Presc > 64)
        u8Presc = 64;
      
      
      if(getPresc(u8Presc, fBaud, clock) != 0xFF)
      { 
        // ok, a valid prescaler was found: get it!
        u8Presc = getPresc(u8Presc, fBaud, clock);
      }
    }
    calculateBitTimings(Tseg1AndTseg2Int, u8Presc, fsamplePoint, fBaud, clock, &tseg1, &tseg2, &presc); 
    
    fNewBaud = clock/(presc*(1+tseg1+tseg2));
    
    if(fabs(fNewBaud-fBaud) > 0.1)
    {  
      // no valid baudrate can be found!
      return CANL2_ERR;  
    }
  }
  
  *ppresc = presc;
  *psjw = sjw;
  *ptseg1 = tseg1;
  *ptseg2 = tseg2;
  *psam = div8x;
  
  return ret;
}


/* +--- initialize_chip(2) -------------------------------------------*/
static INT32 initialize_chip(CAN_HANDLE can_channel,INT32 presc, INT32 sjw, INT32 tseg1, INT32 tseg2, INT32 sam)
{
  INT32 rt = CANLIB_FAIL;
  USIGN32 pram[10];
  L2CONFIG SCIMCfg;
  INT32 ret;
    
  memset(pram,0,sizeof(pram));    
    
  ret = getScimConfig(can_channel, &SCIMCfg);
    
    

  if(ret)
  { // cfg not found!
    pram[0] =  (INT32) presc;
    pram[1] =  (INT32) sjw;
    pram[2] =  (INT32) tseg1;
    pram[3] =  (INT32) tseg2;
    pram[4] =  (INT32) sam;
  }
  else
  {
      
      
    pram[0] =  (INT32) (presc == -1) ? SCIMCfg.s32Prescaler : presc;
    pram[1] =  (INT32) (sjw == -1) ?   SCIMCfg.s32Sjw       : sjw;
    pram[2] =  (INT32) (tseg1 == -1) ? SCIMCfg.s32Tseg1     : tseg1;
    pram[3] =  (INT32) (tseg2 == -1) ? SCIMCfg.s32Tseg2     : tseg2;
    pram[4] =  (INT32) (sam == -1) ?   SCIMCfg.s32Sam       : sam;
      
  }
    


  rt = ioctl(can_channel,CAN_IOC_CANL2_initialize_chip, pram);
  
  if(rt > 0x10000)
  {
    rt = -(rt - 0x10000);
  }



  return rt;


}

static INT32 checkCfg(L2CONFIG *pCfg, L2CONFIG *pUserCfg)
{
  INT32 ret = 0;
  
  if(pUserCfg->bEnableAck != GET_FROM_SCIM)
    pCfg->bEnableAck = pUserCfg->bEnableAck;
  
  if(pUserCfg->bEnableErrorframe != GET_FROM_SCIM)
    pCfg->bEnableErrorframe = pUserCfg->bEnableErrorframe;
  
  if(pUserCfg->s32AccCodeStd != GET_FROM_SCIM)
    pCfg->s32AccCodeStd = pUserCfg->s32AccCodeStd;
  
  if(pUserCfg->s32AccMaskStd != GET_FROM_SCIM)
    pCfg->s32AccMaskStd = pUserCfg->s32AccMaskStd;
  
  if(pUserCfg->s32AccCodeXtd != GET_FROM_SCIM)
    pCfg->s32AccCodeXtd = pUserCfg->s32AccCodeXtd;
  
  if(pUserCfg->s32AccMaskXtd != GET_FROM_SCIM)
    pCfg->s32AccMaskXtd = pUserCfg->s32AccMaskXtd;
  
  if(pUserCfg->s32OutputCtrl != GET_FROM_SCIM)
    pCfg->s32OutputCtrl = pUserCfg->s32OutputCtrl;
  
  if(pUserCfg->s32Prescaler != GET_FROM_SCIM)
  {
    if((pUserCfg->s32Prescaler > 32) || (pUserCfg->s32Prescaler < 1))
    {
      return CANL2_WRONG_PARAM;
    }
    pCfg->s32Prescaler = pUserCfg->s32Prescaler;
  }
  
  if(pUserCfg->s32Sam != GET_FROM_SCIM)
  {
    if(pUserCfg->s32Sam > 3)
    {
      return CANL2_WRONG_PARAM;
    }
    pCfg->s32Sam = pUserCfg->s32Sam;
  }
  
  if(pUserCfg->s32Sjw != GET_FROM_SCIM)
  {
    if((pUserCfg->s32Sjw > 4) || (pUserCfg->s32Sjw < 1))
    {
      return CANL2_WRONG_PARAM;
    }
    pCfg->s32Sjw = pUserCfg->s32Sjw;
  }
  
  if(pUserCfg->s32Tseg1 != GET_FROM_SCIM)
  {
    if((pUserCfg->s32Tseg1 > 16) || (pUserCfg->s32Tseg1 < 1))
    {
      return CANL2_WRONG_PARAM;
    }
    pCfg->s32Tseg1 = pUserCfg->s32Tseg1;
  }
  
  if(pUserCfg->s32Tseg2 != GET_FROM_SCIM)
  {
    if((pUserCfg->s32Tseg2 > 8) || (pUserCfg->s32Tseg2 < 1))
    {
      return CANL2_WRONG_PARAM;
    }
    
    pCfg->s32Tseg2 = pUserCfg->s32Tseg2;
  }
  
  if(pUserCfg->hEvent != INVALID_HANDLE_VALUE)
    pCfg->hEvent = pUserCfg->hEvent;
  
  return ret;
}



INT32 CANL2_initialize_chip(CAN_HANDLE can_channel,INT32 presc, INT32 sjw, INT32 tseg1, INT32 tseg2, INT32 sam)
{

  INT32 ret;
  USIGN32 device_type;
  L2CONFIG ScimConfig, UserCfg;
  
  
              // get the device type
  ret = CANL2_get_device_type(can_channel, &device_type);
  if(ret) return ret;
    
  // printf("presc=%3i,sjw=%3i,tseg1=%3i,tseg2=%3i,sam=%3i\n",presc,sjw,tseg1,tseg2,sam);
  ret = Sja1000ToTwinCAN(&presc, &sjw, &tseg1, &tseg2, &sam, device_type);
  ret = Sja1000ToMultiCAN(&presc, &sjw, &tseg1, &tseg2, &sam, device_type);
  
  UserCfg.s32Prescaler = presc;
  UserCfg.s32Sjw = sjw;
  UserCfg.s32Tseg1 = tseg1;
  UserCfg.s32Tseg2 = tseg2;
  UserCfg.s32Sam = sam;
  
  // printf("presc=%3i,sjw=%3i,tseg1=%3i,tseg2=%3i,sam=%3i,ret=0x%x\n",presc,sjw,tseg1,tseg2,sam,ret);
  // get the configuration build from SCIM 
  ret = getScimConfig(can_channel, &ScimConfig);
  
  if(ret)
  { 
    // error handling use configuration from user
    // no configuration available or corrupt
    memcpy(&ScimConfig, &UserCfg, sizeof(L2CONFIG));
  }
  else
  {
    if(checkCfg(&ScimConfig, &UserCfg))
    {
      return CANL2_WRONG_PARAM;
    }
  }
  
  // set baudrate
  ret = initialize_chip(can_channel, ScimConfig.s32Prescaler, ScimConfig.s32Sjw, 
                        ScimConfig.s32Tseg1, ScimConfig.s32Tseg2, ScimConfig.s32Sam);
  
  return ret;
 
}





/* +--- CANL2_set_mode(2)        -------------------------------------------*/
INT32 CANL2_set_mode(CAN_HANDLE can_channel,INT32 SleepMode, INT32 SpeedMode)
{
    INT32 rt = CANLIB_FAIL;
    
    
    USIGN32 pram[10] = { (INT32) SleepMode,
        (INT32) SpeedMode,
        0, 0, 0, 0, 0, 0, 0, 0
    };


    rt = ioctl(can_channel,CAN_IOC_CANL2_set_mode, pram);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }



    return rt;
}




/* +--- CANL2_set_acceptance(2)  -------------------------------------------*/
INT32 CANL2_set_acceptance(CAN_HANDLE can_channel,USIGN32 AccCodeStd,
                     USIGN32 AccMaskStd,
                     USIGN32 AccCodeXtd, USIGN32 AccMaskXtd)
{
    INT32 rt = CANLIB_FAIL;
    USIGN32 pram[10];
    L2CONFIG SCIMCfg;
    INT32 ret;
    
    memset(pram,0,sizeof(pram));     
    
    ret = getScimConfig(can_channel, &SCIMCfg);
    
 
    if(ret)
    { // cfg not found!
    
      pram[0] = (USIGN32) AccCodeStd;
      pram[1] = (USIGN32) AccMaskStd;
      pram[2] = (USIGN32) AccCodeXtd;
      pram[3] = (USIGN32) AccMaskXtd;
    }
    else
    {
      pram[0] = ((INT32)AccCodeStd == (INT32)-1) ? SCIMCfg.s32AccCodeStd : AccCodeStd;
      pram[1] = ((INT32)AccMaskStd == (INT32)-1) ? SCIMCfg.s32AccMaskStd : AccMaskStd;
      pram[2] = ((INT32)AccCodeXtd == (INT32)-1) ? SCIMCfg.s32AccCodeXtd : AccCodeXtd;
      pram[3] = ((INT32)AccMaskXtd == (INT32)-1) ? SCIMCfg.s32AccMaskXtd : AccMaskXtd;
      
    }


    rt = ioctl(can_channel,CAN_IOC_CANL2_set_acceptance, pram);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }



    return rt;
}



/* +--- CANL2_set_output_control(2)  ---------------------------------------*/
INT32 CANL2_set_output_control(CAN_HANDLE can_channel,INT32 OutputControl)
{

    INT32 rt = CANLIB_FAIL;
    L2CONFIG SCIMCfg;
    INT32 ret;
    
    ret = getScimConfig(can_channel, &SCIMCfg);
    
    if(ret)
    { // no cfg; use value from user!
  
      rt = ioctl(can_channel,CAN_IOC_CANL2_set_output_control,
                 (CAN_ADDRESS) OutputControl);
    }
    else
    {
      if(OutputControl == -1)
      { 
        // -1 means: get it from SCIM-Cfg.
        rt = ioctl(can_channel,CAN_IOC_CANL2_set_output_control,
                   (CAN_ADDRESS) SCIMCfg.s32OutputCtrl);
        
      }
      else
      { 
        rt = ioctl(can_channel,CAN_IOC_CANL2_set_output_control,
                   (CAN_ADDRESS) OutputControl);
        
      }
    }

    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }


    return rt;

}



INT32 CANL2_initialize_interface(CAN_HANDLE can_channel,INT32 ReceiveFifoEnable,
                           INT32 ReceivePollAll,
                           INT32 ReceiveEnableAll,
                           INT32 ReceiveIntEnableAll,
                           INT32 AutoRemoteEnable,
                           INT32 TransmitReqFifoEnable,
                           INT32 TransmitPollAll,
                           INT32 TransmitAckEnableAll,
                           INT32 TransmitAckFifoEnable,
                           INT32 TransmitRmtFifoEnable)
{
    INT32 rt = CANLIB_FAIL;
    USIGN32 pram[10] = { (INT32) ReceiveFifoEnable,
        (INT32) ReceivePollAll,
        (INT32) ReceiveEnableAll,
        (INT32) ReceiveIntEnableAll,
        (INT32) AutoRemoteEnable,
        (INT32) TransmitReqFifoEnable,
        (INT32) TransmitPollAll,
        (INT32) TransmitAckEnableAll,
        (INT32) TransmitAckFifoEnable,
        (INT32) TransmitRmtFifoEnable
    };


    rt = ioctl(can_channel,CAN_IOC_CANL2_initialize_interface, pram);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }


    return rt;
}

/* +--- CANL2_define_object(2)          ------------------------------------*/

INT32 CANL2_define_object(CAN_HANDLE can_channel,USIGN32 Handle,
                    INT32 *ObjectNumber,
                    INT32 Type,
                    INT32 ReceiveIntEnable,
                    INT32 AutoRemoteEnable, INT32 TransmitAckEnable)
{
    INT32 rt = CANLIB_FAIL;
    USIGN32 pram[10] = { (USIGN32) Handle,
        (INT32) *ObjectNumber,
        (INT32) Type,
        (INT32) ReceiveIntEnable,
        (INT32) AutoRemoteEnable,
        (INT32) TransmitAckEnable,
        0, 0, 0, 0
    };


    rt = ioctl(can_channel,CAN_IOC_CANL2_define_object, pram);


    *ObjectNumber = (INT32) pram[1];
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }


    return rt;

}



/* +--- CANL2_define_cyclic(2)        --------------------------------------*/

INT32 CANL2_define_cyclic(CAN_HANDLE can_channel,INT32 ObjectNumber, USIGN32 Rate, USIGN32 Cycles)
{

    INT32 rt = CANLIB_FAIL;
    USIGN32 pram[10] = { (INT32) ObjectNumber,
        (USIGN32) Rate,
        (USIGN32) Cycles,
        0, 0, 0, 0, 0, 0, 0
    };


    rt = ioctl(can_channel,CAN_IOC_CANL2_define_cyclic, pram);


    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }

    return rt;

}



/* +--- CANL2_enable_fifo            ---------------------------------------*/

INT32 CANL2_enable_fifo(CAN_HANDLE can_channel)
{

    INT32 rt = CANLIB_FAIL;


    rt = ioctl(can_channel,CAN_IOC_CANL2_enable_fifo, NULL);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }



    return rt;
}




/* +--- CANL2_enable_dyn_obj_buf     ---------------------------------------*/

INT32 CANL2_enable_dyn_obj_buf(CAN_HANDLE can_channel)
{

    INT32 rt = CANLIB_FAIL;


    rt = ioctl(can_channel,CAN_IOC_CANL2_enable_dyn_obj_buf, NULL);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }



    return rt;
}

/* +--- CANL2_enable_timestamps      ---------------------------------------*/
/* this function has no effect on the CANcard                               */

INT32 CANL2_enable_timestamps(CAN_HANDLE can_channel)
{

    INT32 rt = CANLIB_FAIL;


    rt = ioctl(can_channel,CAN_IOC_CANL2_enable_timestamps, NULL);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }



    return rt;
}

// CAN_IOC_CANL2_trace_level
INT32 CANL2_trace_level(CAN_HANDLE can_channel, USIGN32 level)
{

    INT32 rt = CANLIB_FAIL;
    USIGN32 pram[10] = { (INT32) level,
        0,
        0, 0, 0, 0, 0, 0, 0, 0
    };

    rt = ioctl(can_channel,CAN_IOC_CANL2_trace_level, pram);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }



    return rt;
}

/* +--- CANL2_enable_fifo_transmit_ack  ------------------------------------*/

INT32 CANL2_enable_fifo_transmit_ack(CAN_HANDLE can_channel)
{

    INT32 rt = CANLIB_FAIL;


    rt = ioctl(can_channel,CAN_IOC_CANL2_enable_fifo_transmit_ack, NULL);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }



    return rt;

}



/* +--- CANL2_get_version            ---------------------------------------*/
/* the parameter hw_version is not used with the CAN-AC2                    */

INT32 CANL2_get_version(CAN_HANDLE can_channel,INT32 *sw_version,
                  INT32 *fw_version,
                  INT32 *hw_version, INT32 *license, INT32 *can_chip_type)
{
    INT32 rt = CANLIB_FAIL;
    USIGN32 pram[10] = { (INT32) *sw_version,
        (INT32) *fw_version,
        (INT32) *hw_version,
        (INT32) *license,
        (INT32) *can_chip_type,
        0, 0, 0, 0, 0
    };


    rt = ioctl(can_channel,CAN_IOC_CANL2_get_version, pram);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }



    *sw_version = pram[0];
    *fw_version = pram[1];
    *hw_version = pram[2];
    *license = pram[3];
    *can_chip_type = pram[4];
    can_chip_type[1] = pram[5];

    return rt;

}

/* +--- CANL2_get_serial_number      ---------------------------------------*/
/* this function has no effect on the CAN-AC2                               */

INT32 CANL2_get_serial_number(CAN_HANDLE can_channel,USIGN32 *ser_number)
{

  INT32 rt = CANLIB_FAIL;



  rt = ioctl(can_channel, CAN_IOC_CANL2_get_serial_number, ser_number);
  
  if(rt > 0x10000)
  {
    rt = -(rt - 0x10000);
  }



  return rt;

}
/* +--- CANL2_get_phys_channel      ---------------------------------------*/
/* this function has no effect on the CAN-AC2                               */

INT32 CANL2_get_phys_channel(CAN_HANDLE can_channel, USIGN32 *channel_number)
{

  INT32 rt = CANLIB_FAIL;



  rt = ioctl(can_channel, CAN_IOC_CANL2_get_phys_channel, channel_number);
  
  if(rt > 0x10000)
  {
    rt = -(rt - 0x10000);
  }



  return rt;

}

/* +--- CANL2_get_device_type      ---------------------------------------*/
/* this function has no effect on the CAN-AC2                               */

INT32 CANL2_get_device_type(CAN_HANDLE can_channel, USIGN32 *device_type)
{

  INT32 rt = CANLIB_FAIL;



  rt = ioctl(can_channel, CAN_IOC_CANL2_get_device_type, device_type);
  
  if(rt > 0x10000)
  {
    rt = -(rt - 0x10000);
  }



  return rt;

}

/* +--- CANL2_start_chip             ---------------------------------------*/

INT32 CANL2_start_chip(CAN_HANDLE can_channel)
{

    INT32 rt = CANLIB_FAIL;


    rt = ioctl(can_channel,CAN_IOC_CANL2_start_chip, NULL);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }



    return rt;

}

/* +--- CANL2_send_remote_object(2)  ---------------------------------------*/

INT32 CANL2_send_remote_object(CAN_HANDLE can_channel,INT32 ObjectNumber, INT32 DataLength)
{

    INT32 rt = CANLIB_FAIL;
    USIGN32 pram[10] = { (INT32) ObjectNumber,
        (INT32) DataLength,
        0, 0, 0, 0, 0, 0, 0, 0
    };


    rt = ioctl(can_channel,CAN_IOC_CANL2_send_remote_object, pram);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }



    return rt;

}



/* +--- CANL2_supply_object_data(2)  ---------------------------------------*/

INT32 CANL2_supply_object_data(CAN_HANDLE can_channel,INT32 ObjectNumber,
                         INT32 DataLength, unsigned char *pData)
{
    INT32 rt = CANLIB_FAIL;
    USIGN32 pram[10] = { (INT32) ObjectNumber,
        (INT32) DataLength,
        *(USIGN32 *) pData,
        *(((USIGN32 *) pData) + 1),
        0, 0, 0, 0, 0, 0
    };


    rt = ioctl(can_channel,CAN_IOC_CANL2_supply_object_data, pram);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }


    return rt;


}



/* +--- CANL2_supply_rcv_object_data(2)-------------------------------------*/

INT32 CANL2_supply_rcv_object_data(CAN_HANDLE can_channel,INT32 ObjectNumber,
                             INT32 DataLength, unsigned char *pData)
{
    INT32 rt = CANLIB_FAIL;
    USIGN32 pram[10] = { (INT32) ObjectNumber,
        (INT32) DataLength,
        *(USIGN32 *) pData,
        *(((USIGN32 *) pData) + 1),
        0, 0, 0, 0, 0, 0
    };


    rt = ioctl(can_channel,CAN_IOC_CANL2_supply_rcv_object_data, pram);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }

    return rt;

}



/* +--- CANL2_send_object(2)         ---------------------------------------*/

INT32 CANL2_send_object(CAN_HANDLE can_channel,INT32 ObjectNumber, INT32 DataLength)
{

    INT32 rt = CANLIB_FAIL;
    USIGN32 pram[10] = { (INT32) ObjectNumber,
        (INT32) DataLength,
        0, 0, 0, 0, 0, 0, 0, 0
    };


    rt = ioctl(can_channel,CAN_IOC_CANL2_send_object, pram);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }



    return rt;

}



/* +--- CANL2_write_object(2)        ---------------------------------------*/

INT32 CANL2_write_object(CAN_HANDLE can_channel,INT32 ObjectNumber, INT32 DataLength, unsigned char *pData)
{

    INT32 rt = CANLIB_FAIL;
    USIGN32 pram[10] = { (INT32) ObjectNumber,
        (INT32) DataLength,
        *(USIGN32 *) pData,
        *(((USIGN32 *) pData) + 1),
        0, 0, 0, 0, 0, 0
    };


    rt = ioctl(can_channel,CAN_IOC_CANL2_write_object, pram);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }


    return rt;

}


INT32 CANL2_read_rcv_data(CAN_HANDLE can_channel,INT32 ObjectNumber,
                    unsigned char *pRCV_Data, USIGN32 *Time)
{
    INT32 rt = CANLIB_FAIL;
    USIGN32 pram[10] = { (INT32) ObjectNumber,
        0,                      /* (unsigned char)*pRCV_Data */
        0,                      /* (USIGN32)*Time */
        0, 0, 0, 0, 0, 0, 0
    };


    rt = ioctl(can_channel,CAN_IOC_CANL2_read_rcv_data, pram);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }


    *(USIGN32 *) pRCV_Data = *(USIGN32 *) &pram[1];
    *(USIGN32 *) &pRCV_Data[4] = *(USIGN32 *) &pram[2];
    *Time = (USIGN32) pram[3];

    return rt;

}



INT32 CANL2_read_xmt_data(CAN_HANDLE can_channel,INT32 ObjectNumber,
                    INT32 *pDataLength, unsigned char *pXMT_Data)
{
    INT32 rt = CANLIB_FAIL;
    USIGN32 pram[10] = { (INT32) ObjectNumber,
        0,                      /* (unsigned char)*pDataLength */
        0,                      /* (USIGN32)*pXMT_Data */
        0, 0, 0, 0, 0, 0, 0
    };


    rt = ioctl(can_channel,CAN_IOC_CANL2_read_xmt_data, pram);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }



    *pDataLength = (unsigned char) pram[1];
    *(USIGN32 *) pXMT_Data = *(USIGN32 *) &pram[2];
    *(USIGN32 *) &pXMT_Data[4] = *(USIGN32 *) &pram[3];

    return rt;

}



INT32 CANL2_read_ac(CAN_HANDLE can_channel, PARAM_STRUCT * param)
{

    INT32 rt = CANLIB_FAIL;


    rt = ioctl(can_channel,CAN_IOC_CANL2_read_ac, param);
    param->Can = can_channel;
    
    
   // printf("rt1 = %u\n", rt);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }

    
   // printf("rt2 = %u\n", -rt);

    return rt;

}

/* +--- CANL2_send_data              ---------------------------------------*/

INT32 CANL2_send_data(CAN_HANDLE can_channel,USIGN32 Ident,
                INT32 Xtd, INT32 DataLength, unsigned char *pData)
{
    INT32 rt = CANLIB_FAIL;
    USIGN32 pram[10] = { (USIGN32) Ident,
        (INT32) Xtd,
        (INT32) DataLength,
        *(USIGN32 *) pData,
        *(((USIGN32 *) pData) + 1),
        0, 0, 0, 0, 0
    };


    rt = ioctl(can_channel,CAN_IOC_CANL2_send_data, pram);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }

    
    

    return rt;

}


/* +--- CANL2_send_remote            ---------------------------------------*/

INT32 CANL2_send_remote(CAN_HANDLE can_channel,USIGN32 Ident, INT32 Xtd, INT32 DataLength)
{
    INT32 rt = CANLIB_FAIL;
    USIGN32 pram[10] = { (USIGN32) Ident,
        (INT32) Xtd,
        (INT32) DataLength,
        0, 0, 0, 0, 0, 0, 0
    };


    rt = ioctl(can_channel,CAN_IOC_CANL2_send_remote, pram);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }



    return rt;

}






/* +--- CANL2_reinitialize           ---------------------------------------*/
INT32 CANL2_reinitialize(CAN_HANDLE can_channel)
{

    INT32 rt = CANLIB_FAIL;


    rt = ioctl(can_channel,CAN_IOC_CANL2_reinitialize, NULL);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }



    return rt;

}

/* +--- CANL2_get_time               ---------------------------------------*/
INT32 CANL2_get_time(CAN_HANDLE can_channel,USIGN32 *time)
{

    INT32 rt = CANLIB_FAIL;
    USIGN32 pram[10] = { (INT32) *time,
        0, 0, 0, 0, 0, 0, 0, 0, 0
    };


    rt = ioctl(can_channel,CAN_IOC_CANL2_get_time, pram);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }



    *time = (INT32) pram[0];

    return rt;
}

/* +--- CANL2_get_bus_state          ---------------------------------------*/
INT32 CANL2_get_bus_state(CAN_HANDLE can_channel)
{

    INT32 rt = CANLIB_FAIL;
    USIGN32 pram[10] = { (INT32) 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0
    };


    rt = ioctl(can_channel,CAN_IOC_CANL2_get_bus_state, pram);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }



    return rt;

}



/* +--- CANL2_reset_rcv_fifo     -------------------------------------------*/
INT32 CANL2_reset_rcv_fifo(CAN_HANDLE can_channel)
{
    INT32 rt = CANLIB_FAIL;

    rt = ioctl(can_channel,CAN_IOC_CANL2_reset_rcv_fifo, NULL);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }


    return rt;
}

/* +--- CANL2_reset_xmt_fifo     -------------------------------------------*/
INT32 CANL2_reset_xmt_fifo(CAN_HANDLE can_channel)
{
    INT32 rt = CANLIB_FAIL;

    rt = ioctl(can_channel,CAN_IOC_CANL2_reset_xmt_fifo, NULL);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }


    return rt;
}

/* +--- CANL2_reset_lost_msg_counter ---------------------------------------*/
INT32 CANL2_reset_lost_msg_counter(CAN_HANDLE can_channel)
{
    INT32 rt = CANLIB_FAIL;

    rt = ioctl(can_channel,CAN_IOC_CANL2_reset_lost_msg_counter, NULL);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }


    return rt;
}

/* +--- CANL2_read_xmt_fifo_level ------------------------------------------*/
INT32 CANL2_read_xmt_fifo_level(CAN_HANDLE can_channel)
{
    INT32 rt = CANLIB_FAIL;

    rt = ioctl(can_channel,CAN_IOC_CANL2_read_xmt_fifo_level, NULL);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }


    return rt;
}

/* +--- CANL2_read_xmt_fifo_level ------------------------------------------*/



/* +--- CANL2_read_rcv_fifo_level ------------------------------------------*/
INT32 CANL2_read_rcv_fifo_level(CAN_HANDLE can_channel)
{
    INT32 rt = CANLIB_FAIL;

    rt = ioctl(can_channel,CAN_IOC_CANL2_read_rcv_fifo_level, NULL);
    
    if(rt > 0x10000)
    {
      rt = -(rt - 0x10000);
    }


    return rt;
}

static BOOL getIntVal(char *str, char * format, char *tok, INT32 *pNumber)
{
  char *pStart = strstr(str,tok);
  
  // substring found?
  if(!pStart) return FALSE;
  // read till start of the number
  pStart += strlen(tok);
  
  sscanf(pStart,format,pNumber);
  
  return TRUE;
}

static INT32 getChConfig(USIGN32 u32Serial, USIGN32 u32physChannel, USIGN32 u32DeviceType, char *CfgLine)
{
  USIGN32 u32DeviceType2;
  BOOL bCfgFound = FALSE;
  char lineStart[20];
  FILE *pCfgFile;

  
  memset(CfgLine,0,LINE_SIZE); 
  sprintf(lineStart, "%09u_%u", u32Serial, u32physChannel);
  
 // printf("looking for: %s\n",lineStart);
  
  // open the config file for reading!
  pCfgFile = fopen(CAN_CONFIG_FILE,"r");
  
  if(!pCfgFile)
    return -1;
  
  // looking for the part, which describes that channel
  while(!bCfgFound && !feof(pCfgFile))
  {
    if(fgets(CfgLine,2*LINE_SIZE,pCfgFile))
    {
      // printf("checking %s\n",CfgLine);
      if(strlen(CfgLine) >= strlen(lineStart))
      {
        if(!memcmp(CfgLine,lineStart,strlen(lineStart)))
        {
          if(!getIntVal(CfgLine,"%u","type:",(INT32*)&u32DeviceType2))
          {
            // printf("type not found\n");
            bCfgFound = FALSE;
          }
          else
          {
            // printf("%lu==%lu\n",u32DeviceType2,u32DeviceType);
            bCfgFound = (u32DeviceType2==u32DeviceType);
          }
        }
      }
    }
    
  }
  
  fclose(pCfgFile);
  
  if(!bCfgFound)
  { // channel configuration not found!    
    return -2;
  }
   
  return 0;
  
}
/*
INT32 CANL2_get_device_id(CAN_Handle Can, USIGN32 *puDeviceId)
{

  return CANL2_get_device_type(Can, puDeviceId);
}
*/

// get the channel configruration from config file
static INT32 getScimConfig(CAN_HANDLE Can, L2CONFIG *pScimCfg)
{
  INT32 ret = 0;
  char line[2 * LINE_SIZE];
  INT32 enableAck,enableErrorframe;
  USIGN32 u32Serial, u32physChannel, u32DeviceType1=0;
  
 // get the serial number
  ret = CANL2_get_serial_number(Can, &u32Serial);
  if(ret) return ret;
  
  // get the pysical channel number (1..2)
  ret = CANL2_get_phys_channel(Can, &u32physChannel);
  if(ret) return ret;
  
    // get the device type
  ret = CANL2_get_device_type(Can, &u32DeviceType1);
  if(ret) return ret;

  ret = getChConfig(u32Serial, u32physChannel, u32DeviceType1, line);
  
  if(ret)
  {
   // printf("getChConfig returned %i\n",ret);
    return ret;
  }
      
  // configuration found!
  if(!getIntVal(line,"%u","ack:",&enableAck))
    return -3;
  
  pScimCfg->bEnableAck = (enableAck == 1);
  
  if(!getIntVal(line,"%u","errorframes:",&enableErrorframe))
    return -3;
      
  pScimCfg->bEnableErrorframe = (enableErrorframe == 1);
  
  if(!getIntVal(line,"%x","acccstd:",&pScimCfg->s32AccCodeStd))
    return -3;
  
  if(!getIntVal(line,"%x","accmstd:",&pScimCfg->s32AccMaskStd))
    return -3;
  
  if(!getIntVal(line,"%x","acccxtd:",&pScimCfg->s32AccCodeXtd))
    return -3;
  
  if(!getIntVal(line,"%x","accmxtd:",&pScimCfg->s32AccMaskXtd))
    return -3;
  
  if(!getIntVal(line,"%x","ocr:",&pScimCfg->s32OutputCtrl))
    return -3;
  
  if(!getIntVal(line,"%u","presc:",&pScimCfg->s32Prescaler))
    return -3;
  
  if(!getIntVal(line,"%u","sam:",&pScimCfg->s32Sam))
    return -3;
  
  if(!getIntVal(line,"%u","sjw:",&pScimCfg->s32Sjw))
    return -3;
  
  if(!getIntVal(line,"%u","tseg1:",&pScimCfg->s32Tseg1))
    return -3;
  
  if(!getIntVal(line,"%u","tseg2:",&pScimCfg->s32Tseg2))
    return -3;
  
  return 0;
}




INT32 CANL2_initialize_fifo_mode(CAN_HANDLE Can, L2CONFIG *pUserCfg)
{
  INT32 ret = 0;
  L2CONFIG ScimConfig;
  USIGN32 device_type;
  
                // get the device type
  ret = CANL2_get_device_type(Can, &device_type);
  if(ret) return ret;

    
  // reset can chip
  ret = CANL2_reset_chip(Can);
  if(ret) return ret;
  
  // printf("presc=%3i,sjw=%3i,tseg1=%3i,tseg2=%3i,sam=%3i\n",pUserCfg->s32Prescaler,pUserCfg->s32Sjw,pUserCfg->s32Tseg1,pUserCfg->s32Tseg2,pUserCfg->s32Sam);
  ret = Sja1000ToTwinCAN(&pUserCfg->s32Prescaler, &pUserCfg->s32Sjw, &pUserCfg->s32Tseg1, &pUserCfg->s32Tseg2, &pUserCfg->s32Sam, device_type);
  // printf("presc=%3i,sjw=%3i,tseg1=%3i,tseg2=%3i,sam=%3i,ret=0x%x\n",pUserCfg->s32Prescaler,pUserCfg->s32Sjw,pUserCfg->s32Tseg1,pUserCfg->s32Tseg2,pUserCfg->s32Sam,ret);
  ret = Sja1000ToMultiCAN(&pUserCfg->s32Prescaler, &pUserCfg->s32Sjw, &pUserCfg->s32Tseg1, &pUserCfg->s32Tseg2, &pUserCfg->s32Sam, device_type);
  // printf("presc=%3i,sjw=%3i,tseg1=%3i,tseg2=%3i,sam=%3i,ret=0x%x\n",pUserCfg->s32Prescaler,pUserCfg->s32Sjw,pUserCfg->s32Tseg1,pUserCfg->s32Tseg2,pUserCfg->s32Sam,ret);

  // get the configuration build from SCIM 
  ret = getScimConfig(Can, &ScimConfig);
  
  if(ret)
  { 
    // error handling use configuration from user
    // no configuration available or corrupt
    memcpy(&ScimConfig, pUserCfg, sizeof(L2CONFIG));
  }
  else
  {
    if(checkCfg(&ScimConfig, pUserCfg))
    {
      return CANL2_WRONG_PARAM;
    }
  }

  // set baudrate
  ret = initialize_chip(Can, ScimConfig.s32Prescaler, ScimConfig.s32Sjw, 
                              ScimConfig.s32Tseg1, ScimConfig.s32Tseg2, ScimConfig.s32Sam);
  if(ret) return ret;

  // set acceptance filter
  ret = CANL2_set_acceptance(Can, ScimConfig.s32AccCodeStd, ScimConfig.s32AccMaskStd, 
                             ScimConfig.s32AccCodeXtd, ScimConfig.s32AccMaskXtd);
  if(ret) return ret;

  // define output control register
  ret = CANL2_set_output_control(Can, ScimConfig.s32OutputCtrl);
  if(ret) return ret;

  // select fifo mode
  ret = CANL2_enable_fifo(Can);
  if(ret) return ret;

  // should the appl. be informed, if a message is successful transmitted?
  if(ScimConfig.bEnableAck)
  {
    ret = CANL2_enable_fifo_transmit_ack(Can);
  }

  if(ret) return ret;

  // should the appl. be informed about error frames on the bus?
  if(ScimConfig.bEnableErrorframe)
  {
    ret = CANL2_enable_error_frame_detection(Can);
  }

  if(ret) return ret;

  // go online
  ret = CANL2_start_chip(Can);
  
  return ret;
}

/*HEAD_BEG*******************************************************************/
/*                                                                          */
/* FUNCTION                       get_CAN_channels                          */
/*                                                                          */
/* FUNCTIONAL_DESCRIPTION         retrives all channels in a specified proc */
/*                                directory                                 */
/*                                                                          */
/* INPUTS                                                                   */
/*                                pProcDir: path to the proc file           */
/*                                pBuffer:  buffer where the info about     */
/*                                          the CAN channels will be stored */
/*                                max:      number of free entries in       */
/*                                          pBuffer                         */
/*                                                                          */
/* OUTPUTS                        number of channels found in the directory */
/*                                                                          */
/*HEAD_END*******************************************************************/

static INT32 get_CAN_channels(char *pProcDir, PCHDSNAPSHOT pBuffer, INT32 max)
{
  
  struct dirent **namelist;
  INT32 n, min=0;
  BOOL bCfgNotFound = TRUE;

  n = scandir(pProcDir, &namelist, 0, alphasort);
  if (n < 0)
  {
    
  }
  else 
  {
    while(n--) 
    {
      FILE *pDev = NULL;
      char line[LINE_SIZE];
      
      
      if(
         strcmp(namelist[n]->d_name,"..") && 
         strcmp(namelist[n]->d_name,".")
        )
      {
        strcpy(line,pProcDir);
        strcat(line,namelist[n]->d_name);
        pDev = fopen(line,"r");
        if(pDev)
        {
          USIGN32 u32Serial,u32PhysCh,u32DeviceType,u32FwId,bIsOpen;
          
          memset(line,0,LINE_SIZE);
          if(fgets(line, LINE_SIZE, pDev))
          {
          
            sscanf(namelist[n]->d_name,"%u",&u32Serial);            
            getIntVal(line,"%u", "channel:", (INT32*)&u32PhysCh);
            getIntVal(line,"%u", "type:", (INT32*)&u32DeviceType);
            
            getIntVal(line,"%u", "firmware:", (INT32*)&u32FwId);
            getIntVal(line,"%u", "open:", (INT32*)&bIsOpen);
            bCfgNotFound = getChConfig(u32Serial, u32PhysCh, u32DeviceType, line);
            
            if(min < max)
            {
              char *pCh;
              
              if(!bCfgNotFound)
              {
                pBuffer[min].u32Serial = u32Serial;            
                pBuffer[min].u32PhysCh = u32PhysCh;
                pBuffer[min].u32DeviceType = u32DeviceType;
                
                pBuffer[min].u32FwId = u32FwId;
                pBuffer[min].bIsOpen = bIsOpen;
                
                pCh = strstr(line,"\n");
                if(pCh) *pCh = 0;
                
                pCh = strstr(line,"name:");
                
                if(pCh)
                {
                  pCh += sizeof("name:")-1;
                  if(strlen(pCh) >= sizeof(pBuffer[min].ChannelName))
                  {
                    memcpy(pBuffer[min].ChannelName, pCh, sizeof(pBuffer[min].ChannelName));
                  }
                  else
                  {
                    memcpy(pBuffer[min].ChannelName, pCh, strlen(pCh)+1);
                  }
                }
              }
            }
          }
          fclose(pDev);
          if(!bCfgNotFound) min++;
        }
        else
        {
          //printf("error: can not open file [%s] for reading\n",line);
        }
      }
      
      free(namelist[n]);
    }
    free(namelist);
  }  
  
  return min;
  
}

/*HEAD_BEG*******************************************************************/
/*                                                                          */
/* FUNCTION                       CANL2_get_all_CAN_channels                */
/*                                                                          */
/* FUNCTIONAL_DESCRIPTION         retrives all plugged devices from the     */
/*                                driver                                    */
/*                                                                          */
/* INPUTS                         -                                         */
/* OUTPUTS                        OK_                                       */
/*                                CANL2_FW_TIMEOUT    timeout of firmware             */
/*                                                                          */
/*HEAD_END*******************************************************************/
INT32 CANL2_get_all_CAN_channels(USIGN32 u32ProvidedBufferSize,
                                OUT PUSIGN32 pu32NeededBufferSize,
                                OUT PUSIGN32 pu32NumOfChannels,
                                OUT PCHDSNAPSHOT pBuffer)
                                   
{
  INT32 s32Max = u32ProvidedBufferSize/sizeof(CHDSNAPSHOT);
  INT32 s32Defined;
  
  // the information about CAN channels will be published by the driver
  // as the contents of proc-files in the /proc directory
  
  s32Defined = get_CAN_channels("/proc/softing_can_pci/",pBuffer,s32Max); 
  s32Defined += get_CAN_channels("/proc/softing_can_pcmcia/",&pBuffer[s32Defined],s32Max-s32Defined); 
  s32Defined += get_CAN_channels("/proc/softing_can_104/",&pBuffer[s32Defined],s32Max-s32Defined); 
  s32Defined += get_CAN_channels("/proc/softing_can_eco/",&pBuffer[s32Defined],s32Max-s32Defined); 
  s32Defined += get_CAN_channels("/proc/softing_can_pro/",&pBuffer[s32Defined],s32Max-s32Defined); 
  s32Defined += get_CAN_channels("/proc/softing_can_pcie/",&pBuffer[s32Defined],s32Max-s32Defined); 
  s32Defined += get_CAN_channels("/proc/on_board_can/",&pBuffer[s32Defined],s32Max-s32Defined); 
  s32Defined += get_CAN_channels("/proc/softing_canprousb/",&pBuffer[s32Defined],s32Max-s32Defined); 
  *pu32NumOfChannels = s32Defined;
  *pu32NeededBufferSize = s32Defined * sizeof(CHDSNAPSHOT);
  
  return 0;
}

INT32 INIL2_initialize_channel(CAN_HANDLE *channel_handle, char *channel_name)
{
  INT32 can_fd = -1;
  INT32 rt;
  CANPC_RESOURCES crResources;
  char name[LINE_SIZE];
  
  // all device file names (channel names) start with Softing_AG_
  strcpy(name,DEVFILEPREFIX);
  strcat(name,channel_name);
  
  can_fd = open(name, O_RDONLY);
  //printf(" open %s\n",name);
  
  if(can_fd < 0)
    return VCARD_DEVICENOTFOUND;
  
  rt = ioctl(can_fd, CAN_IOC_INIL2_initialize_board, (USIGN32 *) &crResources);
  
  if(rt > 0x10000)
    return VCARD_DEVICEACCESSERROR;
  
  rt = ioctl(can_fd, CAN_IOC_CANL2_reset_board, NULL);
  
  *channel_handle = can_fd;
  
  return rt;
}

/* +--- INIPC_close_board         ------------------------------------------*/
INT32 INIL2_close_channel(CAN_HANDLE channel_handle)
{
  
  close(channel_handle);
 
  return 0;
}

