
/***************************************************************************
 *   mk_devices.c - generates the device files for the Softing CAN devices *
 *   Copyright (C) 2003-2015 by Softing Industrial Automation GmbH                    *
 *   support.automation@softing.com                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


/**
 *
 *
 * <h2>make devices </h2>
 * \image html softing.jpg
 * \image latex softing.eps
 * <br>
 * Richard-Reitzner-Allee 6 <br>
 * 85540 Haar <br>
 * Phone +49 89/456 56-337 <br>
 * Fax   +49 89/456 56-499 <br>
 *
 * Author: Werner Knorsch
 *
 * 
 *
 *
 * \section Description
 *
 * This document contains the source code for the device configurator
 * The device configurator consists of the following files:
 * mk_devices.c
 * BaseTypes.h 
 * can_def.h 
 * canL2.h
 *
 * 
 * \section 2. section
 *
 *
 * */
 

#include <stdio.h>
#include <pthread.h>
#include <errno.h>
#include <sys/poll.h>
#include <stdlib.h>

// directory
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#include <linux/kdev_t.h>
#include <dirent.h>
#include <string.h>
#include <sys/stat.h>

#include "BaseTypes.h"
#include "can_def.h"
#include "canL2.h"

#define DEVFILEPREFIX "/dev/Softing_AG_"
#define CAN_CONFIG_FILE "/etc/softing/candevices.conf"
#define LINE_SIZE 300

#define SOFTING_VERSION "mk_devices V5.18"

// description of a CAN channel
typedef struct
{
  USIGN32 m_serialNumber; // serial number of the device
  USIGN32 m_physChannel; // physical channel number (1 or 2)
  USIGN32 m_deviceType; // device type (CANcard2 = 5, CAN-ACx-PCI = 7, CAN-ACx-PCI/DN = 8, CAN-ACx-104 = 9, CAN-ECOx-104 = 12, 10 = CANusb, 0x105 = EDICcard2)
  unsigned long m_dpram; // DPRAM Address
  USIGN32 m_ioport; // IoPort Address, valid only for CAN-ACx-104
  USIGN32 m_irq; // interrupt number
  char          m_channelName[80]; // channel name from the config file "/etc/softing/candevices.conf"
} CHANNEL, *PCHANNEL;

char *getDeviceName(unsigned int deviceType);


// gets the device name from the device type
char *getDeviceType(int u32DeviceType)
{
  static char u8Type[100];

  switch(u32DeviceType)
  {
    case 5: 
      // CANcard2
      strcpy(u8Type,"/proc/softing_can_pcmcia");
      break;

    case 7: 
        // CAN-ACx-PCI
      strcpy(u8Type,"/proc/softing_can_pci");
      break;

    case 8: 
      // CAN-ACx-PCI/DN
      strcpy(u8Type,"/proc/softing_can_pci");
      break;

    case 9: 
      // CAN-ACx-104
      strcpy(u8Type,"/proc/softing_can_104");
      break;
      
    case 12:
      // CAN-ECOx-104
      strcpy(u8Type,"/proc/softing_can_eco");
      break;

    case 10:
      // CANusb
      strcpy(u8Type,"/proc/softing_can_usb");
      break;

    case 13: 
       // CAN-PROx-PCIe
      strcpy(u8Type,"/proc/softing_can_pcie");
      break;

    case 14: 
       // CAN-PROx-104+
      strcpy(u8Type,"/proc/softing_can_pro");
      break;

     case 15: 
       // CANpro USB
      strcpy(u8Type,"/proc/softing_canprousb");
      break;

    case 0x105: 
      // EDICcard2
      strcpy(u8Type,"/proc/softing_can_pcmcia");
      break;

    case 0x8000: 
      // PC8 on board CAN 
      strcpy(u8Type,"/proc/on_board_can");
      break;

    default: 
      strcpy(u8Type,"UNKNOWN");
      break;

  }
  return u8Type;
}

// print some error messages, if an error happened
void printMknodError(void)
{
  switch(errno)
  {
    case EPERM:
      printf("operation not permitted!\n");
      break;
    case EINVAL:
      printf("wrong device type!\n");
      break;
    case EEXIST:
      printf("device allready exists!\n");
      break;
    case EFAULT:
      printf("invalid path!\n");
      break;
    case EACCES:
      printf("writing not possible or program can not be executed!\n");
      break;
    case ENAMETOOLONG:
      printf("pathname too long!\n");
      break;
    case ELOOP:
    case ENOENT:
    case ENOTDIR:
      printf("invalid path!\n");
      break;
    case EROFS:
      printf("can not create a file on a read only file system!\n");
      break;
    case ENOMEM:
    case ENOSPC:
      printf("out of memory!\n");
      break;
   
    
    
      default: printf("error %u occurred!\n",errno);
      break;
  }
  
  
}

// get a int value out of the config file
BOOL getIntVal(char *str, char * format, char *tok, INT32 *pNumber)
{
  char *pStart = strstr(str,tok);
  
  // substring found?
  if(!pStart) return FALSE;
  // read till start of the number
  pStart += strlen(tok);
  
  sscanf(pStart,format,pNumber);
  
  return TRUE;
}

// get the channel configuration
static int getChConfig(USIGN32 u32Serial, USIGN32 u32physChannel, USIGN32 u32DeviceType, char *CfgLine)
{
  USIGN32 u32DeviceType2;
  BOOL bCfgFound = FALSE;
  FILE *pCfgFile;
  char lineStart[20];

  
  memset(CfgLine,0,LINE_SIZE); 
  sprintf(lineStart, "%09u_%u", u32Serial, u32physChannel);
  
//  printf("looking for: %s\n",lineStart);
  
  // open the config file for reading!
  pCfgFile = fopen(CAN_CONFIG_FILE,"r");
  
  if(!pCfgFile)
    return -1;
  
  // looking for the part, which describes that channel
  while(!bCfgFound && !feof(pCfgFile))
  {
    fgets(CfgLine,LINE_SIZE,pCfgFile);
   // printf("checking %s\n",CfgLine);
    if(strlen(CfgLine) >= strlen(lineStart))
    {
      if(!memcmp(CfgLine,lineStart,strlen(lineStart)))
      {
        if(!getIntVal(CfgLine,"%u","type:",(INT32*)&u32DeviceType2))
        {
         // printf("type not found\n");
          bCfgFound = FALSE;
        }
        else
        {
         // printf("%lu==%lu\n",u32DeviceType2,u32DeviceType);
          bCfgFound = ((USIGN32)u32DeviceType2==u32DeviceType);
        }
      }
    }
    
  }
  
  fclose(pCfgFile);
  
  if(!bCfgFound)
  { // channel configuration not found!    
    return -2;
  }
   
  return 0;
  
}

// get the name of the board
void getDevName(PCHANNEL channel)
{
  char line[LINE_SIZE];
  char *pName;
  int i=0;
  BOOL exist = FALSE;
  
  FILE *pCfg;
  
  pName = getDeviceName(channel->m_deviceType);
  
  do
  {
    i++;
    sprintf((char*)&channel->m_channelName,"%s_%i",pName,i);
    exist = FALSE;
    pCfg = fopen(CAN_CONFIG_FILE,"r");
  
    if(!pCfg)
    { // configuration file doesn't exist ==> every name is valid!
      sprintf((char*)&channel->m_channelName,"%s_1",pName);
      return;    
    }

    while (!feof(pCfg) && !exist)
    {
      char *pCfgName;
      fgets(line,LINE_SIZE,pCfg);
      pCfgName = strstr(line,"name:");
      if(pCfgName)
      {
        char *pEol = strstr(pCfgName,"\n");
        
        if(pEol)
        { // strip off end of line
          *pEol = 0;
        }
        
        pCfgName += sizeof("name:")-1;
        if(strlen(pCfgName) == strlen(channel->m_channelName))
        {
          exist = !memcmp(pCfgName,channel->m_channelName,strlen(channel->m_channelName));
        }
       
      }
    };
    fclose(pCfg);
  } while (exist);
  
   
}

void saveChannel(PCHANNEL channel)
{
  FILE *pCfg;

  pCfg = fopen(CAN_CONFIG_FILE,"a");
 
  if(!pCfg)
    pCfg = fopen(CAN_CONFIG_FILE,"w");

  if(!pCfg)
   return;


      fprintf(
        pCfg,
        "%09u_%u baud:%.2f presc:%lu tseg1:%lu tseg2:%lu sam:%lu sjw:%lu ocr:0x%lx accmstd:0x%04lx acccstd:0x%04lx accmxtd:0x%08lx acccxtd:0x%08lx ack:%u errorframes:%u type:%u irq:%u ioport:0x%x dpram:0x%lx name:%s\n",
        channel->m_serialNumber,
        channel->m_physChannel,
        125.0,
        4L,
        8L,
        7L,
        1L,
        1L,
        0xFAL,
        0L,
        0L,
        0L,
        0L,
        1,
        0,
        channel->m_deviceType,
        channel->m_irq,
        channel->m_ioport,
        channel->m_dpram,
        channel->m_channelName
      );
    

   fclose(pCfg);

}

int NumOfEntries(char *pProcDir)
{
  struct dirent **namelist;
  int n, ret = 0;

  n = scandir(pProcDir, &namelist, 0, alphasort);
  
  if(n<=0)
  {
    if(n==0)
      free(namelist);
    return ret;
  }

  ret = n;
  while(n--) 
  {
    free(namelist[n]);
  }
  free(namelist);
  
  return (ret > 2) ? ret-2 : 0;
  
}

void updateCfg(char *pProcDir)
{
  struct dirent **namelist;
  int n;
  BOOL bCfgNotFound;

  n = scandir(pProcDir, &namelist, 0, alphasort);

  if(n<=0)
  {
    if(n==0)
      free(namelist);
    return;
  }

  while(n--) 
  {
    FILE *pDev = NULL;
    static char line[LINE_SIZE];
    CHANNEL channel;
      
      
    if(
       strcmp(namelist[n]->d_name,"..") && 
       strcmp(namelist[n]->d_name,".")
      )
    {
      strcpy(line,pProcDir);
      strcat(line,"/");
      strcat(line,namelist[n]->d_name);
      pDev = fopen(line,"r");
      if(pDev)
      {
          
        memset(line,0,LINE_SIZE);
        fgets(line, LINE_SIZE, pDev);
        fclose(pDev);
          
        sscanf(namelist[n]->d_name,"%u",&channel.m_serialNumber);            
        getIntVal(line,"%u", "channel:", (INT32*)&channel.m_physChannel);
        getIntVal(line,"%u", "type:", (INT32*)&channel.m_deviceType);
        getIntVal(line,"%u", "irq:", (INT32*)&channel.m_irq);
        getIntVal(line,"%x", "ioport:", (INT32*)&channel.m_ioport);
        getIntVal(line,"%x", "dpram:", (INT32*)&channel.m_dpram);
            
        bCfgNotFound = getChConfig(channel.m_serialNumber, channel.m_physChannel, channel.m_deviceType, line);
        
        if(bCfgNotFound)
        {
          getDevName(&channel);
          saveChannel(&channel);
          
        }
                 
      }
 
    }
    
    free(namelist[n]);
  }
  free(namelist);
}

    
int main(int argc, char *argv[])
{
  int i, ret;
  USIGN32 u32NeededBufferSize, u32NumOfChannels, u32ProvidedBufferSize, ifc;
  PCHDSNAPSHOT pBuffer;
  DIR * pDir;
  char proc_dir_name[100];
  char startDriverCmd[100];
  int type;
  
  
  
  if(argc < 2)
  {
    printf("only %u arguments\n",argc);
    return -1;
  }
  
 
  sscanf(argv[1],"%i",&type);
  
  for(i=0; i<argc; i++)    
    printf("argv[%i] = %s\n", i, argv[i]);
  
  printf("%s: genarating Softing CAN device files for device type:%u\n",SOFTING_VERSION,type);
 

  
  
  switch(type)
  {
    case 5:
    case 0x105:
      strcpy(proc_dir_name,"/proc/softing_can_pcmcia");
      strcpy(startDriverCmd,"/sbin/modprobe cancard_cs");
      break;
      
    case 7:
    case 8:
      strcpy(proc_dir_name,"/proc/softing_can_pci");
      strcpy(startDriverCmd,"/sbin/modprobe vcanpcid");
      break;
      
    case 9:
      strcpy(proc_dir_name,"/proc/softing_can_104");
      strcpy(startDriverCmd,"/sbin/modprobe vcanacd");
      break;
      
    case 12:
      strcpy(proc_dir_name,"/proc/softing_can_eco");
      strcpy(startDriverCmd,"/sbin/modprobe vcaneco");
      break;
      
    case 13:
      strcpy(proc_dir_name,"/proc/softing_can_pcie");
      strcpy(startDriverCmd,"/sbin/modprobe vcanpcie");
      break;
      
    case 14:
      strcpy(proc_dir_name,"/proc/softing_can_pro");
      strcpy(startDriverCmd,"/sbin/modprobe vcanprod");
      break;
      
    case 15:
      strcpy(proc_dir_name,"/proc/softing_canprousb");
      strcpy(startDriverCmd,"/sbin/modprobe canpro_usb");
      break;
      
    case 0x8000:
      strcpy(proc_dir_name,"/proc/on_board_can");
      strcpy(startDriverCmd,"/sbin/modprobe fanucpc8");
      break;
      
      
    default:
      printf("unknown device %u", type);
      return -2;
      break;
    
  }
  
  
  for(i=0; i<argc; i++)
  {
    printf("arg %u = %s\n",i,argv[i]);
  }
  
  chmod("/etc/softing", 0777);
  chmod(CAN_CONFIG_FILE, 0777);
  // waiting for driver start ...
  for(i=0; ((i<5) && (!NumOfEntries(proc_dir_name))); i++) sleep(1);
  sleep(1);
  
  printf("%u CAN devices of type %u found!\n",NumOfEntries(proc_dir_name), type);
  
  updateCfg(proc_dir_name);
  
  // check if driver is loaded.
  // driver creates a proc. directory, when it was loaded.
  pDir = opendir (proc_dir_name);
  if(pDir)
  {
    closedir(pDir);
  }
  
  if(!pDir && ((type & 0xFF) == 5))
  { // no proc dir no driver!
    // only pcmcia devices 0x105 and 0x05
    pid_t pid;
    int status;
    
    
    pid = fork ();
    
     
    if (pid == 0)
    {
      // This is the child process.  Execute the shell command. 
      execl ("/bin/sh", "/bin/sh", "-c", startDriverCmd, NULL);
      exit (-1);
    }
    else if (pid < 0)
    // The fork failed.  Report failure.  
      status = -1;
    else
    // This is the parent process.  Wait for the child to complete.  
      if (waitpid (pid, &status, 0) != pid)
        status = -1;
    
  
    sleep(10);
  }
  
  
  
  ret = CANL2_get_all_CAN_channels(0, &u32NeededBufferSize, &u32NumOfChannels, NULL);
  
  if(ret)
  {
    printf("ERROR %i\n",ret);
    return -3;
  }
  
  pBuffer = malloc(u32NeededBufferSize);
  u32ProvidedBufferSize = u32NeededBufferSize;
  
  ret = CANL2_get_all_CAN_channels(u32ProvidedBufferSize, &u32NeededBufferSize, &u32NumOfChannels, pBuffer);
  
  if(ret)
  {
    printf("ERROR %i\n",ret);
    return -2;
  }
  
  printf("You have %u Softing CAN channels in your system\n\n", u32NumOfChannels);
  
  printf("\tname\t\t serialnumber\t type\t chan.\t    open\n");
  printf("--------------------------------------------------------------------\n");
  printf("\n");
  
  ifc = 0;

  for(ifc=0; (ifc < u32NumOfChannels); ifc++)
  {
    PCHDSNAPSHOT pCh = &pBuffer[ifc];
 

    printf("%17s\t %09u  %9s\t %2u\t %5s\n",
           pCh->ChannelName,
           pCh->u32Serial,
           getDeviceType(pCh->u32DeviceType),
           pCh->u32PhysCh,
           (pCh->bIsOpen) ? "yes" : "no");
    

    if(!pCh->bIsOpen)
    {
      char name[100];
      
      if((strlen("/dev/")+strlen((char*)pCh->ChannelName))<sizeof(name))
      {
        int ret;
        INT32 major=0, minor=0;
        char CfgLine[LINE_SIZE];
        FILE *pProc;
        
        strcpy(name,DEVFILEPREFIX);
        strcat(name,(char*)pCh->ChannelName);
        
        remove(name);
        
        strcpy(CfgLine, getDeviceType(pCh->u32DeviceType));
        sprintf(CfgLine,"%s/%09u_%u",getDeviceType(pCh->u32DeviceType), pCh->u32Serial, pCh->u32PhysCh);
        pProc = fopen(CfgLine,"r");
        
        if(pProc)
        {
          fgets(CfgLine,LINE_SIZE ,pProc);
          fclose(pProc);
        
          getIntVal(CfgLine, "%u", "major: ", &major);
          getIntVal(CfgLine, "%u", "minor: ", &minor);
        
         // printf("major=%i, minor=%i, file=%s\n",major,minor,name);
        
          if(major)
            ret = mknod(name, S_IFCHR | 0x1ff,  MKDEV(major, minor)); 
         
          if(ret)
          {
            printMknodError();
          }
        }
        else
        {
          printf("file %s not found!\n",CfgLine);
        }

      }
     }

  }
  
  return 0;
}



