/***************************************************************************
 *   Copyright (C) 2007-2015 by Softing Automation GmbH                    *
 *   support.automation@softing.com                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "cnmt.h"

/**
 * *****************************************************************
 * @brief  Constructor
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

nmt::nmt(CAN_HANDLE can, QWidget *parent) : QDialog(parent)
{
    int i;

    m_Can = can;
    
    m_pw = new QDialog(this);
    m_ui.setupUi(m_pw);


    for(i=0; i<128; i++)
    {
        m_slave[i].state = UNUSED;
        m_slave[i].guardTime = 0;
        m_slave[i].heartbeatInterval = 0;
        m_slave[i].retryFactor = 0;
        m_slave[i].b_checked = false;
        m_slave[i].guard_state = GUARD_NOTACTIVE;
        m_slave[i].b_added = false;
    }

    m_ui.s_node->setRange(0,127);
    m_ui.s_retry->setRange(0, 3);

    m_enableUpdate = false;

    connect( &m_nmt_timer, SIGNAL(timeout()), this, SLOT(OnUpdate()) );
    connect( m_ui.bt_add, SIGNAL(clicked(bool)), this, SLOT(OnAdd()));
    connect( m_ui.bt_error_ctrl, SIGNAL(clicked(bool)), this, SLOT(OnErrorCtrl()));
    connect( m_ui.bt_preoperational, SIGNAL(clicked(bool)), this, SLOT(OnPreoperational()));
    connect( m_ui.bt_remove, SIGNAL(clicked(bool)), this, SLOT(OnRemove()));
    connect( m_ui.bt_reset, SIGNAL(clicked(bool)), this, SLOT(OnReset()));
    connect( m_ui.bt_resetcom, SIGNAL(clicked(bool)), this, SLOT(OnResetCom()));
    connect( m_ui.bt_start, SIGNAL(clicked(bool)), this, SLOT(OnStart()));
    connect( m_ui.bt_stop, SIGNAL(clicked(bool)), this, SLOT(OnStop()));
    connect( m_ui.bt_sync, SIGNAL(clicked(bool)), this, SLOT(OnSync()));
    connect( m_ui.bt_update, SIGNAL(clicked(bool)), this, SLOT(OnUpdate()));
    connect( m_ui.bt_ok, SIGNAL(clicked(bool)), this, SLOT(accept()));
    connect( m_ui.s_node, SIGNAL(valueChanged(int)), this, SLOT(OnNodeChanged(int)));

    m_pw->show();

}

/**
 * *****************************************************************
 * @brief  Destructor
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

nmt::~nmt()
{
    deleteAllTabEntries();
    delete m_pw;
}

/**
 * *****************************************************************
 * @brief  delete the live list table
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void nmt::deleteAllTabEntries(void)
{
    int numOfCol = m_ui.tb_livelist->columnCount()-1;
    int i,j;
    QTableWidgetItem *pTableItem;

    for(i=numOfCol; i>=0; i--)
    {
        for(j=6; j>=0; j--)
        {
            pTableItem = m_ui.tb_livelist->item(i,j);
            delete pTableItem;
        }
    }
}


/**
 * *****************************************************************
 * @brief  This slot is called if the add button is pushed.
 * @brief  It adds a new node to the node manager
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void nmt::OnAdd()
{
    CMA_NodeConfig cfg;
    int retCode;

    cfg.ucErrAction = 0;
    cfg.ucNodeID = m_ui.s_node->value(); // node id
    cfg.ucRetryFactor = (USIGN8)m_ui.s_retry->value(); // retry factor
    cfg.usGuardTime = (m_ui.ed_guardtime->text()).toShort(); //guard time in ms
    cfg.ulHeartBeatInterval = (m_ui.ed_heartbeat->text()).toULong(); //heartbeat interval in ms

    // only one error control mechanism can be used at the same time (heartbeat or node guarding)
    // to disable the error control mechanism the time interval must be set to 0
    // heartbeat interval or guard time must be 0, because heartbeat and node guarding at the same time is not possible.

    if((cfg.usGuardTime!=0) && (cfg.ulHeartBeatInterval!=0))
    {
        m_on_Error.information(
            this,
            "Configuration Error",
            "Node guarding and heartbeat cannot be enabled at the same time!\nPlease set the guard time or the heartbeat interval to 0!",
            QMessageBox::Abort,
            QMessageBox::NoButton,
            QMessageBox::NoButton
        );
        return;
    }

    if(m_slave[cfg.ucNodeID].state != UNUSED)
    {
        return;
    }

    retCode = CMA_AddNode(m_Can, cfg);
    if(retCode)
    {
        QString ErrorText;

        ErrorText = ErrorText.sprintf("ERROR: %i in CMA_AddNode", retCode);
        emit(doPrint(ErrorText));
        return;
    }
    else
    {
      QString SuccessText;
      
      SuccessText = SuccessText.sprintf("CMA_AddNode(m_Can, cfg) --> success");
      emit(doPrint(SuccessText));      
    }
    
    m_slave[cfg.ucNodeID].state = USED;
    m_slave[cfg.ucNodeID].retryFactor = cfg.ucRetryFactor;
    m_slave[cfg.ucNodeID].guardTime = cfg.usGuardTime;
    m_slave[cfg.ucNodeID].heartbeatInterval = cfg.ulHeartBeatInterval;
    m_slave[cfg.ucNodeID].b_checked = false;
    m_slave[cfg.ucNodeID].guard_state = GUARD_NOTACTIVE;
    m_slave[cfg.ucNodeID].heartbeatErrors = 0;
    m_slave[cfg.ucNodeID].b_added = true;

    m_ui.ed_heartbeat->setReadOnly(true);
    m_ui.ed_guardtime->setReadOnly(true);
    m_ui.s_retry->setEnabled(false);

    if(m_slave[cfg.ucNodeID].heartbeatInterval == 0)
    {
        if(m_slave[cfg.ucNodeID].guardTime == 0)
        {
            m_ui.bt_error_ctrl->setText("error ctrl");
        }
        else
        {
            m_ui.bt_error_ctrl->setText("guarding");
        }

    }
    else
    {
        m_ui.bt_error_ctrl->setText("heartbeat");
    }


}

/**
 * *****************************************************************
 * @brief  This slot is called if the remove button is pushed.
 * @brief  It removes a node from the node manager
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void nmt::OnRemove()
{
    int retCode;
    USIGN8 nodeId = m_ui.s_node->value();

    if(m_slave[nodeId].state == UNUSED)
    {
        QString ErrorText;

        ErrorText = ErrorText.sprintf("ERROR: node was never added!");
        emit(doPrint(ErrorText));
        return;
    }
    

    retCode = CMA_RemoveNode(m_Can, nodeId, ST_STOPP);
    if(retCode)
    {
        QString ErrorText;

        ErrorText = ErrorText.sprintf("ERROR: %i in CMA_RemoveNode", retCode);
        emit(doPrint(ErrorText));
        return;
    }
    else
    {
      QString SuccessText;
      
      SuccessText = SuccessText.sprintf("CMA_RemoveNode(m_Can, %u, ST_STOPP) --> success", nodeId);
      emit(doPrint(SuccessText));      
    }


    m_slave[nodeId].state = UNUSED;
    m_slave[nodeId].retryFactor = 0;
    m_slave[nodeId].guardTime = 0;
    m_slave[nodeId].heartbeatInterval = 0;
    m_slave[nodeId].b_checked = false;
    m_slave[nodeId].guard_state = GUARD_NOTACTIVE;

    m_ui.ed_heartbeat->setReadOnly(false);
    m_ui.ed_guardtime->setReadOnly(false);
    m_ui.s_retry->setEnabled(true);
    m_slave[nodeId].b_added = false;

}

/**
 * *****************************************************************
 * @brief  This slot is called if the start button is pushed.
 * @brief  It starts the selected slave
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void nmt::OnStart()
{
    int retCode;
    USIGN8 nodeId = m_ui.s_node->value();

    retCode = CMA_ChangeState(m_Can, nodeId, ST_START);
    if(retCode)
    {
        QString ErrorText;

        ErrorText = ErrorText.sprintf("ERROR: %i in CMA_ChangeState", retCode);
        emit(doPrint(ErrorText));
        return;
    }
    m_slave[nodeId].state = OPERATIONAL;
}

/**
 * *****************************************************************
 * @brief  This slot is called if the stop button is pushed.
 * @brief  It stops the selected slave
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void nmt::OnStop()
{
    int retCode;
    USIGN8 nodeId = m_ui.s_node->value();
    retCode = CMA_ChangeState(m_Can, nodeId, ST_STOPP);
    if(retCode)
    {
        QString ErrorText;

        ErrorText = ErrorText.sprintf("ERROR: %i in CMA_ChangeState", retCode);
        emit(doPrint(ErrorText));
        return;
    }
    m_slave[nodeId].state = STOPPED;
}

/**
 * *****************************************************************
 * @brief  This slot is called if the reset button is pushed.
 * @brief  It resets the selected slave
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void nmt::OnReset()
{
    int retCode;
    USIGN8 nodeId = m_ui.s_node->value();
    retCode = CMA_ChangeState(m_Can, nodeId, ST_RESET_NODE);
    if(retCode)
    {
        QString ErrorText;

        ErrorText = ErrorText.sprintf("ERROR: %i in CMA_ChangeState", retCode);
        emit(doPrint(ErrorText));
        return;
    }

    m_slave[nodeId].state = PREOPERATIONAL;

}

/**
 * *****************************************************************
 * @brief  This slot is called if the reset communication button is pushed.
 * @brief  It resets the communication with the selected slave
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void nmt::OnResetCom()
{
    int retCode;
    USIGN8 nodeId = m_ui.s_node->value();

    retCode = CMA_ChangeState(m_Can, nodeId, ST_RESET_COM);
    if(retCode)
    {
        QString ErrorText;

        ErrorText = ErrorText.sprintf("ERROR: %i in CMA_ChangeState", retCode);
        emit(doPrint(ErrorText));
        return;
    }

    m_slave[nodeId].state = PREOPERATIONAL;

}

/**
 * *****************************************************************
 * @brief  This slot is called if the preop button is pushed.
 * @brief  It puts the selected slave into the preoperational state
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void nmt::OnPreoperational()
{
    int retCode;
    USIGN8 nodeId = m_ui.s_node->value();

    retCode = CMA_ChangeState(m_Can, // channel handle
                              nodeId, // node id
                              ST_PRE_OPERATIONAL // new state
                             );
    if(retCode)
    {
        QString ErrorText;

        ErrorText = ErrorText.sprintf("ERROR: %i in CMA_ChangeState", retCode);
        emit(doPrint(ErrorText));
        return;
    }

    m_slave[nodeId].state = PREOPERATIONAL;
}

/**
 * *****************************************************************
 * @brief  This slot is called if the error control button is pushed.
 * @brief  It starts the monitoring the selected slave.
 * @brief  It checks if the slave sends heartbeat messages or it
 * @brief  checks if the slave answers guarding requests.
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void nmt::OnErrorCtrl()
{
    int retCode;
    USIGN8 nodeId = m_ui.s_node->value();

    if(m_slave[nodeId].b_checked)
    {
        retCode = CMA_RequestGuarding(m_Can /*channel handle*/,
                                      nodeId /*node id*/,
                                      0 /*stop heartbeat valdation or node guarding*/);
        if(retCode)
        {
            QString ErrorText;

            ErrorText = ErrorText.sprintf("ERROR: %i in CMA_RequestGuarding", retCode);
            emit(doPrint(ErrorText));
            return;
        }

        m_slave[nodeId].b_checked = false;
        m_slave[nodeId].guard_state = GUARD_NOTACTIVE;
    }
    else
    {
        retCode = CMA_RequestGuarding(m_Can /*channel handle*/,
                                      nodeId /*node id*/,
                                      1 /*start heartbeat valdation or node guarding*/);
        if(retCode)
        {
            QString ErrorText;

            ErrorText = ErrorText.sprintf("ERROR: %i in CMA_RequestGuarding", retCode);
            emit(doPrint(ErrorText));
            return;
        }

        m_slave[nodeId].b_checked = true;
        m_slave[nodeId].guard_state = GUARD_ACTIVE;
    }
}


/**
******************************************************************
* @brief  returns the textual QString representation of the device state.
* @author Werner Knorsch, Softing Automation GmbH
* @date   2013-02-05
* @param  state: device state
*
* @return returns the QString representation of the device state.
*
******************************************************************
*/

QString& nmt::getStateText(NODE_STATE state)
{
    static QString ret;

    switch(state)
    {
    case INITIALISATION:
    case USED:
    case UNUSED:
        ret = "INITIALISATION";
        break;

    case STOPPED:
        ret = "STOPPED";
        break;

    case PREOPERATIONAL:
        ret = "PREOPERATIONAL";
        break;

    case OPERATIONAL:
        ret = "OPERATIONAL";
        break;

    case UNKNOWN:
        ret = "UNKNOWN";
        break;


    default:
        ret.sprintf("unknown %u",state);
        break;

    }
    return ret;
}

/*
typedef enum
{
 GUARD_NOTACTIVE,
 GUARD_ACTIVE,
 GUARD_LOST_RESP,
 GUARD_FAILED,
 GUARD_TOGGLE_ERROR

} GUARD_STATE;
*/

/**
 * *****************************************************************
 * @brief  returns the textual QString representation of the guard state.
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 * @param  state: guardstate state
 *
 * @return returns the QString representation of the guard state.
 *
 ******************************************************************
 */

QString& nmt::getControlState(GUARD_STATE guardstate)
{
    static QString ret;

    switch(guardstate)
    {
    case GUARD_NOTACTIVE:
        ret = "not active";
        break;
    case GUARD_ACTIVE:
        ret = "active";
        break;
    case GUARD_LOST_RESP:
        ret = "lost response";
        break;
    case GUARD_FAILED:
        ret = "failed";
        break;
    case GUARD_TOGGLE_ERROR:
        ret = "error in toggle bit happened";
        break;

    default:
        ret = "unknown";
        break;

    }

    return ret;

}

/**
 * *****************************************************************
 * @brief  This slot is called the first time if the start live list button is pushed.
 * @brief  It starts a timer, and it is called again and again by the timer
 * @brief  It updates the livelist
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void nmt::OnUpdate()
{
    int i,numdef=0,row=0;
    QString itemText;
    QTableWidgetItem *pTableItem;


    for(i=1; i<128; i++)
    {
        if(m_slave[i].b_added)
        {
            numdef++;
        }
    }

    m_ui.tb_livelist->setRowCount(numdef);

    for(i=1; i<128; i++)
    {
        if(m_slave[i].b_added)
        {

            m_ui.tb_livelist->setRowHeight(row,15);
            itemText = itemText.sprintf("%u",i);
            pTableItem = new QTableWidgetItem();
            pTableItem->setText(itemText);
            m_ui.tb_livelist->setItem(row,0,pTableItem);

            pTableItem = new QTableWidgetItem();
            pTableItem->setText(getStateText(m_slave[i].state));
            m_ui.tb_livelist->setItem(row,1,pTableItem);

            itemText = itemText.sprintf("%u",m_slave[i].heartbeatErrors);
            pTableItem = new QTableWidgetItem();
            pTableItem->setText(itemText);
            m_ui.tb_livelist->setItem(row,2,pTableItem);

            pTableItem = new QTableWidgetItem();
            if(m_slave[i].b_checked)
            {
                pTableItem->setText("validated");
                m_ui.tb_livelist->setItem(row,3,pTableItem);
            }
            else
            {
                pTableItem->setText("not validated");
                m_ui.tb_livelist->setItem(row,3,pTableItem);
            }

            pTableItem = new QTableWidgetItem();
            pTableItem->setText(getControlState(m_slave[i].guard_state));

            m_ui.tb_livelist->setItem(row,4,pTableItem);
            row++;
        }
    }

    m_ui.tb_livelist->resizeColumnToContents(1);
    m_ui.tb_livelist->resizeColumnToContents(4);
    m_ui.tb_livelist->horizontalHeader()->setStretchLastSection(true);

    m_enableUpdate = true;
    m_nmt_timer.start( 2000 );

}

/**
 * *****************************************************************
 * @brief  This slot is called if the selected node id changes.
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void nmt::OnNodeChanged(int node)
{
    QString text;

    if(m_slave[node].guardTime == 0)
    {
        m_ui.bt_error_ctrl->setText("heartbeat");
    }
    else
    {
        m_ui.bt_error_ctrl->setText("guarding");
    }

    if(m_slave[node].b_added)
    {
        m_ui.s_retry->setEnabled(false);
        m_ui.ed_heartbeat->setReadOnly(true);
        m_ui.ed_guardtime->setReadOnly(true);
        m_ui.s_retry->setValue(m_slave[node].retryFactor);
        text = QString("%1").arg(m_slave[node].guardTime);
        m_ui.ed_guardtime->setText(text);
        text = QString("%1").arg(m_slave[node].heartbeatInterval);
        m_ui.ed_heartbeat->setText(text);

    }
    else
    {
        m_ui.s_retry->setEnabled(true);
        m_ui.ed_heartbeat->setReadOnly(false);
        m_ui.ed_guardtime->setReadOnly(false);
        m_ui.ed_guardtime->setText("0");
        m_ui.ed_heartbeat->setText("0");
    }
}

/**
 * *****************************************************************
 * @brief  This slot is called if the sync button is pushed.
 * @brief  The sync manager begins or stops to send sync messages after this call.
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void nmt::OnSync()
{
    short retCode;
    static unsigned char syncmode = 1;
    CMA_SyncManConfig cfg;
    QString Text;

    cfg.ucConsumeMode = m_ui.ch_sync->isChecked();
    cfg.ucSendMode = syncmode;
    cfg.ulSyncCycle = m_ui.ed_sync->text().toULong();
    cfg.ulSyncID = 0x80;

    if((retCode = CMA_ConfigSyncMan(m_Can /*channel handle*/, cfg /*sync manager configuration*/)))
    {
        Text.sprintf("ERROR: %i in CMA_ConfigSyncMan", retCode);
        emit(doPrint(Text));
        return;
    }
    ++syncmode &= 1;


}

/**
 * *****************************************************************
 * @brief  This slot is called if the stack detects an guarding/heartbeat error/timeout
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void nmt::OnErrorEvent(unsigned char nodeId, unsigned char eventType, int n_state)
{
    QString cEventText;
    NODE_STATE state = (NODE_STATE) n_state;

    switch(eventType)
    {
    case 1:
        // change of slave state
        m_slave[nodeId].state = (NODE_STATE)state;
        m_slave[nodeId].guard_state = GUARD_ACTIVE;
        cEventText = cEventText.sprintf("slave %3u changed to state ", nodeId) + getStateText(state);
        emit(doPrint(cEventText));
        break;
    case 2:
        // heartbeat/guard timeout
        m_slave[nodeId].state = UNKNOWN;
        m_slave[nodeId].heartbeatErrors++;
        m_slave[nodeId].guard_state = GUARD_LOST_RESP;
        cEventText.sprintf("slave %3u: heartbeat/guard timeout", nodeId);
        emit(doPrint(cEventText));
        break;
    case 3:
        // heartbeat/guard timeout; retry expired!
        m_slave[nodeId].state = UNKNOWN;
        m_slave[nodeId].heartbeatErrors++;
        m_slave[nodeId].b_checked = false;
        m_slave[nodeId].guard_state = GUARD_FAILED;
        cEventText.sprintf("slave %3u: heartbeat or guarding failed - error control disabled", nodeId);
        emit(doPrint(cEventText));
        break;
    case 4:
        // toggle error
        m_slave[nodeId].state = (NODE_STATE)state;
        m_slave[nodeId].heartbeatErrors++;
        m_slave[nodeId].guard_state = GUARD_ACTIVE;
        cEventText.sprintf("slave %3u: toggle error", nodeId);
        emit(doPrint(cEventText));
        break;
    case 5:
        // change of slave state
        m_slave[nodeId].state = (NODE_STATE)state;
        m_slave[nodeId].guard_state = GUARD_ACTIVE;
        cEventText = cEventText.sprintf("slave %3u: changed to state ", nodeId) + getStateText(state);
        emit(doPrint(cEventText));
        break;
    }
}


#include "cnmt.moc"
