/***************************************************************************
 *   Copyright (C) 2007-2015 by Softing Automation GmbH                    *
 *   support.automation@softing.com                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include <stdlib.h>
#include "cothread.h"

/**
 ******************************************************************
 * @brief  Constructor
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

CoThread::CoThread(USIGN8 Can)
{
  // channel handle
  m_Can = Can;
}

/**
 ******************************************************************
 * @brief  returns true, if ch is a printable letter
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return returns true, if ch is a printable letter, false otherwise
 *
 ******************************************************************
 */

BOOLEAN isLetter(unsigned char ch)
{
  if((ch >= 32) && (ch <= 126))
    return TRUE;
  
  return FALSE;
  
}

/**
 ******************************************************************
 * @brief  returns the textual QString representation of the device state.
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 * @param  state: device state
 *
 * @return returns the QString representation of the device state.
 *
 ******************************************************************
 */

QString& CoThread::getStateText(USIGN8 state)
{
    static QString ret;

    switch(state)
    {
    case STOPPED:
        ret = "STOPPED";
        break;

    case PREOPERATIONAL:
        ret = "PREOPERATIONAL";
        break;

    case OPERATIONAL:
        ret = "OPERATIONAL";
        break;

    case UNKNOWN:
        ret = "UNKNOWN";
        break;


    default:
        ret.sprintf("unknown %u",state);
        break;

    }
    return ret;
}

/**
 ******************************************************************
 * @brief  gets the error text from an error code
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 * @param  error code
 *
 * @return error text
 *
 ******************************************************************
 */

QString& CoThread::getErrorText(USIGN32 errorCode)
{
    static QString ret;
    switch(errorCode)
    {
    case 0x00000305:
        ret = "toggle bit not altered";
        break;

    case 0x00000405:
        ret = "SDO timeout";
        break;
    case 0x01000405:
        ret = "command specifier not valid or unknown";
        break;
    case 0x02000405:
        ret = "invalid block size";
        break;
    case 0x03000405:
        ret = "invalid sequence number";
        break;

    case 0x04000405:
        ret = "crc error";
        break;

    case 0x05000405:
        ret = "out of memory";
        break;

    case 0x00000106:
        ret = "unsupported access to an object";
        break;

    case 0x01000106:
        ret = "attempt to read a write only object";
        break;

    case 0x02000106:
        ret = "attempt to write a read only object";
        break;

    case 0x00000206:
        ret = "object does not exist";
        break;

    case 0x41000406:
        ret = "object can not be mapped";
        break;

    case 0x42000406:
        ret = "object can not be mapped, because of the object size";
        break;

    case 0x43000406:
        ret = "parameter incompatibility";
        break;

    case 0x47000406:
        ret = "device incompatibility";
        break;

    case 0x00000606:
        ret = "access failed due to an hardware error";
        break;

    case 0x10000706:
        ret = "length of service parameter does not match";
        break;

    case 0x12000706:
        ret = "length of service parameter to high";
        break;

    case 0x13000706:
        ret = "length of service parameter to low";
        break;

    case 0x11000906:
        ret = "subindex does not exist";
        break;



    default:
        ret = ret.sprintf("unknown SDO error code 0x%08x ...",errorCode);
        break;
    }

    return ret;
}

/**
 ******************************************************************
 * @brief  gets the emergency text from an emergency code
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 * @param  error code
 *
 * @return emergency text
 *
 ******************************************************************
 */

QString& CoThread::getEMCYText(USIGN16 code)
{
    static QString text;

    switch(code & 0xF000)
    {
    case 0x1000:
        text = "generic error";
        break;
    case 0x2000:
        text = "currant error";
        break;
    case 0x3000:
        text = "voltage error";
        break;
    case 0x4000:
        text = "temperature error";
        break;
    case 0x5000:
        text = "device hardware error";
        break;
    case 0x6000:
        text = "device software error";
        break;
    case 0x7000:
        text = "additional modules";
        break;
    case 0x8000:
        switch(code)
        {
        case 0x8110:
            text = "CAN overrun";
            break;
        case 0x8120:
            text = "CAN ERROR PASSIVE";
            break;
        case 0x8130:
            text = "live guard or heartbeat error";
            break;
        case 0x8140:
            text = "recovered from BUS OFF";
            break;
        case 0x8150:
            text = "transmit COB ID collision";
            break;
        case 0x8210:
            text = "PDO not processed due to length error";
            break;
        case 0x8220:
            text = "PDO length exceeded";
            break;

        default:
            if(code > 0x81FF)
            {
                text = "protocol error";
            }
            else
            {
                if(code > 0x80FF)
                {
                    text = "communication error";
                }
                else
                {
                    text = "monitoring error";
                }

            }
            break;
        }

        break;
    case 0x9000:
        text = "external error";
        break;
    case 0xF000:
        if(code < 0xFF00)
        {
            text = "additional functions";
        }
        else
        {
            text = "device specific";
        }
        break;

    default:
        text = "unknown";
        break;
    }

    return text;
}

/**
 ******************************************************************
 * @brief  handles an CANopen event, it's called from the run methode
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 * @param  pEvent: pointer to the event structure
 *
 * @return 0
 *
 ******************************************************************
 */

int CoThread::HandleEvent(CMA_EventStruct *pEvent)
{
    QString cEventText;
    QString textdata;
    QString hexdata;
    QString sval;
    char cError[5][20] = {"  ","API","CAN","SDO","GUARD"};
    USIGN32 i;
    quint64 pdoData, sdoData;

    switch(pEvent->ucEventType)
    {
    case 3: //download complete
        cEventText.sprintf("time:%11u DOWNLOAD COMPLETE node:%03u, subindex:%x, index:%x, %s                                           ",
                           pEvent->ulTimeStamp,
                           pEvent->ucaEventData[0],
                           pEvent->ucaEventData[1],
                           getBufferVal16(&pEvent->ucaEventData[2]),
                           (pEvent->ucaEventData[4] > 0x04) ? "segmented" : "expedited"
                          );
	// print out the success message
        emit(doPrint(cEventText));
        break;
    case 5: //upload complete
        cEventText = cEventText.sprintf("time:%11u UPLOAD COMPLETE node:%03u, subindex:%x, index:%x, length: %u                                                    ",
                                        pEvent->ulTimeStamp,
                                        pEvent->ucaEventData[0],
                                        pEvent->ucaEventData[1],
                                        getBufferVal16(&pEvent->ucaEventData[2]),
                                        (pEvent->ucaEventData[4] == 0xFF) ? getBufferVal32(&pEvent->ucaEventData[16]) : pEvent->ucaEventData[4]
                                       );
        // print the uploaded data
        emit(doPrint(cEventText));

        switch(pEvent->ucaEventData[4])
        {
        case 1:
            textdata.sprintf("data: 0x%02x ", pEvent->ucaEventData[5]);
	    memcpy(&sdoData, &pEvent->ucaEventData[5], pEvent->ucaEventData[4]);
	    emit(sigSdoRcv(pEvent->ucaEventData[4], sdoData));
            emit(doPrint(textdata));
            break;

        case 2:
            textdata.sprintf("data: %5u ", getBufferVal16(&pEvent->ucaEventData[5]));
	    memcpy(&sdoData, &pEvent->ucaEventData[5], pEvent->ucaEventData[4]);
	    emit(sigSdoRcv(pEvent->ucaEventData[4], sdoData));
            emit(doPrint(textdata));
            break;

        case 3:
            textdata.sprintf("data: 0x%02x 0x%02x 0x%02x ", pEvent->ucaEventData[5], pEvent->ucaEventData[6], pEvent->ucaEventData[7]);
	    memcpy(&sdoData, &pEvent->ucaEventData[5], pEvent->ucaEventData[4]);
	    emit(sigSdoRcv(pEvent->ucaEventData[4], sdoData));
            emit(doPrint(textdata));
            break;

        case 4:
            textdata.sprintf("data: %9u ", getBufferVal32(&pEvent->ucaEventData[5]));
	    memcpy(&sdoData, &pEvent->ucaEventData[5], pEvent->ucaEventData[4]);
	    emit(sigSdoRcv(pEvent->ucaEventData[4], sdoData));
            emit(doPrint(textdata));
            break;

        case 0xFF:
        {   
	  // segmented upload
            unsigned char *pData = NULL;
            short retCode;
            unsigned int rest;

            pData = (unsigned char *)malloc(getBufferVal32(&pEvent->ucaEventData[16]));
            if(!pData)
                return 0;

	    // event structure is too small for an segmented upload
            // we have to use a special read function to get the uploaded data
            retCode = CMA_ReadData(
                          m_Can, // channel handle
                          pEvent->ucaEventData[0], // node id of the slave
                          pEvent->ucaEventData[1], // subindex
                          getBufferVal16(&pEvent->ucaEventData[2]), // index
                          getBufferVal32(&pEvent->ucaEventData[16]), // buffer size
                          pData); // data buffer 

            if(retCode)
            {
                QString err = err.sprintf("Error %i in CMA_ReadData()", retCode);
                emit(doPrint(err));
            }
            else
	    {
	      memcpy(&sdoData, pData, 8);
	      emit(sigSdoRcv(getBufferVal32(&pEvent->ucaEventData[16]), sdoData));
	    }
            
            // print the data
            for(i=0; i<getBufferVal32(&pEvent->ucaEventData[16]); i++)
            {
                if(!(i%16)) // 16 bytes per line
                {
                    QString str;
                    str.sprintf("%08x: ",(USIGN32)i);
                    hexdata.append(str);
                    textdata.append(str);

                }

                sval.sprintf("%02x ",pData[i]);
                hexdata.append(sval);


                if((pData[i] >= 32) && (pData[i] < 127))
                {
                    sval.sprintf("%c",pData[i]);
                    textdata.append(sval);
                }
                else
                {
                    textdata.append(".");
                }

            }

            rest = 16 - (i % 16);

            for(i=0; i<rest; i++)
            {
                hexdata.append("   ");
            }

            hexdata.append("");
            textdata.append("");
            
            emit(doPrint(hexdata));
            emit(doPrint(textdata));
            hexdata = "";
            textdata = "";

	    // free data buffer
            free(pData);

        }
        break;

        default:
            textdata = "unknown type at offset 4";
            emit(doPrint(textdata));
            break;


        }


        break;
    case 11: //emergency
        cEventText.sprintf("time:%11u EMERGENCY node:%3u, %02x %02x %02x %02x %02x %02x %02x                                                         ",
                           pEvent->ulTimeStamp,
                           pEvent->ucaEventData[0],
                           pEvent->ucaEventData[1],
                           pEvent->ucaEventData[2],
                           pEvent->ucaEventData[3],
                           pEvent->ucaEventData[4],
                           pEvent->ucaEventData[5],
                           pEvent->ucaEventData[6],
                           pEvent->ucaEventData[7]
                          );
        emit(doPrint(cEventText));

        cEventText = cEventText.sprintf("EMERGENCY node:%3u ", pEvent->ucaEventData[0]) + getEMCYText(getBufferVal16(&pEvent->ucaEventData[1]));
        emit(doPrint(cEventText));

        break;

    case 12:
        // boot up message from a slave - a new slave appeared, may be we should configure or start it
        cEventText.sprintf("slave %3u: boot up!", pEvent->ucaEventData[0]);
        emit(doPrint(cEventText));
        break;

    case 15: //error

        switch (pEvent->ucaEventData[0])
        {
        case 3:
            // SDO abort
            cEventText = cEventText.sprintf("slave %3u: SDO ERROR - ", pEvent->ucaEventData[1]) + getErrorText(getBufferVal32(&pEvent->ucaEventData[2]));
            emit(doPrint(cEventText));
            break;

        case 4:
            // guard error, someone has forgotten to send it's heartbeat, may be it crashed?
            emit(sig_errorEvent(pEvent->ucaEventData[1],pEvent->ucaEventData[2],(int)pEvent->ucaEventData[3]));
            break;

        default:
            cEventText.sprintf("time:%11u %s ERROR: node:%03u data:%02x %02x %02x %02x %02x %02x %02x",
                               pEvent->ulTimeStamp,
                               cError[pEvent->ucaEventData[0]],
                               pEvent->ucaEventData[1],
                               pEvent->ucaEventData[2],
                               pEvent->ucaEventData[3],
                               pEvent->ucaEventData[4],
                               pEvent->ucaEventData[5],
                               pEvent->ucaEventData[6],
                               pEvent->ucaEventData[7],
                               pEvent->ucaEventData[8]
                              );
            emit(doPrint(cEventText));
            break;


        }



        break;

    case 32: //PDO received
        cEventText.sprintf("time:%11u PDO RECEIVED: %x, length:%u, data:%02x %02x %02x %02x %02x %02x %02x                     ",
                           pEvent->ulTimeStamp,
                           getBufferVal16(&pEvent->ucaEventData[0]),
                           pEvent->ucaEventData[2],
                           pEvent->ucaEventData[3],
                           pEvent->ucaEventData[4],
                           pEvent->ucaEventData[5],
                           pEvent->ucaEventData[6],
                           pEvent->ucaEventData[7],
                           pEvent->ucaEventData[8],
                           pEvent->ucaEventData[9]
                          );
//         emit(doPrint(cEventText));
	
	memcpy(&pdoData, &pEvent->ucaEventData[3], pEvent->ucaEventData[2]);
	emit(sigPdoRcv(getBufferVal16(&pEvent->ucaEventData[0]), pEvent->ucaEventData[2], pdoData));

        break;
    case 33: //shutdown
        cEventText.sprintf("time:%11u SHUTDOWN                                                                                 ",pEvent->ulTimeStamp);
        emit(doPrint(cEventText));
        break;
    case 34: //startup
        cEventText=cEventText.sprintf("time:%11u STARTUP                                                                                  ",pEvent->ulTimeStamp);
        emit(doPrint(cEventText));
        break;

    case 0:
        break;

    default:
        textdata = "unknown event                                                                                              ";
        emit(doPrint(textdata));
        break;

    }
    return 0;
}

/**
 ******************************************************************
 * @brief  run methode of the qthread class
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return void
 *
 ******************************************************************
 */
void CoThread::run()
{
    INT16 WaitRet = 0, ret;
    CMA_EventStruct event;
    QString text = "CoThread started, CAN handle = ";

    text.append(text.number(m_Can));

    emit(doPrint(text));


    while(WaitRet != -1)
    {
      // wait for CANopen events
      // this function returns on CANopen events and on program termination
        WaitRet = CMA_WaitForEvent(m_Can);
        switch (WaitRet)
        {
        case 0:
            // it's an CANopen event
            ret = CMA_ReadEvent(m_Can, &event);
            do
            {
                if(ret)
                {
                    QString err = err.sprintf("Error %i in CMA_ReadData()", ret);
                    emit(doPrint(err));
                }
                HandleEvent(&event);
                ret = CMA_ReadEvent(m_Can, &event);
            } while (event.ucEventType);

            break;

        default:
            // program termination or an error!
            break;
        }
    }
}

/**
 ******************************************************************
 * @brief  convert a buffer value to a unsigned short
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 * @param  buffer which contains the value
 *
 * @return value as unsigned short
 *
 ******************************************************************
 */

USIGN16 CoThread::getBufferVal16(void* buffer)
{
    USIGN16 u16val;
    memcpy(&u16val, buffer, sizeof(USIGN16));
    return u16val;
}

/**
 ******************************************************************
 * @brief  convert a buffer value to a unsigned int
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 * @param  buffer which contains the value
 *
 * @return value as unsigned int
 *
 ******************************************************************
 */

USIGN32 CoThread::getBufferVal32(void* buffer)
{
    USIGN32 u32val;
    memcpy(&u32val, buffer, sizeof(USIGN32));
    return u32val;
}
