/***************************************************************************
 *   Copyright (C) 2007-2015 by Softing Automation GmbH                    *
 *   support.automation@softing.com                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "cothread.h"

#include "qtextedit.h"
#include "qlineedit.h"
#include "qstring.h"
#include "qpushbutton.h"

#include "can_def.h"
#include "BaseTypes.h"

#include "canL2.h"
#include "cmadef.h"
#include "cmaret.h"
#include "cnmt.h"
#include "cpdo.h"
#include "csdo.h"
#include "clss.h"

#include "master.h"

/**
 ******************************************************************
 * @brief  Constructor
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

master::master()
    : KMainWindow()
{
    // tell the KXmlGuiWindow that this is indeed the main widget
    QWidget* w = new QWidget(this);
    CHDSNAPSHOT *pSnap = NULL;
    INT32 ret;
    USIGN32 needed, NumOfCh;

    setCentralWidget(w);
    m_ui.setupUi(w);

    connect(m_ui.bt_startup, SIGNAL(pressed()), this, SLOT(OnStartup()));
    connect(m_ui.bt_sdo, SIGNAL(clicked()), this, SLOT(OnSdo()));
    connect(m_ui.bt_nmt, SIGNAL(clicked()), this, SLOT(OnNmt()));
    connect(m_ui.bt_pdo, SIGNAL(clicked()), this, SLOT(OnPdo()));
    connect(m_ui.bt_lss, SIGNAL(clicked()), this, SLOT(OnLss()));
    connect(m_ui.bt_clear, SIGNAL(clicked()), this, SLOT(OnClear()));

    m_thread = NULL;
    m_psdo = NULL;
    m_plss = NULL;
    m_pnmt = NULL;
    m_ppdo = NULL;

    // get the plugged channels and insert it into the combo box
    // we need a function from the CAN Layer2 API for this;
    ret = CANL2_get_all_CAN_channels(0, // in: provided buffer size
                                     &needed, //out: needed buffer size
                                     &NumOfCh, // out: number of channels plugged
                                     NULL); // buffer pointer

    if(!ret)
    {
        unsigned int i;

        // now we know the needed buffer size and can allocate a buffer
        pSnap = (CHDSNAPSHOT *)malloc(needed);

        // and call the function again with the appropriate buffer size
        ret = CANL2_get_all_CAN_channels(needed, &needed, &NumOfCh, pSnap);

        if(ret)
        {
            // error:
            free (pSnap);
            return;
        }

        // and got the information we needed!
        for(i=0; i<NumOfCh; i++)
        {
            QString text = (char*)&pSnap[i].ChannelName[0];
            //we put the channel names into the combo box
            m_ui.cb_channelChoise->addItem(text);
        }

        // the buffer is no more needed now, we can free it.
        free (pSnap);

        m_ui.cb_channelChoise->setSizeAdjustPolicy(QComboBox::AdjustToContents);
    }
}

/**
 ******************************************************************
 * @brief  Destructor
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

master::~master()
{
    // shut down CANopen stack and free thread from wait function
    CMA_Shutdown(m_ucHandle);
    CMA_CloseCard(m_ucHandle);

    if(m_plss)
    {
        delete m_plss;
    }

    if(m_pnmt)
    {
        delete m_pnmt;
    }

    if(m_ppdo)
    {
        delete m_ppdo;
    }

    if(m_psdo)
    {
        delete m_psdo;
    }

    if(m_thread)
    {
        m_thread->wait(4000);
        delete m_thread;
    }

}

/**
 ******************************************************************
 * @brief  slot: startup button clicked
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void master::OnStartup()
{
    USIGN16 retCode;
    CMA_CAN_CONFIG stMstCfg;
    QString ErrorText;
    QString textString;


    OnClear();
    textString = "initialization started!";
    OnPrint(textString);

    // call CMA_InitializeCard with the channel name of the channel, which should be opened.
    retCode = CMA_InitializeCard (
                  0xFFFFFFFF, // device index (use always 0xFFFFFFFF)
                  1,          // functionality allways 1
                  &m_ucHandle, // out: channel handle
                  0,          // license: no license needed -> allways 0
                  (USIGN8*)m_ui.cb_channelChoise->currentText().toAscii().data()); // in: channel name

    if(retCode)
    {
        ErrorText = ErrorText.sprintf("ERROR: %i in CMA_InitializeCard", retCode);
        OnPrint(ErrorText);
        return;
    }

    // create the thread to handle the CANopen events
    if(!m_thread)
    {
        m_thread = new CoThread(m_ucHandle);
        connect(m_thread, SIGNAL(doPrint(QString)), this, SLOT(OnPrint(QString)));
    }

    m_thread->start();

    /* INITIALIZE CANOPEN MASTER */
    stMstCfg.ucBaudRate = (USIGN8)m_ui.ed_baud->text().toShort(); // 4 = 125Kbit/s
    stMstCfg.ucAccCode = 0; // we do not want any filterring
    stMstCfg.ucAccMask = 0; // everything should be received!

    // set baudrate and acceptance filter
    retCode = CMA_ConfigCANChannel(m_ucHandle,stMstCfg);
    if(retCode)
    {
        ErrorText = ErrorText.sprintf("ERROR: %i in CMA_ConfigCANChannel", retCode);
        OnPrint(ErrorText);
        return;
    }

    // assign the masters node id and start the node manager
    retCode = CMA_InitNodeMan(m_ucHandle,m_ui.ed_master_nodeid->text().toShort());
    if(retCode)
    {
        ErrorText = ErrorText.sprintf("ERROR: %i in CMA_InitNodeMan", retCode);
        OnPrint(ErrorText);
        return;
    }

    // start the SDO manager
    retCode = CMA_InitSDOMan(m_ucHandle, // channel handle
                             3000,       // SDO timeout 3 sec.
                             3000);      // Application timeout factor 3000 ms
    if(retCode)
    {
        ErrorText = ErrorText.sprintf("ERROR: %i in CMA_InitSDOMan", retCode);
        OnPrint(ErrorText);
        return;
    }

    // start the stack; after this call the CANopen master is online.
    retCode = CMA_Startup(m_ucHandle);
    if(retCode)
    {
        ErrorText = ErrorText.sprintf("ERROR: %i in CMA_Startup", retCode);
        OnPrint(ErrorText);
        return;
    }

    m_ui.bt_pdo->setEnabled(true);
    m_ui.bt_sdo->setEnabled(true);
    m_ui.bt_nmt->setEnabled(true);
    m_ui.bt_lss->setEnabled(true);

}

/**
 ******************************************************************
 * @brief  slot: lss (Layer Setting Services) button clicked
 * @brief  the lss dialog will be started by this function
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void master::OnLss()
{
    if(m_plss==NULL)
    {
        // ok it's a first start
        m_plss = new lss(m_ucHandle);;
        // connect the signals of the dialog
        connect(m_plss, SIGNAL(finished(int)), this, SLOT(OnLssFinished(int)));
        connect(m_plss, SIGNAL(doPrint(QString)), this, SLOT(OnPrint(QString)));
    }
    else
    {
        // restart
        m_plss->m_pw->show();
    }
}

/**
 ******************************************************************
 * @brief  slot: sdo button clicked
 * @brief  the lss dialog will be started by this function
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void master::OnSdo()
{
    if(m_psdo==NULL)
    {
        m_psdo = new sdo(m_ucHandle);
        connect(m_psdo, SIGNAL(finished(int)), this, SLOT(OnSdoFinished(int)));
        connect(m_psdo, SIGNAL(doPrint(QString)), this, SLOT(OnPrint(QString)));
	connect(m_thread, SIGNAL(sigSdoRcv(unsigned long, quint64)), m_psdo, SLOT(OnSdoUploadComplete(unsigned long, quint64)));
    }
    else
    {
        m_psdo->m_pw->show();
    }
}


/**
 ******************************************************************
 * @brief  slot: nmt button clicked
 * @brief  the lss dialog will be started by this function
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void master::OnNmt()
{
    if(m_pnmt==NULL)
    {
        m_pnmt = new nmt(m_ucHandle);;
        connect(m_pnmt, SIGNAL(finished(int)), this, SLOT(OnNmtFinished(int)));
        connect(m_pnmt, SIGNAL(doPrint(QString)), this, SLOT(OnPrint(QString)));
        connect(m_thread, SIGNAL(sig_errorEvent(unsigned char,unsigned char,int)), m_pnmt, SLOT(OnErrorEvent(unsigned char,unsigned char,int)));
    }
    else
    {
        m_pnmt->m_pw->show();
    }

}

/**
 ******************************************************************
 * @brief  slot: pdo button clicked
 * @brief  the lss dialog will be started by this function
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void master::OnPdo()
{
    if(m_ppdo==NULL)
    {
        m_ppdo = new pdo(m_ucHandle);;
        connect(m_ppdo, SIGNAL(finished(int)), this, SLOT(OnPdoFinished(int)));
        connect(m_ppdo, SIGNAL(doPrint(QString)), this, SLOT(OnPrint(QString)));
	connect(m_thread, SIGNAL(sigPdoRcv(unsigned long, unsigned char, quint64)), m_ppdo, SLOT(OnPdoReceived(unsigned long, unsigned char, quint64)));

    }
    else
    {
        m_ppdo->m_pw->show();
    }

}

/**
 ******************************************************************
 * @brief  slot: sdo dialog finished signal received
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void master::OnSdoFinished(int ret)
{
    m_psdo->m_pw->hide();
    ret = ret;
}

/**
 ******************************************************************
 * @brief  slot: pdo dialog finished signal received
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void master::OnPdoFinished(int ret)
{
    m_ppdo->m_pw->hide();
    ret = ret;
}

/**
 ******************************************************************
 * @brief  slot: lss dialog finished signal received
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void master::OnLssFinished(int ret)
{
    m_plss->m_pw->hide();
    ret = ret;
}

/**
 ******************************************************************
 * @brief  slot: nmt dialog finished signal received
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void master::OnNmtFinished(int ret)
{
    m_pnmt->m_pw->hide();
    ret = ret;
}

/**
 ******************************************************************
 * @brief  print a message on the edit box
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void master::OnPrint(QString PrintIt)
{
    //m_mutex.lock();
    m_ui.ed_message->append(PrintIt);
    //m_mutex.unlock();
}

/**
 ******************************************************************
 * @brief  clear the edit box
 * @author Werner Knorsch, Softing Automation GmbH
 * @date   2013-02-05
 *
 * @return -
 *
 ******************************************************************
 */

void master::OnClear()
{
    m_mutex.lock();
    m_ui.ed_message->clear();
    m_mutex.unlock();
}





#include "master.moc"
