/**
 * @file   LeanCANopen.c
 * @brief  Main module for LeanCANopen API 
 ******************************************************************
 * $Source: /cvs/projekte/CAN/usermode/LeanCANopen/LeanCANopen.c,v $
 * $Author: kno $
 * $Date: 2006/10/30 09:13:30 $
 * $Revision: 1.6 $
 ******************************************************************
 * @verbatim
 * OS       :   Linux kernel 2.6.16
 * Target   :   Linux
 * Compiler :   gcc V. 4.0.2
 * Copyright:   Softing AG
 *              Richard-Reizner-Alle 6
 *              D-85540 Munich
 * @endverbatim
 ******************************************************************
 */

/**
 *
 *
 * <h2>LeanCANopen API </h2>
 * \image html softing.jpg
 * \image latex softing.eps
 * <br>
 * Richard-Reitzner-Allee 6 <br>
 * 85540 Haar <br>
 * Phone +49 89/456 56-337 <br>
 * Fax   +49 89/456 56-499 <br>
 *
 * Author: Werner Knorsch
 *
 * 
 *
 *
 * \section Beschreibung
 *
 * Dieses Dokument beschreibt die Architektur des CANopen API
 * The LeanCANopen API consists of the following files:
 * LeanCANopen.c
 * LeanCANopen.h 
 * BaseTypes.h 
 * CmaDef.h
 * CmaRet.h
 * libSoftingCan-5.x.so
 *
 * 
 * \section 2. section
 *
 *
 * */
 
  /* This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
  * published by the Free Software Foundation; either version 2 of
  * the License, or (at your option) any later version.
  * 
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
  * MA 02111-1307, USA.
  *
  * email: support.automation@softing.com
  * Softing AG, Richard-Reitzner-Allee 6, 85540 Haar, Germany
  */


/****************************************************************************
**  include files
****************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <pthread.h>
#include <errno.h>
#include <sys/poll.h>
#include <unistd.h>


#include "BaseTypes.h"
#include "can_def.h"
#include "canL2.h"

#include "LeanCANopen.h"
#include "cmaret.h"

#include "cmadef.h"

#define UNUSED_PARA(x) x=x


/******************************************************************
**  constants
*******************************************************************/

const char *strBaudrate[9]={"1 Mbaud", "800 kbaud", "500 kbaud", "250 kbaud", "125 kbaud", "100 kbaud", "50 kbaud", "20 kbaud", "10 kbaud"};
const int presc[9] ={ 1, 1, 1, 2, 4, 4, 8,19, 32};
const int sjw[9]   ={ 2, 1, 4, 4, 4, 4, 4, 1, 4};
const int tseg1[9] ={ 4, 6, 8, 8, 8,11,11,15, 16};
const int tseg2[9] ={ 3, 3, 7, 7, 7, 8, 8, 5, 8};
const int sample[9]={ 0, 0, 0, 0, 0, 0, 0, 0, 0};


const USIGN8 BUS_ERR_TOGGLEBIT[4]       = {0x00,0x00,0x03,0x05};
const USIGN8 BUS_ERR_TIMEOUT[4]         = {0x00,0x00,0x04,0x05};
const USIGN8 BUS_ERR_CLIENT_SERVER[4]   = {0x01,0x00,0x04,0x05};
const USIGN8 BUS_ERR_BLOCK_SIZE[4]      = {0x02,0x00,0x04,0x05};
const USIGN8 BUS_ERR_SEQ_NUMBER[4]      = {0x03,0x00,0x04,0x05};
const USIGN8 BUS_ERR_CRC_CODE[4]        = {0x04,0x00,0x04,0x05};
const USIGN8 BUS_ERR_OUT_OF_MEMORY[4]   = {0x05,0x00,0x04,0x05};
const USIGN8 BUS_ERR_OBJ_ACC_NOT_SUP[4] = {0x00,0x00,0x01,0x06};
const USIGN8 BUS_ERR_WRITEONLY[4]       = {0x01,0x00,0x01,0x06};
const USIGN8 BUS_ERR_READONLY[4]        = {0x02,0x00,0x01,0x06};
const USIGN8 BUS_ERR_NO_OBJECT[4]       = {0x00,0x00,0x02,0x06};
const USIGN8 BUS_ERR_MAPPING[4]         = {0x41,0x00,0x04,0x06};
const USIGN8 BUS_ERR_PDO_LEN[4]         = {0x42,0x00,0x04,0x06};
const USIGN8 BUS_ERR_GENERAL_PARA[4]    = {0x43,0x00,0x04,0x06};
const USIGN8 BUS_ERR_INCOMPATIBLE[4]    = {0x47,0x00,0x04,0x06};
const USIGN8 BUS_ERR_HARDWARE[4]        = {0x00,0x00,0x06,0x06};
const USIGN8 BUS_ERR_LENGTH[4]          = {0x10,0x00,0x07,0x06};
const USIGN8 BUS_ERR_GENERAL[4]         = {0x00,0x00,0x00,0x08};


const USIGN8 ERR_TOGGLEBIT[4]           = {0x05,0x03,0x00,0x00};
const USIGN8 ERR_TIMEOUT[4]             = {0x05,0x04,0x00,0x00};
const USIGN8 ERR_CLIENT_SERVER[4]       = {0x05,0x04,0x00,0x01};
const USIGN8 ERR_BLOCK_SIZE[4]          = {0x05,0x04,0x00,0x02};
const USIGN8 ERR_SEQ_NUMBER[4]          = {0x05,0x04,0x00,0x03};
const USIGN8 ERR_CRC_CODE[4]            = {0x05,0x04,0x00,0x04};
const USIGN8 ERR_OUT_OF_MEMORY[4]       = {0x05,0x04,0x00,0x05};
const USIGN8 ERR_OBJ_ACC_NOT_SUP[4]     = {0x06,0x01,0x00,0x00};
const USIGN8 ERR_WRITEONLY[4]           = {0x06,0x01,0x00,0x01};
const USIGN8 ERR_READONLY[4]            = {0x06,0x01,0x00,0x02};
const USIGN8 ERR_NO_OBJECT[4]           = {0x06,0x02,0x00,0x00};
const USIGN8 ERR_MAPPING[4]             = {0x06,0x04,0x00,0x41};
const USIGN8 ERR_PDO_LEN[4]             = {0x06,0x04,0x00,0x42};
const USIGN8 ERR_GENERAL_PARA[4]        = {0x06,0x04,0x00,0x43};
const USIGN8 ERR_INCOMPATIBLE[4]        = {0x06,0x04,0x00,0x47};
const USIGN8 ERR_HARDWARE[4]            = {0x06,0x06,0x00,0x00};
const USIGN8 ERR_LENGTH[4]              = {0x06,0x07,0x00,0x10};
const USIGN8 ERR_GENERAL[4]             = {0x08,0x00,0x00,0x00};
const USIGN8 ERR_SUBINDEX[4]            = {0x06,0x09,0x00,0x11};

/****************************************************************************
**  variables
****************************************************************************/

CHANNELS ch[MAX_CHANNEL+1];

BOOLEAN isDbgEnabled = FALSE;

int retVal = 0;


/******************************************************************
** protoypes  
*******************************************************************/

void removeFromGuardList(USIGN8 u8NodeId, PCHANNELS pCh);
void CO_get_time(CAN_HANDLE u32Channel, USIGN32* pu32Time);

int CANL2_get_device_type(CAN_HANDLE can_handle, USIGN32 *pId);


/**
 ******************************************************************
 * @brief  maps an error code from the CAN L2 API to an error  
 * @brief  code of the CANopen API.
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  errorCode: L2 error code
 * @return CANopen error code
 *
 ******************************************************************
 */

INT16 map_error(int errorCode)
{
  INT16 err = (INT16) errorCode;
  
  switch(errorCode)
  {
    case VCARD_INTERNALERROR:
      err = INIPC_IB_ERR_VC_INTERNALERROR;
      break;
      
   case VCARD_GENERALERROR:
      err = INIPC_IB_ERR_VC_GENERALERROR;
      break;
      
   case VCARD_TIMEOUT:
      err = INIPC_IB_ERR_VC_TIMEOUT;
      break;
      
   case VCARD_ILLEGALCALL:
      err = INIPC_IB_ERR_VC_ILLEGALCALL;
      break;
      
   case VCARD_DRIVERNOTFOUND:
   case VCARD_DRIVERNOTPRESENT:
   case VCARD_DEVICENOTFOUND:
      err = INIPC_IB_PNP_NO_DEVICE_FOUND;
      break;
      
   case VCARD_NOTENOUGHMEMORY:
      err = INIPC_IB_ERR_VC_PHYSMEMORYOVERRUN;
      break;
      
   case INIPC_IB_ERR_ACCESS_DPRAM:
   case VCARD_NODPRAM:
   case VCARD_DPRAMACCESS:
      err = CANPC_RB_ERR_DPRAM_ACCESS;
      break;
      
   case VCARD_CARDREACTION:
      err = CANPC_RB_ERR_CRD_RESP;
      break;
      
   case VCARD_DEVICEISOPEN:
      err = INIPC_IB_ERR_VC_DEVICEACCESSERROR;
      break;
      
   case VCARD_CHANNELNOTOPEN:
   case CANL2_INVALID_CHANNEL_HANDLE:
     err = DDC_ERR_WRONG_DEVICEHANDLE;
     break;
      
   case VCARD_MAXNUMOFCHANNELSOPEN:
   case CANL2_TOO_MANY_CHANNELS:
     err = INIPC_IB_ERR_TOO_MANY_CH_DEFINED;
     break;
     
  case CANL2_ERR:
  case CANL2_BOARD_NOT_INITIALIZED:
  case CANL2_WRONG_PARAM:
     err = INIPC_IB_ERR_VC_GENERALERROR;
     break;
     
  case CANL2_FW_TIMEOUT:
    err = CMA_ERR_FUNCTION_TIMEOUT;
    break;
      
    default:
      
      break;
  }
  return err;
}

/**
 ******************************************************************
 * @brief  inserts an event in the user event queue. 
 * @brief  the user will be informed by setting the appropriate event
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pcEvent: Event which will be queued
 * @param  pCh: pointer to the channel
 * @return NULL if an error occures, otherwise a pointer to the event
 *
 ******************************************************************
 */

PCO_EVENTSTRUCT insertUserEvent(CO_EVENTSTRUCT *pcEvent, PCHANNELS pCh)
{
  COPRINT(("insertUserEvent: \n"));
  
  if(isDbgEnabled)
    SLEEP(10);
  if(pCh->UserEvents.wr == pCh->UserEvents.rd)
  {
    COPRINT(("out of memory...\n"));
    return NULL;
  }
  else
  {
    int rc = 0;
    
    memcpy(&pCh->UserEvents.event[pCh->UserEvents.wr], pcEvent, sizeof(CO_EVENTSTRUCT));
    pCh->UserEvents.wr = (pCh->UserEvents.wr + 1) & FIFOMASK;

   // if(pCh->hQueueEvent)
    {
      pCh->s16WaitRet = 0;
      
      
      rc = pthread_mutex_lock(&pCh->hQueueEvent);
      
      /* The condition has occured. Set the flag and wake up any waiting threads */
      pCh->conditionMet = TRUE;
//       printf("Wake up all waiting threads...\n");
      rc = pthread_cond_signal(&pCh->QueueCondition);
//       checkResults("pthread_cond_broadcast()\n", rc);
      
//       rc = pthread_mutex_unlock(&mutex);
//       checkResults("pthread_mutex_unlock()\n", rc);
      
      pthread_mutex_unlock(&pCh->hQueueEvent); //(pCh->hQueueEvent);
    }
  }
  COPRINT(("insertUserEvent done! time: %lu\n",pcEvent->ulTimeStamp));
  return &pCh->UserEvents.event[pCh->UserEvents.wr];
}


/**
 ******************************************************************
 * @brief  when the user gets the event which is set by insertUserEvent, 
 * @brief  he calls CMA_ReadEvent(). This function calls popUserEvent.
 * @brief  popUserEvent reads the event from the queue and returns it
 * @brief  to the User. 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  u8Channel: channel handle
 * @param  pEvStruct: the event will be copied to this structure
 * @return always 0
 *
 ******************************************************************
 */

USIGN16 popUserEvent(USIGN8 u8Channel, CMA_EventStruct *pEvStruct)
{
  PCHANNELS pCh = &ch[u8Channel];
  USIGN16 rd = (pCh->UserEvents.rd + 1) & FIFOMASK;

  if(pCh->UserEvents.wr == rd)
  {
    USIGN32 time;
    CO_get_time(ch[u8Channel].channel, &time);
    pEvStruct->ucEventType = 0;
    pEvStruct->ulTimeStamp = time;
    if(isDbgEnabled)
    {
      COPRINT(("EMPTY EVENT\n"));
    }
    return 0;
  }
  else
  {
    memcpy(pEvStruct, &pCh->UserEvents.event[rd], sizeof(CMA_EventStruct));
    pCh->UserEvents.rd = rd;
      
  }
  return 0;

}



/**
 ******************************************************************
 * @brief  inserts an error event in the event queue 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel
 * @param  u32TimeStamp: timestamp
 * @param  u8DataLen: data length of the event data
 * @param  isTimeValid: if FALSE the timestamp must be generated by the function, else u32TimeStamp should be used
 * @return nothing
 *
 ******************************************************************
 */

void insertErrorEvent(PCHANNELS pCh, 
                      USIGN32 u32TimeStamp, 
                      USIGN8 u8DataLen,
                      USIGN8 *pData, 
                      CAN_BOOL isTimeValid
                      )
{
  CO_EVENTSTRUCT event;
  USIGN32 u32Time, i;
  
  //COPRINT(("insertErrorEvent:\n"));
  if(isTimeValid)
  {
    u32Time = u32TimeStamp;
  }
  else
  {
    CO_get_time(pCh->channel, &u32Time);
  }

  for(i=0; i<u8DataLen; i++)
  {
    event.ucaEventData[i] = pData[i];
  }
  
  event.ucEventType = EV_ERROR;
  event.ulTimeStamp = u32Time;

  insertUserEvent(&event, pCh);
 // COPRINT(("insertErrorEvent done! time: %lu\n",u32Time));
  
}

/**
 ******************************************************************
 * @brief  inserts an API error event in the event queue 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  u32Channel: channel handle
 * @param  errorCode: error code
 * @return nothing
 *
 ******************************************************************
 */

void insertAPIError(USIGN32 u32Channel, int errorCode)
{
  CO_EVENTSTRUCT event;
  int i;

  COPRINT(("insertAPIError: ch:%lu\n",u32Channel));
  
  CO_get_time(u32Channel, &event.ulTimeStamp);

  event.ucEventType = EV_ERROR;
  event.ucaEventData[0] = 1; // API error

  *(USIGN32*)&event.ucaEventData[2] = errorCode;

  for(i=0; i<MAX_CHANNEL; i++)
  {
    if(ch[i].channel == u32Channel)
    {
      insertUserEvent(&event, &ch[i]);
     // COPRINT(("insertAPIError done! time: %lu\n",event.ulTimeStamp));
      return;
    }
  }

  return;

}

/**
 ******************************************************************
 * @brief  returns the timestamp in pu32Time 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  u32Channel: channel handle
 * @param  pu32Time: actual timestamp
 * @return nothing
 *
 ******************************************************************
 */


void CO_get_time(CAN_HANDLE u32Channel, USIGN32* pu32Time)
{
  int ret;

  ret = CANL2_get_time(u32Channel, pu32Time);

  if(ret)
  {
    COPRINT(("CO_get_time error!\n"));
    insertAPIError(u32Channel, ret);
  }
}

/**
 ******************************************************************
 * @brief  sends the data in pData over the CAN bus 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  u32Channel: channel handle
 * @param  Ident: CAN identifier
 * @param  isXtd: if TRUE CAN 2.0 B protocol is used, else CAN 2.0 A
 * @param  DataLen: data length of the CAN message
 * @param  pData: pointer to the data, which should be sent
 * @return nothing
 *
 ******************************************************************
 */

void CO_send_data(PCHANNELS pCh, USIGN32 u32Channel, USIGN32 Ident, int isXtd, int DataLen, USIGN8 *pData)
{
  int ret = -1,i;
  
  for(i=0; ((i<MAX_XMT_TRIALS) && ret); i++)
  {
    ret = CANL2_send_data(u32Channel, Ident, isXtd, DataLen, pData);
    
    if(ret)
    {
      pthread_mutex_unlock(&pCh->ChSemaphore);
      SLEEP(1);
      pthread_mutex_lock(&pCh->ChSemaphore);
    }
  }
  
  if(ret)
  {
    insertAPIError(u32Channel, ret);
  }
  
}

/**
 ******************************************************************
 * @brief  sends a remote frame over the CAN bus 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  u32Channel: channel handle
 * @param  Ident: CAN identifier
 * @param  isXtd: if TRUE CAN 2.0 B protocol is used, else CAN 2.0 A
 * @param  DataLen: data length of the CAN message
 * @return nothing
 *
 ******************************************************************
 */

void CO_send_remote(PCHANNELS pCh, USIGN32 u32Channel, USIGN32 Ident, int isXtd, int DataLen)
{
  int ret = -1,i;

  for(i=0; ((i<MAX_XMT_TRIALS) && ret); i++)
  {
    ret = CANL2_send_remote(u32Channel, Ident, isXtd, DataLen);
    if(ret)
    {
      pthread_mutex_unlock(&pCh->ChSemaphore);
      SLEEP(1);
      pthread_mutex_lock(&pCh->ChSemaphore);
    }
  }

  if(ret)
  {
    insertAPIError(u32Channel, ret);
  }

}

/**
 ******************************************************************
 * @brief  reads an event from the CAN bus 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  u32Channel: channel handle
 * @param  param: param struct for layer2 see CAN Layer2 manual
 * @return the event type if positive, or an error type if negative.
 *
 ******************************************************************
 */


int CO_read_ac(USIGN32 u32Channel, PARAM_STRUCT *param)
{
  int ret;

  ret = CANL2_read_ac(u32Channel, param);

  if(ret<0)
  {
    insertAPIError(u32Channel, ret);
  }
  return ret;
}

/**
 ******************************************************************
 * @brief  aborts a SDO transfer in cause of an error 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  psdo: SDO for which the transfer should aborted
 * @param  u8Node: node id
 * @param  u32Reason: reason for aborting the transfer
 * @return nothing
 *
 ******************************************************************
 */

void SDOAbort(PCHANNELS pCh, 
                    PSDO psdo, 
                    USIGN8 u8Node, 
                    USIGN32 u32Reason)
{
  USIGN8 ev_data[20], can_data[8];

  memset(can_data, 0, 8);

  ev_data[0] = ERR_SDO;
  ev_data[1] = u8Node;
  
  switch(u32Reason)
  {
  case TOGGLE_ERR:
    // wrong state of toggle bit
    memcpy(&ev_data[2], ERR_TOGGLEBIT, 4);
    
    insertErrorEvent(pCh,0,6,ev_data,FALSE);
    
    memcpy(&can_data[4], BUS_ERR_TOGGLEBIT, 4);
    can_data[0] = CS_ABORT;
    can_data[3] = psdo->u8SubIndex;

    WRT16(&can_data[1], psdo->u16Index);
    CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, can_data);
    break;
    
  case TIMEOUT_ERR:
    // timeout, slave does not respond
    memcpy(&ev_data[2], ERR_TIMEOUT, 4);
    
    insertErrorEvent(pCh,0,6,ev_data,FALSE);
    
    memcpy(&can_data[4], BUS_ERR_TIMEOUT, 4);
    can_data[0] = CS_ABORT;
    can_data[3] = psdo->u8SubIndex;

    WRT16(&can_data[1], psdo->u16Index);
    CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, can_data);


    break;

  case CLIENT_SERVER_ERR:
    memcpy(&ev_data[2], ERR_CLIENT_SERVER, 4);
    
    insertErrorEvent(pCh,0,6,ev_data,FALSE);
    
    memcpy(&can_data[4], BUS_ERR_CLIENT_SERVER, 4);
    can_data[0] = CS_ABORT;
    can_data[3] = psdo->u8SubIndex;

    WRT16(&can_data[1], psdo->u16Index);
    CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, can_data);
    break;
  case BLOCK_SIZE_ERR:
    memcpy(&ev_data[2], ERR_BLOCK_SIZE, 4);
    
    insertErrorEvent(pCh,0,6,ev_data,FALSE);
    
    memcpy(&can_data[4], BUS_ERR_BLOCK_SIZE, 4);
    can_data[0] = CS_ABORT;
    can_data[3] = psdo->u8SubIndex;

    WRT16(&can_data[1], psdo->u16Index);
    CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, can_data);
    break;
  case SEQ_NUMBER_ERR:
    memcpy(&ev_data[2], ERR_SEQ_NUMBER, 4);
    
    insertErrorEvent(pCh,0,6,ev_data,FALSE);
    
    memcpy(&can_data[4], BUS_ERR_SEQ_NUMBER, 4);
    can_data[0] = CS_ABORT;
    can_data[3] = psdo->u8SubIndex;

    WRT16(&can_data[1], psdo->u16Index);
    CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, can_data);
    break;
  case CRC_CODE_ERR:
    memcpy(&ev_data[2], ERR_CRC_CODE, 4);
    
    insertErrorEvent(pCh,0,6,ev_data,FALSE);
    
    memcpy(&can_data[4], BUS_ERR_CRC_CODE, 4);
    can_data[0] = CS_ABORT;
    can_data[3] = psdo->u8SubIndex;

    WRT16(&can_data[1], psdo->u16Index);
    CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, can_data);
    break;
  case OUT_OF_MEMORY_ERR:
    memcpy(&ev_data[2], ERR_OUT_OF_MEMORY, 4);
    
    insertErrorEvent(pCh,0,6,ev_data,FALSE);
    
    memcpy(&can_data[4], BUS_ERR_OUT_OF_MEMORY, 4);
    can_data[0] = CS_ABORT;
    can_data[3] = psdo->u8SubIndex;

    WRT16(&can_data[1], psdo->u16Index);
    CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, can_data);
    break;
  case OBJ_ACC_NOT_SUP_ERR:
    memcpy(&ev_data[2], ERR_OBJ_ACC_NOT_SUP, 4);
    
    insertErrorEvent(pCh,0,6,ev_data,FALSE);
    
    memcpy(&can_data[4], BUS_ERR_OBJ_ACC_NOT_SUP, 4);
    can_data[0] = CS_ABORT;
    can_data[3] = psdo->u8SubIndex;

    WRT16(&can_data[1], psdo->u16Index);
    CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, can_data);
    break;
  case WRITEONLY_ERR:
    memcpy(&ev_data[2], ERR_WRITEONLY, 4);
    
    insertErrorEvent(pCh,0,6,ev_data,FALSE);
    
    memcpy(&can_data[4], BUS_ERR_WRITEONLY, 4);
    can_data[0] = CS_ABORT;
    can_data[3] = psdo->u8SubIndex;

    WRT16(&can_data[1], psdo->u16Index);
    CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, can_data);
    break;
  case READONLY_ERR:
    memcpy(&ev_data[2], ERR_READONLY, 4);
    
    insertErrorEvent(pCh,0,6,ev_data,FALSE);
    
    memcpy(&can_data[4], BUS_ERR_READONLY, 4);
    can_data[0] = CS_ABORT;
    can_data[3] = psdo->u8SubIndex;

    WRT16(&can_data[1], psdo->u16Index);
    CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, can_data);
    break;
  case NO_OBJECT_ERR:
    memcpy(&ev_data[2], ERR_NO_OBJECT, 4);
    
    insertErrorEvent(pCh,0,6,ev_data,FALSE);
    
    memcpy(&can_data[4], BUS_ERR_NO_OBJECT, 4);
    can_data[0] = CS_ABORT;
    can_data[3] = psdo->u8SubIndex;

    WRT16(&can_data[1], psdo->u16Index);
    CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, can_data);
    break;
  case MAPPING_ERR:
    memcpy(&ev_data[2], ERR_MAPPING, 4);
    
    insertErrorEvent(pCh,0,6,ev_data,FALSE);
    
    memcpy(&can_data[4], BUS_ERR_MAPPING, 4);
    can_data[0] = CS_ABORT;
    can_data[3] = psdo->u8SubIndex;

    WRT16(&can_data[1], psdo->u16Index);
    CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, can_data);
    break;
  case PDO_LEN_ERR:
    memcpy(&ev_data[2], ERR_PDO_LEN, 4);
    
    insertErrorEvent(pCh,0,6,ev_data,FALSE);
    
    memcpy(&can_data[4], BUS_ERR_PDO_LEN, 4);
    can_data[0] = CS_ABORT;
    can_data[3] = psdo->u8SubIndex;

    WRT16(&can_data[1], psdo->u16Index);
    CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, can_data);
    break;
    
  case GENERAL_PARA_ERR:
    memcpy(&ev_data[2], ERR_GENERAL_PARA, 4);
    
    insertErrorEvent(pCh,0,6,ev_data,FALSE);
    
    memcpy(&can_data[4], BUS_ERR_GENERAL_PARA, 4);
    can_data[0] = CS_ABORT;
    can_data[3] = psdo->u8SubIndex;

    WRT16(&can_data[1], psdo->u16Index);
    CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, can_data);
    break;
    
  case INCOMPATIBLE_ERR:
    memcpy(&ev_data[2], ERR_INCOMPATIBLE, 4);
    
    insertErrorEvent(pCh,0,6,ev_data,FALSE);
    
    memcpy(&can_data[4], BUS_ERR_INCOMPATIBLE, 4);
    can_data[0] = CS_ABORT;
    can_data[3] = psdo->u8SubIndex;

    WRT16(&can_data[1], psdo->u16Index);
    CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, can_data);
    break;
    
  case HARDWARE_ERR:
    memcpy(&ev_data[2], ERR_HARDWARE, 4);
    
    insertErrorEvent(pCh,0,6,ev_data,FALSE);
    
    memcpy(&can_data[4], BUS_ERR_HARDWARE, 4);
    can_data[0] = CS_ABORT;
    can_data[3] = psdo->u8SubIndex;

    WRT16(&can_data[1], psdo->u16Index);
    CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, can_data);
    break;
    
  case LENGTH_ERR:
    memcpy(&ev_data[2], ERR_LENGTH, 4);
    
    insertErrorEvent(pCh,0,6,ev_data,FALSE);
    
    memcpy(&can_data[4], BUS_ERR_LENGTH, 4);
    can_data[0] = CS_ABORT;
    can_data[3] = psdo->u8SubIndex;

    WRT16(&can_data[1], psdo->u16Index);
    CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, can_data);
    break;
    
  case LEN_TOO_HIGH_ERR:
        
    break;
    
  case LEN_TOO_LOW_ERR:
        
    break;
  case SUBINDEX_ERR:
        
    break;
  case VALUE_RANGE_ERR:
        
    break;
  case VAL_TOO_HIGH_ERR:
        
    break;
  case VAL_TOO_LOW_ERR:
        
    break;
  case MINMAX_ERR:
        
    break;
  case GENERAL_ERR:
    memcpy(&ev_data[2], ERR_GENERAL, 4);
    
    insertErrorEvent(pCh,0,6,ev_data,FALSE);
    
    memcpy(&can_data[4], BUS_ERR_GENERAL, 4);
    can_data[0] = CS_ABORT;
    can_data[3] = psdo->u8SubIndex;

    WRT16(&can_data[1], psdo->u16Index);
    CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, can_data);
    break;
  case DATA_ACCESS_ERR:
        
    break;
  case DEVICE_STATE_ERR:
        
    break;
  case NO_OD_ERR:
    
    break;
    
  }  
  if((psdo->DataBuffer.pBuffer) && psdo->DataBuffer.isAllocated)
  {
    free(psdo->DataBuffer.pBuffer);
    psdo->DataBuffer.isAllocated = FALSE;
    psdo->DataBuffer.pBuffer = NULL;
    psdo->DataBuffer.u32Size = 0;
    psdo->DataBuffer.u32BufferPtr = 0;
  }
  psdo->isUsed = FALSE;
  psdo->isWaiting = FALSE;
  psdo->isLocked = FALSE;
}

/**
 ******************************************************************
 * @brief  returns a pointer to the PDO from CobId 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  u16CobId: CobId of the PDO
 * @return a pointer to the PDO from the PDO-Index
 *
 ******************************************************************
 */

PPDO getPdo(PCHANNELS pCh, USIGN32 u16IntCobId)
{

  if(u16IntCobId <= 0x7FF)
    return &pCh->pdo[u16IntCobId-0x181];

  return NULL;

}

USIGN8 getPdoType(USIGN16 u16CobId)
{

  if((u16CobId > 0x180) && (u16CobId <= 0x200))
    return TX_PDO;

  if((u16CobId > 0x200) && (u16CobId <= 0x280))
    return RX_PDO;

  if((u16CobId > 0x280) && (u16CobId <= 0x300))
    return TX_PDO;

  if((u16CobId > 0x300) && (u16CobId <= 0x380))
    return RX_PDO;

  if((u16CobId > 0x380) && (u16CobId <= 0x400))
    return TX_PDO;

  if((u16CobId > 0x400) && (u16CobId <= 0x480))
    return RX_PDO;

  if((u16CobId > 0x480) && (u16CobId <= 0x500))
    return TX_PDO;

  if((u16CobId > 0x500) && (u16CobId <= 0x580))
    return RX_PDO;

  return 2;

}

/**
 ******************************************************************
 * @brief  returns the PDO-Index   
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  u16CobId: CobId of the PDO
 * @return returns the PDO-Index
 *
 ******************************************************************
 */

static USIGN16 getPdoIndex(PCHANNELS pCh, USIGN16 u16ExtCobId, BOOLEAN bRxPdo)
{
  USIGN16 index;

  if(u16ExtCobId > 0x7FF)
    return (USIGN16)0xFFFF;
 
  for(index=0; index <= MAX_PDO; index++)
  {
    if((!pCh->pdo[index].isUsed) || (pCh->u16CobInt2Ext[index+0x181] != pCh->pdo[index].u16CobId))
    {
      if(bRxPdo)
      {
        if(getPdoType((USIGN16)(index+0x181)) == RX_PDO)
        {
          index += 0x181;
          pCh->u16CobInt2Ext[index] = u16ExtCobId;
          pCh->u16CobExt2Int[u16ExtCobId] = index;
          return (USIGN16)index;
        }
      }
      else
      {
        if(getPdoType((USIGN16)(index+0x181)) == TX_PDO)
        {
          index += 0x181;
          pCh->u16CobInt2Ext[index] = u16ExtCobId;
          pCh->u16CobExt2Int[u16ExtCobId] = index;

          return (USIGN16)index;
        }
      }
    }

  }

  return (USIGN16) 0xffff;

}

/**
 ******************************************************************
 * @brief  returns a pointer to the PDO from the PDO-Index 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  index: PDO-Index
 * @return a pointer to the PDO from the PDO-Index
 *
 ******************************************************************
 */

PPDO getPdoFromIndex(PCHANNELS pCh, USIGN32 index)
{

  if((index-0x181) <= 0x1024)
    return &pCh->pdo[index-0x181];

  return NULL;

}


/**
 ******************************************************************
 * @brief  handles an SD0 upload from master to masters Object Dictionary 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  u8NodeId: node Id of the master
 * @param  usIndex: index of the object
 * @param  ucSubIndex: subindex of the object
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */
int HandleMasterUpload(PCHANNELS pCh, USIGN8 u8NodeId, USIGN16 usIndex, USIGN8 ucSubIndex)
{
  CO_EVENTSTRUCT event;
  PSDO psdo;
  USIGN8 *pIndex = (USIGN8*)&usIndex;
  
  CO_get_time(pCh->channel, &event.ulTimeStamp);
  event.ucEventType = EV_UPLOAD_COMPLETE;

  event.ucaEventData[0] = (USIGN8)pCh->MasterNodeId;
  event.ucaEventData[1] = ucSubIndex;

  event.ucaEventData[2] = *pIndex;
  event.ucaEventData[3] = pIndex[1];

  psdo = &pCh->node[pCh->MasterNodeId].Sdo;

  if(psdo->isUsed)
    return CMA_ERR_LAST_REQ_PENDING;

  if(psdo->isLocked)
    return CMA_ERR_LAST_REQ_PENDING;

  
  
  psdo->u16Index = usIndex;
  psdo->u8SubIndex = ucSubIndex;
  
  switch(usIndex)
  {
  case 0x1000:
    event.ucaEventData[4] = 4;
    memcpy(&event.ucaEventData[5], pCh->pObjects[0x1000], 4);
    insertUserEvent(&event, pCh);
    break;
  case 0x1001:
    event.ucaEventData[4] = 1;
    memcpy(&event.ucaEventData[5], pCh->pObjects[0x1001], 1);
    insertUserEvent(&event, pCh);
    break;
  case 0x1005:
    event.ucaEventData[4] = 4;
    memcpy(&event.ucaEventData[5], pCh->pObjects[0x1005], 4);
    insertUserEvent(&event, pCh);
    break;
  case 0x1006:
    event.ucaEventData[4] = 4;
    memcpy(&event.ucaEventData[5], pCh->pObjects[0x1006], 4);
    insertUserEvent(&event, pCh);
    break;
  case 0x1008:
    psdo->DataBuffer.pBuffer = (USIGN8*)malloc(1000);
    
    
    if(!psdo->DataBuffer.pBuffer)
    { //error handling
      return CMA_ERR_MALLOC;
    }
    psdo->DataBuffer.isAllocated = TRUE;
    psdo->DataBuffer.u32BufferPtr = 0;
    psdo->DataBuffer.u32Size = 1000;
    
    psdo->isLocked = TRUE;
    event.ucaEventData[4] = 0xFF;
    memcpy((void*)psdo->DataBuffer.pBuffer, (void*)pCh->pObjects[0x1008], strlen((const char*)pCh->pObjects[0x1008]));
    psdo->DataBuffer.u32BufferPtr = strlen((const char*)pCh->pObjects[0x1008]);
    ((CO_EVENTSTRUCT_E*)&event)->ulDataLenSegmented = strlen((const char*)pCh->pObjects[0x1008]);
    insertUserEvent(&event, pCh);
    break;
  case 0x1009:
    psdo->DataBuffer.pBuffer = (USIGN8*)malloc(1000);
    
    if(!psdo->DataBuffer.pBuffer)
    { //error handling
      return CMA_ERR_MALLOC;
    }
    psdo->DataBuffer.isAllocated = TRUE;
    psdo->DataBuffer.u32BufferPtr = 0;
    psdo->DataBuffer.u32Size = 1000;
    
    psdo->isLocked = TRUE;
    event.ucaEventData[4] = 0xFF;
    memcpy(psdo->DataBuffer.pBuffer, pCh->pObjects[0x1009], strlen((const char*)pCh->pObjects[0x1009]));
    psdo->DataBuffer.u32BufferPtr = strlen((const char*)pCh->pObjects[0x1009]);
    ((CO_EVENTSTRUCT_E*)&event)->ulDataLenSegmented = strlen((const char*)pCh->pObjects[0x1009]);
    insertUserEvent(&event, pCh);
    break;
  case 0x100A:
    psdo->DataBuffer.pBuffer = (USIGN8*)malloc(1000);
    
    if(!psdo->DataBuffer.pBuffer)
    { //error handling
      return CMA_ERR_MALLOC;
    }
    psdo->DataBuffer.isAllocated = TRUE;
    psdo->DataBuffer.u32BufferPtr = 0;
    psdo->DataBuffer.u32Size = 1000;
    
    psdo->isLocked = TRUE;
    event.ucaEventData[4] = 0xFF;
    memcpy(psdo->DataBuffer.pBuffer, pCh->pObjects[0x100A], strlen((const char*)pCh->pObjects[0x100A]));
    psdo->DataBuffer.u32BufferPtr = strlen((const char*)pCh->pObjects[0x100A]);
    ((CO_EVENTSTRUCT_E*)&event)->ulDataLenSegmented = strlen((const char*)pCh->pObjects[0x100A]);
    insertUserEvent(&event, pCh);
    break;

  case 0x1018:
    if(ucSubIndex<=4)
    {
	if(ucSubIndex)
	{
	 IDENTIFY* pid = (IDENTIFY*)pCh->pObjects[0x1018];
	 event.ucaEventData[4] = 4;
	 //memcpy(&event.ucaEventData[5], ((USIGN32 *)pCh->pObjects[0x1018])[ucSubIndex], 4);
	 memcpy(&event.ucaEventData[5], &pid->id[ucSubIndex], 4);
	}
	else
	{
	 event.ucaEventData[4] = 1;
	 // event.ucaEventData[5] = (USIGN8) ((USIGN32 *)pCh->pObjects[0x1018])[0];
	 event.ucaEventData[5] = (USIGN8) 4;
	}
	insertUserEvent(&event, pCh);
    }
    else
    {
      USIGN8 data[20];

      data[0] = ERR_SDO;
      data[1] = u8NodeId;
      memcpy(&data[2], ERR_SUBINDEX, 4);
    
      insertErrorEvent(pCh,0,6,data,FALSE);

      
    }
    break;

  default:
    {
      USIGN8 data[20];

     data[0] = ERR_SDO;
     data[1] = u8NodeId;
     data[2] = 6;
     data[3] = 2;
     insertErrorEvent(pCh,0,4,data,FALSE);
    }
    break;



  }

  return 0;

}

/**
 ******************************************************************
 * @brief  handles an SD0 upload from masters Object Dictionary 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  u8NodeId: node Id of the master
 * @param  usIndex: index of the object
 * @param  ucSubIndex: subindex of the object
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */
int HandleMasterOdRd(PCHANNELS pCh, USIGN16 usIndex, USIGN8 ucSubIndex)
{
  PSDO psdo;
 // USIGN8 *pIndex = (USIGN8*)&usIndex;

  psdo = &pCh->node[pCh->MasterNodeId].Sdo;

  if(psdo->isUsed)
    return CMA_ERR_LAST_REQ_PENDING;

  if(psdo->isLocked)
    return CMA_ERR_LAST_REQ_PENDING;
  
  psdo->u16Index = usIndex;
  psdo->u8SubIndex = ucSubIndex;
  
  switch(usIndex)
  {
  case 0x1000:
  case 0x1005:
  case 0x1F80:
    psdo->DataBuffer.u32Size = 4;
    psdo->DataBuffer.u32BufferPtr = 0;
    psdo->DataBuffer.pBuffer = (USIGN8*)pCh->pObjects[usIndex];
    break;
    
  case 0x1006:
  {
   
    
    psdo->DataBuffer.u32Size = 4;
    psdo->DataBuffer.u32BufferPtr = 0;
    psdo->DataBuffer.pBuffer = (USIGN8*)pCh->pObjects[usIndex];
  }
    break;
      
  case 0x1001:
    psdo->DataBuffer.u32Size = 1;
    psdo->DataBuffer.u32BufferPtr = 0;
    psdo->DataBuffer.pBuffer = (USIGN8*)pCh->pObjects[usIndex];
    break;
  case 0x1008:
  case 0x1009:
  case 0x100A:
    psdo->DataBuffer.u32BufferPtr = 0;
    psdo->DataBuffer.u32Size = strlen((const char*)pCh->pObjects[usIndex])+1;
    psdo->isLocked = TRUE;
    psdo->DataBuffer.pBuffer = (USIGN8*)pCh->pObjects[usIndex];
    break;

  case 0x1018:
    psdo->DataBuffer.u32BufferPtr = 0;
    if(ucSubIndex<=4)
    {
      if(ucSubIndex)
      {
        IDENTIFY* pid = (IDENTIFY*)pCh->pObjects[0x1018];
        psdo->DataBuffer.u32Size = 4;
       // psdo->DataBuffer.pBuffer =  (USIGN8*) ((CAN_ADDRESS *)pCh->pObjects[usIndex])[ucSubIndex];
        psdo->DataBuffer.pBuffer =  (USIGN8*) &pid->id[ucSubIndex];
      }
      else
      {
        IDENTIFY* pid = (IDENTIFY*)pCh->pObjects[0x1018];
        psdo->DataBuffer.u32Size = 1;
        psdo->DataBuffer.pBuffer =  (USIGN8*) &pid->id[ucSubIndex];
      }
    }
    else
    { 
      return -1;
    }
    break;


    
#ifdef TESTSDOTRANSFER

  case 0x2FFF:
    psdo->DataBuffer.u32Size = 1024*64;
    psdo->DataBuffer.u32BufferPtr = 0;
    psdo->DataBuffer.pBuffer = pCh->pObjects[usIndex];
    break;
#endif // TESTSDOTRANSFER

  default:
     return -1;
    break;

  }

  return 0;

}

/**
 ******************************************************************
 * @brief  handles an SD0 upload from masters Object Dictionary 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  u8NodeId: node Id of the master
 * @param  usIndex: index of the object
 * @param  ucSubIndex: subindex of the object
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */
int HandleMasterOdWrt(PCHANNELS pCh, USIGN16 usIndex, USIGN8 ucSubIndex)
{
  PSDO psdo;
 // USIGN8 *pIndex = (USIGN8*)&usIndex;

  psdo = &pCh->node[pCh->MasterNodeId].Sdo;

  if(psdo->isUsed)
    return CMA_ERR_LAST_REQ_PENDING;

  if(psdo->isLocked)
    return CMA_ERR_LAST_REQ_PENDING;
  
  psdo->u16Index = usIndex;
  psdo->u8SubIndex = ucSubIndex;
  
  switch(usIndex)
  {
  case 0x1000:
  case 0x1005:
    psdo->DataBuffer.u32Size = 4;
    psdo->DataBuffer.u32BufferPtr = 0;
    psdo->DataBuffer.pBuffer = (USIGN8*)pCh->pObjects[usIndex];
    break;
    
  case 0x1006:
  {
    USIGN32 time;
    
    CO_get_time(pCh->channel,&time);
    pCh->u32LastSyncTime = MK_TIME(time);
 
  }
    break;
  case 0x1001:
    psdo->DataBuffer.u32Size = 1;
    psdo->DataBuffer.u32BufferPtr = 0;
    psdo->DataBuffer.pBuffer = (USIGN8*)pCh->pObjects[usIndex];
    break;
  case 0x1008:
  case 0x1009:
  case 0x100A:
    psdo->DataBuffer.u32BufferPtr = 0;
    psdo->DataBuffer.u32Size = 80;
    psdo->isLocked = TRUE;
    psdo->DataBuffer.pBuffer = (USIGN8*)pCh->pObjects[usIndex];
    break;

    
    case 0x1018:
      psdo->DataBuffer.u32BufferPtr = 0;
      if(ucSubIndex<=4)
      {
        if(ucSubIndex)
        {
          IDENTIFY* pid = (IDENTIFY*)pCh->pObjects[0x1018];
          psdo->DataBuffer.u32Size = 4;
          psdo->DataBuffer.pBuffer =  (USIGN8*) &pid->id[ucSubIndex];
        }
        else
        {
          IDENTIFY* pid = (IDENTIFY*)pCh->pObjects[0x1018];
          psdo->DataBuffer.u32Size = 1;
          psdo->DataBuffer.pBuffer =  (USIGN8*) &pid->id[ucSubIndex];
        }
      }
      else
      {
        return -1;
      }
      break;

#ifdef TESTSDOTRANSFER

  case 0x2FFF:
    psdo->DataBuffer.u32Size = 1024*64;
    psdo->DataBuffer.u32BufferPtr = 0;
    psdo->DataBuffer.pBuffer = pCh->pObjects[usIndex];
    break;
#endif // TESTSDOTRANSFER

  default:
     return -1;
    break;

  }

  return 0;

}

/**
 ******************************************************************
 * @brief  handles an SD0 download to the masters Object Dictionary 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  u8NodeId: node Id of the master
 * @param  usIndex: index of the object
 * @param  ucSubIndex: subindex of the object
 * @param  u32Size: size of the data
 * @param  pData: new object data
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */

void HandleMasterDownload(PCHANNELS pCh, 
                          USIGN8 u8NodeId, 
                          USIGN16 usIndex, 
                          USIGN8 ucSubIndex,
                          USIGN32 u32Size,
                          USIGN8 *pData)
{
  CO_EVENTSTRUCT event;
  USIGN8 *pIndex = (USIGN8*)&usIndex;
  
  CO_get_time(pCh->channel, &event.ulTimeStamp);
  event.ucEventType = EV_DOWNLOAD_COMPLETE;

  event.ucaEventData[0] = (USIGN8)pCh->MasterNodeId;
  event.ucaEventData[1] = ucSubIndex;
  
  event.ucaEventData[2] = *pIndex;
  event.ucaEventData[3] = pIndex[1];
  
  
  
  switch(usIndex)
  {
  case 0x1000:
    if(u32Size<=4)
    {
      memcpy(pCh->pObjects[0x1000], pData, u32Size);
      insertUserEvent(&event, pCh);
    }
    else
    {
      USIGN8 data[20];

     data[0] = ERR_SDO;
     data[1] = u8NodeId;
     data[2] = 6;
     data[3] = 7;
     data[4] = 0;
     data[5] = 0x12;
     insertErrorEvent(pCh,0,6,data,FALSE);
    }
    break;
  case 0x1001:
    if(u32Size==1)
    {
      memcpy(pCh->pObjects[0x1001], pData, u32Size);
      insertUserEvent(&event, pCh);
    }
    else
    {
      USIGN8 data[20];

     data[0] = ERR_SDO;
     data[1] = u8NodeId;
     data[2] = 6;
     data[3] = 7;
     data[4] = 0;
     data[5] = 0x12;
     insertErrorEvent(pCh,0,6,data,FALSE);
    }
    break;
  case 0x1005:
    if(u32Size<=4)
    {
      memcpy(pCh->pObjects[0x1005], pData, u32Size);
      insertUserEvent(&event, pCh);
    }
    else
    {
      USIGN8 data[20];

     data[0] = ERR_SDO;
     data[1] = u8NodeId;
     data[2] = 6;
     data[3] = 7;
     data[4] = 0;
     data[5] = 0x12;
     insertErrorEvent(pCh,0,6,data,FALSE);
    }
    break;
  case 0x1006:
    if(u32Size<=4)
    {
      USIGN32 time;
      
      CO_get_time(pCh->channel, &time);
      pCh->u32LastSyncTime = MK_TIME(time);
          
      memcpy(pCh->pObjects[0x1006], pData, u32Size);
      insertUserEvent(&event, pCh);
    }
    else
    {
      USIGN8 data[20];

     data[0] = ERR_SDO;
     data[1] = u8NodeId;
     data[2] = 6;
     data[3] = 7;
     data[4] = 0;
     data[5] = 0x12;
     insertErrorEvent(pCh,0,6,data,FALSE);
    }
    break;
  case 0x1008:
    {
      USIGN8 data[20];

     data[0] = ERR_SDO;
     data[1] = u8NodeId;
     data[2] = 6;
     data[3] = 1;
     data[4] = 0;
     data[5] = 2;
     insertErrorEvent(pCh,0,6,data,FALSE);
    }
    break;
  case 0x1009:
    {
      USIGN8 data[20];

     data[0] = ERR_SDO;
     data[1] = u8NodeId;
     data[2] = 6;
     data[3] = 1;
     data[4] = 0;
     data[5] = 2;
     insertErrorEvent(pCh,0,6,data,FALSE);
    }
    break;
  case 0x100A:
    {
      USIGN8 data[20];

     data[0] = ERR_SDO;
     data[1] = u8NodeId;
     data[2] = 6;
     data[3] = 1;
     data[4] = 0;
     data[5] = 2;
     insertErrorEvent(pCh,0,6,data,FALSE);
    }
    break;

  default:
    {
      USIGN8 data[20];

     data[0] = ERR_SDO;
     data[1] = u8NodeId;
     data[2] = 6;
     data[3] = 2;
     data[4] = 0;
     data[5] = 0;
     insertErrorEvent(pCh,0,6,data,FALSE);
    }
    break;



  }

}

/**
 ******************************************************************
 * @brief  initializes a SDO block download 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  u8NodeId: node Id 
 * @param  usIndex: index of the object
 * @param  ucSubIndex: subindex of the object
 * @param  u32Size: size of the data
 * @param  pData: new object data
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */

INT16 InitSdoBlockDownload(PCHANNELS pCh, 
                      USIGN8 u8NodeId,
                      USIGN16 u16Index,
                      USIGN8  u8SubIndex,
                      USIGN32 u32Size,
                      USIGN8  *pData)
{
  USIGN16 u16CobId = SDO_RX + u8NodeId;
  USIGN8 data[8];
  USIGN32 u32Time;
  PSDO  psdo;
  int ret=0;

  psdo = &pCh->node[u8NodeId].Sdo;

  if(psdo->isUsed)
    return MAP_ERROR(CMA_ERR_LAST_REQ_PENDING);

  if(psdo->isLocked)
    return MAP_ERROR(CMA_ERR_LAST_REQ_PENDING);

  CO_get_time(pCh->channel, &u32Time);

  psdo->u32LastEvent = MK_TIME(u32Time);

  if((u8NodeId == pCh->MasterNodeId) || (u8NodeId == 0))
  {
    HandleMasterDownload(pCh, u8NodeId, u16Index, u8SubIndex, u32Size, pData);
    return 0;
  }

  psdo->DataBuffer.pBuffer = (USIGN8*)malloc(u32Size);

  if(!psdo->DataBuffer.pBuffer)
  {

    return MAP_ERROR(CMA_ERR_MALLOC);
  }
  psdo->DataBuffer.isAllocated = TRUE;
  psdo->DataBuffer.u32BufferPtr = (u32Size > 4) ? 0 : u32Size;
  psdo->DataBuffer.u32Size = u32Size;
  psdo->Direction = BLOCKDOWNLOAD;
  psdo->isMasterInvolved = TRUE;
  psdo->isUsed = TRUE;
  psdo->u16CobId = u16CobId;
  psdo->u8NodeId = u8NodeId;
  psdo->isInit = TRUE;
  psdo->u8OldDataByte0 = 0;
  psdo->u8SubIndex = u8SubIndex;
  psdo->u16Index = u16Index;
  psdo->isWaiting = TRUE;
  psdo->u8BlockSeqNum = 1;
  psdo->SdoBlockState = BLK_TRANSFER_INIT;
  psdo->u16CrcCode = 0;

  memcpy(psdo->DataBuffer.pBuffer, pData, u32Size);


  memset(data, 0, 8);
  data[0] = CS_INIT_BLOCKDOWNLOAD | SUPPORT_CRC_CODE | SIZE_VALID;

  WRT16(&data[1],u16Index);
  data[3] = u8SubIndex;
  WRT32(&data[4],u32Size);

  CO_send_data(pCh, pCh->channel, u16CobId, pCh->isXtd, 8, data);

  
  return MAP_ERROR(ret);
}


/**
 ******************************************************************
 * @brief  initializes a SDO download 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  u8NodeId: node Id 
 * @param  usIndex: index of the object
 * @param  ucSubIndex: subindex of the object
 * @param  u32Size: size of the data
 * @param  pData: new object data
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */


INT16 InitSdoDownload(PCHANNELS pCh, 
                      USIGN8 u8NodeId,
                      USIGN16 u16Index,
                      USIGN8  u8SubIndex,
                      USIGN32 u32Size,
                      USIGN8  *pData)
{
  USIGN16 u16CobId = SDO_RX + u8NodeId;
  USIGN8 data[8], u8CanMsgLen;
  USIGN32 u32Time;
  PSDO  psdo;
  USIGN8 i;
  int ret=0;

  
  psdo = &pCh->node[u8NodeId].Sdo;

  if(psdo->isUsed)
    return MAP_ERROR(CMA_ERR_LAST_REQ_PENDING);

  if(psdo->isLocked)
    return MAP_ERROR(CMA_ERR_LAST_REQ_PENDING);

  
  CO_get_time(pCh->channel, &u32Time);

  psdo->u32LastEvent = MK_TIME(u32Time);
  

  if((u8NodeId == pCh->MasterNodeId) || (u8NodeId == 0))
  {
    HandleMasterDownload(pCh, u8NodeId, u16Index, u8SubIndex, u32Size, pData);
    return 0;
  }

  psdo->DataBuffer.u32BufferPtr = (u32Size > 4) ? 0 : u32Size;
  psdo->DataBuffer.u32Size = u32Size;
  psdo->DataBuffer.pBuffer = (USIGN8*)malloc(u32Size);
  psdo->Direction = DOWNLOAD;
  psdo->isMasterInvolved = TRUE;
  psdo->isUsed = TRUE;
  psdo->u16CobId = u16CobId;
  psdo->u8NodeId = u8NodeId;
  psdo->isInit = TRUE;
  psdo->u8OldDataByte0 = 0;
  psdo->u8SubIndex = u8SubIndex;
  psdo->u16Index = u16Index;
  psdo->isWaiting = TRUE;

  if(!psdo->DataBuffer.pBuffer)
    return MAP_ERROR(CMA_ERR_MALLOC);

  psdo->DataBuffer.isAllocated = TRUE;
  memcpy(psdo->DataBuffer.pBuffer, pData, u32Size);


  memset(data, 0, 8);
  data[0] = CS_INIT_DOWNLOAD;

  if(u32Size <= 4)
  { //expedited
    data[0] |= NUMOFBYTES(u32Size) | EXPEDITED;
    for(i=0; ((i<u32Size) && (i<4)); i++)
    {
      data[i+4] = pData[i];
    }
    u8CanMsgLen = (USIGN8) 8;
  }
  else
  {
    data[0] |= SEGMENTED;
    *(USIGN32*) &data[4] = u32Size;

    u8CanMsgLen = 8;
  }

  WRT16(&data[1],u16Index);


  data[3] = u8SubIndex;

  CO_send_data(pCh, pCh->channel, u16CobId, pCh->isXtd, u8CanMsgLen, data);

  
  return MAP_ERROR(ret);
}





void UploadNextSegment(PCHANNELS pCh, PSDO psdo, PARAM_STRUCT *param)
{
  USIGN8 data[8], byte0 = param->RCV_data[0];
  USIGN8 u8DataLen = 7 - ((byte0 >> 1) & 0x07);
  PCO_BUFFER pBuf = &psdo->DataBuffer;
  CO_EVENTSTRUCT_E event;

  if(pBuf->u32Size < (pBuf->u32BufferPtr + u8DataLen))
  { 
    if((byte0 >> 1) & 0x3)
    {// buffer too small
      SDOAbort(pCh, psdo, psdo->u8NodeId, OUT_OF_MEMORY_ERR);
      return;
    }
    else
    { // no size indicated.
      u8DataLen = (USIGN8) (pBuf->u32Size - pBuf->u32BufferPtr);
    }
  }

  if(byte0 & 0x01)
  { // last segment
    memcpy(&pBuf->pBuffer[pBuf->u32BufferPtr], &param->RCV_data[1], u8DataLen);
    pBuf->u32BufferPtr += u8DataLen;
    psdo->isUsed = FALSE;
    event.ucEventType = EV_UPLOAD_COMPLETE;
    event.ulTimeStamp = param->Time;
    event.ucaEventData[0] = psdo->u8NodeId;
    event.ucaEventData[1] = psdo->u8SubIndex;
    *(USIGN16*)&event.ucaEventData[2] = psdo->u16Index;
    event.ucaEventData[4] = 0xFF; //segmented
    event.ulDataLenSegmented = pBuf->u32BufferPtr;
    insertUserEvent((PCO_EVENTSTRUCT)&event,pCh);
  }
  else
  { // further bytes must be uploaded
    memcpy(&pBuf->pBuffer[pBuf->u32BufferPtr], &param->RCV_data[1], u8DataLen);
    pBuf->u32BufferPtr += u8DataLen;

    memset(data,0,8);
    data[0] = psdo->u8OldDataByte0;
    
    CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, data);
    
  }


}

/**
 ******************************************************************
 * @brief  handles SDO upload, called from receive thread 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  param: param struct from Layer2, contains the received message 
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */

void HandleUpload(PCHANNELS pCh, PARAM_STRUCT *param)
{
  USIGN8 u8NodeId = (USIGN8)(param->Ident - SDO_TX);
  PSDO psdo = &pCh->node[u8NodeId].Sdo;
  CO_EVENTSTRUCT event;
  
  // abort SDO transfer ?
  
  if(param->RCV_data[0] == CS_ABORT)
  {
    USIGN8 data[20];
    int i;
    
    data[0] = ERR_SDO;
    data[1] = u8NodeId;
    
    for(i=0; i<4; i++)
      data[2+i] = param->RCV_data[7-i];
    
    insertErrorEvent(pCh,param->Time,6,data,TRUE);
    psdo->isUsed = FALSE;
    psdo->isLocked = FALSE;
    return;
  }
  
  
  if(!psdo->isUsed)
  { // should never happen!
    USIGN8 data[20];
    
    data[0] = ERR_SDO;
    data[1] = u8NodeId;
    memcpy(&data[2], ERR_TOGGLEBIT, 4);
    
    
    insertErrorEvent(pCh,param->Time,6,data,TRUE);
    return;
  }

  psdo->u32LastEvent = MK_TIME(param->Time);
  
  if(psdo->isInit)
  {
    USIGN8 e_s;
    USIGN32 u32Size;
    
    if((param->RCV_data[0] & 0xC0) != CS_INIT_UPLOAD)
    {
      SDOAbort(pCh, psdo, u8NodeId, CLIENT_SERVER_ERR);
      return;
    }
    
    // e and s in spec.
    e_s = param->RCV_data[0] & 3;

    
    
    switch(e_s)
    {
    case 2:
    case 3:
      // expedited
      u32Size = (e_s == 2) ? 4 : 4 - ((param->RCV_data[0] >> 2) & 3);
      psdo->DataBuffer.pBuffer = (USIGN8*)malloc(u32Size);
      if(!psdo->DataBuffer.pBuffer)
      {
       SDOAbort(pCh, psdo, u8NodeId, OUT_OF_MEMORY_ERR);
       return;
      }
      psdo->DataBuffer.isAllocated = TRUE;
      psdo->DataBuffer.u32Size = u32Size;
      psdo->DataBuffer.u32BufferPtr = u32Size;
      memcpy(psdo->DataBuffer.pBuffer, &param->RCV_data[4], u32Size);
      
      event.ucEventType = EV_UPLOAD_COMPLETE;
      event.ulTimeStamp = param->Time;
      event.ucaEventData[0] = u8NodeId;
      event.ucaEventData[1] = psdo->u8SubIndex;

      *(USIGN16*) &event.ucaEventData[2] = psdo->u16Index;
      event.ucaEventData[4] = (USIGN8) u32Size&7;

      
      memcpy(&event.ucaEventData[5], &param->RCV_data[4], u32Size);
      insertUserEvent(&event, pCh);
      
      psdo->isUsed = FALSE;

      psdo->isWaiting = FALSE;
      psdo->isLocked = FALSE;

      break;
      
    case 1:
      // segmented
      // check command response
      {
        USIGN8 byte0 = param->RCV_data[0], data[20];
        
        // toggle bit is always 0 for the first response!
        if(byte0 & TOGGLE_BIT_UPLOAD)
        { // toggle bit error
          SDOAbort(pCh, psdo, u8NodeId, TOGGLE_ERR);
          return;
        }
        
        
        // check toggle bit "t"
        
        Rd32(&u32Size, &param->RCV_data[4]);
        psdo->DataBuffer.pBuffer = (USIGN8*)malloc(u32Size);
        if(!psdo->DataBuffer.pBuffer)
        {
          
          SDOAbort(pCh, psdo, u8NodeId, OUT_OF_MEMORY_ERR);
          return;
        }
        psdo->DataBuffer.isAllocated = TRUE;
        psdo->DataBuffer.u32Size = u32Size;
        psdo->DataBuffer.u32BufferPtr = 0;
        
        
        memset(data,0,8);
        data[0] = psdo->u8OldDataByte0;
        
        CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, data);
      }
      

      
      break;
      
    default:
      {
        SDOAbort(pCh, psdo, u8NodeId, GENERAL_ERR);
        return;
      }
      
      return;
      //error
      break;
      
      
    }
    
    
    
    psdo->isInit = FALSE;
  }
  else
  {
    USIGN8 byte0 = param->RCV_data[0];
    // check "scs"; scs must always be 0
    if(byte0 & 0xE0)
    { //scs error
      SDOAbort(pCh, psdo, u8NodeId, GENERAL_ERR);
      return;
    }
    
    // toggle bit is always 0 for the first response!
    if((byte0 & TOGGLE_BIT_UPLOAD) != (psdo->u8OldDataByte0 & TOGGLE_BIT_UPLOAD))
    { // toggle bit error
      SDOAbort(pCh, psdo, u8NodeId, TOGGLE_ERR);
      return;
    }
    
    
    
    psdo->u8OldDataByte0 ^= TOGGLE_BIT_UPLOAD;
    
    
    UploadNextSegment(pCh, psdo, param);
  }
}

void DownloadNextSegment(PCHANNELS pCh, PSDO psdo, PARAM_STRUCT *param)
{
  USIGN32 u32RestData = psdo->DataBuffer.u32Size - psdo->DataBuffer.u32BufferPtr;
  USIGN8 data[8];
  CO_EVENTSTRUCT event;
  
  if(u32RestData >= 7)
  {
    data[0] = psdo->u8OldDataByte0;
    memcpy(&data[1], &psdo->DataBuffer.pBuffer[psdo->DataBuffer.u32BufferPtr],7);
    psdo->DataBuffer.u32BufferPtr += 7;

    if(u32RestData == 7)
    {
      data[0] |= LAST_SEGMENT;
    }
    
  }
  else
  {
    USIGN8 u8BytesNotUsed = (USIGN8) (7-u32RestData);

    if(u32RestData)
    {
      memcpy(&data[1], &psdo->DataBuffer.pBuffer[psdo->DataBuffer.u32BufferPtr],u32RestData);
      psdo->DataBuffer.u32BufferPtr += u32RestData;
      
      data[0] = psdo->u8OldDataByte0;
      data[0] |= LAST_SEGMENT | (u8BytesNotUsed<<1);

    }
    else
    {
      psdo->isUsed = FALSE;
      event.ucEventType = EV_DOWNLOAD_COMPLETE;
      event.ulTimeStamp = param->Time;
      event.ucaEventData[0] = psdo->u8NodeId;
      event.ucaEventData[1] = psdo->u8SubIndex;
      *(USIGN16*)&event.ucaEventData[2] = psdo->u16Index;
      event.ucaEventData[4] = (psdo->DataBuffer.u32Size>4) ? 0xFF : psdo->DataBuffer.u32Size;
      insertUserEvent(&event,pCh);
      free(psdo->DataBuffer.pBuffer);
      psdo->DataBuffer.isAllocated = FALSE;
      psdo->DataBuffer.pBuffer = NULL;
      psdo->DataBuffer.u32BufferPtr = 0;
      psdo->DataBuffer.u32Size = 0;
      psdo->isUsed = FALSE;
      
      return;
    }
    
  }

  CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, data);

}

/**
 ******************************************************************
 * @brief  handles SDO download, called from receive thread 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  param: param struct from Layer2, contains the received message 
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */

void HandleDownload(PCHANNELS pCh, PARAM_STRUCT *param)
{
  USIGN8 u8NodeId = (USIGN8)(param->Ident - SDO_TX);
  PSDO psdo = &pCh->node[u8NodeId].Sdo;
  
  psdo->u32LastEvent = MK_TIME(param->Time);
  if(!psdo->isUsed)
  {
    USIGN8 data[20];
    
    data[0] = ERR_SDO;
    data[1] = u8NodeId;
    memcpy(&data[2], ERR_TOGGLEBIT, 4);
    
    
    insertErrorEvent(pCh,param->Time,6,data,TRUE);
    return;
  }
  
  // abort SDO transfer ?
  
  if(param->RCV_data[0] == CS_ABORT)
  {
    USIGN8 data[20];
    int i;
    
    data[0] = ERR_SDO;
    data[1] = u8NodeId;
    
    for(i=0; i<4; i++)
      data[2+i] = param->RCV_data[7-i];
    
    insertErrorEvent(pCh,param->Time,6,data,TRUE);
    psdo->isUsed = FALSE;
    psdo->isLocked = FALSE;
    return;
  }
  
  if(psdo->isInit)
  {
    USIGN16 u16Index;
    USIGN8  u8SubIndex = param->RCV_data[3];
    
    Rd16(&u16Index, &param->RCV_data[1]);
   
    if((param->RCV_data[0] & SCS_INIT_DOWNLOAD) != SCS_INIT_DOWNLOAD)
    {
      
      SDOAbort(pCh, psdo, u8NodeId, CLIENT_SERVER_ERR);
      psdo->isUsed = FALSE;
      psdo->isLocked = FALSE;
      return;
    }
    if(u16Index != psdo->u16Index)
    {

      SDOAbort(pCh, psdo, u8NodeId, GENERAL_ERR);
      psdo->isUsed = FALSE;
      psdo->isLocked = FALSE;
      return;
    }
    if(u8SubIndex != psdo->u8SubIndex)
    {
      SDOAbort(pCh, psdo, u8NodeId, GENERAL_ERR);
      psdo->isUsed = FALSE;
      psdo->isLocked = FALSE;
      
      return;
    }
    
    psdo->isInit = FALSE;
    
    
    DownloadNextSegment(pCh, psdo, param);    
    
  }
  else
  {
    USIGN8 byte0 = param->RCV_data[0];
    
    
    if((byte0 & 0xE0) != 0x20)
    {
      SDOAbort(pCh, psdo, u8NodeId, CLIENT_SERVER_ERR);
      return;
    }
    
    if((byte0 & TOGGLE_BIT) != (psdo->u8OldDataByte0 & TOGGLE_BIT))
    { // toggle bit error
      SDOAbort(pCh, psdo, u8NodeId, TOGGLE_ERR);
      return;
    }
    
    psdo->u8OldDataByte0 ^= TOGGLE_BIT;
    DownloadNextSegment(pCh, psdo, param);
  }
  
  
}



/**
 ******************************************************************
 * @brief  handles SDO block upload, called from receive thread 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  param: param struct from Layer2, contains the received message 
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */

void HandleBlockUpload(PCHANNELS pCh, PARAM_STRUCT *param)
{
  USIGN8 u8NodeId = (USIGN8)(param->Ident - SDO_TX);
  PSDO psdo = &pCh->node[u8NodeId].Sdo;
  USIGN8 byte0 = param->RCV_data[0];
      
  USIGN8 data[20];



  memset(data,0,8);
  psdo->u32LastEvent = MK_TIME(param->Time);

  // abort SDO transfer ?
  
  if(byte0 == CS_ABORT)
  {
    int i;
    
    data[0] = ERR_SDO;
    data[1] = u8NodeId;
    
    for(i=0; i<4; i++)
      data[2+i] = param->RCV_data[7-i];
    
    insertErrorEvent(pCh,param->Time,6,data,TRUE);
    psdo->isUsed = FALSE;
    psdo->isLocked = FALSE;
    return;
  }
  
  
  if(!psdo->isUsed)
  {
    
    data[0] = ERR_SDO;
    data[1] = u8NodeId;
    memcpy(&data[2], ERR_TOGGLEBIT, 4);
    
    
    insertErrorEvent(pCh,param->Time,6,data,TRUE);
    return;
  }  

  switch(psdo->SdoBlockState)
  {

  case BLK_TRANSFER_INIT:
    {
    USIGN32 u32Size;
    
    if((byte0 & 0xE0) != SCS_INIT_BLOCKUPLOAD)
    {
      SDOAbort(pCh, psdo, u8NodeId, CLIENT_SERVER_ERR);
      return;
    }

    psdo->isCrcProvided = byte0 & SUPPORT_CRC_CODE;

    if(byte0 & 0x02)
    {
      Rd32(&u32Size, &param->RCV_data[4]);
      psdo->DataBuffer.pBuffer = (USIGN8*)malloc(u32Size+8);
      if(!psdo->DataBuffer.pBuffer)
      {
        SDOAbort(pCh, psdo, u8NodeId, OUT_OF_MEMORY_ERR);
        return;
      }
      psdo->DataBuffer.isAllocated = TRUE;
      psdo->DataBuffer.u32Size = u32Size+8;
      psdo->DataBuffer.u32BufferPtr = 0;
      
      
    }
    else
    {
      psdo->DataBuffer.pBuffer = (USIGN8*)malloc(MAX_MALLOCSIZE);
      psdo->DataBuffer.u32Size = MAX_MALLOCSIZE;
      psdo->DataBuffer.u32BufferPtr = 0;
      if(!psdo->DataBuffer.pBuffer)
      {
        SDOAbort(pCh, psdo, u8NodeId, OUT_OF_MEMORY_ERR);
        return;
      }
      psdo->DataBuffer.isAllocated = TRUE;
    }
    psdo->u8BlockSeqNum = 0;

    memset(data, 0, 8);
    data[0] = CS_INIT_BLOCKUPLOAD | 0x03;
    
    CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, data);

    psdo->SdoBlockState = BLK_TRANSFER_DATA;
    }
    break;

  case BLK_TRANSFER_DATA:
    {

      psdo->u8BlockSeqNum++;

      if((byte0 & 0x7F) != psdo->u8BlockSeqNum)
      { // error

        SDOAbort(pCh, psdo, u8NodeId, SEQ_NUMBER_ERR);
        return;
      }


      if(byte0 & 0x80)
      {
        psdo->SdoBlockState = BLK_TRANSFER_END;
        
        data[0] = CS_INIT_BLOCKUPLOAD | 2;
        data[1] = psdo->u8BlockSeqNum;
        data[2] = psdo->u8BlockSize;


        CO_send_data(pCh, pCh->channel,psdo->u16CobId,pCh->isXtd, 8, data);

      }

      if(psdo->DataBuffer.u32BufferPtr < psdo->DataBuffer.u32Size)
      {
        memcpy(&psdo->DataBuffer.pBuffer[psdo->DataBuffer.u32BufferPtr], &param->RCV_data[1], 7);
      }
      else
      {
        if(MAX_MALLOCSIZE == psdo->DataBuffer.u32Size)
        {
          SDOAbort(pCh, psdo, u8NodeId, OUT_OF_MEMORY_ERR);
          return;
        }
        else
        {
          SDOAbort(pCh, psdo, u8NodeId, BLOCK_SIZE_ERR);
          return;
        }
        break;
      }

      psdo->DataBuffer.u32BufferPtr += 7;
      if((psdo->u8BlockSeqNum == psdo->u8BlockSize) && !(byte0 & 0x80))
      {
        psdo->u8BlockSeqNum = 0;
        data[0] = CS_INIT_BLOCKUPLOAD | 2;
        data[1] = psdo->u8BlockSize;
        data[2] = psdo->u8BlockSize;


        CO_send_data(pCh, pCh->channel,psdo->u16CobId,pCh->isXtd, 8, data);

      }

    }
    break;

  case BLK_TRANSFER_END:
    {
      USIGN16 u16Crc;
      USIGN8 data[8];
      CO_EVENTSTRUCT_E event;
      
      if((byte0 & 0xE0) != SCS_END_BLOCKUPLOAD)
      {
        SDOAbort(pCh, psdo, u8NodeId, CLIENT_SERVER_ERR);
        return;
      }
      
      psdo->DataBuffer.u32BufferPtr -= ((byte0 >> 2) & 7);
      

      Rd16(&psdo->u16CrcCode, &param->RCV_data[1]);
      
      u16Crc = 0;
      
        u16Crc = calcCRC16r(psdo->DataBuffer.pBuffer, psdo->DataBuffer.u32BufferPtr);
      
      if((u16Crc != psdo->u16CrcCode) && psdo->isCrcProvided)
      { // CRC Error
        SDOAbort(pCh, psdo, u8NodeId, CRC_CODE_ERR);

      }
      else
      {
        memset(data, 0, 8);
        data[0] = CCS_END_BLOCKUPLOAD;
        CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, data);
        
        psdo->isUsed = FALSE;
        psdo->isWaiting = FALSE;
        
        event.ucEventType = EV_UPLOAD_COMPLETE;
        event.ulTimeStamp = param->Time;
        event.ucaEventData[0] = psdo->u8NodeId;
        event.ucaEventData[1] = psdo->u8SubIndex;
        *(USIGN16*)&event.ucaEventData[2] = psdo->u16Index;
        event.ucaEventData[4] = 0xFF; //segmented
        event.ulDataLenSegmented = psdo->DataBuffer.u32BufferPtr;
        insertUserEvent((PCO_EVENTSTRUCT)&event,pCh);
      }
      
    }
    
    break;
    
    default:
      
      break;
    
    
  }

}

/**
 ******************************************************************
 * @brief  handles SDO block download, called from receive thread 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  param: param struct from Layer2, contains the received message 
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */

void HandleBlockDownload(PCHANNELS pCh, PARAM_STRUCT *param)
{
  USIGN8 u8NodeId = (USIGN8)(param->Ident - SDO_TX);
  PSDO psdo = &pCh->node[u8NodeId].Sdo;
  USIGN8 byte0 = param->RCV_data[0];
    
  USIGN32 u32SegSize,i;
  USIGN8 data[20];

  psdo->u32LastEvent = MK_TIME(param->Time);
  // abort SDO transfer ?
  
  if(byte0 == CS_ABORT)
  {
    int i;
    
    data[0] = ERR_SDO;
    data[1] = u8NodeId;
    
    for(i=0; i<4; i++)
      data[2+i] = param->RCV_data[7-i];
    
    insertErrorEvent(pCh,param->Time,6,data,TRUE);
    psdo->isUsed = FALSE;
    psdo->isLocked = FALSE;
    return;
  }
  
  
  if(!psdo->isUsed)
  {
    
    data[0] = ERR_SDO;
    data[1] = u8NodeId;
    memcpy(&data[2], ERR_TOGGLEBIT, 4);
    
    
    insertErrorEvent(pCh,param->Time,6,data,TRUE);
    return;
  }  

  switch(psdo->SdoBlockState)
  {

  case BLK_TRANSFER_INIT:
    {
    USIGN16 u16Index;
    USIGN32 i;

    
    if((byte0 & 0xE0) != SCS_INIT_BLOCKDOWNLOAD)
    {
      SDOAbort(pCh, psdo, u8NodeId, CLIENT_SERVER_ERR);
      return;
    }

    psdo->isCrcProvided = byte0 & SUPPORT_CRC_CODE;

    if(psdo->isCrcProvided)
    {
      USIGN16 u8Crc = 0;
      
     // for(i=0; i<psdo->DataBuffer.u32Size; i++)
     // {
        
      u8Crc = calcCRC16r(psdo->DataBuffer.pBuffer, psdo->DataBuffer.u32Size);
    //  }
      psdo->u16CrcCode = u8Crc;
      
    }

    if(byte0 & 3)
    { // error
      SDOAbort(pCh, psdo, u8NodeId, GENERAL_ERR);
      return;
    }
 
    Rd16(&u16Index, &param->RCV_data[1]);

    if(u16Index != psdo->u16Index)
    {
      SDOAbort(pCh, psdo, u8NodeId, GENERAL_ERR);
      return;
    }

    if(psdo->u8SubIndex != param->RCV_data[3])
    {
      SDOAbort(pCh, psdo, u8NodeId, GENERAL_ERR);
      return;
    }

    psdo->u8BlockSeqNum = 1;
    psdo->u8BlockSize = param->RCV_data[4];
    psdo->u32BlockStart = psdo->DataBuffer.u32BufferPtr;

    memset(data, 0, 8);

    data[0] = psdo->u8BlockSeqNum;

    if ((psdo->DataBuffer.u32Size - psdo->DataBuffer.u32BufferPtr) > 7)
    {
      u32SegSize = 7; 
    }
    else
    {
      data[0] |= 0x80;
      u32SegSize = psdo->DataBuffer.u32Size;
    }

    for(i=0; i<u32SegSize; i++)
    {
      data[i+1] = psdo->DataBuffer.pBuffer[psdo->DataBuffer.u32BufferPtr++];
      
    }

    if(data[0] & 0x80)
    {
      psdo->SdoBlockState = BLK_TRANSFER_BLK_CONFIRM;
      CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, data);
      break;
    }
    else
    {
      psdo->SdoBlockState = BLK_TRANSFER_DATA;
      CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, data);
    }
    

    
    }
  //  break;

  case BLK_TRANSFER_BLK_CONFIRM:

    if(psdo->SdoBlockState == BLK_TRANSFER_BLK_CONFIRM)
    {
      if(byte0 != 0xA2)
      {
        SDOAbort(pCh, psdo, u8NodeId, CLIENT_SERVER_ERR); 
        return;
      }
      psdo->SdoBlockState = BLK_TRANSFER_DATA;
      if(psdo->u8BlockSeqNum != param->RCV_data[1])
      {  // not all messages received, retry sending
         psdo->DataBuffer.u32BufferPtr = psdo->DataBuffer.u32BufferPtr - (psdo->u8BlockSize - param->RCV_data[1]) * 7;
      }
      psdo->u8BlockSeqNum = 0;
      if(psdo->DataBuffer.u32BufferPtr >= psdo->DataBuffer.u32Size)
      { // last block ?
        USIGN8 u8ValidBytes = (USIGN8) 7 - (USIGN8)(psdo->DataBuffer.u32Size % 7);

        psdo->SdoBlockState = BLK_TRANSFER_END;
        memset(data,0,8);
        data[0] = 0xC1 | (u8ValidBytes << 2);
        WRT16(&data[1], psdo->u16CrcCode);
        CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, data);
        
        psdo->u8BlockSize = param->RCV_data[2];
      }
    }

  //  break;


  case BLK_TRANSFER_DATA:

    while((psdo->SdoBlockState == BLK_TRANSFER_DATA) && psdo->DataBuffer.pBuffer)
    {
      psdo->u8BlockSeqNum++;
      

      memset(data, 0, 8);
      data[0] = psdo->u8BlockSeqNum;

      if ((psdo->DataBuffer.u32Size - psdo->DataBuffer.u32BufferPtr) > 7)
      {
        u32SegSize = 7; 
      }
      else
      { // last segment of all data blocks
        data[0] |= 0x80;
        u32SegSize = psdo->DataBuffer.u32Size - psdo->DataBuffer.u32BufferPtr;
        psdo->SdoBlockState = BLK_TRANSFER_BLK_CONFIRM;
      }

      if(psdo->u8BlockSeqNum >= psdo->u8BlockSize)
      { // last segment of the block
        psdo->SdoBlockState = BLK_TRANSFER_BLK_CONFIRM;
        if(psdo->u8BlockSeqNum > psdo->u8BlockSize)
        {
          break;
        }
      }

      
      for(i=0; i<u32SegSize; i++)
      {
        data[i+1] = psdo->DataBuffer.pBuffer[psdo->DataBuffer.u32BufferPtr++];
      }
      
      
      CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, data);
    }
    
    break;


  case BLK_TRANSFER_END:
    {
      CO_EVENTSTRUCT event;

      psdo->isUsed = FALSE;
      psdo->isWaiting = FALSE;

      event.ucEventType = EV_DOWNLOAD_COMPLETE;
      event.ulTimeStamp = param->Time;
      event.ucaEventData[0] = psdo->u8NodeId;
      event.ucaEventData[1] = psdo->u8SubIndex;
      *(USIGN16*)&event.ucaEventData[2] = psdo->u16Index;
      insertUserEvent(&event,pCh);
      free(psdo->DataBuffer.pBuffer);
      psdo->DataBuffer.isAllocated = FALSE;
      psdo->DataBuffer.pBuffer = NULL;
      psdo->DataBuffer.u32BufferPtr = 0;
      psdo->DataBuffer.u32Size = 0;
    }


    
    break;


  }
}

/**
 ******************************************************************
 * @brief  checks the kind of SDO transfer (upload,download,block) and calls the handler 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  param: Layer2 param structure
 * @return nothing
 *
 ******************************************************************
 */
void HandleTxSdo(PCHANNELS pCh, PARAM_STRUCT *param)
{
  USIGN8 u8NodeId = (USIGN8)(param->Ident - SDO_TX);

  switch(pCh->node[u8NodeId].Sdo.Direction)
  {
  case UPLOAD:
    HandleUpload(pCh,param);
    break;

  case DOWNLOAD:
    HandleDownload(pCh,param);
    break;

  case BLOCKUPLOAD:
    HandleBlockUpload(pCh,param);
    break;

  case BLOCKDOWNLOAD:
    HandleBlockDownload(pCh,param);
    break;

  }
}

//#ifdef RXSDOTRANSFER

void HandleSInitUpload(PCHANNELS pCh, PARAM_STRUCT *param)
{
  USIGN16 u16Index;
  PSDO psdo;
  USIGN8  u8SubIndex = param->RCV_data[3];
  USIGN8 data[8];
  
  memset(data,0,8);
  Rd16(&u16Index, &param->RCV_data[1]);
  
  if(HandleMasterOdRd(pCh, u16Index, u8SubIndex))
  {
    return;
  }
  
  psdo = &pCh->node[pCh->MasterNodeId].Sdo;
  
  // multiplexor
  data[1] = param->RCV_data[1];
  data[2] = param->RCV_data[2];
  data[3] = param->RCV_data[3];
  
  psdo->u16CobId = (USIGN16)pCh->MasterNodeId + 0x580;
  if(psdo->DataBuffer.u32Size <= 4)
  { 
    // expedited transfer
    USIGN8 n=(4-(USIGN8)psdo->DataBuffer.u32Size)<<2;
    
    //command
    data[0] = 0x43 | n;
    
    
    // data bytes
    memcpy(&data[4],psdo->DataBuffer.pBuffer,psdo->DataBuffer.u32Size);
    
    psdo->isUsed = FALSE;
    psdo->isWaiting = FALSE;
    psdo->isLocked = FALSE;
    
  }
  else
  { // normal transfer

    data[0] = 0x41;
    WRT32(&data[4], psdo->DataBuffer.u32Size);
    psdo->isInit = FALSE;
    psdo->Direction = UPLOAD;
    psdo->u8OldDataByte0 = 0x10;

    psdo->u32LastEvent = MK_TIME(param->Time);

    psdo->isUsed = TRUE;
    psdo->isWaiting = TRUE;
    
  }

  CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, data);
  
}

void HandleSUpload(PCHANNELS pCh, PARAM_STRUCT *param)
{
  USIGN8 byte0 = param->RCV_data[0];
  USIGN8 data[8];
  PSDO psdo = &pCh->node[pCh->MasterNodeId].Sdo;
  
  memset(data,0,8);
  psdo->u32LastEvent = MK_TIME(param->Time);

  if(!psdo->isWaiting)
  {
    psdo->DataBuffer.pBuffer = NULL;
    psdo->DataBuffer.u32Size = 0;
    psdo->DataBuffer.u32BufferPtr = 0;
    SDOAbort(pCh, psdo, (USIGN8)pCh->MasterNodeId, GENERAL_ERR);
  }

  psdo->u8OldDataByte0 ^= 0x10; // toggle bit

  if((byte0 & 0x10) != psdo->u8OldDataByte0)
  {
    psdo->DataBuffer.pBuffer = NULL;
    psdo->DataBuffer.u32Size = 0;
    psdo->DataBuffer.u32BufferPtr = 0;
    SDOAbort(pCh, psdo, (USIGN8)pCh->MasterNodeId, TOGGLE_ERR);
    return;
  }

  

  if((psdo->DataBuffer.u32Size - psdo->DataBuffer.u32BufferPtr) > 7)
  {
    int i;
    data[0] = psdo->u8OldDataByte0;
    for(i=1; i<=7; i++)
    {
      data[i] = psdo->DataBuffer.pBuffer[psdo->DataBuffer.u32BufferPtr++];
    }

  }
  else
  {
    int i;
    USIGN8 u8Rest = (USIGN8) (psdo->DataBuffer.u32Size - psdo->DataBuffer.u32BufferPtr);
    data[0] = psdo->u8OldDataByte0 | 0x01; // last segment

    for(i=1; i<=u8Rest; i++)
    {
      data[i] = psdo->DataBuffer.pBuffer[psdo->DataBuffer.u32BufferPtr++];
    }

    // n not used! only with expedited transfer!
   // data[0] |= ((7-u8Rest)<<1);

    psdo->isUsed = FALSE;
    psdo->isWaiting = FALSE;
    psdo->isLocked = FALSE;


  }

  CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, data);
}

void HandleSInitDownload(PCHANNELS pCh, PARAM_STRUCT *param)
{
  USIGN16 u16Index;
  USIGN8  u8SubIndex = param->RCV_data[3];
  USIGN8  data[8],byte0 = param->RCV_data[0];
  PSDO psdo = &pCh->node[pCh->MasterNodeId].Sdo;

  memset(data,0,8);
  
  Rd16(&u16Index, &param->RCV_data[1]);

  if(HandleMasterOdWrt(pCh, u16Index, u8SubIndex))
  {
    return;
  }
  
  // multiplexor
  data[1] = param->RCV_data[1];
  data[2] = param->RCV_data[2];
  data[3] = param->RCV_data[3];

  psdo->u16CobId = (USIGN16)pCh->MasterNodeId + 0x580;

  if(byte0 & 0x02)
  { // expedited transfer
    USIGN8 u8Length = (USIGN8) ((byte0 & 0x01) ? 4-((byte0>>2)&3) : (psdo->DataBuffer.u32Size<4) ? psdo->DataBuffer.u32Size : 4);
    int i;

    for(i=0; i<u8Length; i++)
    {
      psdo->DataBuffer.pBuffer[i] = param->RCV_data[4+i];
    }
    
    data[0] = 0x60;
  }
  else
  { // normal transfer
    psdo->u32LastEvent = MK_TIME(param->Time);
    psdo->isUsed = TRUE;
    psdo->isWaiting = TRUE;
    psdo->isInit = FALSE;
    psdo->u8OldDataByte0 = 0x10;

    if(byte0 & 0x01)
    {
      USIGN32 u32size;

      Rd32(&u32size, &param->RCV_data[4]);
      if(u32size > psdo->DataBuffer.u32Size)
      { // error
        psdo->DataBuffer.pBuffer = NULL;
        psdo->DataBuffer.u32Size = 0;
        SDOAbort(pCh,psdo,(USIGN8)pCh->MasterNodeId,LENGTH_ERR);
        return;
      }
    }

    data[0] = 0x60;
  }
  
  CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, data);
}

void HandleSDownload(PCHANNELS pCh, PARAM_STRUCT *param)
{
  USIGN8 byte0 = param->RCV_data[0];
  USIGN8 data[8];
  PSDO psdo = &pCh->node[pCh->MasterNodeId].Sdo;
  
  memset(data,0,8);
  psdo->u32LastEvent = MK_TIME(param->Time);

  if(!psdo->isWaiting)
  {
    SDOAbort(pCh, psdo, (USIGN8)pCh->MasterNodeId, GENERAL_ERR);
    return;
  }

  psdo->u8OldDataByte0 ^= 0x10; // toggle bit

  if((byte0 & 0x10) != psdo->u8OldDataByte0)
  {
    SDOAbort(pCh, psdo, (USIGN8)pCh->MasterNodeId, TOGGLE_ERR);
    return;
  }

  if(byte0 & 0x01)
  { // last segment
    USIGN8 u8Rest = (byte0 >> 1) & 7;
    int i;

    if(!u8Rest)
    {
      u8Rest = (USIGN8)(((psdo->DataBuffer.u32Size - psdo->DataBuffer.u32BufferPtr) >= 7) ? 0 : 7-(psdo->DataBuffer.u32Size - psdo->DataBuffer.u32BufferPtr));
    }

    u8Rest = 7-u8Rest;

    for(i=0; i<u8Rest; i++)
    {
      psdo->DataBuffer.pBuffer[psdo->DataBuffer.u32BufferPtr++] = param->RCV_data[1+i];
    }

    psdo->isUsed = FALSE;
    psdo->isWaiting = FALSE;
    psdo->isLocked = FALSE;
    data[0] = 0x20 | psdo->u8OldDataByte0;
  }
  else
  {
    int i;
    if((psdo->DataBuffer.u32Size - psdo->DataBuffer.u32BufferPtr) < 7)
    {
      SDOAbort(pCh, psdo, (USIGN8)pCh->MasterNodeId, LENGTH_ERR);
      return;
    }

    for(i=0; i<7; i++)
    {
      psdo->DataBuffer.pBuffer[psdo->DataBuffer.u32BufferPtr++] = param->RCV_data[1+i];
    }

    data[0] = 0x20 | psdo->u8OldDataByte0;
  }

  CO_send_data(pCh, pCh->channel, psdo->u16CobId, pCh->isXtd, 8, data);

}
//#endif // RXSDOTRANSFER

/**
 ******************************************************************
 * @brief  checks the kind of SDO transfer (upload,download,block) and calls the handler 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  param: Layer2 param structure
 * @return nothing
 *
 ******************************************************************
 */

void HandleRxSdo(PCHANNELS pCh, PARAM_STRUCT *param)
{
//#ifdef RXSDOTRANSFER
  if(param->Ident != (0x600+pCh->MasterNodeId))
  { // sdo for an other slave
    return;
  }
  switch(param->RCV_data[0] & 0xE0)
  {
  case 0x40: // init upload
    HandleSInitUpload(pCh, param);
    break;

  case 0x60: // upload segmented
    HandleSUpload(pCh, param);
    break;

  case 0x80: // abort
    {
      PSDO psdo = &pCh->node[pCh->MasterNodeId].Sdo;
      USIGN8 ev_data[20];
      
      ev_data[0] = ERR_SDO;
      ev_data[1] = (USIGN8)pCh->MasterNodeId;
      ev_data[2] = param->RCV_data[7];
      ev_data[3] = param->RCV_data[6];
      ev_data[4] = param->RCV_data[5];
      ev_data[5] = param->RCV_data[4];

      insertErrorEvent(pCh, param->Time, 6, ev_data, TRUE);
      
      psdo->isUsed = FALSE;
      psdo->isWaiting = FALSE;
      psdo->isLocked = FALSE;
    }
    break;

  case 0x20: // init download
     HandleSInitDownload(pCh, param);
    break;

  case 0x00: // download segmented
     HandleSDownload(pCh, param);
    break;


  }
//#endif //RXSDOTRANSFER
}

/**
 ******************************************************************
 * @brief  handles a LSS response 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  param: Layer2 param structure
 * @return nothing
 *
 ******************************************************************
 */

void HandleLSS(PCHANNELS pCh, PARAM_STRUCT *param)
{
  CO_EVENTSTRUCT event;

  switch(param->RCV_data[0])
  {

  case LSS_RES_SWITCH_MODE:
    pCh->Lss.u8LSSPendingService = LSS_NO_SERVICE;
    SET_EVENT(pCh->Lss.hLSSEvent);

    break;

  case LSS_CONFIGURE_NODE_ID:

    if(param->RCV_data[1])
    {
      event.ucEventType = EV_ERROR;
      event.ucaEventData[0] = ERR_LSS;
      event.ucaEventData[1] = param->RCV_data[1];
      event.ucaEventData[2] = param->RCV_data[2];
      insertUserEvent(&event, pCh);

    }

    pCh->Lss.u8LSSPendingService = LSS_NO_SERVICE;
    SET_EVENT(pCh->Lss.hLSSEvent);
    break;

  case LSS_CONFIGURE_BIT_TIMING:
    if(param->RCV_data[1])
    {
      event.ucEventType = EV_ERROR;
      event.ucaEventData[0] = ERR_LSS;
      event.ucaEventData[1] = param->RCV_data[1];
      event.ucaEventData[2] = param->RCV_data[2];
      insertUserEvent(&event, pCh);

    }

    pCh->Lss.u8LSSPendingService = LSS_NO_SERVICE;
    SET_EVENT(pCh->Lss.hLSSEvent);

    break;
  case LSS_STORE_CONFIGURATION:

    if(param->RCV_data[1])
    {
      event.ucEventType = EV_ERROR;
      event.ucaEventData[0] = ERR_LSS;
      event.ucaEventData[1] = param->RCV_data[1];
      event.ucaEventData[2] = param->RCV_data[2];
      insertUserEvent(&event, pCh);

    }
    pCh->Lss.u8LSSPendingService = LSS_NO_SERVICE;
    SET_EVENT(pCh->Lss.hLSSEvent);

    break;
  case LSS_INQ_ID_VID:
    Rd32(&pCh->Lss.u32Vid, &param->RCV_data[1]);
    pCh->Lss.u8LSSPendingService = LSS_NO_SERVICE;

    break;
  case LSS_INQ_ID_PC:
    Rd32(&pCh->Lss.u32Pc, &param->RCV_data[1]);
    pCh->Lss.u8LSSPendingService = LSS_NO_SERVICE;

    break;
  case LSS_INQ_ID_RN:
    Rd32(&pCh->Lss.u32Rn, &param->RCV_data[1]);
    pCh->Lss.u8LSSPendingService = LSS_NO_SERVICE;

    break;
  case LSS_INQ_ID_SN:
    Rd32(&pCh->Lss.u32Sn, &param->RCV_data[1]);
    pCh->Lss.u8LSSPendingService = LSS_NO_SERVICE;

    break;
  case LSS_INQ_ID_NODE_ID:
    pCh->Lss.u8NodeId = param->RCV_data[1];
    pCh->Lss.u8LSSPendingService = LSS_NO_SERVICE;

    break;
  case LSS_RES_IDENTITY:
    pCh->Lss.u8LSSPendingService = LSS_NO_SERVICE;
    SET_EVENT(pCh->Lss.hLSSEvent);

    break;

  case LSS_RES_IDENTITY_NONCFGSLVS:
    pCh->Lss.u8LSSPendingService = LSS_NO_SERVICE;
    SET_EVENT(pCh->Lss.hLSSEvent);

    break;


  }
}


/**
 ******************************************************************
 * @brief  called when a SYNC is received: checks which PDOs must be sent 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @return nothing
 *
 ******************************************************************
 */

void CheckCyclicPDOs(PCHANNELS pCh)
{
  PPDOLIST pToDoList = pCh->pPdoList;

  // PDOs only in state opreational possible!
  if(pCh->state != OPERATIONAL)
    return;
  
  while(pToDoList)
  {
    PPDO ppdo = pToDoList->pPdo; 
    
    if(ppdo->u8Type != RX_PDO)
    {
      if((ppdo->u8XmtType > 0) && (ppdo->u8XmtType <= 240))
      {
        if(--ppdo->u8NumOfSyncs == 0)
        {
          ppdo->u8NumOfSyncs = ppdo->u8XmtType;
          ppdo->bSndReq = TRUE;
          pCh->bSendPdo = TRUE;
        }       
      }

      if(ppdo->u8XmtType == 252)
      {
        memcpy(ppdo->data, ppdo->data252, 8);
      }

      if(ppdo->bSendAtNextSync)
      {
        ppdo->bSndReq = TRUE;
        ppdo->bSendAtNextSync = FALSE;
        pCh->bSendPdo = TRUE;
      }
    }
    
    pToDoList = (PDOLIST*)pToDoList->pNext;
    
  }

}

/**
 ******************************************************************
 * @brief  sends all PDOs, which (bSndReq)-flag is set. 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @return nothing
 *
 ******************************************************************
 */

void CheckPDOs(PCHANNELS pCh)
{
  PPDOLIST pToDoList = pCh->pPdoList;

  // PDOs only in state opreational possible!
  if(pCh->state != OPERATIONAL)
    return ;
  
  while(pToDoList)
  {
    PPDO ppdo = pToDoList->pPdo; 

    if(ppdo->bSndReq)
    {
      CO_send_data(pCh, 
        pCh->channel, 
        ppdo->u16CobId, 
        pCh->isXtd,
        ppdo->u8DataLen,
        ppdo->data);


      ppdo->bSndReq = FALSE;
    }

    
    pToDoList = (PDOLIST*)pToDoList->pNext;
    
  }
  
  return ;

}

/**
 ******************************************************************
 * @brief  called, when a L2 remote frame is received. 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  param: Layer2 param structure
 * @return nothing
 *
 ******************************************************************
 */

void DispatchRemote(PCHANNELS pCh, PARAM_STRUCT *param)
{
  USIGN32 Ident = pCh->u16CobExt2Int[param->Ident & 0x7FF];

      switch(Ident & 0x780)
      {

      case PDO1_RX:
      case PDO1_TX:
      case PDO2_RX:
      case PDO2_TX:
      case PDO3_RX:
      case PDO3_TX:
      case PDO4_RX:
      case PDO4_TX:
        {
          PPDO pPdo = getPdo(pCh, Ident);
          
          if(!pPdo)
            break;

          if(pPdo->u16CobId != param->Ident)
            break;
          
          if((pPdo->u8Type == TX_PDO) && (pCh->state == OPERATIONAL))
          {
            if(pPdo->isUsed)
            {
              if(pPdo->u8XmtType >= 252)
              {
                pPdo->bSndReq = TRUE;
                CheckPDOs(pCh);
              }
            }
          }
        }
        break;


      default:
        if(Ident == LSS_RESPONSE)
          HandleLSS(pCh, param);

        break;

      }
}

void HandleRxPdo(PCHANNELS pCh, USIGN32 u32Ident, USIGN8 *pData, USIGN32 u32Time, USIGN8 u8Length)
{
  CO_EVENTSTRUCT event;
  USIGN32 u32IntIdent = pCh->u16CobExt2Int[u32Ident & 0x7FF];
  PPDO pPdo = getPdo(pCh, u32IntIdent);
  
  if(!pPdo)
    return;

  if(pPdo->u16CobId != u32Ident)
    return;
  
  if((pPdo->u8Type == RX_PDO) && (pCh->state == OPERATIONAL))
  {
    memcpy(pPdo->data, pData, 8);
    if((pPdo->isUsed) && (pPdo->bEvent))
    {
      event.ucEventType = EV_PDO;
      event.ulTimeStamp = u32Time;
      *(USIGN16*)&event.ucaEventData[0] = (USIGN16)u32IntIdent;
      event.ucaEventData[2] = u8Length;
      memcpy(&event.ucaEventData[3], pData, u8Length);
      insertUserEvent(&event, pCh);
    }
  }
  
}

/**
 ******************************************************************
 * @brief  called, when a L2 data frame is received. 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  param: Layer2 param structure
 * @return nothing
 *
 ******************************************************************
 */

void DispatchEvent(PCHANNELS pCh, PARAM_STRUCT *param)
{
  USIGN32 u32NodeId;
            
  CO_EVENTSTRUCT event;

  USIGN32 Ident = pCh->u16CobExt2Int[param->Ident & 0x7FF];


      switch(Ident & 0x780)
      {
      case NMT:

        break;
      case SYNC_EMRC:
        {
          int i;

          if(Ident == 0x80)
          {
            if(pCh->isSyncConsumer)
            {
              CheckCyclicPDOs(pCh);
            }

            break;
            
          }
          else
          {
            event.ulTimeStamp = param->Time;
            event.ucEventType = EV_EMCY;
            event.ucaEventData[0] = (USIGN8) Ident - 0x80; //0x6e;
            
            for(i=0; i<8; i++)
            {
              event.ucaEventData[i+1] = param->RCV_data[i];
            }
          }

         insertUserEvent(&event, pCh);
        }
        break;
      case TIME_STAMP:

        break;
      case PDO1_RX:
      case PDO1_TX:
      case PDO2_RX:
      case PDO2_TX:
      case PDO3_RX:
      case PDO3_TX:
      case PDO4_TX:
      case PDO4_RX:
          
          HandleRxPdo(pCh, param->Ident, param->RCV_data, param->Time, (USIGN8)param->DataLength);
        break;
      case SDO_TX:
          param->Ident = Ident;
          HandleTxSdo(pCh, param);
        break;
      case SDO_RX:
          HandleRxSdo(pCh, param);
        break;
      case ERROR_HANDLING:

        u32NodeId = Ident - 0x700;

        if(u32NodeId <= MAX_NODE)
        {
          PCO_NODE pNode = &pCh->node[u32NodeId];

          if(pNode->isDefined)
          {
            if(pNode->u32HeartbeatInterval)
            { // Heartbeat defined
              pNode->u32LastHeartbeat = MK_TIME(param->Time);

////////////////////////////////////////////////////////////
                //  pNode->u32GuardResTime = MK_TIME(param->Time);
                  if((param->RCV_data[0] & 0x7F) != pNode->u8GuardingState)
                  {
                    CO_EVENTSTRUCT event;
                    USIGN32 u32Time;
                    CO_get_time(pCh->channel, &u32Time);
                    
                    event.ucEventType = ERROR_EVENT;
                    event.ulTimeStamp = u32Time;
                    event.ucaEventData[0] = ERR_GUARD;
                    event.ucaEventData[1] = (USIGN8)u32NodeId;
                    event.ucaEventData[2] = 1;
                    event.ucaEventData[3] = param->RCV_data[0] & 0x7F;
                    insertUserEvent(&event,pCh);
                    pNode->u8GuardingState = param->RCV_data[0] & 0x7F;

                   }

///////////////////////////////////////////////////////////
            }
            else
            { 
              if(pNode->u16GuardTime)
              {
                // Node Guarding
                USIGN8 toggleCheck = pNode->u8ToggleBit ^ (param->RCV_data[0] & 0x80);
                if(toggleCheck)
                { // toggle bit check ok!
                  pNode->u8ToggleBit = param->RCV_data[0] & 0x80;
                  pNode->u32GuardResTime = MK_TIME(param->Time);
                  if((param->RCV_data[0] & 0x7F) != pNode->u8GuardingState)
                  {
                    CO_EVENTSTRUCT event;
                    USIGN32 u32Time;
                    CO_get_time(pCh->channel, &u32Time);
                    
                    event.ucEventType = ERROR_EVENT;
                    event.ulTimeStamp = u32Time;
                    event.ucaEventData[0] = ERR_GUARD;
                    event.ucaEventData[1] = (USIGN8)u32NodeId;
                    event.ucaEventData[2] = 1;
                    event.ucaEventData[3] = param->RCV_data[0] & 0x7F;
                    insertUserEvent(&event,pCh);
                    pNode->u8GuardingState = param->RCV_data[0] & 0x7F;
                  }
                  else
                  {  // guarding successful, reset retries
                     pNode->u8Retries = 0;
                  }
                }
                else
                {
                  CO_EVENTSTRUCT event;
                  USIGN32 u32Time;
                  CO_get_time(pCh->channel, &u32Time);

                  event.ucEventType = ERROR_EVENT;
                  event.ulTimeStamp = u32Time;
                  event.ucaEventData[0] = ERR_GUARD;
                  event.ucaEventData[1] = (USIGN8)u32NodeId;
                  event.ucaEventData[2] = 4;
                  event.ucaEventData[3] = param->RCV_data[0] & 0x7F;
                  insertUserEvent(&event,pCh);
                  // ERROR Toggle bit
                }
              }

            }
          }

        }



        break;
        /*
      case LSS_RESPONSE:

        HandleLSS(pCh, param);

        break;
        */

      default:
        if(Ident == LSS_RESPONSE)
          HandleLSS(pCh, param);

        break;

      }
}

/**
 ******************************************************************
 * @brief  called in case of an L2 event 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  param: Layer2 param structure
 * @return nothing
 *
 ******************************************************************
 */

void HandleRcvEvent(PCHANNELS pCh)
{
  int ret;
  PARAM_STRUCT param;
  

  do
  {
    ret = CO_read_ac(pCh->channel, &param);

    switch(ret)
    {
    case CANL2_RA_DATAFRAME:
      DispatchEvent(pCh, &param);
      break;

    case CANL2_RA_REMOTEFRAME:
      DispatchRemote(pCh, &param);
     break;
      
    case CANL2_RA_XTD_DATAFRAME:
      DispatchEvent(pCh, &param);
      break;
      
    case CANL2_RA_XTD_REMOTEFRAME:
      DispatchRemote(pCh, &param);
      break;
      
    case CANL2_RA_TX_DATAFRAME:
      break;

    case CANL2_RA_TX_REMOTEFRAME:
      break;

    case CANL2_RA_XTD_TX_DATAFRAME:
      break;

    case CANL2_RA_XTD_TX_REMOTEFRAME:
      DispatchRemote(pCh, &param);
      break;

    case CANL2_RA_CHG_BUS_STATE:
      {
        USIGN8 data[20];

        data[0] = ERR_CAN;
        data[2] = 1;
        data[3] = 3 - param.Bus_state;

       insertErrorEvent(pCh, param.Time, 4, &data[0], TRUE);
      }
      break;

    default:
      break;
    }

  } while (ret);

}

/**
 ******************************************************************
 * @brief  interrupt service thread 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  context: pointer to context
 * @return nothing
 *
 ******************************************************************
 */

int RcvThread(void *context)
{
  
  PCHANNELS pCh = (PCHANNELS)context;
  int ret;
  
  struct pollfd can_poll;
  
  while(!pCh->bStopRcvThread)
  {
    // in this implementation the CAN handle is the same as the file descriptor
    // for driver access; Please use the following macro CANL2_handle_to_descriptor 
    // for compatibility with later versions of this API
    
    can_poll.fd = CANL2_handle_to_descriptor(pCh->channel);
    can_poll.events = POLLIN | POLLHUP;
    
    do
    {
      ret = poll(&can_poll, 1, -1);
      //  printf("interruptThread1: poll returned %u, events=0x%x. revents=0x%x, POLLHUP=0x%x\n",ret,can_poll.events,can_poll.revents,POLLHUP);
      
      if(can_poll.revents & POLLHUP)
      {
        return 0;
      }
      
      
      
      if (ret == -1 && errno != EINTR)
      {
        // error
      }
    } while (ret <= 0);
    
    
    pthread_mutex_lock(&pCh->ChSemaphore);
    HandleRcvEvent(pCh);
    pthread_mutex_unlock(&pCh->ChSemaphore);
    
  } // while(!end)
  return 0;
}


/**
 ******************************************************************
 * @brief  calculates the time between lastTime an now in us. 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @param  lastTime: reference time: time in us.
 * @return time difference
 *
 ******************************************************************
 */

USIGN32 GetDifTime(PCHANNELS pCh, USIGN32 lastTime)
{
  
  USIGN32 interval, maxValue = TIMEHACK;
  USIGN32 time;

  CO_get_time(pCh->channel, &time);
  time &= TIMEHACK; // abhacken
  
  if(time >= lastTime)
  {
    interval = time - lastTime;
  }
  else
  {
    interval = 1 + time + maxValue - lastTime;
  }
  
  if(interval>(TIMEHACK-100))
  { // this is because the firmware updates the time in the DPRAM
    // every 100 us. But the rcv routine does not use this timestamp.
    // The rcv routine exactly timestamps the message when it is received.
    // If the timestamp is read by CO_get_time before it is updated, but
    // after the message was received, this contition works.
    interval = 0;
  }
  
  return interval;
}

/**
 ******************************************************************
 * @brief  checks all slaves about timeout of an heartbeat event 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @return max. time till next call
 *
 ******************************************************************
 */

USIGN32 CheckHeartbeat(PCHANNELS pCh)
{
  PGUARDLIST pToDoList = pCh->pGuardList;
  USIGN32 u32MaxTime = 1000;
  
  while(pToDoList)
  {
    PCO_NODE pNode = &pCh->node[pToDoList->u8NodeId];

    if(pNode->u32HeartbeatInterval)
    {
      USIGN32 u32DifTime = GetDifTime(pCh, pNode->u32LastHeartbeat);

      if(u32DifTime > pNode->u32HeartbeatInterval)
      {
        USIGN8 data[20];

        if(pNode->u8NodeId == pCh->MasterNodeId)
        {
          USIGN32 time;
          CO_get_time(pCh->channel, &time);
          pNode->u32LastHeartbeat = MK_TIME(time);
          if(pCh->state > RESET)
          {
            data[0] = pCh->state;              
            CO_send_data(pCh,pCh->channel,0x700+pCh->MasterNodeId,pCh->isXtd,1,data);       
          }   
        }
        else
        {
          pNode->u8Retries++;
          data[0] = ERR_GUARD;
          data[1] = pToDoList->u8NodeId;
          
          
          if(pNode->u8RetryFactor >= pNode->u8Retries)
          {
            data[2] = 2;            
          }
          else
          {
            data[2] = 3;
            pNode->u8Retries = 0;
            pToDoList->bRemove = TRUE;
          }
          

          insertErrorEvent(pCh, 0, 3, data, FALSE);
        }
        
        if(pNode->u32HeartbeatInterval < u32MaxTime)
        {
          u32MaxTime = pNode->u32HeartbeatInterval;
        }
        

      }
      else
      {
        USIGN32 u32Diff = pNode->u32HeartbeatInterval - u32DifTime;
        
        if(u32Diff < u32MaxTime)
          u32MaxTime = u32Diff;
      }

    }
    
    pToDoList = (GUARDLIST*)pToDoList->pNext;
    
  }
  
  pToDoList = pCh->pGuardList;
  
  while(pToDoList)
  {
    if(pToDoList->bRemove)
    {
      removeFromGuardList(pToDoList->u8NodeId, pCh);
    }
    pToDoList = (GUARDLIST*)pToDoList->pNext;
  }
  
  return u32MaxTime;
}



/**
 ******************************************************************
 * @brief  checks all slaves about timeout of an node guarding response 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @return nothing
 *
 ******************************************************************
 */
              
USIGN32 CheckNodeGuarding(PCHANNELS pCh)
{
  PGUARDLIST pToDoList = pCh->pGuardList;
  USIGN32 u32TicksTillNextCall = 1000000;

  pToDoList = pCh->pGuardList;
  while(pToDoList)
  {
    PCO_NODE pNode = &pCh->node[pToDoList->u8NodeId];
    USIGN32 u32DifReqTime = GetDifTime(pCh,pNode->u32GuardReqTime);
    USIGN32 u32DifResTime = GetDifTime(pCh,pNode->u32GuardResTime);

    if(pNode->u16GuardTime)
    { // guarding is active
      if(u32DifResTime > u32DifReqTime)
      { // manager waits for response
        if((u32DifReqTime > ((USIGN32)pNode->u16GuardTime * (USIGN32)1000)) && (!pNode->isFirstGuardingReq))
          { // timeout
            
            USIGN8 data[20];
            
            data[0] = ERR_GUARD;
            data[1] = pToDoList->u8NodeId;
            if(pNode->u8RetryFactor > pNode->u8Retries)
            {
              data[2] = 2;            
            }
            else
            {
              data[2] = 3;
              pNode->u16GuardTime = 0;
            }
            
            insertErrorEvent(pCh, 0, 3, data, FALSE);
            
            if(((USIGN32)pNode->u16GuardTime * (USIGN32)1000) < u32TicksTillNextCall)
            {
              u32TicksTillNextCall = ((USIGN32)pNode->u16GuardTime * (USIGN32)1000);
            }

            
            pNode->u8Retries++;
            
                     
          }
          else
          {
            if(u32DifReqTime < u32TicksTillNextCall)
            {
              u32TicksTillNextCall = u32DifReqTime;
            }
          }

        
      }
       
      
      if(pNode->u8Retries <= pNode->u8RetryFactor)
      { // check sending next request
        if(u32DifReqTime > ((USIGN32)pNode->u16GuardTime * (USIGN32)1000) - 900)
        {
          
          USIGN32 u32Time;
       //   USIGN32 u32TimeDelta = u32DifReqTime - ((USIGN32)pNode->u16GuardTime * (USIGN32)1000);
          
          CO_get_time(pCh->channel, &u32Time);
          pNode->u32GuardReqTime = MK_TIME((u32Time));
          CO_send_remote(pCh, pCh->channel, 0x700+pToDoList->u8NodeId, pCh->isXtd, 1); 
          if(pNode->isFirstGuardingReq)
          {
            pNode->isFirstGuardingReq = FALSE;
            pNode->u32GuardResTime = pNode->u32GuardReqTime - 1;
          }
          
          if(((USIGN32)pNode->u16GuardTime * (USIGN32)1000) < u32TicksTillNextCall)
          {
            u32TicksTillNextCall = ((USIGN32)pNode->u16GuardTime * (USIGN32)1000);
          }

        }
        else
        {
          if(u32DifReqTime < u32TicksTillNextCall)
          {
            u32TicksTillNextCall = u32DifReqTime;
          }
          
        }
      }
    }
    pToDoList = (GUARDLIST*)pToDoList->pNext;
    
  }
  
  return u32TicksTillNextCall;
}



/**
 ******************************************************************
 * @brief  checks for SDO response timeout 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @return nothing
 *
 ******************************************************************
 */

USIGN32 CheckSDOTimeout(PCHANNELS pCh)
{
  USIGN32 node;
  USIGN32 u32TicksTillNextCall = 1000000; // initialized with 1s

  // no SDO transfer in state init or stopped!
  if((pCh->state == STOPPED) || (pCh->state <= RESET))
    return u32TicksTillNextCall;


  for(node = 0; node<=MAX_NODE; node++)
  {
    PSDO psdo = &pCh->node[node].Sdo;
    if(psdo->isUsed)
    {
      if(psdo->isWaiting)
      {
        USIGN32 diffTime = GetDifTime(pCh, psdo->u32LastEvent)  & TIMEHACK;
        if(diffTime > (pCh->u32SdoTimeout & TIMEHACK))
        { // timeout
          COPRINT(("timeout: diffTime=%lu,last evt:%lu timeout:%lu\n", diffTime, psdo->u32LastEvent, pCh->u32SdoTimeout));
          SDOAbort(pCh, psdo,(USIGN8) node, TIMEOUT_ERR );
          psdo->isUsed = FALSE;
          psdo->isLocked = FALSE;
          
          if((pCh->u32SdoTimeout & TIMEHACK) < u32TicksTillNextCall)
          {
            u32TicksTillNextCall = (pCh->u32SdoTimeout & TIMEHACK);
          }

        }
        else
        { // calculate max. time till next call of this function
          if(diffTime < u32TicksTillNextCall)
          {
            u32TicksTillNextCall = diffTime;
          }
        }
        
      }
    }

  }
  
  return u32TicksTillNextCall;
}

/**
 ******************************************************************
 * @brief  checks for sending a sync message
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @return nothing
 *
 ******************************************************************
 */

USIGN32 CheckSync(PCHANNELS pCh)
{
  USIGN32 u32CommCycle, diffTime;
  USIGN8 data[8];
  USIGN32 u32TicksTillNextCall = 1000000; // initialized with 1s

  // no sync in state init or stopped!
  if((pCh->state == STOPPED) || (pCh->state <= RESET))
    return u32TicksTillNextCall;

  u32CommCycle = *(USIGN32*)pCh->pObjects[0x1006];

  if(!u32CommCycle)
    return u32TicksTillNextCall;
  
  diffTime = GetDifTime(pCh, pCh->u32LastSyncTime);

  if(diffTime >= u32CommCycle)
  { 
    USIGN32 time;
    USIGN32 u32ToLate = diffTime - u32CommCycle; 
    
    if(u32ToLate > 1000) u32ToLate = 1000;
    
   // COPRINT(("CheckSync: diffTime=%u, u32CommCycle=%u, 
    memset(data,0,0);
    CO_send_data(pCh, pCh->channel, *(USIGN32*)pCh->pObjects[0x1005], pCh->isXtd ,0, data);
        
    CO_get_time(pCh->channel, &time);
    pCh->u32LastSyncTime = MK_TIME((time-u32ToLate));
    
    if(u32CommCycle < u32TicksTillNextCall)
    {
      u32TicksTillNextCall = u32CommCycle;
    }

    if(pCh->isSyncConsumer)
      CheckCyclicPDOs(pCh);

  }
  else
  {
    if(diffTime < u32TicksTillNextCall)
    {
      u32TicksTillNextCall = diffTime;
    }
  }

  return u32TicksTillNextCall;
}

/**
 ******************************************************************
 * @brief  handles all cyclic jobs
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @return nothing
 *
 ******************************************************************
 */

USIGN32 DoCyclicJobs(PCHANNELS pCh)
{
  USIGN32 u32TicksTillNextCall = 1000000; // initialized with 1s
  USIGN32 u32Time1, u32Time2;
  
  u32Time1 = CheckHeartbeat(pCh);  
  u32Time2 = CheckNodeGuarding(pCh);
  u32TicksTillNextCall = (u32Time1 < u32Time2) ? u32Time1 : u32Time2;
  u32Time1 = CheckSDOTimeout(pCh);  
  u32Time2 = CheckSync(pCh);
  
  u32Time2 = (u32Time1 < u32Time2) ? u32Time1 : u32Time2;
  u32Time1 = 2;
      
  CheckPDOs(pCh);
  
  u32TicksTillNextCall = (u32Time1 < u32Time2) ? u32Time1 : u32Time2;
  
//  if(pCh->bSendPdo) u32TicksTillNextCall = 0;
 
      
  return u32TicksTillNextCall;
}


/**
 ******************************************************************
 * @brief  cyclic timer thread
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the context data
 * @return 0
 *
 ******************************************************************
 */
int timerThread(void * context)
{
  
 // DWORD ret;
  USIGN32 u32timeout = 1;
  PCHANNELS pCh = (PCHANNELS)context;

  
  while(!pCh->bStopTimerThread)
  {
    usleep(u32timeout*1000);
    pthread_mutex_lock(&pCh->ChSemaphore);
    if(pCh->bStopTimerThread)
    {
      pCh->bStopTimerThread = STOP_ACK;    
      pthread_mutex_unlock(&pCh->ChSemaphore);   
      return 0;      
    }
    
    u32timeout = DoCyclicJobs(pCh);
    pthread_mutex_unlock(&pCh->ChSemaphore);

  } // while(!end)
  
  pCh->bStopTimerThread = STOP_ACK;
  
 // pthread_exit(NULL);
  return 0;
}

/**
 ******************************************************************
 * @brief  starts the receive interrupt thread
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @return error code
 *
 ******************************************************************
 */
int startRcvThread(PCHANNELS pCh)
{
  int RetCode;

  COPRINT(("starting Rcv Thread\n"));
  pCh->s16WaitRet = 0;
  pCh->bStopRcvThread = RUNNING;
  
  if ((RetCode =
       pthread_create(&pCh->hRcvThread, 0, (void* (*)(void*)) RcvThread, pCh)))
  {
    return (INIPC_OS_ERR);
  }
  

  
  return(0);
}

void stopRcvThread(PCHANNELS pCh)
{
  void *pRet;
  
  COPRINT(("stop Rcv Thread\n"));
  pCh->bStopRcvThread = STOP_REQ;
  CANL2_reset_chip(pCh->channel);
  pthread_join(pCh->hRcvThread, &pRet);
  pthread_detach(pCh->hRcvThread);

  
 // pthread_cancel(pCh->hRcvThread);
}

/**
 ******************************************************************
 * @brief  starts the cyclic timer thread
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @return error code
 *
 ******************************************************************
 */

int startTimerThread(PCHANNELS pCh)
{
  int RetCode;
  
  /*
  struct sched_param pp;
  
  pthread_attr_t attr;
  int ret;
  
  pp.sched_priority = 140;
  
  ret = pthread_attr_init(&attr);
  if(ret) printf("error pthread_attr_init\n");
        
  ret = pthread_attr_setschedpolicy(&attr, SCHED_FIFO);
  if(ret) printf("error pthread_attr_setschedpolicy\n");
  
  ret = pthread_attr_setschedparam(&attr, &pp);
  if(ret) printf("error pthread_attr_setschedparam\n");
  */
  
  pCh->bStopTimerThread = RUNNING;
  
  if ((RetCode =
       pthread_create(&pCh->hTimerThread, 0, (void* (*)(void*)) timerThread, pCh)))
  {
   // printf("RetCode = %u %s\n",RetCode,strerror(RetCode));
    return (INIPC_OS_ERR);
  }

 // printf("leaving starttimerth\n");
  
 /* RetCode = pthread_setschedparam(pCh->hTimerThread, SCHED_RR, &pp);
  if(RetCode) printf("error pthread_setschedparam %u\n",RetCode);
 */

  
  return(0);
}

void stopTimerThread(PCHANNELS pCh)
{
  void *pRet;
  pCh->bStopTimerThread = STOP_REQ;
  
 
 // printf("stopp timer thread!\n");
  pthread_join(pCh->hTimerThread, &pRet);
  pthread_detach(pCh->hTimerThread);
//  printf("timer thread is stopped!\n");
  
   
}


/**
 ******************************************************************
 * @brief  enter critical section for channel access
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  u8Channel: channel handle
 * @return TRUE if successful, else FALSE
 *
 ******************************************************************
 */

CAN_BOOL GetChannelAccess(USIGN8 u8Channel)
{
  int ret;
  if(u8Channel > MAX_CHANNEL)
    return CAN_FALSE;

  if(!ch[u8Channel].isDefined)
    return CAN_FALSE;

  ret = pthread_mutex_lock(&ch[u8Channel].ChSemaphore);

  return (ret) ? CAN_FALSE : CAN_TRUE;
}

/**
 ******************************************************************
 * @brief  leave critical section for channel access
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  u8Channel: channel handle
 * @return nothing
 *
 ******************************************************************
 */

void ReleaseChannelAccess(USIGN8 u8Channel)
{
  pthread_mutex_unlock(&ch[u8Channel].ChSemaphore);
}

/**
 ******************************************************************
 * @brief  inserts a node in the list of the guarded nodes
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  u8NodeId: node id
 * @param  u8Channel: channel handle
 * @return nothing
 *
 ******************************************************************
 */

PGUARDLIST insertInGuardList(USIGN8 u8NodeId, USIGN8 u8Channel)
{
  PGUARDLIST pEl = NULL;

  pEl = (GUARDLIST*)malloc(sizeof(GUARDLIST));

  if(pEl)
  {
    USIGN32 u32Time;
    CO_get_time(ch[u8Channel].channel, &u32Time);
    u32Time = MK_TIME(u32Time);

    ch[u8Channel].node[u8NodeId].u32GuardReqTime = u32Time;
    ch[u8Channel].node[u8NodeId].u32GuardResTime = u32Time;
    ch[u8Channel].node[u8NodeId].isFirstGuardingReq = TRUE;

    pEl->u8NodeId = u8NodeId;
    pEl->bRemove = FALSE;
    pEl->pNext = ch[u8Channel].pGuardList;
    ch[u8Channel].pGuardList = pEl;
  }

  return pEl;
}

/**
 ******************************************************************
 * @brief  remove a node from the list of the guarded nodes
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  u8NodeId: node id
 * @param  u8Channel: channel handle
 * @return nothing
 *
 ******************************************************************
 */

void removeFromGuardList(USIGN8 u8NodeId, PCHANNELS pCh)
{
  PGUARDLIST pEl = pCh->pGuardList, pFather = NULL;
  BOOLEAN isRemoved = FALSE;

  if(!pEl)
    return;

  do
  {
    if(pEl->u8NodeId == u8NodeId)
    { // remove element
      if(pFather)
      {
        pFather->pNext = pEl->pNext;
      }
      else
      {
        pCh->pGuardList = (GUARDLIST*)pEl->pNext;
      }
      isRemoved = TRUE;
      free(pEl);
     // return;
    }
    else
    {
      pFather = pEl;
      pEl = (GUARDLIST*)pEl->pNext;
    }

  } while (pEl  && !isRemoved);
}

/**
 ******************************************************************
 * @brief  insets a PDO in the list of the installed PDOs
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pPdo: PDO
 * @param  u8Channel: channel handle
 * @return nothing
 *
 ******************************************************************
 */

PPDOLIST insertPdo(PPDO pPdo, USIGN8 u8Channel)
{
  PPDOLIST pEl = NULL, pEnd;

  pEl = (PDOLIST*)malloc(sizeof(PDOLIST));

  if(pEl)
  {


    pEl->pPdo = pPdo;
    pEl->pNext = NULL;
    pEnd = ch[u8Channel].pPdoList;
    if(!pEnd)
    {
      ch[u8Channel].pPdoList = pEl;
    }
    else
    {
      while(pEnd->pNext)
        pEnd = (PDOLIST*)pEnd->pNext;

      pEnd->pNext = pEl;

    }
   // pEl->pNext = ch[u8Channel].pPdoList;
  //  ch[u8Channel].pPdoList = pEl;
  }

  return pEl;
}
/**
 ******************************************************************
 * @brief  remove a PDO from the list of the installed PDOs
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pPdo: PDO
 * @param  u8Channel: channel handle
 * @return nothing
 *
 ******************************************************************
 */

void removePdo(PPDO pPdo, USIGN8 u8Channel)
{
  PPDOLIST pEl = ch[u8Channel].pPdoList, pFather = NULL;
  BOOLEAN isRemoved = FALSE;

  if(!pEl)
    return;

  do
  {
    if(pEl->pPdo == pPdo)
    { // remove element
      if(pFather)
      {
        pFather->pNext = pEl->pNext;
      }
      else
      {
        ch[u8Channel].pPdoList = (PDOLIST*)pEl->pNext;
      }
      isRemoved = TRUE;
      free(pEl);
     // return;
    }
    else
    {
      pFather = pEl;
      pEl = (PDOLIST*) pEl->pNext;
    }

  } while (pEl  && !isRemoved);
}

/**
 ******************************************************************
 * @brief  initializes the CANopen Object Dictionary
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pCh: pointer to the channel data
 * @return nothing
 *
 ******************************************************************
 */

void initObjectDir(PCHANNELS pCh)
{
  USIGN32 value;
  int sw_version, fw_version, hw_version, license, chip[2];
  
  CANL2_get_version(
    pCh->channel,
    &sw_version,
    &fw_version,
    &hw_version,
    &license,
    chip
    );


  pCh->ObjDict.u32DevType_1000 = 301;
  pCh->ObjDict.u8ErrorRegister_1001 = 0;
  pCh->ObjDict.u32SyncCobId_1005 = 0x80;
  pCh->ObjDict.u32ComCyclePeriode_1006 = 0;
  strcpy((char*)pCh->ObjDict.DeviceName_1008, "Softing CAN Interface");
  strcpy((char*)pCh->ObjDict.HwVersion_1009,"1.00");
  strcpy((char*)pCh->ObjDict.SwVersion_100A,"1.00");
  pCh->ObjDict.Identify_1018.id[0] = 4;
  pCh->ObjDict.Identify_1018.id[1] = VENDOR_ID;
  CANL2_get_serial_number(pCh->channel,&value);
  pCh->ObjDict.Identify_1018.id[2] = value;
  pCh->ObjDict.Identify_1018.id[3] = hw_version;
  CANL2_reset_chip(pCh->channel);
  CANL2_get_serial_number(pCh->channel, &pCh->ObjDict.Identify_1018.id[4]);
  pCh->ObjDict.NMTStartup_1F80 = ISMASTER | NOOPERATIONAL | APPSTARTNODE;


  pCh->pObjects[0x1000] = &pCh->ObjDict.u32DevType_1000;
  pCh->pObjects[0x1001] = &pCh->ObjDict.u8ErrorRegister_1001;
  pCh->pObjects[0x1005] = &pCh->ObjDict.u32SyncCobId_1005;
  pCh->pObjects[0x1006] = &pCh->ObjDict.u32ComCyclePeriode_1006;
  pCh->pObjects[0x1008] = &pCh->ObjDict.DeviceName_1008;
  pCh->pObjects[0x1009] = &pCh->ObjDict.HwVersion_1009;
  pCh->pObjects[0x100A] = &pCh->ObjDict.SwVersion_100A;
  pCh->pObjects[0x1018] = &pCh->ObjDict.Identify_1018;
  pCh->pObjects[0x1F80] = &pCh->ObjDict.NMTStartup_1F80;
#ifdef TESTSDOTRANSFER

  pCh->pObjects[0x2FFF] = NULL;
  pCh->pObjects[0x2FFF] = malloc(64*1024);

#endif // TESTSDOTRANSFER
}

/**
 ******************************************************************
 * @brief  initializes the CAN device
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  see CANopen manual
 * @return error code
 *
 ******************************************************************
 */
COL_API INT16 CALLING_CONV CMA_InitializeCard (USIGN32 ulDeviceIndex, 
                                         USIGN8 ucCAN1Functionality,
                                         USIGN8 *ucpCAN1Handle,
                                         USIGN8 ucLicence,
                                         USIGN8 *pChannelName)
{

  CANL2_CH_STRUCT ch_str;
  INT32 ret,i=0, j;
 
  ulDeviceIndex=ulDeviceIndex;
  ucCAN1Functionality=ucCAN1Functionality;
  ucLicence=ucLicence;
 
  if(strlen((char*)pChannelName) >= 80)
    return MAP_ERROR(INIPC_IB_ERR_CH_NAME_TOO_LONG);

  strcpy((char*)ch_str.sChannelName, (char*)pChannelName);

  while((i<MAX_CHANNEL) && ch[i].isDefined)
  {
    i++;
  }
  if(i==MAX_CHANNEL)
    return MAP_ERROR(INIPC_IB_ERR_TOO_MANY_CH_DEFINED);
  
  pthread_mutex_init(&ch[i].ChSemaphore, NULL);
 
  
  pthread_mutex_init(&ch[i].hQueueEvent, NULL);
  pthread_cond_init(&ch[i].QueueCondition, NULL);
  ch[i].conditionMet = FALSE;

//   pthread_mutex_lock(&ch[i].hQueueEvent);
  
  ch[i].bStopRcvThread = STOP_ACK;
  ch[i].bStopTimerThread = STOP_ACK;

 
  
  
  ret = INIL2_initialize_channel(&ch_str.ulChannelHandle, (char*)&ch_str.sChannelName);

  if(ret) return MAP_ERROR(ret);

  *ucpCAN1Handle = i;
  ch[i].channel = ch_str.ulChannelHandle;
  ch[i].isDefined = CAN_TRUE;

  initObjectDir(&ch[i]);
  for(j=0; j<0x800; j++)
  {
    ch[i].u16CobExt2Int[j] = j;
    ch[i].u16CobInt2Ext[j] = j;

  }

  ch[i].MasterNodeId = MAX_NODE;
    
  ch[i].UserEvents.rd = FIFOMASK;
  ch[i].UserEvents.wr = 0;

  ch[i].u32ApplTimeout = 2000000;
  ch[i].u32SdoTimeout  = 100000;

  return 0;

} ///CMA_InitializeCard



/****************************************************************************/
/*                                                                          */
/*  CMA_CloseCard                                                           */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_CloseCard (USIGN8 ucChannelHandle )
{
  PCHANNELS pCh;
   

  if(ucChannelHandle > MAX_CHANNEL)
  {
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE); 
  }

  pCh = &ch[ucChannelHandle];

  if(pCh->bStopTimerThread != STOP_ACK)
  {
    stopTimerThread(pCh);
  }
  
  if(pCh->bStopRcvThread != STOP_ACK)
  {
    pCh->s16WaitRet = -1;
    stopRcvThread(pCh);
  }

  
#ifdef TESTSDOTRANSFER
  if(pCh->pObjects[0x2FFF])
  {
    free(pCh->pObjects[0x2FFF]);
    pCh->pObjects[0x2FFF] = NULL;
  }

#endif // TESTSDOTRANSFER


  pCh->isDefined = FALSE;
  return MAP_ERROR(INIL2_close_channel(pCh->channel));
} //CMA_CloseCard


/****************************************************************************/
/*                                                                          */
/*  CMA_Startup                                                             */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_Startup (USIGN8 ucChannelHandle)
{
  INT32 ret,i;
  USIGN8 data[8];
  CO_EVENTSTRUCT event;


  if(!GetChannelAccess(ucChannelHandle))
  {
   return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);
  }

  ch[ucChannelHandle].UserEvents.rd = FIFOMASK;
  ch[ucChannelHandle].UserEvents.wr = 0;



  if(ch[ucChannelHandle].bStopTimerThread == STOP_ACK)
  {
    ret = startTimerThread(&ch[ucChannelHandle]);
    
    if (ret) 
    {
     // printf("startTimerThread returned %li\n",ret);
      ReleaseChannelAccess(ucChannelHandle);
      return MAP_ERROR(ret);
    }
  }
  else
  {
    ch[ucChannelHandle].bStopTimerThread = RUNNING;
  }

 // if(ch[ucChannelHandle].state == INITIALISATION)
  {
    // start SJA1000
    ret = CANL2_start_chip(ch[ucChannelHandle].channel);
    
    if (ret) 
    {
    //  printf("CANL2_start_chip returned %li\n",ret);
      ReleaseChannelAccess(ucChannelHandle);
      return MAP_ERROR(ret);
    }
  }
  
  
  if(ch[ucChannelHandle].bStopRcvThread == STOP_ACK)
  {
    
    ret = startRcvThread(&ch[ucChannelHandle]);
    
    if (ret) 
    {
      // printf("startRcvThread returned %li\n",ret);
      ReleaseChannelAccess(ucChannelHandle);
      return MAP_ERROR(ret);
    }
  }
  else
  {
    ch[ucChannelHandle].bStopRcvThread = RUNNING;
  }
  
  

  for(i=0; i<MAX_NODE; i++)
  {
    ch[ucChannelHandle].node[i].Sdo.isUsed = FALSE;
  }

  ch[ucChannelHandle].state = PREOPERATIONAL;

  // send bootup message
  CO_send_data(&ch[ucChannelHandle],
    ch[ucChannelHandle].channel, 
    ch[ucChannelHandle].MasterNodeId + 0x80, 
    ch[ucChannelHandle].isXtd, 
    0, 
    data);

  event.ucEventType = EV_STARTUP;
  event.ulTimeStamp = 0;

  insertUserEvent(&event, &ch[ucChannelHandle]);


  ReleaseChannelAccess(ucChannelHandle);

  return MAP_ERROR(ret);

} //CMA_Startup( );


/****************************************************************************/
/*                                                                          */
/*  CMA_ReadEvent                                                           */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_ReadEvent (USIGN8 ucChannelHandle,
                                    CMA_EventStruct *stpEvent)
{
  if(!GetChannelAccess(ucChannelHandle))
   return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);

  popUserEvent(ucChannelHandle, stpEvent);

  ReleaseChannelAccess(ucChannelHandle);

  /*
  if(isDbgEnabled)
  {
    COPRINT((
      "\nCMA_ReadEvent: %u, %02x %02x %02x %02x %02x \n", 
      stpEvent->ucEventType,
      stpEvent->ucaEventData[0],
      stpEvent->ucaEventData[1],
      stpEvent->ucaEventData[2],
      stpEvent->ucaEventData[3],
      stpEvent->ucaEventData[4]
      ));
  }
    
   */


  
  return 0;

} //CMA_ReadEvent( );


/****************************************************************************/
/*                                                                          */
/*  CMA_AddNode                                                             */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_AddNode (USIGN8 ucChannelHandle,
                                  CMA_NodeConfig stNodeConfig)
{
  PCO_NODE pNode;
  USIGN32 time;

  COPRINT(("CMA_AddNode: %u state:%u\n",stNodeConfig.ucNodeID,ch[ucChannelHandle].state));
  if(stNodeConfig.ucNodeID > MAX_NODE)
    return MAP_ERROR(CMA_ERR_SCT_NODE_ID);

  if(ch[ucChannelHandle].state == INITIALISATION)
    return MAP_ERROR(CMA_ERR_MASTER_STATE);

  if(!GetChannelAccess(ucChannelHandle))
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);


  if(ch[ucChannelHandle].node[stNodeConfig.ucNodeID].isDefined)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(CMA_ERR_SCT_NODE_ID);
  }

  pNode = &ch[ucChannelHandle].node[stNodeConfig.ucNodeID];

  pNode->u8GuardingState = INITIALISATION;
  pNode->u8NodeId = stNodeConfig.ucNodeID;
  pNode->u16GuardTime = stNodeConfig.usGuardTime;
//  COPRINT(("Guard Time: %u ms, for Node: %u\n", pNode->u16GuardTime, pNode->u8NodeId));
  pNode->u8ErrorAction = stNodeConfig.ucErrAction;
  pNode->u8RetryFactor = stNodeConfig.ucRetryFactor;
  pNode->u32HeartbeatInterval = stNodeConfig.ulHeartBeatInterval * 1000;
  pNode->state = STOPPED;

  CO_get_time(ch[ucChannelHandle].channel, &time);
  pNode->u32LastHeartbeat = MK_TIME(time);
  pNode->u8ToggleBit = 0x80;

  pNode->u8Retries = 0;




  pNode->isDefined = CAN_TRUE;

 // COPRINT(("error action: %x\n", pNode->u8ErrorAction));

  ReleaseChannelAccess(ucChannelHandle);
  COPRINT(("CMA_AddNode: done!\n"));
  return 0;
} //CMA_AddNode( );


/****************************************************************************/
/*                                                                          */
/*  CMA_RequestGuarding                                                     */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_RequestGuarding (USIGN8 ucChannelHandle,
                                          USIGN8 ucNodeID,
                                          USIGN8 ucReqGuard)
{
  COPRINT(("CMA_RequestGuarding: %u guard:%u\n",ucNodeID,ucReqGuard));
  if(!GetChannelAccess(ucChannelHandle))
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);

  if(ucReqGuard)
  {
    if(!insertInGuardList(ucNodeID, ucChannelHandle))
    {
      ReleaseChannelAccess(ucChannelHandle);
      return MAP_ERROR(CMA_ERR_MALLOC);
    }
  }
  else
  {
    removeFromGuardList(ucNodeID, &ch[ucChannelHandle]);
  }

  ReleaseChannelAccess(ucChannelHandle);
  COPRINT(("CMA_RequestGuarding: done!\n"));
  return 0;

} //CMA_RequestGuarding( );


/****************************************************************************/
/*                                                                          */
/*  CMA_ChangeState                                                         */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_ChangeState (USIGN8 ucChannelHandle,
                                      USIGN8 ucNodeID,
                                      USIGN8 ucState) 
{
  int ret = 0;
  USIGN8 data[8], state;
  COPRINT(("CMA_ChangeState: %u state:%u\n",ucNodeID,ucState));
  if(ucNodeID > MAX_NODE+1)
    return MAP_ERROR(CMA_ERR_SCT_NODE_ID);
  
  
  
  if(!GetChannelAccess(ucChannelHandle))
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);

  if(ch[ucChannelHandle].state == INITIALISATION)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(CMA_ERR_MASTER_STATE);
  }


  
  if((ucNodeID != 128) && ucNodeID)
  {
    if(!ch[ucChannelHandle].node[ucNodeID].isDefined)
    {
      ReleaseChannelAccess(ucChannelHandle);
      return MAP_ERROR(CMA_ERR_PAR_NODE_ID);
    }
  }

  
  
  switch(ucState)
  {
    
  case ST_RESET_COM:
    data[0] = 130;
    state = RESET;
    break;
    
  case ST_RESET_NODE:
    data[0] = 129;
    state = RESET;
    break;
    
  case ST_PRE_OPERATIONAL:
    data[0] = 128;
    state = PREOPERATIONAL;
    break;
    
  case ST_STOPP:
    data[0] = 2;
    state = STOPPED;
    break;
    
  case ST_START:
    data[0] = 1;
    state = OPERATIONAL;
    break;
    
  default:
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(CMA_ERR_PAR_STATE);
    
  }

  if((ucNodeID == ch[ucChannelHandle].MasterNodeId) || !ucNodeID)
  { // master change state ( node id 0 or master node id )
    if(ch[ucChannelHandle].state != state)
    { 
      
      USIGN8 boot[8];
      if(ch[ucChannelHandle].state <= RESET)
      {
        // master boot up
        
        memset(boot,0,8);
        
        CO_send_data(&ch[ucChannelHandle],
          ch[ucChannelHandle].channel, 
          ch[ucChannelHandle].MasterNodeId | 0x80,
          ch[ucChannelHandle].isXtd,
          0,
          boot);
      }
      
      ch[ucChannelHandle].state = (CO_STATE)state;

      
      if(state == OPERATIONAL)
      {
        CheckPDOs(&ch[ucChannelHandle]);
      }
      
    }
    
    
    ReleaseChannelAccess(ucChannelHandle);
    return 0;
  }
  
  if(ucNodeID == 128)
  {
    int i;
    PCO_NODE pNode = ch[ucChannelHandle].node;
    
    if(ch[ucChannelHandle].state != state)
    { 
      
      USIGN8 boot[8];
      PPDOLIST pdoList;
      if(ch[ucChannelHandle].state <= RESET)
      {
        // master boot up
        
        memset(boot,0,8);
        
        CO_send_data(&ch[ucChannelHandle],
          ch[ucChannelHandle].channel, 
          ch[ucChannelHandle].MasterNodeId | 0x80,
          ch[ucChannelHandle].isXtd,
          0,
          boot);
      }
      
      
      if(ch[ucChannelHandle].state == OPERATIONAL)
      {
        
        pdoList = ch[ucChannelHandle].pPdoList;
        
        while(pdoList)
        {
          PPDO pdo = pdoList->pPdo;
          if(pdo->u8Type)
          {
            pdo->bSndReq = TRUE;
          }
          pdoList = (PDOLIST*)pdoList->pNext;
        }
      }
      
    }


    ch[ucChannelHandle].state = (CO_STATE)state;

    data[1] = 0;
    CO_send_data(&ch[ucChannelHandle],
      ch[ucChannelHandle].channel, 
      0,
      ch[ucChannelHandle].isXtd,
      2,
      data);
    
    
    
    for(i=0; i<MAX_NODE; i++)
    {
      pNode[i].state = (CO_STATE)state;
    }
    
    if(state == OPERATIONAL)
    {
      
      DoCyclicJobs(&ch[ucChannelHandle]);
    }
    
    
    
    
  }
  else
  {
    
    data[1] = ucNodeID;
    
    if(!ucNodeID)
    { // master
      CO_send_data(
        &ch[ucChannelHandle],
        ch[ucChannelHandle].channel, 
        0xFF,
        ch[ucChannelHandle].isXtd,
        0,
        data);  
      ch[ucChannelHandle].state = (CO_STATE)state;

      if(state == OPERATIONAL)
        DoCyclicJobs(&ch[ucChannelHandle]);


      COPRINT(("master state: %u\n",state));
    }
    else
    {
      
      CO_send_data(
        &ch[ucChannelHandle],
        ch[ucChannelHandle].channel, 
        0,
        ch[ucChannelHandle].isXtd,
        2,
        data);
    }
    
    
    if(!ret)
    {
      ch[ucChannelHandle].node[ucNodeID].state = (CO_STATE)state;
    }
  }
  
  ReleaseChannelAccess(ucChannelHandle);
  COPRINT(("CMA_ChangeState: done!\n"));
  return MAP_ERROR(ret);
  
} //CMA_ChangeState( );


/****************************************************************************/
/*                                                                          */
/*  CMA_RemoveNode                                                          */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_RemoveNode (USIGN8 ucChannelHandle,
                                     USIGN8 ucNodeID,
                                     USIGN8 ucState) 
{
  PCO_NODE pNode;
  INT16 ret;

  COPRINT(("CMA_RemoveNode: %u state:%u\n",ucNodeID,ucState));
  ret = CMA_ChangeState(ucChannelHandle, ucNodeID, ucState);
  
  if(ret)
  {
    return ret;
  }

  if(ucNodeID > MAX_NODE)
    return MAP_ERROR(CMA_ERR_PAR_NODE_ID);

  if(!GetChannelAccess(ucChannelHandle))
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);


  if(!ch[ucChannelHandle].node[ucNodeID].isDefined)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(CMA_ERR_PAR_NODE_ID);
  }

  pNode = &ch[ucChannelHandle].node[ucNodeID];
  pNode->state = (CO_STATE)ucState;
  pNode->u8NodeId = ucNodeID;
  ch[ucChannelHandle].node[ucNodeID].isDefined = FALSE;


  removeFromGuardList(ucNodeID, &ch[ucChannelHandle]);

  pNode->isDefined = CAN_FALSE;

  ReleaseChannelAccess(ucChannelHandle);
  COPRINT(("CMA_RemoveNode: done!\n"));
  return 0;

} //CMA_RemoveNode( );


/****************************************************************************/
/*                                                                          */
/*  CMA_ConfigCANChannel                                                    */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_ConfigCANChannel (USIGN8 ucChannelHandle,
                                                 CMA_CAN_CONFIG stCANConfig)
{
  USIGN8 baudrate;
  int ret;
  CMA_VersionStruct version;
  PCHANNELS pCh;

  

  if(!GetChannelAccess(ucChannelHandle))
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);
  
  pCh = &ch[ucChannelHandle];

  // Reset CAN chips of CAN 1 and 2
  if ((ret = CANL2_reset_chip(pCh->channel)) != 0)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(ret);
  }
  
  ret = CMA_GetVersion(ucChannelHandle, &version);
  if(ret)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(ret);
  }

  sprintf((char*)pCh->ObjDict.HwVersion_1009, "%2u.%02u", version.ucHWMainRevision, version.ucHWSubRevision);
  sprintf((char*)pCh->ObjDict.SwVersion_100A, "%2u.%03u", version.ucSWMajorVersion, version.ucSWMinorVersion);
  sprintf((char*)pCh->ObjDict.DeviceName_1008, "Softing AG CANopen Manager %s", version.ucHWIdentString);

  pCh->ObjDict.Identify_1018.id[3] = version.ucSWMajorVersion*1000+version.ucSWMinorVersion;
  pCh->ObjDict.Identify_1018.id[4] = version.ulHWSerialNumber;
  
  baudrate = stCANConfig.ucBaudRate;
  if(baudrate != 0xFF)
  {
    // Set Bit timing for CAN 1 and 2
    if(baudrate > MAX_BAUD_RATE)
    {
      ReleaseChannelAccess(ucChannelHandle);
      return MAP_ERROR(CMA_ERR_SCT_BAUD_RATE);
    }
    
    if ((ret = CANL2_initialize_chip(ch[ucChannelHandle].channel, presc[baudrate], sjw[baudrate], tseg1[baudrate], tseg2[baudrate], sample[baudrate])) != 0)
    {
      ReleaseChannelAccess(ucChannelHandle);
      return MAP_ERROR(ret);
    }
  }
  else
  {
    if ((ret = CANL2_initialize_chip(ch[ucChannelHandle].channel, stCANConfig.ucPresc, stCANConfig.ucSjw, stCANConfig.ucTSeg1, stCANConfig.ucTSeg2, 0)) != 0)
    {
      ReleaseChannelAccess(ucChannelHandle);
      return MAP_ERROR(ret);
    }
  }
  
  // Set acceptance filter for CAN 1 and 2
  if ((ret = CANL2_set_acceptance(ch[ucChannelHandle].channel, stCANConfig.ucAccCode, stCANConfig.ucAccMask, stCANConfig.ucAccCode, stCANConfig.ucAccMask)) != 0)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(ret);
  }
  
  
  // Set resynchronization
/*
  if ((ret = CANL2_set_mode(ch[ucChannelHandle].channel, 0, 0)) != 0)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(ret);
  }
*/
  
  // Set output control (phys. layer -1: default  CAN High Speed)
  if ((ret = CANL2_set_output_control(ch[ucChannelHandle].channel, 0xFA)) != 0)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(ret);
  }
  
  if ((ret = CANL2_enable_fifo(ch[ucChannelHandle].channel)) != 0)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(ret);
  }

  ReleaseChannelAccess(ucChannelHandle);
  return 0;
  
} //CMA_ConfigCANChannel( );


/****************************************************************************/
/*                                                                          */
/*  CMA_InitSDOMan                                                          */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_InitSDOMan (USIGN8 ucChannelHandle, 
                                           USIGN16 usSDOTimeout, 
                                           USIGN16 usApplTimeoutFactor)
{
  
  COPRINT(("\nCMA_InitSDOMan: %u, %u\n", usSDOTimeout, usApplTimeoutFactor));
  
  if(!GetChannelAccess(ucChannelHandle))
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);
  
  ch[ucChannelHandle].u32ApplTimeout = usApplTimeoutFactor * 1000;
  if(ch[ucChannelHandle].u32ApplTimeout < 20000)
    ch[ucChannelHandle].u32ApplTimeout = 20000;
  
  ch[ucChannelHandle].u32SdoTimeout = usSDOTimeout * 1000;

  ReleaseChannelAccess(ucChannelHandle);
  return 0;
  
} //CMA_InitSDOMan( );


/****************************************************************************/
/*                                                                          */
/*  CMA_InitNodeMan                                                         */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_InitNodeMan (USIGN8 ucChannelHandle, 
                                      USIGN8 ucNodeID)
{
  COPRINT(("CMA_InitNodeMan_E: \n"));

    
  if(ucChannelHandle > MAX_CHANNEL)
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE); 


  ch[ucChannelHandle].MasterNodeId = ucNodeID;

  return 0;

} //CMA_InitNodeMan( );


                                                                /*Ewtg NmtSlv*/
#ifdef NMT_SLV
/*****************************************************************************/
/*                                                                           */
/*  CMA_InitNodeMan_E                                                        */
/*                                                                           */
/*****************************************************************************/
COL_API INT16 CALLING_CONV CMA_InitNodeMan_E (USIGN8 ucChannelHandle, 
                                        USIGN8 ucNodeID,
                                        USIGN8 ucNMTState,
                                        USIGN8 ucRestartMode)
{

  COPRINT(("CMA_InitNodeMan_E: \n"));

  if(ucChannelHandle > MAX_CHANNEL)
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE); 

  ch[ucChannelHandle].MasterNodeId = ucNodeID;

} //CMA_InitNodeMan_E( );
#endif //NMT_SLV


/****************************************************************************/
/*                                                                          */
/*  CMA_Shutdown                                                            */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_Shutdown (USIGN8 ucChannelHandle)
{

  CO_EVENTSTRUCT event;
  USIGN32 u32Time;

  if(ucChannelHandle > MAX_CHANNEL)
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE); 


  CANL2_get_time(ch[ucChannelHandle].channel, &u32Time);

  ch[ucChannelHandle].state = INITIALISATION;

  stopTimerThread(&ch[ucChannelHandle]);
 
  stopRcvThread(&ch[ucChannelHandle]);
  
  event.ucEventType = EV_SHUTDOWN;
  event.ulTimeStamp = u32Time;
  insertUserEvent(&event,&ch[ucChannelHandle]);

  /* entry not free */
  return 0;

} //CMA_Shutdown( );


/****************************************************************************/
/*                                                                          */
/*  CMA_ConfigSyncMan                                                       */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_ConfigSyncMan (USIGN8 ucChannelHandle, 
                                        CMA_SyncManConfig stSyncManConfig)
{
  USIGN32 u32Time;

  if(!GetChannelAccess(ucChannelHandle))
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);

  *(USIGN32*)ch[ucChannelHandle].pObjects[0x1005] = stSyncManConfig.ulSyncID;

  if(stSyncManConfig.ucSendMode)
  {
    *(USIGN32*)ch[ucChannelHandle].pObjects[0x1006] = stSyncManConfig.ulSyncCycle;
  }
  else
  {
    *(USIGN32*)ch[ucChannelHandle].pObjects[0x1006] = 0;
  }
  
  ch[ucChannelHandle].isSyncConsumer = stSyncManConfig.ucConsumeMode;

  CO_get_time(ch[ucChannelHandle].channel, &u32Time);

  ch[ucChannelHandle].u32LastSyncTime = MK_TIME(u32Time);

  ReleaseChannelAccess(ucChannelHandle);

  return 0;

} //CMA_ConfigSyncMan( );



/****************************************************************************/
/*                                                                          */
/*  CMA_SendSyncObject                                                      */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_SendSyncObject (USIGN8 ucChannelHandle, 
                                         CMA_SyncManConfig stSyncManConfig)
{

  ucChannelHandle=ucChannelHandle;
  stSyncManConfig=stSyncManConfig;

  return (0);

} //CMA_SendSyncObject( );


/****************************************************************************/
/*                                                                          */
/*  CMA_AbortSDOTransmission                                                */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_AbortSDOTransmission (USIGN8 ucChannelHandle, 
                                               USIGN8 ucNodeID)
{
  if(ucNodeID > MAX_NODE)
    return MAP_ERROR(CMA_ERR_SCT_NODE_ID);

  if(!GetChannelAccess(ucChannelHandle))
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);

  if(ch[ucChannelHandle].node[ucNodeID].Sdo.isUsed)
  {
   SDOAbort(&ch[ucChannelHandle], &ch[ucChannelHandle].node[ucNodeID].Sdo, ucNodeID, GENERAL_ERR);
  }

    
  ch[ucChannelHandle].node[ucNodeID].Sdo.isLocked = FALSE;

  
  ReleaseChannelAccess(ucChannelHandle);
  return 0;

} //CMA_AbortSDOTransmission( );


/****************************************************************************/
/*                                                                          */
/*  CMA_InitDownload                                                        */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_InitDownload (USIGN8  ucChannelHandle, 
                                       USIGN8  ucNodeID,
                                       USIGN16 usIndex,
                                       USIGN8  ucSubIndex,
                                       USIGN32  ulTotalDataSize,
                                       USIGN8  *ucpData)
{

  INT16 ret;

  //COPRINT(("CMA_InitDownload(%u,%u,0x%x,0x%x,%u)\n",ucChannelHandle,ucNodeID,usIndex,ucSubIndex,ulTotalDataSize));

  if(!GetChannelAccess(ucChannelHandle))
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);

  // no SDO transfer in state init or stopped!
  if((ch[ucChannelHandle].state == STOPPED) || (ch[ucChannelHandle].state <= RESET))
  {
    ReleaseChannelAccess(ucChannelHandle);
    return CMA_ERR_MASTER_STATE;
  }



  ret = InitSdoDownload(&ch[ucChannelHandle], ucNodeID, usIndex, ucSubIndex, ulTotalDataSize, ucpData);
  /* entry not free */

  ReleaseChannelAccess(ucChannelHandle);

  return MAP_ERROR(ret);

} //CMA_InitDownload( );

/****************************************************************************/
/*                                                                          */
/*  CMA_InitDownload                                                        */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_InitBlockDownload (USIGN8  ucChannelHandle, 
                                       USIGN8  ucNodeID,
                                       USIGN16 usIndex,
                                       USIGN8  ucSubIndex,
                                       USIGN32  ulTotalDataSize,
                                       USIGN8  *ucpData)
{

  INT16 ret;

//  COPRINT(("CMA_InitDownload(%u,%u,0x%x,0x%x,%u)\n",ucChannelHandle,ucNodeID,usIndex,ucSubIndex,ulTotalDataSize));


  if(!GetChannelAccess(ucChannelHandle))
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);

  // no SDO transfer in state init or stopped!
  if((ch[ucChannelHandle].state == STOPPED) || (ch[ucChannelHandle].state <= RESET))
  {
    ReleaseChannelAccess(ucChannelHandle);
    return CMA_ERR_MASTER_STATE;
  }

  ret = InitSdoBlockDownload(&ch[ucChannelHandle], ucNodeID, usIndex, ucSubIndex, ulTotalDataSize, ucpData);
  /* entry not free */

  ReleaseChannelAccess(ucChannelHandle);

  return MAP_ERROR(ret);

} //CMA_InitDownload( );



/****************************************************************************/
/*                                                                          */
/*  CMA_WriteNextSegment                                                        */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_WriteNextSegment (USIGN8  ucChannelHandle, 
                                           USIGN8  ucNodeID,
                                           USIGN16 usIndex,
                                           USIGN8  ucSubIndex,
                                           USIGN8  ucDataSize,
                                           USIGN8  *ucpData)
{

  UNUSED_PARA(ucChannelHandle);
  UNUSED_PARA(ucNodeID);
  UNUSED_PARA(usIndex);
  UNUSED_PARA(ucSubIndex);
  UNUSED_PARA(ucDataSize);
  UNUSED_PARA(ucpData);
  /* entry not free */
  return 0;

} //CMA_WriteNextSegment( );



/****************************************************************************/
/*                                                                          */
/*  CMA_InitUpload                                                          */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_InitUpload (USIGN8  ucChannelHandle, 
                                     USIGN8  ucNodeID,
                                     USIGN16 usIndex,
                                     USIGN8  ucSubIndex)
{

  USIGN16 u16CobId = SDO_RX + ucNodeID;
  USIGN8 data[8];
  USIGN32 u32Time;
  PSDO  psdo;
  int ret=0;


  COPRINT(("CMA_InitUpload(%u,%u,0x%x,0x%x)\n",ucChannelHandle,ucNodeID,usIndex,ucSubIndex));


  if(!GetChannelAccess(ucChannelHandle))
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);


  // no SDO transfer in state init or stopped!
  if((ch[ucChannelHandle].state == STOPPED) || (ch[ucChannelHandle].state <= RESET))
  {
    ReleaseChannelAccess(ucChannelHandle);
    return CMA_ERR_MASTER_STATE;
  }


  if((ucNodeID == ch[ucChannelHandle].MasterNodeId) || (ucNodeID == 0))
  {
    HandleMasterUpload(&ch[ucChannelHandle], ucNodeID, usIndex, ucSubIndex);
    ReleaseChannelAccess(ucChannelHandle);
    return 0;
  }

  if(ucNodeID>MAX_NODE)
    return MAP_ERROR(CMA_ERR_PAR_NODE_ID);


  memset(data,0,8);
  
  psdo = &ch[ucChannelHandle].node[ucNodeID].Sdo;

  if(psdo->isUsed)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(CMA_ERR_LAST_REQ_PENDING);
  }

  if(psdo->isLocked)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(CMA_ERR_LAST_REQ_PENDING);
  }

  CANL2_get_time(ch[ucChannelHandle].channel, &u32Time);

  psdo->u32LastEvent = MK_TIME(u32Time);
  COPRINT(("CMA_InitUpload: time:%lu\n",psdo->u32LastEvent));

  psdo->DataBuffer.u32BufferPtr = 0;
  psdo->DataBuffer.u32Size = 0;
  psdo->DataBuffer.pBuffer = NULL;
  psdo->Direction = UPLOAD;
  psdo->isMasterInvolved = TRUE;
  psdo->isUsed = TRUE;
  psdo->isLocked = TRUE;
  psdo->u16CobId = u16CobId;
  psdo->u8NodeId = ucNodeID;
  psdo->isInit = TRUE;
  psdo->u8OldDataByte0 = 0;
  psdo->u16Index = usIndex;
  psdo->u8SubIndex = ucSubIndex;
  psdo->isWaiting = TRUE;
  psdo->u8OldDataByte0 = CS_NEXT_SEGMENT;



  data[0] = CS_INIT_UPLOAD;


  WRT16(&data[1], usIndex)

  data[3] = ucSubIndex;

  CO_send_data(&ch[ucChannelHandle], ch[ucChannelHandle].channel, u16CobId, ch[ucChannelHandle].isXtd, 8, data);

  ReleaseChannelAccess(ucChannelHandle);
  return MAP_ERROR(ret);  

  return 0;

} //CMA_InitUpload( )

/****************************************************************************/
/*                                                                          */
/*  CMA_InitUpload                                                          */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_InitBlockUpload (USIGN8  ucChannelHandle, 
                                     USIGN8  ucNodeID,
                                     USIGN16 usIndex,
                                     USIGN8  ucSubIndex,
                                     USIGN8  ucBlockSize)
{

  USIGN16 u16CobId = SDO_RX + ucNodeID;
  USIGN8 data[8];
  USIGN32 u32Time;
  PSDO  psdo;
  int ret=0;


  COPRINT(("CMA_InitBlockUpload(%u,%u,0x%x,0x%x)\n",ucChannelHandle,ucNodeID,usIndex,ucSubIndex));


  if(ucBlockSize > 127)
    return MAP_ERROR(INIPC_IB_ERR_BLK_SIZE);

  if(!GetChannelAccess(ucChannelHandle))
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);


  // no SDO transfer in state init or stopped!
  if((ch[ucChannelHandle].state == STOPPED) || (ch[ucChannelHandle].state <= RESET))
  {
    ReleaseChannelAccess(ucChannelHandle);
    return CMA_ERR_MASTER_STATE;
  }




  if((ucNodeID == ch[ucChannelHandle].MasterNodeId) || (ucNodeID == 0))
  {
    HandleMasterUpload(&ch[ucChannelHandle], ucNodeID, usIndex, ucSubIndex);
    ReleaseChannelAccess(ucChannelHandle);
    return 0;
  }

  if(ucNodeID>MAX_NODE)
    return MAP_ERROR(CMA_ERR_PAR_NODE_ID);


  memset(data,0,8);
  
  psdo = &ch[ucChannelHandle].node[ucNodeID].Sdo;

  if(psdo->isUsed)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(CMA_ERR_LAST_REQ_PENDING);
  }

  if(psdo->isLocked)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(CMA_ERR_LAST_REQ_PENDING);
  }

  CANL2_get_time(ch[ucChannelHandle].channel, &u32Time);

  psdo->u32LastEvent = MK_TIME(u32Time);

  psdo->DataBuffer.u32BufferPtr = 0;
  psdo->DataBuffer.u32Size = 0;
  psdo->DataBuffer.pBuffer = NULL;
  psdo->Direction = BLOCKUPLOAD;
  psdo->isMasterInvolved = TRUE;
  psdo->isUsed = TRUE;
  psdo->isLocked = TRUE;
  psdo->u16CobId = u16CobId;
  psdo->u8NodeId = ucNodeID;
  psdo->SdoBlockState = BLK_TRANSFER_INIT;
  psdo->u8OldDataByte0 = 0;
  psdo->u16Index = usIndex;
  psdo->u8SubIndex = ucSubIndex;
  psdo->isWaiting = TRUE;
  psdo->u8OldDataByte0 = CS_NEXT_SEGMENT;
  psdo->u8BlockSize = ucBlockSize;



  data[0] = CS_INIT_BLOCKUPLOAD | SUPPORT_CRC_CODE;

  WRT16(&data[1], usIndex)

  data[3] = ucSubIndex;
  data[4] = ucBlockSize;
  data[5] = NO_FALLBACK;

  CO_send_data(&ch[ucChannelHandle], ch[ucChannelHandle].channel, u16CobId, ch[ucChannelHandle].isXtd, 8, data);

  ReleaseChannelAccess(ucChannelHandle);
  return MAP_ERROR(ret);  

  return 0;

} //CMA_InitBlockUpload( )

/****************************************************************************/
/*                                                                          */
/*  CMA_ReadData                                                     */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_ReadData (USIGN8  ucChannelHandle, 
                                                USIGN8  ucNodeID,
                                                USIGN8  ucSubIndex,
                                                USIGN16 usIndex,
                                                USIGN32 ulBufferSize,
                                                USIGN8 *pData)
{

  PSDO psdo;

  if(!GetChannelAccess(ucChannelHandle))
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);


  psdo = &ch[ucChannelHandle].node[ucNodeID].Sdo;

  if(!psdo->isLocked)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(CMA_ERR_LAST_REQ_PENDING);
  }

  if(psdo->isUsed)
  { // upload/download in progress
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(CMA_ERR_LAST_REQ_PENDING);
  }

  if(psdo->u16Index != usIndex)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(CMA_ERR_PAR_INDEX);
  }
  if(psdo->u8SubIndex != ucSubIndex)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(CMA_ERR_PAR_SUBINDEX);
  }

  if(psdo->DataBuffer.u32BufferPtr > ulBufferSize)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(CMA_ERR_PAR_DATA_SIZE);
  }

  memcpy(pData, psdo->DataBuffer.pBuffer, psdo->DataBuffer.u32BufferPtr);

  
  psdo->isLocked = FALSE;
  free(psdo->DataBuffer.pBuffer);
  psdo->DataBuffer.isAllocated = FALSE;
  psdo->DataBuffer.pBuffer = NULL;
  psdo->DataBuffer.u32BufferPtr = 0;
  psdo->DataBuffer.u32Size = 0;

  ReleaseChannelAccess(ucChannelHandle);
  return 0;

} //CMA_ReadNextSegment( );



/****************************************************************************/
/*                                                                          */
/*  CMA_ReadNextSegment                                                     */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_ReadNextSegment (USIGN8  ucChannelHandle, 
                                          CMA_SDOSeg *stpSDOSeg)
{
  UNUSED_PARA(ucChannelHandle);
  UNUSED_PARA(stpSDOSeg);
  /* no entry found */
  return 0;

} //CMA_ReadNextSegment( );


/****************************************************************************/
/*                                                                          */
/*  CMA_InstallPDO                                                          */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_InstallPDO (USIGN8  ucChannelHandle, 
                                     CMA_PDOConfig  stPDOConfig, 
                                     USIGN16 *uspPDOHandle)
{
  USIGN16 ret = 0, IntCobId;
  PPDO pPdo = NULL;
  BOOLEAN bCheck = FALSE;


  if(stPDOConfig.ulCOBID > 0x7FF)
    return MAP_ERROR(CMA_ERR_SCT_COB_ID);

  

  if(!GetChannelAccess(ucChannelHandle))
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);

    
  IntCobId = getPdoIndex(
    &ch[ucChannelHandle], 
    (USIGN16)stPDOConfig.ulCOBID, 
    (BOOLEAN)(stPDOConfig.ucPDOType == 0));

  if(IntCobId == 0xFFFF)
  {
    ReleaseChannelAccess(ucChannelHandle);
    if(stPDOConfig.ucPDOType == RX_PDO)
    {
      if(stPDOConfig.ucAccType)
        return MAP_ERROR(CMA_ERR_FIFO_RX_PDO_MAX);
      else
        return MAP_ERROR(CMA_ERR_DPRAM_RX_PDO_MAX);
    }
    else
    {
      if(stPDOConfig.ucAccType)
        return MAP_ERROR(CMA_ERR_FIFO_TX_PDO_MAX);
      else
        return MAP_ERROR(CMA_ERR_DPRAM_TX_PDO_MAX);
    }  
  }

  *uspPDOHandle = IntCobId;

  pPdo = getPdo(&ch[ucChannelHandle], IntCobId);

  if(pPdo == NULL)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(-1);
  }

  if(pPdo->isUsed)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(-1);
  }



  if(stPDOConfig.ucPDOType == RX_PDO)
  {
    memcpy(pPdo->data,stPDOConfig.ucaDefVal, 8);
    pPdo->u8Type = RX_PDO;
    pPdo->u16CobId = stPDOConfig.ulCOBID;
  }
  else
  {
    memcpy(pPdo->data,stPDOConfig.ucaDefVal, 8);
    memcpy(pPdo->data252,stPDOConfig.ucaDefVal, 8);

    pPdo->u8Type = TX_PDO;
    pPdo->u16CobId = stPDOConfig.ulCOBID;

    if((stPDOConfig.ucTransType <= 240) && (stPDOConfig.ucTransType > 0))
    {
      pPdo->u8NumOfSyncs = stPDOConfig.ucTransType;
    }

    {
      pPdo->bSndReq = TRUE;
      bCheck = TRUE;
    }
  }

  insertPdo(pPdo, ucChannelHandle);

    
  pPdo->u8XmtType = stPDOConfig.ucTransType;
  pPdo->bEvent = stPDOConfig.ucEventNotif;
  pPdo->bSendAtNextSync = FALSE;

  pPdo->isUsed = TRUE;
  pPdo->u8DataLen = 8;

  if(bCheck) CheckPDOs(&ch[ucChannelHandle]);


  ReleaseChannelAccess(ucChannelHandle);

  COPRINT(("%s PDO %x, COB %lx, XMT-Type %3u installed!\n",(pPdo->u8Type)?"XMT":"RCV", IntCobId, pPdo->u16CobId, pPdo->u8XmtType));



  return MAP_ERROR(ret);

} //CMA_InstallPDO( );



/*Ewtg V1.20*/
/****************************************************************************/
/*                                                                          */
/*  CMA_InstallPDO_E                                                        */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_InstallPDO_E (USIGN8    ucChannelHandle, 
                                       CMA_PDOConfig_E  stPDOConfig_E, 
                                       USIGN16   *uspPDOHandle)
{
  USIGN16 ret = 0, IntCobId;
  PPDO pPdo = NULL;
  BOOLEAN bCheck = FALSE;


  if(stPDOConfig_E.ulCOBID > 0x7FF)
    return MAP_ERROR(CMA_ERR_SCT_COB_ID);
    


  if(!GetChannelAccess(ucChannelHandle))
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);

  IntCobId = getPdoIndex(
    &ch[ucChannelHandle], 
    (USIGN16)stPDOConfig_E.ulCOBID, 
    (BOOLEAN)(stPDOConfig_E.ucPDOType == 0));

  if(IntCobId == 0xFFFF)
  {
    ReleaseChannelAccess(ucChannelHandle);
    if(stPDOConfig_E.ucPDOType == RX_PDO)
    {
      if(stPDOConfig_E.ucAccType)
        return MAP_ERROR(CMA_ERR_FIFO_RX_PDO_MAX);
      else
        return MAP_ERROR(CMA_ERR_DPRAM_RX_PDO_MAX);
    }
    else
    {
      if(stPDOConfig_E.ucAccType)
        return MAP_ERROR(CMA_ERR_FIFO_TX_PDO_MAX);
      else
        return MAP_ERROR(CMA_ERR_DPRAM_TX_PDO_MAX);
    }
  }


  *uspPDOHandle = IntCobId;


  pPdo = getPdo(&ch[ucChannelHandle], IntCobId);

  if(pPdo == NULL)
    return MAP_ERROR(-1);

  if(pPdo->isUsed)
  {
    return MAP_ERROR(-1);
  }



  if(stPDOConfig_E.ucPDOType == RX_PDO)
  {
    memcpy(pPdo->data,stPDOConfig_E.ucaDefVal, stPDOConfig_E.ucNrOfByte);
    pPdo->u8Type = RX_PDO;
    pPdo->u16CobId = stPDOConfig_E.ulCOBID;
  }
  else
  {
    memcpy(pPdo->data,stPDOConfig_E.ucaDefVal, stPDOConfig_E.ucNrOfByte);
    memcpy(pPdo->data252,stPDOConfig_E.ucaDefVal, stPDOConfig_E.ucNrOfByte);
    pPdo->u8Type = TX_PDO;
    pPdo->u16CobId = stPDOConfig_E.ulCOBID;

    if((stPDOConfig_E.ucTransType <= 240) && (stPDOConfig_E.ucTransType > 0))
    {
      pPdo->u8NumOfSyncs = stPDOConfig_E.ucTransType;
    }

    {
      if(!stPDOConfig_E.ucNoInitTrans)
      {
        pPdo->bSndReq = TRUE;
        bCheck = TRUE;
        
      }
      else
      {
        pPdo->bSndReq = FALSE;
      }
    }
  }

  insertPdo(pPdo, ucChannelHandle);

  pPdo->bSendAtNextSync = FALSE;
  pPdo->u8XmtType = stPDOConfig_E.ucTransType;
  pPdo->bEvent = stPDOConfig_E.ucEventNotif;
  pPdo->isUsed = TRUE;
  pPdo->u8DataLen = stPDOConfig_E.ucNrOfByte;

  if(bCheck) CheckPDOs(&ch[ucChannelHandle]);


  ReleaseChannelAccess(ucChannelHandle);

  COPRINT(("%s PDO %x, COB %lx, XMT-Type %3u installed! E\n",(pPdo->u8Type)?"XMT":"RCV", IntCobId, pPdo->u16CobId,pPdo->u8XmtType));

  return MAP_ERROR(ret);

} //CMA_InstallPDO_E( );



                                                                /*Ewtg V1.20*/
/****************************************************************************/
/*                                                                          */
/*  CMA_RemovePDO                                                           */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_RemovePDO  (USIGN8  ucChannelHandle, 
                                     USIGN16 usPDOHandle) 
{
 // USIGN16 u16PdoHandle = usPDOHandle;
  PPDO pPdo;
  USIGN16 ExtCobId;


  if(!GetChannelAccess(ucChannelHandle))
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);

  pPdo = getPdoFromIndex(&ch[ucChannelHandle], usPDOHandle);

  if(!pPdo)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);
  }

  pPdo->isUsed = FALSE;

  ExtCobId = ch[ucChannelHandle].u16CobInt2Ext[usPDOHandle];
  ch[ucChannelHandle].u16CobInt2Ext[usPDOHandle] = usPDOHandle;
  ch[ucChannelHandle].u16CobExt2Int[ExtCobId] = ExtCobId;

   
  removePdo(pPdo, ucChannelHandle);

  pPdo->u16CobId = 0;
  
  ReleaseChannelAccess(ucChannelHandle);

  COPRINT(("%s PDO %x, COB %x, XMT-Type %3u removed!\n",(pPdo->u8Type)?"XMT":"RCV",usPDOHandle, ExtCobId, pPdo->u8XmtType));

  return 0;

} //CMA_RemovePDO( );



                                                                /*Ewtg V1.20*/
/****************************************************************************/
/*                                                                          */
/*  CMA_WritePDO                                                            */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_WritePDO  (USIGN8  ucChannelHandle, 
                                    USIGN16 usPDOHandle,
                                    USIGN8  ucDataOffset, 
                                    USIGN8  ucDataSize, 
                                    USIGN8  *ucpData) 
{

 // USIGN16 u16PdoHandle = usPDOHandle;
  PPDO pPdo;


  if((ucDataOffset + ucDataSize) > 8)
    return MAP_ERROR(CMA_ERR_PAR_DATA_SIZE);

  if(usPDOHandle > 0x7FF)
    return MAP_ERROR(CMA_ERR_PAR_PDO_HANDLE);

  if(!GetChannelAccess(ucChannelHandle))
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);

  pPdo = getPdoFromIndex(&ch[ucChannelHandle], usPDOHandle);

  if(!pPdo)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(CMA_ERR_PAR_PDO_HANDLE);
  }

  if(pPdo->u8Type == RX_PDO)
  {
    memcpy(&pPdo->data[ucDataOffset], ucpData, ucDataSize);
  }
  else
  {
    
    
    if((pPdo->u8XmtType != 252) && (pPdo->u8XmtType != 253))
    {
      if(!pPdo->u8XmtType)
      { // transmission type null!
        pPdo->bSendAtNextSync = TRUE;
      }
      else
      {
        pPdo->bSendAtNextSync = FALSE;
        if((pPdo->u8XmtType > 0) && (pPdo->u8XmtType < 240))
        {
          
        }
        else
        {
          pPdo->bSndReq = TRUE;
        }
      }
    }

    if(pPdo->u8XmtType == 252)
    {
      memcpy(&pPdo->data252[ucDataOffset], ucpData, ucDataSize);
    }
    else
    {
      memcpy(&pPdo->data[ucDataOffset], ucpData, ucDataSize);
    }
    
  }


    
  ReleaseChannelAccess(ucChannelHandle);

  COPRINT(("write %s PDO %x, COB %lx, XMT-Type %3u\n",(pPdo->u8Type)?"XMT":"RCV",usPDOHandle, pPdo->u16CobId, pPdo->u8XmtType));
  
  return (0);

} //CMA_WritePDO( );



                                                                /*Ewtg V1.20*/
/****************************************************************************/
/*                                                                          */
/*  CMA_WritePDOBit                                                         */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_WritePDOBit  (USIGN8  ucChannelHandle, 
                                       USIGN16 usPDOHandle,
                                       USIGN8  ucBitPosition, 
                                       USIGN8  ucBitValue) 
{
 // USIGN16 u16PdoHandle = usPDOHandle;
  PPDO pPdo;


  if(ucBitPosition > 63)
    return -1;


  if(usPDOHandle > 0x7FF)
    return MAP_ERROR(CMA_ERR_PAR_PDO_HANDLE);

  if(!GetChannelAccess(ucChannelHandle))
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);

  if(ucBitValue > 1) ucBitValue = 1;

  pPdo = getPdoFromIndex(&ch[ucChannelHandle], usPDOHandle);

  if(!pPdo)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(CMA_ERR_PAR_PDO_HANDLE);
  }

  if(pPdo->u8Type == RX_PDO)
  {
    USIGN8 byte, offset;
    byte = ucBitPosition / 8;
    offset = ucBitPosition % 8;
    if(offset)
    {
      if(ucBitValue)
      {
        pPdo->data[byte] |= (ucBitValue << offset);
      }
      else
      {
        pPdo->data[byte] &= ~(1 << offset);
      }
    }
    else
    {
      if(ucBitValue)
      {
        pPdo->data[byte] |= 1;
      }
      else
      {
        pPdo->data[byte] &= ~(1);
      }
    }


  }
  else
  {
    USIGN8 byte, offset;
    byte = ucBitPosition / 8;
    offset = ucBitPosition % 8;
    if(offset)
    {
      if(ucBitValue)
      {
        pPdo->data[byte] |= (ucBitValue << offset);
      }
      else
      {
        pPdo->data[byte] &= ~(1 << offset);
      }
    }
    else
    {
      if(ucBitValue)
      {
        pPdo->data[byte] |= 1;
      }
      else
      {
        pPdo->data[byte] &= ~(1);
      }
    }
    
    if((pPdo->u8XmtType != 252) && (pPdo->u8XmtType != 253))
    {
      if(!pPdo->u8XmtType)
      { // transmission type null!
        pPdo->bSendAtNextSync = TRUE;
      }
      else
      {
        pPdo->bSendAtNextSync = FALSE;
        if((pPdo->u8XmtType > 0) && (pPdo->u8XmtType < 240))
        {
          
        }
        else
        {
          pPdo->bSndReq = TRUE;
        }
      }
    }
    
  }


    
  ReleaseChannelAccess(ucChannelHandle);

  return 0;

} //CMA_WritePDOBit( );



                                                                /*Ewtg V1.20*/
/****************************************************************************/
/*                                                                          */
/*  CMA_ReadPDO                                                             */
/*                                                                          */
/****************************************************************************/
/* !!!                                                                      */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_ReadPDO  (USIGN8  ucChannelHandle, 
                                   USIGN16 usPDOHandle,
                                   USIGN8  *ucpData) 
{

 // USIGN16 u16PdoHandle = usPDOHandle;
  PPDO  pPdo;

  if(usPDOHandle > 0x7FF)
    return MAP_ERROR(CMA_ERR_PAR_PDO_HANDLE);

  if(!GetChannelAccess(ucChannelHandle))
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);

  pPdo = getPdoFromIndex(&ch[ucChannelHandle], usPDOHandle);

  if(!pPdo)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(CMA_ERR_PAR_PDO_HANDLE);
  }

  memset(ucpData, 0, 8);

  if(pPdo->u8Type == RX_PDO)
  {
    memcpy(ucpData, &pPdo->data[0], pPdo->u8DataLen);
  }
  else
  {
    memcpy(ucpData, &pPdo->data[0], pPdo->u8DataLen);
  }

  ReleaseChannelAccess(ucChannelHandle);

  return (0);

} //CMA_ReadPDO( );



/****************************************************************************/
/*                                                                          */
/*  CMA_SendRemotePDO                                                       */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_SendRemotePDO  (USIGN8  ucChannelHandle, 
                                         USIGN16 usPDOHandle) 
{
 // USIGN16 u16PdoHandle = usPDOHandle;
  PPDO pPdo;

  if(usPDOHandle > 0x7FF)
    return MAP_ERROR(CMA_ERR_PAR_PDO_HANDLE);

  if(!GetChannelAccess(ucChannelHandle))
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE);

  pPdo = getPdoFromIndex(&ch[ucChannelHandle], usPDOHandle);

  if(!pPdo)
  {
    ReleaseChannelAccess(ucChannelHandle);
    return MAP_ERROR(CMA_ERR_PAR_PDO_HANDLE);
  }

  if(pPdo->u8Type == RX_PDO)
  {    
     CO_send_remote(
       &ch[ucChannelHandle], 
       ch[ucChannelHandle].channel, 
       pPdo->u16CobId, 
       ch[ucChannelHandle].isXtd,
       8);
  }
    
  ReleaseChannelAccess(ucChannelHandle);

  // return success
  return (0);

} //CMA_SendRemotePDO( );


/****************************************************************************/
/*                                                                          */
/*  CMA_SendL2Message                                                       */
/*                                                                          */
/****************************************************************************/
COL_API INT16 CALLING_CONV CMA_SendL2Message  (USIGN8  ucChannelHandle, 
                                         USIGN32  ulCanID,
                                         USIGN8  ucConfig,
                                         USIGN8  ucDLC,
                                         USIGN8  *ucpData) 
{
  int ret;

  if(ucChannelHandle > MAX_CHANNEL)
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE); 

  ret = CANL2_send_data(ch[ucChannelHandle].channel, ulCanID, ucConfig, ucDLC, ucpData);
  
  return MAP_ERROR(ret);

} //CMA_SendL2Message( );




/****************************************************************************/
/*                                                                          */
/*  CMA_GetVersion                                                          */
/*                                                                          */
/****************************************************************************/


COL_API INT16 CALLING_CONV CMA_GetVersion (USIGN8     ucChannelHandle,
                                     CMA_VersionStruct *stpVersion)
{ // used by CMA_ConfigCANChannel, be careful with synchronisation!
  int sw_version, fw_version, hw_version, license, chip[2], ret;
  USIGN32 serial, u32deviceId;

    
  if(ucChannelHandle > MAX_CHANNEL)
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE); 


  ret = CANL2_get_version(
    ch[ucChannelHandle].channel,
    &sw_version,
    &fw_version,
    &hw_version,
    &license,
    chip
    );

  if(ret) return MAP_ERROR(ret);

  ret = CANL2_get_serial_number(ch[ucChannelHandle].channel, &serial);

  stpVersion->ulHWSerialNumber = serial;
  stpVersion->ucHWMainRevision = hw_version;
  stpVersion->ucHWSubRevision = hw_version;
  stpVersion->ucSWMajorVersion = SW_MAJ_VERSION;
  stpVersion->ucSWMinorVersion = SW_MIN_VERSION;
  stpVersion->ucSWBuild = SW_BUILD;

  if(ret) return MAP_ERROR(ret);


  ret = CANL2_get_device_type(ch[ucChannelHandle].channel, &u32deviceId);

  switch(u32deviceId)
  {
    case CANCARD2:
      memcpy(&stpVersion->ucHWIdentString, HW_STRING_CC2, sizeof(HW_STRING_CC2));
      break;
    case EDICCARD2:
      memcpy(&stpVersion->ucHWIdentString, HW_STRING_ED2, sizeof(HW_STRING_ED2));
      break;
    case CANPCI:
      memcpy(&stpVersion->ucHWIdentString, HW_STRING_PCI, sizeof(HW_STRING_PCI));
      break;
    case CAN_104:
      memcpy(&stpVersion->ucHWIdentString, HW_STRING_104, sizeof(HW_STRING_104));
      break;
    case CANUSB:
      memcpy(&stpVersion->ucHWIdentString, HW_STRING_USB, sizeof(HW_STRING_USB));
      break;
    case CANPROPCIE:
      memcpy(&stpVersion->ucHWIdentString, HW_STRING_PCIE, sizeof(HW_STRING_PCIE));
      break;
    case CAN104PLUS:
      memcpy(&stpVersion->ucHWIdentString, HW_STRING_104P, sizeof(HW_STRING_104P));
      break;
    
    
    default:
      memcpy(&stpVersion->ucHWIdentString, "unknown", sizeof("unknown"));
      break;

  }  

  return(ret);

} //CMA_GetVersion( );





COL_API INT16 CALLING_CONV CMA_WaitForEvent(USIGN8 ucChannelHandle)
{
  int rc;
  
  if(ucChannelHandle > MAX_CHANNEL)
    return MAP_ERROR(CMA_ERR_PAR_CHAN_HANDLE); 
  
  if(ch[ucChannelHandle].s16WaitRet)
    return ch[ucChannelHandle].s16WaitRet;
  
  if(!ch[ucChannelHandle].isDefined)
    return -1;
  
  pthread_mutex_lock(&ch[ucChannelHandle].hQueueEvent);
  
  while (!ch[ucChannelHandle].conditionMet)
  {
//     printf("Thread blocked\n");
    pthread_cond_wait(&ch[ucChannelHandle].QueueCondition, &ch[ucChannelHandle].hQueueEvent);
//     checkResults("pthread_cond_wait()\n", rc);
  }
  ch[ucChannelHandle].conditionMet = FALSE;
  rc = pthread_mutex_unlock(&ch[ucChannelHandle].hQueueEvent);
  
  
  return ch[ucChannelHandle].s16WaitRet;
 
}


/**
 ******************************************************************
 * @brief  waits for a response from the slave 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  pLss: pointer to an LSS data structure
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */
INT16 LSS_WaitForResponse(LSS *pLss)
{
 // return WAIT_FOR_EVENT(pLss->hLSSEvent, pLss->u32Timeout);
  USIGN32 timeslices = 0;
  
  while(timeslices < pLss->u32Timeout)
  {
    usleep(20*1000); // 20 ms
    timeslices += 20*1000;
    if(!pLss->u8LSSPendingService)
    {
      return 0;
    }
  }
  
  return CMA_ERR_FUNCTION_TIMEOUT;
}

/**
 ******************************************************************
 * @brief  initialize LSS Manager 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  ucChannelHandle: channel handle
 * @param  usLssTimeout: timout for slave responses
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */

COL_API INT16 CALLING_CONV LSS_InitLSSMan(USIGN8 ucChannelHandle, USIGN16 usLssTimeout)
{
  ch[ucChannelHandle].Lss.u32Timeout = usLssTimeout * 1000;

  /*
  if (NULL == (ch[ucChannelHandle].Lss.hLSSEvent = CREATE_EVENT(NULL, FALSE, FALSE, NULL)))
  {
    return MAP_ERROR(INIPC_OS_ERR);
  }

  RESET_EVENT(ch[ucChannelHandle].Lss.hLSSEvent);
  */

  return 0;

}

/**
 ******************************************************************
 * @brief  switch mode global 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  ucChannelHandle: channel handle
 * @param  ucMode: mode -> 1 = configuration mode, 0 = operation mode;
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */

COL_API INT16 CALLING_CONV LSS_SwitchModeGlobal(USIGN8 ucChannelHandle, USIGN8 ucMode)
{
  USIGN8 data[8];
  int ret;

  if(ucMode > 1)
    return MAP_ERROR(LSS_WRONG_MODE);

  memset(data,0,8);

  data[0] = LSS_SWITCH_GLOBAL_MODE;
  data[1] = ucMode;

  ret = CANL2_send_data(
    ch[ucChannelHandle].channel, 
    LSS_MASTER_ID, 
    ch[ucChannelHandle].isXtd, 
    8,
    data);

  return ret;


}

/**
 ******************************************************************
 * @brief  switch mode selective 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  ucChannelHandle: channel handle
 * @param  ulVendorId: vendor id
 * @param  ulProductCode: product code
 * @param  ulRevisionNumber: revision number
 * @param  ulSerialNumber: serial number
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */

COL_API INT16 CALLING_CONV LSS_SwitchModeSelective(USIGN8 ucChannelHandle, 
                                                   USIGN32 ulVendorId,
                                                   USIGN32 ulProductCode,
                                                   USIGN32 ulRevisionNumber,
                                                   USIGN32 ulSerialNumber)
{
  USIGN8 data[8];
  int ret;


  memset(data,0,8);

  data[0] = LSS_SWITCH_MODE_VID;

  WRT32(&data[1], ulVendorId);

  ret = CANL2_send_data(
    ch[ucChannelHandle].channel, 
    LSS_MASTER_ID, 
    ch[ucChannelHandle].isXtd, 
    8,
    data);

  if(ret) return ret;

  data[0] = LSS_SWITCH_MODE_PC;
  WRT32(&data[1], ulProductCode);

  ret = CANL2_send_data(
    ch[ucChannelHandle].channel, 
    LSS_MASTER_ID, 
    ch[ucChannelHandle].isXtd, 
    8,
    data);

  if(ret) return ret;

  data[0] = LSS_SWITCH_MODE_RN;
  WRT32(&data[1], ulRevisionNumber);

  ret = CANL2_send_data(
    ch[ucChannelHandle].channel, 
    LSS_MASTER_ID, 
    ch[ucChannelHandle].isXtd, 
    8,
    data);

  if(ret) return ret;

  data[0] = LSS_SWITCH_MODE_SN;
  WRT32(&data[1], ulSerialNumber);

  ret = CANL2_send_data(
    ch[ucChannelHandle].channel, 
    LSS_MASTER_ID, 
    ch[ucChannelHandle].isXtd, 
    8,
    data);

  ch[ucChannelHandle].Lss.u8LSSPendingService = LSS_RES_SWITCH_MODE;

  ret = LSS_WaitForResponse(&ch[ucChannelHandle].Lss);

  return ret;


}

/**
 ******************************************************************
 * @brief  configure node id 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  ucChannelHandle: channel handle
 * @param  ucNodeId: new node id
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */
COL_API INT16 CALLING_CONV LSS_ConfigureNodeID(USIGN8 ucChannelHandle, USIGN8 ucNodeId)
{
  USIGN8 data[8];
  int ret;


  memset(data,0,8);

  data[0] = LSS_CONFIGURE_NODE_ID;
  data[1] = ucNodeId;

  ret = CANL2_send_data(
    ch[ucChannelHandle].channel, 
    LSS_MASTER_ID, 
    ch[ucChannelHandle].isXtd, 
    8,
    data);

  ch[ucChannelHandle].Lss.u8LSSPendingService = LSS_CONFIGURE_NODE_ID;

  ret = LSS_WaitForResponse(&ch[ucChannelHandle].Lss);

  return ret;


}

/**
 ******************************************************************
 * @brief  configure bit timing 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  ucChannelHandle: channel handle
 * @param  ucTable: bit timing table; must always be 0
 * @param  ucIndex: index in the table 0; see 4.2.2, DSP305
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */

COL_API INT16 CALLING_CONV LSS_ConfigureBitTiming(USIGN8 ucChannelHandle, 
                                                  USIGN8 ucTable,
                                                  USIGN8 ucIndex)
{
  USIGN8 data[8];
  int ret;


  memset(data,0,8);

  data[0] = LSS_CONFIGURE_BIT_TIMING;
  data[1] = ucTable;
  data[2] = ucIndex;

  ret = CANL2_send_data(
    ch[ucChannelHandle].channel, 
    LSS_MASTER_ID, 
    ch[ucChannelHandle].isXtd, 
    8,
    data);

  ch[ucChannelHandle].Lss.u8LSSPendingService = LSS_CONFIGURE_BIT_TIMING;

  ret = LSS_WaitForResponse(&ch[ucChannelHandle].Lss);

  return ret;


}

/**
 ******************************************************************
 * @brief  activates a configured bit timing 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  ucChannelHandle: channel handle
 * @param  usDelay: delay for switching the bit timing and for sending the first message
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */
COL_API INT16 CALLING_CONV LSS_ActivateBitTiming(USIGN8 ucChannelHandle, 
                                                  USIGN16 usDelay)
{
  USIGN8 data[8];
  int ret;


  memset(data,0,8);

  data[0] = LSS_ACTIVATE_BIT_TIMING;
  WRT16(&data[1], usDelay)

  ret = CANL2_send_data(
    ch[ucChannelHandle].channel, 
    LSS_MASTER_ID, 
    ch[ucChannelHandle].isXtd, 
    8,
    data);

  ch[ucChannelHandle].Lss.u8LSSPendingService = LSS_ACTIVATE_BIT_TIMING;

  return ret;


}

/**
 ******************************************************************
 * @brief  store configuration 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  ucChannelHandle: channel handle
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */

COL_API INT16 CALLING_CONV LSS_StoreConfiguration(USIGN8 ucChannelHandle)
{
  USIGN8 data[8];
  int ret;


  memset(data,0,8);

  data[0] = LSS_STORE_CONFIGURATION;

  ret = CANL2_send_data(
    ch[ucChannelHandle].channel, 
    LSS_MASTER_ID, 
    ch[ucChannelHandle].isXtd, 
    8,
    data);

  ch[ucChannelHandle].Lss.u8LSSPendingService = LSS_STORE_CONFIGURATION;

  ret = LSS_WaitForResponse(&ch[ucChannelHandle].Lss);

  return ret;


}

/**
 ******************************************************************
 * @brief  inquire vendor id 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  ucChannelHandle: channel handle
 * @param  pNodeId: pointer to the vendor id
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */
COL_API INT16 CALLING_CONV LSS_InquireVendorID(USIGN8 ucChannelHandle, 
                                               USIGN32 *pVendorId)
{
  USIGN8 data[8];
  int ret;


  memset(data,0,8);

  data[0] = LSS_INQ_ID_VID;

  ret = CANL2_send_data(
    ch[ucChannelHandle].channel, 
    LSS_MASTER_ID, 
    ch[ucChannelHandle].isXtd, 
    8,
    data);

  ch[ucChannelHandle].Lss.u8LSSPendingService = LSS_INQ_ID_VID;

  ret = LSS_WaitForResponse(&ch[ucChannelHandle].Lss);

  if(ret) return MAP_ERROR(LSS_TIMEOUT);

  *pVendorId = ch[ucChannelHandle].Lss.u32Vid;

  ch[ucChannelHandle].Lss.u8LSSPendingService = LSS_NO_SERVICE;
  return ret;


}

/**
 ******************************************************************
 * @brief  inquire product code 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  ucChannelHandle: channel handle
 * @param  pNodeId: pointer to the product code
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */

COL_API INT16 CALLING_CONV LSS_InquireProductCode(USIGN8 ucChannelHandle,
                                                  USIGN32 *pProductCode)
{
  USIGN8 data[8];
  int ret;


  memset(data,0,8);

  data[0] = LSS_INQ_ID_PC;

  ret = CANL2_send_data(
    ch[ucChannelHandle].channel, 
    LSS_MASTER_ID, 
    ch[ucChannelHandle].isXtd, 
    8,
    data);

  ch[ucChannelHandle].Lss.u8LSSPendingService = LSS_INQ_ID_PC;

  ret = LSS_WaitForResponse(&ch[ucChannelHandle].Lss);

  if(ret) return MAP_ERROR(LSS_TIMEOUT);

  *pProductCode = ch[ucChannelHandle].Lss.u32Pc;

  ch[ucChannelHandle].Lss.u8LSSPendingService = LSS_NO_SERVICE;


  return ret;


}

/**
 ******************************************************************
 * @brief  inquire serial number 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  ucChannelHandle: channel handle
 * @param  pNodeId: pointer to the serial number
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */

COL_API INT16 CALLING_CONV LSS_InquireSerialNr(unsigned char ucChannelHandle,
                                               USIGN32 *pSerialnumber)
{
  USIGN8 data[8];
  int ret;


  memset(data,0,8);

  data[0] = LSS_INQ_ID_SN;

  ret = CANL2_send_data(
    ch[ucChannelHandle].channel, 
    LSS_MASTER_ID, 
    ch[ucChannelHandle].isXtd, 
    8,
    data);

  ch[ucChannelHandle].Lss.u8LSSPendingService = LSS_INQ_ID_SN;

  ret = LSS_WaitForResponse(&ch[ucChannelHandle].Lss);

  if(ret) return MAP_ERROR(LSS_TIMEOUT);

  *pSerialnumber = ch[ucChannelHandle].Lss.u32Sn;

  ch[ucChannelHandle].Lss.u8LSSPendingService = LSS_NO_SERVICE;


  return ret;


}

/**
 ******************************************************************
 * @brief  inquire revision number 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  ucChannelHandle: channel handle
 * @param  pNodeId: pointer to the revision number
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */

COL_API INT16 CALLING_CONV LSS_InquireRevNr(USIGN8 ucChannelHandle,
                                            USIGN32 *pRevisionNr)
{
  USIGN8 data[8];
  int ret;


  memset(data,0,8);

  data[0] = LSS_INQ_ID_RN;

  ret = CANL2_send_data(
    ch[ucChannelHandle].channel, 
    LSS_MASTER_ID, 
    ch[ucChannelHandle].isXtd, 
    8,
    data);

  ch[ucChannelHandle].Lss.u8LSSPendingService = LSS_INQ_ID_RN;

  ret = LSS_WaitForResponse(&ch[ucChannelHandle].Lss);

  if(ret) return MAP_ERROR(LSS_TIMEOUT);

  *pRevisionNr = ch[ucChannelHandle].Lss.u32Rn;

  ch[ucChannelHandle].Lss.u8LSSPendingService = LSS_NO_SERVICE;

  return ret;


}

/**
 ******************************************************************
 * @brief  inquire node id 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  ucChannelHandle: channel handle
 * @param  pNodeId: pointer to the node id
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */

COL_API INT16 CALLING_CONV LSS_InquireNodeID(USIGN8 ucChannelHandle,
                                             USIGN8 *pNodeId)
{
  USIGN8 data[8];
  int ret;


  memset(data,0,8);

  data[0] = LSS_INQ_ID_NODE_ID;

  ret = CANL2_send_data(
    ch[ucChannelHandle].channel, 
    LSS_MASTER_ID, 
    ch[ucChannelHandle].isXtd, 
    8,
    data);

  ch[ucChannelHandle].Lss.u8LSSPendingService = LSS_INQ_ID_NODE_ID;

  ret = LSS_WaitForResponse(&ch[ucChannelHandle].Lss);

  if(ret) return MAP_ERROR(LSS_TIMEOUT);

  *pNodeId = ch[ucChannelHandle].Lss.u8NodeId;

  ch[ucChannelHandle].Lss.u8LSSPendingService = LSS_NO_SERVICE;

  return ret;


}

/**
 ******************************************************************
 * @brief  sends the LSS service identify remote slaves 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  ucChannelHandle: channel handle
 * @param  ulVendorId: vendor id
 * @param  ulProductCode: product code
 * @param  ulRevisionNumberLow: lower boundary of the requested revision number
 * @param  ulRevisionNumberHigh: higher boundary of the requested revision number
 * @param  ulSerialNumberLow: lower boundary of the requested serial number
 * @param  ulSerialNumberHigh: higher boundary of the requested serial number
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */

COL_API INT16 CALLING_CONV LSS_IdentityRemoteSlaves(USIGN8 ucChannelHandle, 
                                                   USIGN32 ulVendorId,
                                                   USIGN32 ulProductCode,
                                                   USIGN32 ulRevisionNumberLow,
                                                   USIGN32 ulRevisionNumberHigh,
                                                   USIGN32 ulSerialNumberLow,
                                                   USIGN32 ulSerialNumberHigh)
{
  USIGN8 data[8];
  int ret;


  memset(data,0,8);

  data[0] = LSS_IDENTITY_VID;

  WRT32(&data[1], ulVendorId);

  ret = CANL2_send_data(
    ch[ucChannelHandle].channel, 
    LSS_MASTER_ID, 
    ch[ucChannelHandle].isXtd, 
    8,
    data);

  if(ret) return ret;

  data[0] = LSS_IDENTITY_PC;
  WRT32(&data[1], ulProductCode);

  ret = CANL2_send_data(
    ch[ucChannelHandle].channel, 
    LSS_MASTER_ID, 
    ch[ucChannelHandle].isXtd, 
    8,
    data);

  if(ret) return ret;

  data[0] = LSS_IDENTITY_RNL;
  WRT32(&data[1], ulRevisionNumberLow);

  ret = CANL2_send_data(
    ch[ucChannelHandle].channel, 
    LSS_MASTER_ID, 
    ch[ucChannelHandle].isXtd, 
    8,
    data);

  if(ret) return ret;

  data[0] = LSS_IDENTITY_RNH;
  WRT32(&data[1], ulRevisionNumberHigh);

  ret = CANL2_send_data(
    ch[ucChannelHandle].channel, 
    LSS_MASTER_ID, 
    ch[ucChannelHandle].isXtd, 
    8,
    data);

  if(ret) return ret;

  data[0] = LSS_IDENTITY_SNL;
  WRT32(&data[1], ulSerialNumberLow);

  ret = CANL2_send_data(
    ch[ucChannelHandle].channel, 
    LSS_MASTER_ID, 
    ch[ucChannelHandle].isXtd, 
    8,
    data);

  if(ret) return ret;

  data[0] = LSS_IDENTITY_SNH;
  WRT32(&data[1], ulSerialNumberHigh);

  ret = CANL2_send_data(
    ch[ucChannelHandle].channel, 
    LSS_MASTER_ID, 
    ch[ucChannelHandle].isXtd, 
    8,
    data);

  ch[ucChannelHandle].Lss.u8LSSPendingService = LSS_RES_IDENTITY;

  ret = LSS_WaitForResponse(&ch[ucChannelHandle].Lss);

  return ret;


}

/**
 ******************************************************************
 * @brief  sends the LSS service identify non configured slaves 
 * @author Werner Knorsch, Softing AG
 * @date   2005-01-13
 *
 * @param  ucChannelHandle: channel handle
 * @return 0 if successful, or an error code
 *
 ******************************************************************
 */

COL_API INT16 CALLING_CONV LSS_IdentifyNonConfiguredSlaves(USIGN8 ucChannelHandle)
{
  USIGN8 data[8];
  int ret;


  memset(data,0,8);

  data[0] = LSS_IDENTIFY_NONCFGSLVS;

  ret = CANL2_send_data(
    ch[ucChannelHandle].channel, 
    LSS_MASTER_ID, 
    ch[ucChannelHandle].isXtd, 
    8,
    data);

  ch[ucChannelHandle].Lss.u8LSSPendingService = LSS_IDENTIFY_NONCFGSLVS;

  ret = LSS_WaitForResponse(&ch[ucChannelHandle].Lss);

  return ret;


}

COL_API INT16 CALLING_CONV DebugEnable(BOOLEAN bEnable)
{
  isDbgEnabled = bEnable;

  return 0;
}
